#ifdef __MVS__
                   ??=ifdef __COMPILER_VER__
                     ??=pragma filetag("IBM-1047")
                   ??=endif
                   #pragma nomargins nosequence
                   #pragma checkout(suspend)
#endif

#ifndef _IBUFFER_
#define _IBUFFER_
/*******************************************************************************
* FILE NAME: ibuffer.hpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*     IBuffer                                                                  *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #include <stddef.h>
  #include <stdio.h>
  #include <string.h>
  }

#include <ivbase.hpp>
#include <istrenum.hpp>

#pragma pack(4)


#if defined (__SOM_ENABLED__)
#if defined (__RRBC_LIB__)
#pragma SOMAsDefault (on)
#else
#pragma SOMAsDefault (off)
#endif
#endif

class IStringTest;

class IBuffer : public IVBase {
public:
/*------------------------------- Reallocation -------------------------------*/
IBuffer
 *newBuffer           ( const void* p1,
                        unsigned    len1,
                        const void* p2      = 0,
                        unsigned    len2    = 0,
                        const void* p3      = 0,
                        unsigned    len3    = 0,
                        char        padChar = 0 ) const;

static unsigned
  checkAddition       ( unsigned addend1,
                        unsigned addend2 ),
  checkMultiplication ( unsigned factor1,
                        unsigned factor2 ),
  overflow            ( );

static void
  setDefaultBuffer    ( IBuffer* newDefaultBuffer );

/*--------------------------------- Queries ----------------------------------*/
static IBuffer
 *defaultBuffer ( ),
 *fromContents  ( const char *pBuffer );

IBuffer
 *null          ( ) const;

unsigned
  useCount      ( ) const,
  length        ( ) const;

const char
 *contents      ( ) const;

char
 *contents      ( );

virtual IStringEnum::CharType
  charType      ( unsigned index ) const;

virtual char
 *next          ( const char *prev );

virtual const char
 *next          ( const char *prev ) const;

/*---------------------------- Reference Counting ----------------------------*/
void
  addRef    ( ),
  removeRef ( );

/*------------------------------- Comparisons --------------------------------*/
typedef enum {
  equal,
  greaterThan,
  lessThan
  } Comparison;

virtual Comparison
  compare ( const void* p,
            unsigned    len ) const;
/*---------------------------------- Subset ----------------------------------*/
virtual IBuffer
 *subString ( unsigned startPos,
              unsigned len,
              char     padCharacter ) const;

/*-----------------Internationalization functions----------------------*/
#ifdef IC_NLS
virtual IBuffer
 *subString ( unsigned startPos ) const;
#endif

/*------------------------------- Diagnostics --------------------------------*/
virtual IString
  asDebugInfo ( ) const;

/*--------------------------------- Testing ----------------------------------*/
virtual Boolean
  isAlphanumeric ( ) const,
  isAlphabetic   ( ) const,
  isASCII        ( ) const,
  isControl      ( ) const,
  isDigits       ( ) const,
  isGraphics     ( ) const,
  isHexDigits    ( ) const,
  isLowerCase    ( ) const,
  isPrintable    ( ) const,
  isPunctuation  ( ) const,
  isUpperCase    ( ) const,
  isWhiteSpace   ( ) const;

/*------------------------------- NLS Testing --------------------------------*/
virtual Boolean
  isMBCS       ( ) const,
  isDBCS       ( ) const,
  isSBCS       ( ) const,
  isValidMBCS  ( ) const,
  isValidDBCS  ( ) const,
  includesMBCS ( ) const,
  includesDBCS ( ) const,
  includesSBCS ( ) const;

/*--------------------------------- Searches ---------------------------------*/
virtual unsigned
  indexOf           ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 1 ) const,
  indexOf           ( const IStringTest& aTest,
                      unsigned           startPos = 1 ) const,
  indexOfAnyBut     ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 1 ) const,
  indexOfAnyBut     ( const IStringTest& aTest,
                      unsigned           startPos = 1 ) const,
  indexOfAnyOf      ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 1 ) const,
  indexOfAnyOf      ( const IStringTest& aTest,
                      unsigned           startPos = 1 ) const,
  lastIndexOf       ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 0 ) const,
  lastIndexOf       ( const IStringTest& aTest,
                      unsigned           startPos = 0 ) const,
  lastIndexOfAnyBut ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 0 ) const,
  lastIndexOfAnyBut ( const IStringTest& aTest,
                      unsigned           startPos = 0 ) const,
  lastIndexOfAnyOf  ( const char*        pString,
                      unsigned           len,
                      unsigned           startPos = 0 ) const,
  lastIndexOfAnyOf  ( const IStringTest& aTest,
                      unsigned           startPos = 0 ) const;

/*--------------------------------- Editing ----------------------------------*/
virtual IBuffer
 *center       ( unsigned newLen,
                 char     padCharacter ),
 *change       ( const char* pSource,
                 unsigned    sourceLen,
                 const char* pTarget,
                 unsigned    targetLen,
                 unsigned    startPos,
                 unsigned    numChanges ),
 *copy         ( unsigned numCopies ),
 *insert       ( const char* pInsert,
                 unsigned    insertLen,
                 unsigned    pos,
                 char        padCharacter ),
 *leftJustify  ( unsigned newLen,
                 char     padCharacter ),
 *lowerCase    ( ),
 *overlayWith  ( const char* overlay,
                 unsigned    len,
                 unsigned    pos,
                 char        padCharacter ),
 *remove       ( unsigned startPos,
                 unsigned numChars ),
 *reverse      ( ),
 *rightJustify ( unsigned newLen,
                 char     padCharacter ),
 *strip        ( const char*            pChars,
                 unsigned               len,
                 IStringEnum::StripMode mode ),
 *strip        ( const IStringTest&     aTest,
                 IStringEnum::StripMode mode ),
 *translate    ( const char* pInputChars,
                 unsigned    inputLen,
                 const char* pOutputChars,
                 unsigned    outputLen,
                 char        padCharacter ),
 *upperCase    ( );

protected:

#ifdef IC_PMWIN
static Boolean isDBCSLead(char inByte);
#endif


/*-------------------------- Search Initialization ---------------------------*/
virtual unsigned
  startSearch          ( unsigned startPos,
                         unsigned searchLen ) const,
  startBackwardsSearch ( unsigned startPos,
                         unsigned searchLen ) const;

/*-------------------------------- Allocation --------------------------------*/
virtual IBuffer
 *allocate        ( unsigned bufLength ) const;

#ifdef __DEBUG_ALLOC__
void
 *operator new    ( size_t      t,
                    const char* filename,
                    size_t      linenum,
                    unsigned    bufLen );
void
  operator delete ( void*       p,
                    const char* filename,
                    size_t      linenum );
#else
void
 *operator new    ( size_t   t,
                    unsigned bufLen );
void
  operator delete ( void*    p );
#endif

/*-------------------------------- Initialize --------------------------------*/
static IBuffer
 *initialize ( );

/*---------------------------- Protected Queries -----------------------------*/
virtual const char
 *className ( ) const;

/*-------------------------- Protected Constructors --------------------------*/
  IBuffer ( unsigned newLen );
 ~IBuffer ( );

private:
/*--------------------------------- Private ----------------------------------*/
unsigned
  refs,
  len;
char
  data[1];
static IBuffer
 *nullBuffer;

static IBuffer
 *nullIBuffer;

#ifdef IC_PMWIN
#ifdef __BORLANDC__
 public:
#endif

/*-------------------------------- DBCS Table --------------------------------*/
static char
  dbcsTable[ 256 ];
#endif

#if defined (__SOM_ENABLED__ ) && defined (__RRBC_LIB__)

#pragma SOMClassVersion  (*, 3, 0)

#ifdef __MVS__

#pragma SOMReleaseOrder( \
/*   1 */       newBuffer(const void*,unsigned int,const void*,  \
                          unsigned int,const void*,unsigned int,char) const,\
/*   2 */       checkAddition(unsigned int,unsigned int),\
/*   3 */       checkMultiplication(unsigned int,unsigned int),\
/*   4 */       overflow(),\
/*   5 */       setDefaultBuffer(IBuffer*),\
/*   6 */       defaultBuffer(),\
/*   7 */       fromContents(const char*),\
/*   8 */       null() const,\
/*   9 */       useCount() const,\
/*  10 */       length() const,\
/*  11 */       contents() const,\
/*  12 */       contents(),\
/*  13 */       charType(unsigned int) const,\
/*  14 */       next(const char*),\
/*  15 */       next(const char*) const,\
/*  16 */       addRef(),\
/*  17 */       removeRef(),\
/*  18 */       compare(const void*,unsigned int) const,\
/*  19 */       subString(unsigned int,unsigned int,char) const,\
/*  20 */       isAlphanumeric() const,\
/*  21 */       isAlphabetic() const,\
/*  22 */       isASCII() const,\
/*  23 */       isControl() const,\
/*  24 */       isDigits() const,\
/*  25 */       isGraphics() const,\
/*  26 */       isHexDigits() const,\
/*  27 */       isLowerCase() const,\
/*  28 */       isPrintable() const,\
/*  29 */       isPunctuation() const,\
/*  30 */       isUpperCase() const,\
/*  31 */       isWhiteSpace() const,\
/*  32 */       isMBCS() const,\
/*  33 */       isDBCS() const,\
/*  34 */       isSBCS() const,\
/*  35 */       isValidMBCS() const,\
/*  36 */       isValidDBCS() const,\
/*  37 */       includesMBCS() const,\
/*  38 */       includesDBCS() const,\
/*  39 */       includesSBCS() const,\
/*  40 */       indexOf(const char*,unsigned int,unsigned int) const,\
/*  41 */       indexOf(const IStringTest&,unsigned int) const,\
/*  42 */       indexOfAnyBut(const char*,unsigned int,unsigned int) const,\
/*  43 */       indexOfAnyBut(const IStringTest&,unsigned int) const,\
/*  44 */       indexOfAnyOf(const char*,unsigned int,unsigned int) const,\
/*  45 */       indexOfAnyOf(const IStringTest&,unsigned int) const,\
/*  46 */       lastIndexOf(const char*,unsigned int,unsigned int) const,\
/*  47 */       lastIndexOf(const IStringTest&,unsigned int) const,\
/*  48 */       lastIndexOfAnyBut(const char*, \
                                  unsigned int,unsigned int) const,\
/*  49 */       lastIndexOfAnyBut(const IStringTest&,unsigned int) const,\
/*  50 */       lastIndexOfAnyOf(const char*,   \
                                 unsigned int,unsigned int) const,\
/*  51 */       lastIndexOfAnyOf(const IStringTest&,unsigned int) const,\
/*  52 */       center(unsigned int,char),\
/*  53 */       change(const char*,unsigned int,const char*,  \
                       unsigned int,unsigned int,unsigned int),\
/*  54 */       copy(unsigned int),\
/*  55 */       insert(const char*,unsigned int,unsigned int,char),\
/*  56 */       leftJustify(unsigned int,char),\
/*  57 */       lowerCase(),\
/*  58 */       overlayWith(const char*,unsigned int,unsigned int,char),\
/*  59 */       remove(unsigned int,unsigned int),\
/*  60 */       reverse(),\
/*  61 */       rightJustify(unsigned int,char),\
/*  62 */       strip(const char*,unsigned int,IStringEnum::StripMode),\
/*  63 */       strip(const IStringTest&,IStringEnum::StripMode),\
/*  64 */       translate(const char*,unsigned int,   \
                          const char*,unsigned int,char),\
/*  65 */       upperCase(),\
/*  66 */       startSearch(unsigned int,unsigned int) const,\
/*  67 */       startBackwardsSearch(unsigned int,unsigned int) const,\
/*  68 */       allocate(unsigned int) const,\
/*  69 */       operator new(SOMClass*,size_t,unsigned int),\
/*  70 */       operator delete(void*),\
/*  71 */   /*  dbcsTable,  OS/2 speecific NOT Supported on MVS */   \
/*  72 */       initialize(),\
/*  73 */       className() const,\
/*  74 */       IBuffer(unsigned int),\
/*  75 */       nullBuffer,\
/*  76 */       nullIBuffer,\
/*  77 */       subString (unsigned int) const,\
/*  78 */   /*  isDBCSLead(char), OS/2 speecific NOT Supported on MVS */ )

#else

#pragma SOMReleaseOrder( \
/*   1 */       newBuffer(const void*,unsigned int,const void*,  \
                          unsigned int,const void*,unsigned int,char) const,\
/*   2 */       checkAddition(unsigned int,unsigned int),\
/*   3 */       checkMultiplication(unsigned int,unsigned int),\
/*   4 */       overflow(),\
/*   5 */       setDefaultBuffer(IBuffer*),\
/*   6 */       defaultBuffer(),\
/*   7 */       fromContents(const char*),\
/*   8 */       null() const,\
/*   9 */       useCount() const,\
/*  10 */       length() const,\
/*  11 */       contents() const,\
/*  12 */       contents(),\
/*  13 */       charType(unsigned int) const,\
/*  14 */       next(const char*),\
/*  15 */       next(const char*) const,\
/*  16 */       addRef(),\
/*  17 */       removeRef(),\
/*  18 */       compare(const void*,unsigned int) const,\
/*  19 */       subString(unsigned int,unsigned int,char) const,\
/*  20 */       isAlphanumeric() const,\
/*  21 */       isAlphabetic() const,\
/*  22 */       isASCII() const,\
/*  23 */       isControl() const,\
/*  24 */       isDigits() const,\
/*  25 */       isGraphics() const,\
/*  26 */       isHexDigits() const,\
/*  27 */       isLowerCase() const,\
/*  28 */       isPrintable() const,\
/*  29 */       isPunctuation() const,\
/*  30 */       isUpperCase() const,\
/*  31 */       isWhiteSpace() const,\
/*  32 */       isMBCS() const,\
/*  33 */       isDBCS() const,\
/*  34 */       isSBCS() const,\
/*  35 */       isValidMBCS() const,\
/*  36 */       isValidDBCS() const,\
/*  37 */       includesMBCS() const,\
/*  38 */       includesDBCS() const,\
/*  39 */       includesSBCS() const,\
/*  40 */       indexOf(const char*,unsigned int,unsigned int) const,\
/*  41 */       indexOf(const IStringTest&,unsigned int) const,\
/*  42 */       indexOfAnyBut(const char*,unsigned int,unsigned int) const,\
/*  43 */       indexOfAnyBut(const IStringTest&,unsigned int) const,\
/*  44 */       indexOfAnyOf(const char*,unsigned int,unsigned int) const,\
/*  45 */       indexOfAnyOf(const IStringTest&,unsigned int) const,\
/*  46 */       lastIndexOf(const char*,unsigned int,unsigned int) const,\
/*  47 */       lastIndexOf(const IStringTest&,unsigned int) const,\
/*  48 */       lastIndexOfAnyBut(const char*, \
                                  unsigned int,unsigned int) const,\
/*  49 */       lastIndexOfAnyBut(const IStringTest&,unsigned int) const,\
/*  50 */       lastIndexOfAnyOf(const char*,   \
                                 unsigned int,unsigned int) const,\
/*  51 */       lastIndexOfAnyOf(const IStringTest&,unsigned int) const,\
/*  52 */       center(unsigned int,char),\
/*  53 */       change(const char*,unsigned int,const char*,  \
                       unsigned int,unsigned int,unsigned int),\
/*  54 */       copy(unsigned int),\
/*  55 */       insert(const char*,unsigned int,unsigned int,char),\
/*  56 */       leftJustify(unsigned int,char),\
/*  57 */       lowerCase(),\
/*  58 */       overlayWith(const char*,unsigned int,unsigned int,char),\
/*  59 */       remove(unsigned int,unsigned int),\
/*  60 */       reverse(),\
/*  61 */       rightJustify(unsigned int,char),\
/*  62 */       strip(const char*,unsigned int,IStringEnum::StripMode),\
/*  63 */       strip(const IStringTest&,IStringEnum::StripMode),\
/*  64 */       translate(const char*,unsigned int,   \
                          const char*,unsigned int,char),\
/*  65 */       upperCase(),\
/*  66 */       startSearch(unsigned int,unsigned int) const,\
/*  67 */       startBackwardsSearch(unsigned int,unsigned int) const,\
/*  68 */       allocate(unsigned int) const,\
/*  69 */       operator new(SOMClass*,size_t,unsigned int),\
/*  70 */       operator delete(void*),\
/*  71 */       dbcsTable,      \
/*  72 */       initialize(),\
/*  73 */       className() const,\
/*  74 */       IBuffer(unsigned int),\
/*  75 */       nullBuffer,\
/*  76 */       nullIBuffer,\
/*  77 */       subString (unsigned int) const,\
/*  78 */       isDBCSLead(char))


#endif   // for release order pragma


#endif    // SOM_ENABLED & RRBC_LIB


};


#if defined (__SOM_ENABLED__)
#pragma SOMAsDefault (pop)
#endif

#pragma pack()

#ifndef I_NO_INLINES
  #include <ibuffer.inl>
  #include <ibuffern.inl>
#endif

#endif


#ifdef __MVS__
                   #pragma checkout(resume)
#endif

