/**********************************************************************
*                                                                     *
*  IBM(R) VisualAge(TM) for C++ for Windows(R), Version 3.5           *
*                                                                     *
*  PID: 5622-880                                                      *
*  - Licensed Material - Program-Property of IBM                      *
*  (C) Copyright IBM Corp. 1991, 1995 - All Right Reserved.           *
*                                                                     *
*  US Government Users Restricted Rights - Use, duplication or        *
*  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.  *
*                                                                     *
*  VisualAge, and IBM are trademarks or registered trademarks of      *
*  International Business Machines Corporation.                       *
*  Windows is a registered trademark of Microsoft Corporation.        *
*                                                                     *
**********************************************************************/

#include <iiasb.h>

#pragma info (nocls, nocnd, nocns, nocnv, noext, nognr, novft)
#pragma pack (4)


#pragma SOMAsDefault (off)

class IPointer {
  IACollectionImpl* ivImpl;
public:
  IPointer (IACollectionImpl* impl)
  : ivImpl (impl) { ivImpl->Initialize (); }
 ~IPointer () { ivImpl->Delete (); }
  IACollectionImpl& operator* () { return *ivImpl; }
};

#pragma SOMAsDefault (pop)


// ---
// IASortedBagImpl
// ---

// public members

IASortedBagImpl::
~IASortedBagImpl ()
{
}

void
IASortedBagImpl::
AddAntivalence (IASortedBagImpl const& collection1,
                IASortedBagImpl const& collection2)
{
}

void
IASortedBagImpl::
AddDifference (IASortedBagImpl const& collection1,
               IASortedBagImpl const& collection2)
{
  if (this == &collection1) {
    if (this != &collection2) {
      Union (*this);
      ICursorImpl& cursor2 = CursorOf (collection2);
      IBoolean hasNext;
      for (hasNext = collection2.SetToFirst (cursor2);
           hasNext;
           hasNext = collection2.SetToNextDifferentElement (cursor2)) {
        void const* element = collection2.ElementAt (cursor2);
        INumber n1 = NumberOfOccurrences (element) / 2,
                n2 = collection2.NumberOfOccurrences (element),
                n  = (n1 < n2 ? n1 : n2);
        while (n--) {
          Remove (element);
        }
      }
    }
  }
  else {
    ICursorImpl& cursor1 = CursorOf (collection1);
    IBoolean hasNext;
    for (hasNext = collection1.SetToFirst (cursor1);
         hasNext;
         hasNext = collection1.SetToNextDifferentElement (cursor1)) {
      void const* element = collection1.ElementAt (cursor1);
      INumber n1 = collection1.NumberOfOccurrences (element),
              n2 = collection2.NumberOfOccurrences (element),
              n  = (n1 < n2 ? 0 : n1 - n2);
      while (n--) {
        Add (element, CursorOf (*this));
      }
    }
  }
}

void
IASortedBagImpl::
AddIntersection (IASortedBagImpl const& collection1,
                 IASortedBagImpl const& collection2)
{
  if (this == &collection1) {
    if (this == &collection2)
      Union (*this);
    else
      AddIntersection (collection2, collection1);
  }
  else {
    ICursorImpl& cursor2 = CursorOf (collection2);
    IBoolean hasNext;
    for (hasNext = collection2.SetToFirst (cursor2);
         hasNext;
         hasNext = collection2.SetToNextDifferentElement (cursor2)) {
      void const* element = collection2.ElementAt (cursor2);
      INumber n1 = collection1.NumberOfOccurrences (element),
              n2 = collection2.NumberOfOccurrences (element),
              n  = (n1 < n2 ? n1 : n2);
      for ( ; n != 0; n--)
        Add (element, CursorOf (*this));
    }
  }
}

void
IASortedBagImpl::
AddUnion (IASortedBagImpl const& collection1,
          IASortedBagImpl const& collection2)
{
  if (this == &collection1) {
    if (this == &collection2) {
      IPointer bag (Clone ());
      AddAllFrom ((IASortedBagImpl&)*bag);
      AddAllFrom ((IASortedBagImpl&)*bag);
    }
    else {
      Union (collection1);
      Union (collection2);
    }
  }
  else {
    Union (collection2);
    Union (collection1);
  }
}

void
IASortedBagImpl::
Antivalence (IASortedBagImpl const& collection)
{
}

long
IASortedBagImpl::
Compare (IASortedBagImpl const& collection,
         ICompFunc compFunc, void* addArg) const
{ long result = 0;

  if (this != &collection) {
    ICursorImpl& thisCursor = CursorOf (*this),
               & thatCursor = CursorOf (collection);
    IBoolean thisHasNext,
             thatHasNext;

    for (thisHasNext = SetToFirst (thisCursor),
         thatHasNext = collection.SetToFirst (thatCursor);
         result == 0 && thisHasNext && thatHasNext;
         thisHasNext = SetToNext (thisCursor),
         thatHasNext = collection.SetToNext (thatCursor)) {
      result = (*compFunc) (ElementAt (thisCursor),
                            collection.ElementAt (thatCursor), addArg);
    }

    if (result == 0) {
      if (thisHasNext && ! thatHasNext)
        result = 1;
      else if (! thisHasNext && thatHasNext)
        result = -1;
    }
  }
  return result;
}

void
IASortedBagImpl::
Difference (IASortedBagImpl const& collection)
{
  if (this != &collection) {
    ICursorImpl& cursor = CursorOf (collection);
    for (collection.SetToFirst (cursor);
         cursor.IsValid ();
         collection.SetToNext (cursor)) {
      Remove (collection.ElementAt (cursor));
    }
  }
  else
    RemoveAll ();
}

void
IASortedBagImpl::
Intersection (IASortedBagImpl const& collection)
{ if (this != &collection) {
    IACollectionImpl const* eq = &collection;
    RemoveAll (IACollectionImpl::Predicate::IsNotContained, (void*)eq);
  }
}

void
IASortedBagImpl::
Union (IASortedBagImpl const& collection)
{ if (this != &collection) {
    AddAllFrom (collection);
  }
  else {
    IPointer bag (Clone ());
    AddAllFrom ((IASortedBagImpl&)*bag);
  }
}

IBoolean
IASortedBagImpl::
operator!= (IASortedBagImpl const& collection) const
{ return ! operator== (collection);
}

// protected members

IASortedBagImpl::
IASortedBagImpl (Ops& ops)
: IACollectionImpl (ops)
{
}

#pragma info (restore)
#pragma pack ()
