/**********************************************************************
*                                                                     *
*  IBM(R) VisualAge(TM) for C++ for Windows(R), Version 3.5           *
*                                                                     *
*  PID: 5622-880                                                      *
*  - Licensed Material - Program-Property of IBM                      *
*  (C) Copyright IBM Corp. 1991, 1995 - All Right Reserved.           *
*                                                                     *
*  US Government Users Restricted Rights - Use, duplication or        *
*  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.  *
*                                                                     *
*  VisualAge, and IBM are trademarks or registered trademarks of      *
*  International Business Machines Corporation.                       *
*  Windows is a registered trademark of Microsoft Corporation.        *
*                                                                     *
**********************************************************************/

#include <iiexc.h>

// --------------------------------
// IAKeySortedSetAsBstTreeINodeImpl
// --------------------------------

// public members

inline
IAKeySortedSetAsBstTreeINodeImpl::
IAKeySortedSetAsBstTreeINodeImpl ()
: ivm (0)
{
}

inline
IAKeySortedSetAsBstTreeINodeImpl::
IAKeySortedSetAsBstTreeINodeImpl
  (IAKeySortedSetAsBstTreeINodeImpl const& inode)
: ivm (inode.ivm)
{
}

// -------------------------------
// IKeySortedSetAsBstTreeINodeImpl
// -------------------------------

// public members

template <class Key, size_t nodeSize>
inline
IKeySortedSetAsBstTreeINodeImpl <Key, nodeSize>::
IKeySortedSetAsBstTreeINodeImpl ()
{ ivIsLeaf = false;
  Inherited::ivPtrs = ivPtrs;
  Inherited::ivKeys = ivKeys;
}

template <class Key, size_t nodeSize>
inline
IKeySortedSetAsBstTreeINodeImpl <Key, nodeSize>::
IKeySortedSetAsBstTreeINodeImpl
  (IKeySortedSetAsBstTreeINodeImpl <Key, nodeSize > const& inode)
: Inherited (inode)
{ ivIsLeaf = false;
  Inherited::ivPtrs = ivPtrs;
  Inherited::ivKeys = ivKeys;
}

template <class Key, size_t nodeSize>
inline
IKeySortedSetAsBstTreeINodeImpl <Key, nodeSize>::
~IKeySortedSetAsBstTreeINodeImpl ()
{
}

// -------------------------------
// IKeySortedSetAsBstTreeLNodeImpl
// -------------------------------

// public members

template <class Element>
inline
IKeySortedSetAsBstTreeLNodeImpl <Element>::
IKeySortedSetAsBstTreeLNodeImpl (Element const& element)
: ivElement (element)
{ ivIsLeaf = true;
}

template <class Element>
inline
IKeySortedSetAsBstTreeLNodeImpl <Element>::
~IKeySortedSetAsBstTreeLNodeImpl ()
{
}

// --------------------------------
// IKeySortedSetAsBstTreeCursorImpl
// --------------------------------

// public members

inline void
IKeySortedSetAsBstTreeCursorImpl::
Invalidate ()
{ ivNode = 0;
}

inline IBoolean
IKeySortedSetAsBstTreeCursorImpl::
IsValid () const
{ return (ivNode != 0);
}

inline IBoolean
IKeySortedSetAsBstTreeCursorImpl::
IsInBetween () const
{ return false;
}

inline void
IKeySortedSetAsBstTreeCursorImpl::
operator= (IKeySortedSetAsBstTreeCursorImpl const& cursor)
{ Inherited::operator= (cursor);
  ivNode = cursor.ivNode;
}

inline IBoolean
IKeySortedSetAsBstTreeCursorImpl::
operator== (ICursorImpl const& cursor) const
{ return (Inherited::operator== (cursor) &&
          ivNode == ((Self const&)cursor).ivNode);
}


// --------------------------
// IKeySortedSetAsBstTreeImpl
// --------------------------

// public members

#if defined (__LIB_BUILD__)

#if ! defined (__2nd_LEVEL__)
#define inline
#endif

inline IBoolean
IKeySortedSetAsBstTreeImpl::
AllElementsDo (IApplFunc applFunc, void* addArg)
{ if (ivRoot != 0)
    return AllElementsDoSubtree (applFunc, addArg, ivRoot);
  else
    return true;
}

inline void*
IKeySortedSetAsBstTreeImpl::
Any () const
{ return First ();
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
CheckReplacement (ICursorImpl const& cursor, void const* element) const
{ return (CompareKeys (Self::ElementAt (cursor),
                       element, kElementElement) == 0);
}

inline char const*
IKeySortedSetAsBstTreeImpl::
ClassName () const
{ return cvClassName;
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
ContainsElementWithKey (void const* key) const
{ return Self::LocateElementWithKey (key, CursorOf (*this));
}

inline void*
IKeySortedSetAsBstTreeImpl::
ElementAt (ICursorImpl const& cursor) const
{ Node const* node = NodeOf (cursor);
  IASSERT (node->ivIsLeaf);
  return ElementOf ((LNode const*)node);
}

inline void*
IKeySortedSetAsBstTreeImpl::
ElementAtPosition (IPosition position) const
{ Self::SetToPosition (position, CursorOf (*this));
  return Self::ElementAt (CursorOf (*this));
}

inline void*
IKeySortedSetAsBstTreeImpl::
First () const
{ Self::SetToFirst (CursorOf (*this));
  return Self::ElementAt (CursorOf (*this));
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsBounded () const
{ return false;
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsEmpty () const
{ return (ivRoot == 0);
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsFirstAt (ICursorImpl const& cursor) const
{ Self::SetToFirst (CursorOf (*this));
  return (NodeOf (cursor) == NodeOf (CursorOf (*this)));
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsFull () const
{ return false;
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsLastAt (ICursorImpl const& cursor) const
{ Self::SetToLast (CursorOf (*this));
  return (NodeOf (cursor) == NodeOf (CursorOf (*this)));
}

inline void*
IKeySortedSetAsBstTreeImpl::
Key (void const* element) const
{ return OpsOf (*this).Key (element);
}

inline void*
IKeySortedSetAsBstTreeImpl::
Last () const
{ Self::SetToLast (CursorOf (*this));
  return Self::ElementAt (CursorOf (*this));
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
LocateOrAddElementWithKey (void const* element, ICursorImpl& cursor)
{ return ! Self::Add (element, cursor);
}

inline INumber
IKeySortedSetAsBstTreeImpl::
NumberOfElements () const
{ return ivNumberOfElements;
}

inline void
IKeySortedSetAsBstTreeImpl::
RemoveAtPosition (IPosition position)
{ Self::SetToPosition (position, CursorOf (*this));
  Self::RemoveAt (CursorOf (*this));
}

inline void
IKeySortedSetAsBstTreeImpl::
RemoveFirst ()
{ Self::SetToFirst (CursorOf (*this));
  Self::RemoveAt (CursorOf (*this));
}

inline void
IKeySortedSetAsBstTreeImpl::
RemoveLast ()
{ Self::SetToLast (CursorOf (*this));
  Self::RemoveAt (CursorOf (*this));
}

inline void
IKeySortedSetAsBstTreeImpl::
ReplaceAt (ICursorImpl const& cursor, void const* element)
{ Node const* node = NodeOf (cursor);
  IASSERT (node->ivIsLeaf);
  Assign (ElementOf ((LNode const*)node), element);
}

#if ! defined (__2nd_LEVEL__)
#undef inline
#endif

// protected members

inline void
IKeySortedSetAsBstTreeImpl::
Assign (void* element1, void const* element2) const
{ OpsOf (*this).Assign (element1, element2);
}

inline long
IKeySortedSetAsBstTreeImpl::
CompareKeys (void const* arg1, void const* arg2, IArgType argType) const
{ return OpsOf (*this).CompareKeys (arg1, arg2, argType);
}

inline long
IKeySortedSetAsBstTreeImpl::
CompareKeys (void const* key, INumber index,
             void const* arg, IArgType argType) const
{ return OpsOf (*this).CompareKeys (key, index, arg, argType);
}

inline void
IKeySortedSetAsBstTreeImpl::
CopyKey (void* to, INumber toIndex,
         void const* from, INumber fromIndex) const
{ OpsOf (*this).CopyKey (to, toIndex, from, fromIndex);
}

inline void*
IKeySortedSetAsBstTreeImpl::
CreateINode (void const* inode) const
{ return OpsOf (*this).CreateINode (inode);
}

inline void*
IKeySortedSetAsBstTreeImpl::
CreateKey () const
{ return OpsOf (*this).CreateKey ();
}

inline void*
IKeySortedSetAsBstTreeImpl::
CreateLNode (void const* element) const
{ return OpsOf (*this).CreateLNode (element);
}

inline void
IKeySortedSetAsBstTreeImpl::
DeleteINode (void* inode) const
{ OpsOf (*this).DeleteINode (inode);
}

inline void
IKeySortedSetAsBstTreeImpl::
DeleteKey (void* key) const
{ OpsOf (*this).DeleteKey (key);
}

inline void
IKeySortedSetAsBstTreeImpl::
DeleteLNode (void* lnode) const
{ OpsOf (*this).DeleteLNode (lnode);
}

inline void*
IKeySortedSetAsBstTreeImpl::
StreamIn (TStream& fromWhere) const
{ return OpsOf (*this).StreamIn (fromWhere);
}

inline void
IKeySortedSetAsBstTreeImpl::
StreamOut (void const* element, TStream& toWhere) const
{ OpsOf (*this).StreamOut (element, toWhere);
}

// private members

inline void*
IKeySortedSetAsBstTreeImpl::
ElementOf (LNode const* lnode) const
{ typedef IKeySortedSetAsBstTreeLNodeImpl <long> TypedLNode;
  return (OpsOf (*this).ivContainsDTSObjects ?
    (void*) ((TypedLNode const*)lnode)->ivElement :
    (void*)&((TypedLNode const*)lnode)->ivElement);
}

inline IBoolean
IKeySortedSetAsBstTreeImpl::
IsInBetween (ICursorImpl const& cursor)
{
  return false;
}

inline IKeySortedSetAsBstTreeImpl::LNode*
IKeySortedSetAsBstTreeImpl::
MaximumLeaf (Node* node)
{ IASSERT (node != 0);

  while (! node->ivIsLeaf) {
    node = ((INode*)node)->ivPtrs [((INode*)node)->ivm - 1];
  }

  return (LNode*)node;
}

inline IKeySortedSetAsBstTreeImpl::LNode*
IKeySortedSetAsBstTreeImpl::
MinimumLeaf (Node* node)
{ IASSERT (node != 0);

  while (! node->ivIsLeaf) {
    node = ((INode*)node)->ivPtrs [0];
  }

  return (LNode*)node;
}

inline IKeySortedSetAsBstTreeImpl::Node*&
IKeySortedSetAsBstTreeImpl::
NextNodeOf (ICursorImpl const& cursor)
{
  return ((Cursor&)cursor).ivNode; // just a dummy
}

inline IKeySortedSetAsBstTreeImpl::Node*&
IKeySortedSetAsBstTreeImpl::
NodeOf (ICursorImpl const& cursor)
{ return ((Cursor&)cursor).ivNode;
}

inline IKeySortedSetAsBstTreeImpl::Ops&
IKeySortedSetAsBstTreeImpl::
OpsOf (IKeySortedSetAsBstTreeImpl const& collection)
{ return (Ops&) Inherited::OpsOf (collection);
}

inline IKeySortedSetAsBstTreeImpl::Node*&
IKeySortedSetAsBstTreeImpl::
PreviousNodeOf (ICursorImpl const& cursor)
{
  return ((Cursor&)cursor).ivNode; // just a dummy
}

#endif

// --------------------------
// IAKeySortedSetAsBstTreeOps
// --------------------------

// public members

inline
IAKeySortedSetAsBstTreeOps::
~IAKeySortedSetAsBstTreeOps ()
{
}

// protected members

inline
IAKeySortedSetAsBstTreeOps::
IAKeySortedSetAsBstTreeOps (IBoolean containsDTSObjects)
: ivContainsDTSObjects (containsDTSObjects)
{
}
