/*******************************************************************************
* FILE NAME: istring6.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the searching functions of        *
*   the class IString.                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <istring.hpp>

#include <ibuffer.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING6_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IString::indexOf                                                             |
| IString::indexOfAnyBut                                                       |
| IString::indexOfAnyOf                                                        |
|                                                                              |
| Each of these functions is implemented by delegating to the string's         |
| IBuffer contents.                                                            |
------------------------------------------------------------------------------*/
// See istring.inl

unsigned IString :: indexOf ( char     aChar,
                              unsigned startPos ) const
  {
  return this->buffer()->indexOf( &aChar,
                                  1,
                                  startPos );
  }

/*------------------------------------------------------------------------------
| IString::lastIndexOf                                                         |
| IString::lastIndexOfAnyBut                                                   |
| IString::lastIndexOfAnyOf                                                    |
|                                                                              |
| Each of these functions is implemented by delegating to the string's         |
| IBuffer contents.                                                            |
------------------------------------------------------------------------------*/
// See istring.inl

unsigned IString :: lastIndexOf ( char     aChar,
                                  unsigned startPos ) const
  {
  return this->buffer()->lastIndexOf( &aChar,
                                      1,
                                      startPos );
  }

/*------------------------------------------------------------------------------
| IString::occurrencesOf                                                       |
|                                                                              |
| The public versions of this function that accept char* and char              |
| arguments are implementd via a common protected version that accepts         |
| char* and length.  It in turn implements the function by repetitive          |
| calls to IBuffer::indexOf, keeping count of the number of instances          |
| of the search string that are located.                                       |
|                                                                              |
| The public function accepting an IStringTest argument implements the         |
| function by repetitive calls to the corresponding IBuffer::indexOf           |
| function, counting the number of characters for which the "test" is          |
| successful.                                                                  |
------------------------------------------------------------------------------*/
unsigned IString :: occurrencesOf ( const char *pSearchString,
                                    unsigned    searchLen,
                                    unsigned    startPos ) const
  {
  unsigned count = 0;
  while ( ( startPos = this->buffer()->indexOf( pSearchString,
                                                searchLen,
                                                startPos ) ) != 0 )
    {
    count++;
    startPos += searchLen;
    }
  return count;
  }

/*------------------------------------------------------------------------------
| IString::occurrencesOf                                                       |
------------------------------------------------------------------------------*/
unsigned IString :: occurrencesOf ( const IStringTest &aTest,
                                    unsigned           startPos ) const
  {
  unsigned count = 0;
  while ( ( startPos = this->buffer()->indexOf( aTest, startPos ) ) != 0 )
    {
    count++;
    startPos++;
    }
  return count;
  }

/*------------------------------------------------------------------------------
| IString::occurrencesOf                                                       |
------------------------------------------------------------------------------*/
unsigned IString :: occurrencesOf ( char     searchChar,
                                    unsigned startPos ) const
  {
  return this->occurrencesOf( &searchChar,
                              1,
                              startPos );
  }
