/* NOSHIP */
#ifndef _IDDEINFO_
#define _IDDEINFO_
/*******************************************************************************
* FILE NAME: iddeinfo.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*     Iddeinfo                                                                 *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #include <iwindefs.h>
}
#include <ibase.hpp>
#include <istring.hpp>
#include <ireslock.hpp>

#pragma pack(4)

#ifdef IC_PM
  struct _DDESTRUCT;
#endif
#ifdef IC_WIN
  enum IDDESTRUCT { IDDEData, IDDEPoke, IDDEAdvise, IDDEAck, IDDECommands };
#endif

class IDDEAckQueue;
class IDDEServerAcknowledgeEvent;
class IDDEAtomsSet ;

class IDDEFormat {
public:
  IDDEFormat ( unsigned long ddeFormat );
 ~IDDEFormat ( );
unsigned long
  ulClFormat;
IDDEFormat
 *pNextCl;
};  // IDDEFormat

class IDDEFormatSet {
public:
   IDDEFormatSet ( );
  ~IDDEFormatSet ( );
void
  add ( unsigned long ddeFormat );
IDDEFormat
 *pFirstCl;
};  // IDDEFormatSet

#ifdef IC_WIN
  class IDDEItemAtom {
  public:
    IDDEItemAtom ( unsigned long ddeItem );
   ~IDDEItemAtom ( );
  unsigned long
    ulClItemAtom;
  IDDEItemAtom
   *pNextCl;
  };  // IDDEItemAtom

  class IDDEItemAtomSet {
  public:
     IDDEItemAtomSet ( );
    ~IDDEItemAtomSet ( );
  void
    add ( unsigned long ddeItem );
  IDDEItemAtom
   *pFirstCl;
  };  // IDDEItemAtomSet

  class IDDEInfo : public IBase {
  typedef IBase
    Inherited;
  public:
  /*-------------------------- DDE Predefined Formats --------------------------*/
  typedef const unsigned short Status;

  static Status const
    acknowledge,
    busy,
    noData,
    acknowledgeRequested,
    responseFromReq,
    notProcessed,
    applicationSpecific,
    releaseData;

  private:
  /*------------------------------ Hidden Members ------------------------------*/
    IDDEInfo ( );

  }; // IDDEInfo
#endif   /* IC_WIN */

/*-------------------------------- Accessors -----------------------------------
| W freeAtom       - Frees the atom from system.                               |
|   freeMemory     - Frees the shared memory from used to send data between    |
|                    processes.                                                |
|   buildDDEStruct - Builds a DDESTRUCT.                                       |
|   atomFromString - Converts a string to its associated atom.  The atom will  |
|                    be added to the atom table if it does not already exist.  |
| W formatFromString - Converts a data format from string to Windows Id form.  |
|   stringFromAtom - Converts an atom to its associated string.                |
| W loWord         - Extract the low order word from a packed DDE lParam.      |
| W hiWord         - Extract the high order word from a packed DDE lParam.     |
------------------------------------------------------------------------------*/
unsigned long
  IDDEInfo__atomFromString ( const char* atomAsString,
                             Boolean*    atomAdded  );
#ifdef IC_WIN
UINT
  IDDEInfo__formatFromString ( const char* formatAsString );
#endif

IString
  IDDEInfo__stringFromAtom ( unsigned long atom );

#ifdef IC_PM
void
  IDDEInfo__freeMemory ( _DDESTRUCT* ddeInformation );

_DDESTRUCT
 *IDDEInfo__buildDDEStruct ( const char*    itemName,
                             const char*    dataFormat,
                             unsigned short status,
                             const void *   transferData,
                             unsigned long  dataLength,
                             Boolean*       atomAdded );

#endif  /* IC_PM */

#ifdef IC_WIN
void
  IDDEInfo__freeAtom ( ATOM aAtom );

void
  IDDEInfo__freeMemory ( HGLOBAL hDdeStruct );

// TTD
UINT
  IDDEInfo__buildDDEAck( unsigned short status ) ;

HGLOBAL
  IDDEInfo__buildDDEStruct ( int            ddeStructType,
                             const char*    dataFormat,
                             unsigned short status,
                             const void    *transferData,
                             unsigned long  dataLength,
                             Boolean       *atomAdded );

UINT
  IDDEInfo__loWord ( UINT msgId, LPARAM lParam );

UINT
  IDDEInfo__hiWord ( UINT msgId, LPARAM lParam );
#endif   /* IC_WIN */

class IDDEServerConversation : public IBase
{
public:
  IDDEServerConversation  ( unsigned long clientHwnd )
       : ulClClientHwnd(clientHwnd), pAckQCl(0)  { }
  ~IDDEServerConversation ( );
unsigned long const&
  conversationId ( ) const
        { return ulClClientHwnd; }
Boolean
  isAckOutstanding ( IString item,
                     IString format ) const;
void
  addEvent ( IDDEServerAcknowledgeEvent* pEvt );
IDDEServerAcknowledgeEvent*
  removeEvent ( );
unsigned long
  eventCount ( ) const;

private:
unsigned long
  ulClClientHwnd;
IDDEAckQueue
 *pAckQCl;
}; // IDDEServerConversation

class IDDEStatics : public IBase {
public:
  ~IDDEStatics();
static IPrivateResource
   &semaphor ( );
static IDDEAtomsSet
   *setOfAtoms() ;
private:
static IPrivateResource
   *pPriResCl;
static IDDEAtomsSet
   *pSetOfAtoms ;
};  // IDDEStatics

inline unsigned long const& key (IDDEServerConversation* const& conversation)
       { return conversation->conversationId(); }

#endif // _IDDEINFO_
