/* NOSHIP */
/*****************************************************************************
 *
 * asconmgr.h
 *
 * (C) Copyright Lotus Development Coporation 1994, 1996. All Rights Reserved.
 *
 * LIST OF CLASSES IN THIS FILE
 *      CAswConnMgr, CAswAutoConnMgr
 *
 * Description:
 *      The connection manager classes in this file are used by classes which
 * provide an outgoing interface, and need a common mechanism for pushing and
 * managing the connection.
 *
 *****************************************************************************/

#ifndef ASCONMGR_H
#define ASCONMGR_H
#include <aspack.h>
#ifdef ASWANDEBUG
#define CMNAME(MEMBER, NAME, CHANNEL)                                      \
   {                                                                       \
      MEMBER.SetName(NAME);                                                \
                         \
   }
#else
#define CMNAME(MEMBER, NAME, CHANNEL)
#endif 
class ASWPUBCLASS CAswConnMgr : public CAswBase
{
public:
   ASWPUBMETHOD CAswConnMgr( PASWCONTROL pControl );
   ASWPUBMETHOD ~CAswConnMgr( void );
   CAswResult ASWPUBMETHOD Setup( void );
   CAswResult ASWPUBMETHOD Shutdown( BOOL fEmergency = FALSE );
   CAswResult ASWPUBMETHOD Connect( PASWBASE pConnectingComponent,
      REFIID riid, LPUNKNOWN punkOutgoing );
   CAswResult ASWPUBMETHOD Disconnect(void);
#ifdef ASWANDEBUG
   virtual LPCSTR GetAswanClassName( void ) const;
   void SetName( LPSTR szName );
#endif
private:
   inline BOOL IsConnected(void)
   {
      return( m_dwConnKey != 0);
   }
   inline PASWREFHOLDER GetConnPointRH(void)
   {
      return &m_rhConnPoint;
   }
   inline LPCONNECTIONPOINT GetConnPoint(void)
   {
      ASWMETHOD("GetConnPoint", (CH_CLASS_ASWBASE, CH_OCX_CASWBASE));
      PCASWREFHOLDER prhConnPoint = GetConnPointRH();
      #ifdef ASWANASSERT
      if (prhConnPoint->IsSet())
      {
         ASWASSERT_RPTROKLEN(prhConnPoint->Get(), sizeof(IConnectionPoint));
         prhConnPoint->CheckIID( IID_IConnectionPoint, TRUE );
      }
      #endif
      ASWEXITX();
      return( (LPCONNECTIONPOINT)prhConnPoint->Get() );
   }
   CAswRefHolder m_rhConnPoint;  
   DWORD m_dwConnKey;            
   #ifdef ASWANDEBUG
   char m_szName[ MAX_ASWCLASSNAME + MAX_ASWCLASSNAME + 1 ];
   #endif 
};
class ASWPUBCLASS CAswAutoConnMgr : public CAswConnMgr
{
public:
   ASWPUBMETHOD CAswAutoConnMgr( PASWCONTROL pControl );
   ASWPUBMETHOD ~CAswAutoConnMgr( void );
   CAswResult ASWPUBMETHOD Setup( PASWBASE pConnectingComponent,
      PASWBASE pOuterClass = NULL);
   CAswResult ASWPUBMETHOD Shutdown( BOOL fEmergency = FALSE );
   CAswResult ASWPUBMETHOD Connect( REFIID riid, LPUNKNOWN punkOutgoing );
   inline PASWBASE ASWPUBMETHOD GetParent( void )
   {
      return(m_pConnectingComponent);
   }
   inline PASWBASE GetConnectingComponent(void)
   {
      return(m_pConnectingComponent);
   }
   CAswListEl m_leShutdownItem;  
private:
   inline BOOL IsSetup(void)
   {
      return( NULL != GetConnectingComponent() );
   }
   PASWBASE m_pOuterClass;       
   PASWBASE m_pConnectingComponent;
};
MAKELISTHEADCLASS(CAswAutoConnMgr);
#include <aspop.h>
#endif 
