/* NOSHIP */
/*****************************************************************************
 *
 * Source Module: asdebtst.h
 *
 * (C) Copyright Lotus Development Coporation 1994, 1996. All Rights Reserved.
 *
 * Purpose:     Testability hook extensions to Aswan Debug Trace Facility.
 *		This code came from asdebug.h, which got too big, so
 *		it lives in this file now.
 *
 *****************************************************************************/

#ifndef ASDEBTST_H
#define ASDEBTST_H
#include <aspack.h>
#ifdef ASWANDEBUG                   
#if !defined (BASED_FILTER)
   #if defined (WIN16) && defined (__MSC__)
      #define BASED_FILTER __based(__segname("ASW_D_F_TEXT"))
   #else
      #define BASED_FILTER
   #endif
#endif
#define NFILTER0(x,p) (x) << (p)
#define NFILTER1(x,sz,p1) {\
 static char BASED_FILTER _sz[] = sz;\
 (x) << _sz << (p1);\
}
#define SFILTER0(p) NFILTER0(Stream,p)
#define SFILTER1(sz, p1) NFILTER1(Stream,sz,p1);
#define FILTER0(p) NFILTER0(this->GetOStream,p)
#define FILTER1(sz, p1) NFILTER1(this->GetOStream,sz,p1)
class ASWPUBCLASS CAswOStream : public CAswMemoryBase
{
public:
   ASWPUBMETHOD CAswOStream( PASWDEBUGOUT pDebugOut = NULL );
   ASWPUBMETHOD ~CAswOStream();
   inline PASWDEBUGOUT SetDebugOut( PASWDEBUGOUT pDebugOut )
   {
      PASWDEBUGOUT pOldDebugOut = m_pDebugOut;
      m_pDebugOut = pDebugOut;
      return ( pOldDebugOut );
   }
   CAswOStream & operator<<( LPSTR lpsz );
   CAswOStream & operator<<( const void FAR * lp );
#ifdef _NEARDATA
   CAswOStream & operator<<( const void NEAR * np );
#endif
   CAswOStream & operator<<( BYTE by );
   CAswOStream & operator<<( WORD w );
   CAswOStream & operator<<( UINT u );
   CAswOStream & operator<<( LONG l );
   CAswOStream & operator<<( DWORD dw );
   CAswOStream & operator<<( float f );
   CAswOStream & operator<<( double d );
   CAswOStream & operator<<( int n );
   void ASWPUBMETHOD HexDump( const char * pszLine,
                 BYTE *       pby, 
                 int          nBytes, 
                 int          nWidth );
   void ASWPUBMETHOD Flush( void );
private:
   CAswOStream( const CAswOStream & dcSrc );
   void operator=( const CAswOStream & dcSrc );
   void OutputString( LPSTR lpsz );
   PASWDEBUGOUT m_pDebugOut;
};
class ASWPUBCLASS CAswDebugFriend : public CAswMemoryBase
{
public:
   ASWPUBMETHOD CAswDebugFriend( LPCSTR       lpszClassName = NULL, 
                    PASWDEBUGOUT pDebugOut = NULL );
   inline void SetClassName( LPCSTR lpszClassName )
   {
      if ( lpszClassName )
         ::lstrcpy( m_szClassName, lpszClassName );
      else
         m_szClassName[ 0 ] = '\0';
   }
   inline PASWDEBUGOUT SetDebugOut( PASWDEBUGOUT pDebugOut )
   {
      return ( m_CAswOStream.SetDebugOut( pDebugOut ) );
   }
   inline CAswOStream & GetOStream( void )
   {
      return ( m_CAswOStream );
   }
   inline const ASWAN_CLASS_ID GetClassID( void ) const
   {
      return( m_szClassName );
   }
   inline BOOL IsDebugFriendOfClass( PASWDEBUGBASE pDebugBase ) const
   {
      return ( pDebugBase ? (0 == ::lstrcmp( pDebugBase->GetClassID(),
                                             GetClassID() ))
                          : FALSE );
   }
   BOOL ASWPUBMETHOD Check( PASWDEBUGBASE pObject, 
      void ASWPTR * pvoidPrivateInfo ) const;
   virtual void Dump( PASWDEBUGBASE              pObject,
                      void ASWPTR * pvoidPrivateInfo ) const PURE;
   inline void Poke( PASWDEBUGBASE              pObject,
                     void ASWPTR * pvoidPrivateInfo,
                     DWORD                      dwPokeID ) const
   {
      if ( pvoidPrivateInfo && pObject )
      {
         if ( pObject->IsClassIDEqual( GetClassID() ) )
         {
            pObject->HasBeenPoked();
            DoPoke( pObject, pvoidPrivateInfo, dwPokeID );
         }
      }
   }
   virtual void SelfOutput( void ) PURE;
private:
   virtual void DoPoke( PASWDEBUGBASE              pObject,
                        void ASWPTR * pvoidPrivateInfo,
                        DWORD                      dwPokeID ) const PURE;
   ASWPUBCLASS CAswDebugFriend( const CAswDebugFriend & refObjectToCopy );
   void operator=(const CAswDebugFriend & refObjectToAssign );
   char m_szClassName[ MAX_ASWCLASSNAME + 1 ];
   CAswOStream m_CAswOStream;
};
typedef CAswDebugFriend ASWPTR * PASWDEBUGFRIEND;
#define DEFINE_DEBUG_FRIEND_CLASS(name) \
class ASWPUBCLASS D_F_CLASSNAME(name) : public CAswDebugFriend\
{\
public:\
 D_F_CLASSNAME(name)(PASWDEBUGOUT pDebugOut,D_F_PRIVATEINFOPTR(name) pPrivateInfo=NULL);\
 D_F_CLASSNAME(name)(LPCSTR lpszName=NULL,PASWDEBUGOUT pDebugOut=NULL,D_F_PRIVATEINFOPTR(name) pPrivateInfo=NULL);\
\
 void Dump( PASWDEBUGBASE pDebugBase,\
            void ASWPTR * pvoidPrivateInfo ) const;\
\
 D_F_PRIVATEINFOPTR(name) GetPrivateInfo( void )\
 {\
   return ( m_pPrivateInfo );\
 }\
 void SetPrivateInfo( D_F_PRIVATEINFOPTR(name) pPrivateInfo )\
 {\
   m_pPrivateInfo = pPrivateInfo;\
 }\
\
 virtual void SelfOutput( void )\
 {\
   GetOStream() << GetPrivateInfo();\
 }\
\
 friend CAswOStream & operator<<( CAswOStream & Stream,\
                                  const D_F_PRIVATEINFOPTR(name) pPrivateInfo );\
\
private:\
 void DoPoke( PASWDEBUGBASE pDebugBase,\
              void ASWPTR * pvoidPrivateInfo,\
              DWORD dwPokeID ) const;\
\
 D_F_PRIVATEINFOPTR(name) m_pPrivateInfo;\
}
#define DEBUG_FRIEND_DUMP(name) void D_F_CLASSNAME(name) :: \
 Dump( PASWDEBUGBASE              pDebugBase,\
       void ASWPTR * pvoidPrivateInfo ) const
#define DEBUG_FRIEND_CTOR(name) D_F_CLASSNAME(name) :: D_F_CLASSNAME(name)(PASWDEBUGOUT pDebugOut,\
D_F_PRIVATEINFOPTR(name) pPrivateInfo)\
  : CAswDebugFriend(#name,pDebugOut)\
 {\
   SetPrivateInfo(pPrivateInfo);\
 }\
 D_F_CLASSNAME(name) :: D_F_CLASSNAME(name)(LPCSTR lpszName,PASWDEBUGOUT pDebugOut,\
 D_F_PRIVATEINFOPTR(name) pPrivateInfo)\
  : CAswDebugFriend(lpszName,pDebugOut)\
 {\
   SetPrivateInfo(pPrivateInfo);\
 }\
#define DEBUG_FRIEND_DOPOKE(name) void D_F_CLASSNAME(name) :: \
 DoPoke( PASWDEBUGBASE              pDebugBase,\
         void ASWPTR * pvoidPrivateInfo,\
         DWORD                      dwPokeID ) const
#define D_F_PINFOLOCAL(name) D_F_PRIVATEINFOPTR(name) pPrivateInfo = (D_F_PRIVATEINFOPTR(name)) pvoidPrivateInfo;
#define D_F_LOCALS(name) D_F_PINFOLOCAL(name);\
                         ##name ASWPTR * pObject = (##name ASWPTR *) pDebugBase;
#define DUMP_LOCALS(name)   D_F_LOCALS(name)
#define DOPOKE_LOCALS(name) D_F_LOCALS(name)
#define DEBUG_FRIEND_FILTER(name) CAswOStream & operator<< \
( CAswOStream & Stream, const D_F_PRIVATEINFOPTR(name) pPrivateInfo )
#define CHECK_D_F() Check( pObject, ((void ASWPTR *) pPrivateInfo) )
#define CHECK_D_F_DUMP() CHECK_D_F()
#define CHECK_D_F_DOPOKE() CHECK_D_F()
#define D_F_POKEID_ALL  ((DWORD) -1)
struct D_F_PRIVATEINFO(CAswDebugBase)
{
   CAswListEl m_leNeedsShutdown;
   BOOL  m_fNeedsShutdown;
   BOOL  m_fSetup;
   PASWCONTROL m_pControl;  
   ULONG m_ulInstance;            
   char m_szClassName[ MAX_ASWCLASSNAME + 1 ];
   BOOL  m_bHasBeenPoked;  
};
DEFINE_DEBUG_FRIEND_CLASS( CAswDebugBase );
#endif 	
#include <aspop.h>
#endif   
