/*******************************************************************************
* FILE NAME: iolestrm.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IOLEStream                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1996             *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "iolestrm.hpp"

/*******************************************************************************
*                                                                              *
*                        IOLEStream                                            *
*                                                                              *
*******************************************************************************/

IOLEStream::IOLEStream(const CAswStream& aswStream)
: IBaseStream(), fAswStream((CAswStream*)&aswStream)
{
}

IOLEStream::IOLEStream() : fAswStream(NULL)
{
}

IOLEStream::~IOLEStream()
{
}

IOLEStream&
IOLEStream::seek (IBaseStream::Position position)
{
    //** Check if the new position is within the file.
    if (position > logicalEndOfStream())
    {
        setLogicalEndOfStream(position);        
    }
    else {
        LARGE_INTEGER li = { position, 0 };
        CAswResult result = fAswStream->Seek( li, STREAM_SEEK_SET, NULL );
        if ( ASWFAILED(result) )
            throw IAswException( result );
    }
    return *this;
}

IBaseStream::Position  
IOLEStream::position () const
{
    ULARGE_INTEGER pos;
    LARGE_INTEGER zero = { 0, 0 };
    CAswResult result = fAswStream->Seek( zero, STREAM_SEEK_SET, &pos );
    if ( ASWFAILED(result) )
        throw IAswException( result );

    return (IBaseStream::Position) pos.LowPart;
}

Boolean
IOLEStream::isWriteable () const
{
    ASWSTATSTG stbuf;
    CAswResult result = fAswStream->Stat( &stbuf );
    if ( ASWFAILED(result) )
        throw IAswException( result );

    return !( stbuf.grfMode & (STGM_WRITE|STGM_READWRITE) );
}

IOLEStream&
IOLEStream::setLogicalEndOfStream (IBaseStream::Position newEndOfStream)
{
    ULARGE_INTEGER newEnd = { newEndOfStream, 0 };
    CAswResult result = fAswStream->SetSize( newEnd );
    if ( ASWFAILED(result) )
        throw IAswException( result );
    
    LARGE_INTEGER li = { newEndOfStream, 0 };
    result = fAswStream->Seek( li, STREAM_SEEK_SET, NULL );
    if ( ASWFAILED(result) )
        throw IAswException( result );

    return *this;
}

IBaseStream::Position  
IOLEStream::logicalEndOfStream () const
{
    ASWSTATSTG stbuf;
    CAswResult result = fAswStream->Stat( &stbuf );
    if ( ASWFAILED(result) )
        throw IAswException( result );
    
    return (IBaseStream::Position) stbuf.cbSize.LowPart;
}

IOLEStream&
IOLEStream::setPhysicalEndOfStream(IBaseStream::Position physicalEndOfStream)
{
    return setLogicalEndOfStream(physicalEndOfStream);
}

IBaseStream::Position  
IOLEStream::physicalEndOfStream () const
{
    return logicalEndOfStream();
}

void 
IOLEStream::handleWriteBufferFull (const void* source, size_t sourceSize)
{
    // Second argument is the size of the elements being written (1 == 1 byte).
    CAswResult result = fAswStream->Write( (LPVOID)source, (ULONG)sourceSize, NULL );
    if ( ASWFAILED(result) )
        throw IAswException( result );
}

void 
IOLEStream::handleReadBufferEmpty (void* destination, size_t destinationSize)
{
    // Second argument is the size of the elements being read (1 == 1 byte).
    CAswResult result = fAswStream->Read( (LPVOID)destination, (ULONG)destinationSize, NULL );
    if ( ASWFAILED(result) )
        throw IAswException( result );
}

