/* NOSHIP */
/*******************************************************************************
* FILE NAME: precomp.hpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Precompiled implementation headers                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define IUSING_OLE

// aswan headers
#pragma info(none)
#include <aswan.h>
#pragma info(restore)

// IBM Open Class headers
#include <assert.h>
#include <inotify.hpp>
#include <iwindow.hpp>
#include <istring.hpp>
#include <iexcbase.hpp>

// Exception class for returning Aswan errors
class IAswException : public IException {
public:
                    IAswException( const CAswResult& r )
                        : IException( "Aswan Error" ), fResult( r ) {}

    CAswResult      aswResult() const { return fResult; }
    void            setAswResult( const CAswResult& r ) { fResult = r; }

private:
    CAswResult      fResult;
};

// IBM <--> Microsoft compatibility
#define ASSERT  assert
#define NULL    0

#define bool    int
#define BOOL    int

// Handy Macros

#define SHUTDOWNANDDELETE( p ) \
{ \
    if( p ) \
    { \
        p->Shutdown(); \
        delete p; \
        p = 0; \
    } \
}

#define IASWASSERTSTATE( ar )\
    if( !ASWSUCCEEDED( ar ) )\
    {\
        char szMsg[1024];\
        sprintf( szMsg, "ASWSUCCEEDED( ar ), SCODE = %X", ar.Scode() );\
        IExcept__throwAccess( szMsg, ar.Scode(), IEXCEPTION_LOCATION() );\
    }

#define IASWSHOWERROR( ar )\
    if( ASWFAILED( ar ) )\
    {\
        ShowAswResult( ar, IEXCEPTION_LOCATION() );\
    }

// function for showing Aswan error codes to the
// cout stream and a dialog box (implemented in istatnery.cpp)
void ShowAswResult( CAswResult& ar, const IExceptionLocation& loc );

// function for throwing Aswan error codes.
// (implemented in istatnery.cpp)
void IExcept__throwAccess ( const char* errorText,
                            unsigned long errorCode,
                            const IExceptionLocation& location );
