/*******************************************************************************
* FILE NAME: ibctlobs.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of functions that have been replaced *
*   and are now obsolete.  They have been placed here to provide compatibility *
*   with prior releases.                                                       *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <iwindefs.h>
#include <islider.hpp>
#include <imle.hpp>
#include <ientryfd.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBCTLOBS_CPP_
  #include <ipagetun.h>
#endif

#if (IC_OBSOLETE <= IC_OBSOLETE_2)

//
//
//  Block of obsolete function for obsolete level 2
//
//

/******************************************************************************/
/* Slider Notification strings                                                */
/******************************************************************************/
#pragma data_seg(ICLStaticConst)
INotificationId const IProgressIndicator::armTickOffsetId  =
                                    "IProgressIndicator::armTickOffset";
INotificationId const IProgressIndicator::armPixelOffsetId =
                                    "IProgressIndicator::armPixelOffset";
#pragma data_seg()



/*------------------------------------------------------------------------------
| IMultiLineEdit::setChangedFlag                                               |
|                                                                              |
| Set a flag to indicate that the MLE contents have changed.                   |
------------------------------------------------------------------------------*/
IMultiLineEdit&   IMultiLineEdit:: setChangedFlag( Boolean changed )
{
   handle().sendEvent(MLM_SETCHANGED, IEventParameter1(changed),
                                      IEventParameter2(0));
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::resetChangedFlag                                             |
|                                                                              |
| Reset the changed flag to indicate no changes have been made.                |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::resetChangedFlag()
{
  setChangedFlag( false );
  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::isChanged                                                    |
|                                                                              |
| Queries whether the MLE text has changed since the last time the changed     |
| flag was cleared.                                                            |
------------------------------------------------------------------------------*/
Boolean  IMultiLineEdit::isChanged() const
{
  return ( hasTextChanged() );
}


/*------------------------------------------------------------------------------
| IEntryField::hasChanged                                                      |
|                                                                              |
| Queries whether the entry field has changed since the last query.            |
------------------------------------------------------------------------------*/
Boolean IEntryField::hasChanged( ) const
{
  IEventResult evt = handle().sendEvent(EM_QUERYCHANGED,
                                        IEventParameter1(0),
                                        IEventParameter2(0));

  if (evt.asUnsignedLong())
    return( true );
  else
    return( false );
}

/*------------------------------------------------------------------------------
| IEntryField::setChangedFlag                                                  |
------------------------------------------------------------------------------*/
IEntryField& IEntryField::setChangedFlag( Boolean changeFlag )
{
#ifdef IC_PM
  if (changeFlag)
    ISETWINDOWTEXT(handle(), (PSZ)text());
  else
    hasChanged();
#endif

#ifdef IC_WIN
  handle().sendEvent(EM_SETMODIFY, IEventParameter1(changeFlag),
                                   IEventParameter2(0));
#endif

  return *this;
}



//
//
//  End of obsolete function for obsolete level 2
//
//
#endif
