/*******************************************************************************
* FILE NAME: ibmphdr.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ibmphdr.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_GPIBITMAPS
  #include <iwindefs.h>
  }

#include <ibmphdr.hpp>

#include <ipainevt.hpp>
#include <iswp.hpp>
#include <itrace.hpp>
#include <icontrol.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <ibmpctl.hpp>
#include <igrafctx.hpp>
#include <igbitmap.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBMPHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IBitmapHandler::IBitmapHandler                                               |
------------------------------------------------------------------------------*/
IBitmapHandler :: IBitmapHandler() {;}

/*------------------------------------------------------------------------------
| IBitmapHandler::~IBitmapHandler                                              |
------------------------------------------------------------------------------*/
IBitmapHandler :: ~IBitmapHandler() {;}

/*------------------------------------------------------------------------------
| IBitmapHandler::paintWindow                                                  |
| This inherited function is overridden:                                       |
|   paintWindow - Process the paint events for a bitmap control window.        |
------------------------------------------------------------------------------*/
Boolean IBitmapHandler :: paintWindow          ( IPaintEvent& evt )
{
  IBitmapControl* pBitmap = (IBitmapControl*)evt.window();
  ISize winsize = pBitmap->size();
  // Get a graphic context
  IGraphicContext gc(evt.presSpaceHandle());

  // draw the bitmap if we have one
  if ( pBitmap->bitmap() )
    {
    // Get a pointer to the current bitmap
    IGBitmap bitmap( pBitmap->bitmap() );
    ISize    bitmapSize = bitmap.size();
    // draw the bitmap
    bitmap.drawOn(gc,
                  IPoint(0,0),
                  IPoint( winsize.width(), winsize.height() ),
                  IPoint(0,0),
                  IPoint( bitmapSize.width(), bitmapSize.height() ),
                  (!pBitmap->isEnabled() || pBitmap->isHalftone()) ?
                     IGBitmap::halftone : IGBitmap::normal );
    }
  else
    {
    // paint the current background color
    evt.clearBackground();
    }

  // done if there is no text
  return ( pBitmap->textLength() ? false : true );
}
