/*******************************************************************************
* FILE NAME: icnrevt.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icnrevt.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINSTDCNR

#include <icnrrec.hpp>  // Must be first for OS flags

#include <icnrobj.hpp>
#include <icnrevt.hpp>
#include <icnrctl.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICNREVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| ICnrEvent::~ICnrEvent                                                        |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrEvent :: ~ICnrEvent      ( ) {}

/*------------------------------------------------------------------------------
| ICnrEnterEvent::~ICnrEnterEvent                                              |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrEnterEvent :: ~ICnrEnterEvent () {}

/*------------------------------------------------------------------------------
| ICnrEmphasisEvent::~ICnrEmphasisEvent                                        |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrEmphasisEvent :: ~ICnrEmphasisEvent  ( ) {}

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::~ICnrQueryDeltaEvent                                    |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrQueryDeltaEvent :: ~ICnrQueryDeltaEvent ( ) {}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::~ICnrScrollEvent                                            |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrScrollEvent :: ~ICnrScrollEvent ( ) {}

/*------------------------------------------------------------------------------
| ICnrHelpEvent::~ICnrHelpEvent                                                |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
ICnrHelpEvent :: ~ICnrHelpEvent ( ) {}

/*------------------------------------------------------------------------------
| ICnrEvent::ICnrEvent                                                         |
------------------------------------------------------------------------------*/
ICnrEvent :: ICnrEvent( const IControlEvent& evt)
        : IControlEvent(evt) {}

/*------------------------------------------------------------------------------
| ICnrEnterEvent::ICnrEnterEvent                                               |
------------------------------------------------------------------------------*/
ICnrEnterEvent ::  ICnrEnterEvent( const IControlEvent& evt)
        : ICnrEvent(evt)  {}

/*------------------------------------------------------------------------------
| ICnrEmphasisEvent::ICnrEmphasisEvent                                         |
------------------------------------------------------------------------------*/
ICnrEmphasisEvent :: ICnrEmphasisEvent( const IControlEvent& evt)
        : ICnrEvent(evt)  {}

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::ICnrQueryDeltaEvent                                     |
------------------------------------------------------------------------------*/
ICnrQueryDeltaEvent :: ICnrQueryDeltaEvent( const IControlEvent& evt)
        : ICnrEvent(evt) {}

/*------------------------------------------------------------------------------
| ICnrHelpEvent::ICnrHelpEvent                                                 |
------------------------------------------------------------------------------*/
ICnrHelpEvent :: ICnrHelpEvent( const IControlEvent& chpevt)
        : ICnrEvent(chpevt) {}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::ICnrScrollEvent                                             |
------------------------------------------------------------------------------*/
ICnrScrollEvent :: ICnrScrollEvent( const IControlEvent& chpevt)
        : ICnrEvent(chpevt) {}

/*------------------------------------------------------------------------------
| ICnrEvent::containerId                                                       |
|                                                                              |
| Returns the control identifier for the container.                            |
------------------------------------------------------------------------------*/
unsigned long ICnrEvent :: containerId() const {
  return  parameter1().number1();
}

/*-----------------------------------------------------------------------------
| ICnrEnterEvent Functions                                                    |
-----------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| ICnrEnterEvent::enterPressed                                                 |
|                                                                              |
| Determine if the enter key was pressed (as opposed to a mouse double-click). |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrEnterEvent :: enterPressed() const
{
  return ((PNOTIFYRECORDENTER)(void*)parameter2())->fKey;
}

/*------------------------------------------------------------------------------
| ICnrEnterEvent::validObject                                                  |
|                                                                              |
| Return true if the event occurred over an object.                            |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrEnterEvent :: validObject() const
{
  return (((PNOTIFYRECORDENTER)(void*)parameter2())->pRecord !=0);
}

/*------------------------------------------------------------------------------
| ICnrEnterEvent::object                                                       |
|                                                                              |
| Return the object over which the Enter occurred.                             |
------------------------------------------------------------------------------*/
IContainerObject* ICnrEnterEvent :: object()  const
{
  IMiniCnrRecord* prec = (IMiniCnrRecord*)((PNOTIFYRECORDENTER)
                                                 (void*)parameter2())->pRecord;
  return IObjFromRec(prec);
}

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent Functions                                                |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::atTopDelta                                              |
|                                                                              |
| Return true if the object that represents the delta value scrolls into view  |
| at the top of the client area.                                               |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrQueryDeltaEvent :: atTopDelta() const {
  return ((((PNOTIFYDELTA)(void*)parameter2())->fDelta &
                                           CMA_DELTATOP)) ? true : false; };

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::atBottomDelta                                           |
|                                                                              |
| Returns true if the object that represents the delta value  scrolls into     |
| view at the bottom of the client area.                                       |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrQueryDeltaEvent :: atBottomDelta() const
{
  return ((((PNOTIFYDELTA)(void*)parameter2())->fDelta &
                                           CMA_DELTABOT)? true : false);
}

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::atFirstObject                                           |
|                                                                              |
| Returns true if the container has scrolled to the first object in the        |
| container.                                                                   |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrQueryDeltaEvent :: atFirstObject() const
{
  return ((((PNOTIFYDELTA)(void*)parameter2())->fDelta &
                                           CMA_DELTAHOME)? true : false);
}

/*------------------------------------------------------------------------------
| ICnrQueryDeltaEvent::atLastObject                                            |
|                                                                              |
| Returns true if the container has scrolled to the last object in the         |
| container.                                                                   |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrQueryDeltaEvent :: atLastObject() const
{
  return ((((PNOTIFYDELTA)(void*)parameter2())->fDelta & CMA_DELTAEND)? true : false);
}


/*------------------------------------------------------------------------------
| ICnrEmphasisEvent Functions                                                  |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| ICnrEmphasisEvent::object                                                    |
|                                                                              |
| Return the object which has changed emphasis.                                |
------------------------------------------------------------------------------*/
IContainerObject* ICnrEmphasisEvent :: object()  const
{
  IMiniCnrRecord* prec = (IMiniCnrRecord*)((PNOTIFYRECORDEMPHASIS)
                                                  (void*)parameter2())->pRecord;
  return IObjFromRec(prec);
}

/*------------------------------------------------------------------------------
| ICnrEmphasisEvent::changed                                                   |
|                                                                              |
| Determine it the passed emphasis has changed.                                |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrEmphasisEvent :: changed(IContainerObject::Emphasis emphasis) const
{
  unsigned long ulEmphasisChanged = ((PNOTIFYRECORDEMPHASIS)
                                           (void*)parameter2())->fEmphasisMask;

  switch(emphasis)
  {
     case IContainerObject::inuse    :
        if(ulEmphasisChanged & CRA_INUSE)
           return true;
        break;

     case IContainerObject::selected :
        if(ulEmphasisChanged & CRA_SELECTED)
           return true;
        break;

     case IContainerObject::cursored :
        if(ulEmphasisChanged & CRA_CURSORED)
           return true;
        break;
  }
  return false;
};

/*------------------------------------------------------------------------------
| ICnrScrollEvent Functions                                                    |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| ICnrScrollEvent::amount                                                      |
|                                                                              |
| Returns the amount the window has scrolled (in pixels).                      |
------------------------------------------------------------------------------*/
long ICnrScrollEvent :: amount() const
{
  return ((PNOTIFYSCROLL)(void*)parameter2())->lScrollInc;
}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::isLeftDetails                                               |
|                                                                              |
| Return true if the scrolling occurred in the left (or only) side of the      |
| details view.                                                                |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrScrollEvent :: isLeftDetails() const
{
  return ((((PNOTIFYSCROLL)(void*)parameter2())->fScroll &
                                           CMA_LEFT)? true : false);
}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::isRightDetails                                              |
|                                                                              |
| Returns true if the scrolling occurred in the right side of a split details  |
| view.                                                                        |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrScrollEvent :: isRightDetails() const
{
  return ((((PNOTIFYSCROLL)(void*)parameter2())->fScroll &
                                           CMA_RIGHT)? true : false);
}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::isHorizontal                                                |
|                                                                              |
| Return true if scrolling horizontally.                                       |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrScrollEvent :: isHorizontal() const
{
  return ((((PNOTIFYSCROLL)(void*)parameter2())->fScroll &
                                           CMA_HORIZONTAL)? true : false);
}

/*------------------------------------------------------------------------------
| ICnrScrollEvent::isVertical                                                  |
|                                                                              |
| Return true if scrolling vertically.                                         |
------------------------------------------------------------------------------*/
IBase::Boolean ICnrScrollEvent :: isVertical() const
{
  return ((((PNOTIFYSCROLL)(void*)parameter2())->fScroll &
                                           CMA_VERTICAL) ? true : false);
}

/*------------------------------------------------------------------------------
| ICnrHelpEvent Functions                                                      |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| ICnrHelpEvent::object                                                        |
|                                                                              |
| Return the object over which help is requested.                              |
------------------------------------------------------------------------------*/
IContainerObject* ICnrHelpEvent :: object()  const
{
  /******************************************************************/
  /* The "object" that is passed in a CN_HELP event is the cursored */
  /* object.  If we have an mle open, we want to return the mle     */
  /* object so long as we are editing an object, otherwise return   */
  /* the cursored object.                                           */
  /******************************************************************/
  IContainerControl* pcnrctl = (IContainerControl*)controlWindow();
  IContainerObject* pcnrobj = 0;

  if(pcnrctl)
  {
     pcnrobj = pcnrctl->currentEditObject();
     // if there was no cursored object, then -1 was passed as parm2.
     IMiniCnrRecord *recordFromEvent = 0;
     if ( (void*)(this->parameter2()) != (void*)(-1) )
       recordFromEvent = (IMiniCnrRecord*)(void*)(this->parameter2());

     if (!pcnrobj && !pcnrctl->currentEditMLE() && recordFromEvent)
       pcnrobj = IObjFromRec(recordFromEvent);

  }
  return pcnrobj;
}

/*------------------------------------------------------------------------------
| ICnrHelpEvent::column                                                        |
|                                                                              |
| Returns the column that was being edited if help was requested within a      |
| multiple-line entry field; otherwise zero is returned.                       |
------------------------------------------------------------------------------*/
IContainerColumn* ICnrHelpEvent :: column()  const
{
  IContainerControl* pcnrctl = (IContainerControl*)controlWindow();
  if(pcnrctl)
    return pcnrctl->currentEditColumn();
  else
    return 0;
}
