/* NOSHIP */
#ifndef _IDMDNFO_
#define _IDMDNFO_
/*******************************************************************************
* FILE NAME: idmdnfo.hpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*     IDragInfo                                                                *
*     IDragInfoWrapper                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <ihandle.hpp>
#include <ipoint.hpp>
#include <idmditm.hpp>
#include <imarsha.hpp>
#include <ievtdata.hpp>
#include <idmcomm.hpp>

#pragma pack(4)

class IDMOperation;

#if defined(IC_MOTIF) || defined(IC_WIN)

class IDragInfo : public IDMMarshall {
typedef IVBase
  Inherited;
public:
/*------------------------------- Constructors -------------------------------*/
  IDragInfo ( unsigned short numDItems );

virtual
 ~IDragInfo ( );

virtual unsigned long
  marshall              ( unsigned long startPos=0, char* outStr=0 ) const;

virtual const char
 *unmarshall            ( const char* inStr );

IDM::DropIndicator
  dropIndicator         ( ) const;

unsigned short
  dropOperation         ( ) const;

IDragInfo
 &setDragResult         ( const IEventResult& result );

private:
/*--------------------------------- Private ----------------------------------*/
friend class IDragInfoWrapper;
// Dissallow copy and assignment.
  IDragInfo( const IDragInfo& );
IDragInfo
 &operator = ( const IDragInfo& );

IPoint
  pointDrop;

IWindowHandle
  hwndSource;

IProcessId
  pidSource;    // PID of source window.

unsigned short
  usDrop,
  usDefaultOp,
  usOperation,
  cditem;       // Number of IDragItems in array dragItems.

IDragItem
 *pDragItems;    // Pointer to array of IDragItem.

}; // IDragInfo

  // This allows us to keep the IC_PM interfaces that use these.
  typedef IDragInfo _DRAGINFO;
//  typedef IDragInfo DRAGINFO;
#endif
#ifdef IC_PM
  struct _DRAGINFO;
#endif

class IDragInfoWrapper {
public:
/*------------------------------- Constructors -------------------------------*/
  IDragInfoWrapper ( _DRAGINFO& aDragInfo );
  IDragInfoWrapper ( const IDMOperation& anIDMOperation );

  IDragInfoWrapper ( const IDragInfoWrapper& diw );

virtual
 ~IDragInfoWrapper ( );

_DRAGINFO
 *asDragInfoPtr         ( ) const;

_DRAGITEM
 &operator[]            ( unsigned long ulItem ) const;

IPoint
  dropPoint             ( ) const;

IWindowHandle
  sourceWindowHandle    ( ) const;

IProcessId
 sourcePID              ( ) const;

unsigned short
  dragOperation         ( ) const,
  numDragItems          ( ) const;

IDragInfoWrapper
 &setDropPoint          ( const IPoint& dropPoint ),
 &setSourceWindowHandle ( const IWindowHandle& window ),
 &setDragOperation      ( unsigned short dragOp ),
 &deleteTargetResources ( ),
 &deleteSourceResources ( ),
 &access                ( ),
 &release               ( );

static IDragInfoWrapper*
 setDragInfoWrapper( IDMSourceOperation *pSrcOp, _DRAGINFO *pDInfo );

static IDragInfoWrapper*
 dragInfoWrapper( IDataObject *pDataObject );

private:
/*--------------------------------- Private ----------------------------------*/
// Dissallow assignment.
IDragInfoWrapper
 &operator = ( const IDragInfoWrapper& item );

_DRAGINFO
 &theDragInfo;
}; // IDragInfoWrapper

#pragma pack()

#ifndef I_NO_INLINES
  #include <idmdnfo.inl>
#endif

#endif // _IDMDNFO_
