/* NOSHIP */
/*******************************************************************************
* FILE NAME: idmdragh.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMDragHandler                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define IUSING_OLE
  #ifdef WIN32_LEAN_AND_MEAN
    #undef WIN32_LEAN_AND_MEAN
  #endif
  #ifdef _OLE2_H_
    #undef _OLE2_H_
  #endif
  #include <iwindefs.h>
  }

#ifndef _IDMDRAGH_
  #include <idmdragh.hpp>
#endif

#ifndef _IDMTGTRN_
  #include <idmtgtrn.hpp>
#endif

// Trace Facility
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#ifndef _IWINDOW_
  #include <iwindow.hpp>
#endif

#ifndef _IDMOIPRV_
  #include <idmoiprv.hpp>
#endif

#ifndef _ITHREAD_
  #include <ithread.hpp>
#endif
#include <imcevt.hpp>
// Segment definitions

#ifdef IC_PAGETUNE
  #define _IDMDRAGH_CPP_
  #include <ipagetun.h>
#endif

IDMThreadIdElement :: IDMThreadIdElement( unsigned long threadId,
                                      int refCnt )
  : threadIdKey( threadId )
  , refCount( refCnt )
  {;}

IDMThreadIdElement::~IDMThreadIdElement( )
  {;}


  IDMThreadIdSet::IDMThreadIdSet()
  {
  } // IDMThreadIdSet

  IDMThreadIdSet::~IDMThreadIdSet()
  {
  } // ~IDMThreadIdSet

class IDMDragHandlerData
  {
  public:
    IDMDragHandlerData()
    {;}
    virtual
      ~IDMDragHandlerData()
    {;}
    static IDMThreadIdSet
     *defaultThreadIdSet();
     IPoint
       startingPoint;
  private:
    static IDMThreadIdSet
     *threadIdSet;
  };

IDMThreadIdSet* IDMDragHandlerData::threadIdSet = 0;

IDMThreadIdSet* IDMDragHandlerData::defaultThreadIdSet()
{
  if (threadIdSet == 0)
  {
    threadIdSet = new IDMThreadIdSet();
  }
  return threadIdSet;
}

IDMDragHandler :: IDMDragHandler ()
  : IHandler()
  , fDMDragHandlerData( new IDMDragHandlerData() )

  {
  }


IDMDragHandler :: ~IDMDragHandler ()

  {
  if (fDMDragHandlerData)
    {
    delete fDMDragHandlerData;
    }
  }

/*------------------------------------------------------------------------------
| IDMDragHandler::dispatchHandlerEvent                                         |
|                                                                              |
| Overload of Handler's dispatcher.                                            |
------------------------------------------------------------------------------*/
IBase::Boolean IDMDragHandler :: dispatchHandlerEvent(IEvent& event)

  {
  Boolean bProcessed = false;

  switch(event.eventId())
    {
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
      {
      // Save the starting point so that it can be used to start drag.
      IMouseClickEvent mevt( event );
      fDMDragHandlerData->startingPoint = mevt.mousePosition();
      }
    case WM_LBUTTONUP:
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MOUSEMOVE:
      {
      if ( event.window()->isDragStarting( event ) )
        {
        // Use the original point that started the drag, not the point
        // associated with the move.
        event.window()->sendEvent( WM_BEGINDRAG,
                    IEventParameter1( fDMDragHandlerData->startingPoint.x(),
                                      fDMDragHandlerData->startingPoint.y()),
                    IEventParameter2( 0 ));
        }
      if ( event.result().asUnsignedLong() )
        bProcessed = true;
      break;
      }
    case WM_DESTROY:
      {
      IDMDragHandler::uninitOle();
      break;
      }
    }
  return bProcessed;
  }


/*------------------------------------------------------------------------------
| IDMDragHandler::initOle                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
void  IDMDragHandler :: initOle ()
{
  IMODTRACE_DEVELOP("IDMDragHandler::initOle");

  IDMThreadIdSet *threadIdSet = IDMDragHandlerData::defaultThreadIdSet();
  // look for threadId to see if it is already in list.

  unsigned long currentId = IThread::current().id().asUnsigned();
  if ( threadIdSet->containsElementWithKey( currentId ) )
  {
    IDMThreadIdElement* threadElement =
        threadIdSet->elementWithKey( currentId );

    // increment the ref count
    threadElement->refCount++;
    ITRACE_DEVELOP("ref count is " + IString(threadElement->refCount) );
  }
  else
  {
    ITRACE_DEVELOP("OleInitialize on thread " +
        IString( currentId ));
    ::OleInitialize( NULL );
    threadIdSet->add(
      new IDMThreadIdElement( currentId, 1));
  }
}

/*------------------------------------------------------------------------------
| IDMDragHandler::uninitOle                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
void  IDMDragHandler :: uninitOle ()
{
  IMODTRACE_DEVELOP("IDMDragHandler::uninitOle");

  unsigned long currentId = IThread::current().id().asUnsigned();
  IDMThreadIdSet *threadIdSet = IDMDragHandlerData::defaultThreadIdSet();
  // look for threadId to see if it is already in list.
  if ( threadIdSet->containsElementWithKey( currentId ) )
  {
    IDMThreadIdElement* threadElement =
        threadIdSet->elementWithKey( currentId );
    threadElement->refCount--;
    ITRACE_DEVELOP("ref count is " + IString(threadElement->refCount) );
    if ( threadElement->refCount == 0 )
    {
      ITRACE_DEVELOP("OleUninitialize on thread " +
          IString( currentId ));
      ::OleUninitialize( );
      threadIdSet->removeElementWithKey( currentId );
    }
  }
  else
    ITRACE_DEVELOP("Not supposed to happen");
}
