/*******************************************************************************
* FILE NAME: ifilehdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ifilehdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
  #define INCL_WINMESSAGEMGR
  #define INCL_WINSTDFILE
  #define INCL_WINWINDOWMGR
  #define IUSING_OLE
  #ifdef WIN32_LEAN_AND_MEAN
    #undef WIN32_LEAN_AND_MEAN
  #endif
  #include <iwindefs.h>
}

#include <iexcept.hpp>
#include <ifiledlg.hpp>
#include <ifilehdr.hpp>
#ifdef IC_PM
#include <ipmexsta.hpp>
#endif
#include <istring.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IFILEHDR_CPP_
  #include <ipagetun.h>
#endif

//
// Implementation for IFileDialogHandler.
//

/*------------------------------------------------------------------------------
| IFileDialogHandler::IFileDialogHandler                                       |
------------------------------------------------------------------------------*/
IFileDialogHandler::IFileDialogHandler ( )
  : IFileDialogHandler::Inherited()
{
#ifdef IC_WIN
  fRegisteredOkMsg = RegisterWindowMessage( FILEOKSTRING );
#endif
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::~IFileDialogHandler                                      |
------------------------------------------------------------------------------*/
IFileDialogHandler::~IFileDialogHandler ( )
{ }

/*------------------------------------------------------------------------------
| IFileDialogHandler::handleEventsFor                                          |
------------------------------------------------------------------------------*/
IFileDialogHandler& IFileDialogHandler::handleEventsFor ( IFileDialog* fileDialog )
{
  IASSERTPARM( fileDialog != 0 );
  Inherited::handleEventsFor( fileDialog );
  return *this;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::stopHandlingEventsFor                                    |
------------------------------------------------------------------------------*/
IFileDialogHandler& IFileDialogHandler::stopHandlingEventsFor
                                                   ( IFileDialog* fileDialog )
{
  IASSERTPARM( fileDialog != 0 );
  Inherited::stopHandlingEventsFor( fileDialog );
  return *this;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::handleEventsFor                                          |
------------------------------------------------------------------------------*/
IHandler& IFileDialogHandler::handleEventsFor( IWindow *window )
{
  Inherited::handleEventsFor( window );
  return *this;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::stopHandlingEventsFor                                    |
------------------------------------------------------------------------------*/
IHandler& IFileDialogHandler::stopHandlingEventsFor( IWindow *window )
{
  Inherited::stopHandlingEventsFor( window );
  return *this;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::dispatchHandlerEvent                                     |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::dispatchHandlerEvent ( IEvent& event )
{
  Boolean retcond = false;

  IFileDialogEvent fileEvt( event );
  fileEvt.setResult( true );
  unsigned long idOfEvent = event.eventId();
  switch ( idOfEvent )
  {
#ifdef IC_PM
    case FDM_FILTER:
      retcond = this->filter( fileEvt );
      break;
    case FDM_VALIDATE:
      retcond = this->validate( fileEvt );
      break;
#endif
#ifdef IC_WIN
    case IC_UM_DLG_APPLY_PRESSED:
#endif
    case WM_DESTROY:
    {
#ifdef IC_PM
      /*****************************************************************/
      /* Get the "this" pointer for window object from the window word */
      /* which is set during dialog initialization                     */
      /*****************************************************************/
      FILEDLG* pfildData = (FILEDLG*)WinQueryWindowPtr(event.handle(),QWL_USER);
      IFileDialog* pfildlg = (IFileDialog*)pfildData->ulUser;
#else
      IFileDialog* pfildlg = (IFileDialog*)
                            IQUERYWINDOWULONG( event.handle(), DWL_USER );
#endif
      /*****************************************************************/
      /* Verify we have a non-null object to reference before using    */
      /*****************************************************************/
      if (pfildlg != (IFileDialog*)0)
      {
        /***************************************************************/
        /* If modeless, call handler function, dependent on event type */
        /***************************************************************/
        if (pfildlg->isModeless())
        {
          /*************************************************************/
          /* Even if dialog cancelled, still call results on destroy   */
          /*************************************************************/
          if (idOfEvent == WM_DESTROY)
            modelessResults( pfildlg );

          /*************************************************************/
          /* otherwise this is apply button, so notify of file to use  */
          /*************************************************************/
          else if (idOfEvent == IC_UM_DLG_APPLY_PRESSED)
            modelessApply( pfildlg, (const char*)event.parameter1() );
        }
      }
    }
      break;

    default:
      break;
  } /* endswitch */

#ifdef IC_WIN
  if ( idOfEvent == fRegisteredOkMsg )
  {
    IFileDialog* pfildlg = (IFileDialog*)
                           IQUERYWINDOWULONG( event.handle(), DWL_USER );
    // Get the filename and add it to the private data of the event
    //  so that queries to the events fileName() function can be resolved
    fileEvt.fFileNameSelected = pfildlg->pFileDlgData->lpstrFile;
    retcond = this->validate( fileEvt );
  }
   // Result is defaulted to true, therefore only on a validate message
   // will it change.  Therefore default result to false and return false
   // on Windows.  If validate returns false, then continue dialog so set
   // result to true and return true.
   event.setResult( !(Boolean)(fileEvt.result()) );
   if ( (Boolean)event.result() )
#endif
#ifdef IC_PM
   // filter() and validate() will call setResult on fileEvt to indicate if
   // the file is OK to process.  We have to pass this on to the original event.
   event.setResult( fileEvt.result() );

   if ( !(Boolean)fileEvt.result() )
#endif
   {
      // If validate() or filter() setResult to false, the file should not be
      // processed, and therefore no further handlers should be called.
      return true;
   }
   else
   {
      return retcond;
   }
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::filter                                                   |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::filter ( IFileDialogEvent& event )
{
  event.setResult( this->filterName( event.fileName(),
                                     (IFileDialog*)event.window() ));
  return false;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::filterName                                               |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::filterName ( const IString& fileName,
                                         IFileDialog* dialog )
{
  return true;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::validate                                                 |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::validate ( IFileDialogEvent& event )
{
  event.setResult( this->validateName( event.fileName(),
                                       (IFileDialog*)event.window() ));
  return false;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::validateName                                             |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::validateName ( const IString& fileName,
                                           IFileDialog* dialog )
{
  return true;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::modelessResults                                          |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::modelessResults ( IFileDialog* endingDialog )
{
  return true;
}

/*------------------------------------------------------------------------------
| IFileDialogHandler::modelessApply                                            |
------------------------------------------------------------------------------*/
Boolean IFileDialogHandler::modelessApply ( IFileDialog*   modelessDialog,
                                            const IString& fileName )
{
  return true;
}


//
// Implementation for IFileDialogEvent.
//

/*------------------------------------------------------------------------------
| IFileDialogEvent::~IFileDialogEvent                                          |
------------------------------------------------------------------------------*/
IFileDialogEvent::~IFileDialogEvent ( )
{ }

/*------------------------------------------------------------------------------
| IFileDialogEvent::IFileDialogEvent                                           |
------------------------------------------------------------------------------*/
IFileDialogEvent::IFileDialogEvent ( IEvent& event )
  : IEvent( event )
{ }

/*------------------------------------------------------------------------------
| IFileDialogEvent::fileName                                                   |
------------------------------------------------------------------------------*/
IString IFileDialogEvent::fileName ( ) const
{
#ifdef IC_PM
  char* temp = this->parameter1();
  return temp;
#endif
#ifdef IC_WIN
  return this->fFileNameSelected;
#endif
}

/*------------------------------------------------------------------------------
| IFileDialogEvent::fileLength                                                 |
------------------------------------------------------------------------------*/
unsigned long IFileDialogEvent::fileLength ( ) const
{
#ifdef IC_PM
  char* temp = this->parameter1();
  IString str(temp);
  return str.length();
#endif
#ifdef IC_WIN
  return this->fFileNameSelected.length();
#endif
}
