/*******************************************************************************
* FILE NAME: igraphdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igraphdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
#define INCL_WINBUTTONS
#define INCL_WINENTRYFIELDS
#define INCL_WINLISTBOXES
#define INCL_WINSTDCNR
#define INCL_WINSTDVALSET
#define INCL_WINWINDOWMGR

extern "C" {
  #include <iwindefs.h>
  #include <stdlib.h>
}

#include <igraphdr.hpp>
#include <igraphbt.hpp>
#include <ievent.hpp>
#include <iwindow.hpp>
#include <iiconctl.hpp>
#include <icoordsy.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGRAPHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGraphicPushButtonHandler::IGraphicPushButtonHandler                         |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IGraphicPushButtonHandler::IGraphicPushButtonHandler ( )
{ ; }


/*------------------------------------------------------------------------------
| IGraphicPushButtonHandler::~IGraphicPushButtonHandler                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IGraphicPushButtonHandler::~IGraphicPushButtonHandler ( )
{ ; }


/************************************************************************/
/* dispatchHandler Event - Dispatch command functions for this handler. */
/************************************************************************/
IBase::Boolean IGraphicPushButtonHandler :: dispatchHandlerEvent(IEvent & evt)
{
   IGraphicPushButton* graphicbutton (0);
   Boolean bMovedImage (false);
   switch (evt.eventId())
   {
      case BM_SETHILITE:
         {
           // We keep track of whether the button is up or down in
           // the control's private data.  We need this since there are
           // scenarios where multiple BM_SETSTATE (BM_SETHILITE) messages
           // will be received - such as mouse movement w/ button 1 down.
           // Without this state data the graphic moves off the button.
           int yDeltaDown = 1;
           if (ICoordinateSystem::applicationOrientation() ==
               ICoordinateSystem::originLowerLeft )
             yDeltaDown = -1;
           graphicbutton = (IGraphicPushButton*)evt.window();
           IPoint location = graphicbutton->graphicWindow().position();
           if (evt.parameter1().asUnsignedLong() == 0)
              {
              // draw button up
              if (graphicbutton->fGraphicPushButtonData->fDrawDown)
                 {
                 graphicbutton->fGraphicPushButtonData->fDrawDown = false;
                 bMovedImage = true;
                 location.setX(location.x()-1);
                 location.setY(location.y()-yDeltaDown);
                 }
#ifdef IC_PM
              // There is an OS/2 bug that does not repaint the top
              // and left outermost edge of the graphic when no-border
              // and no-margin is set.  So... we need to be sure that
              // the entire graphic is invalidated.
              if ( !graphicbutton->hasBorder() &&
                 ( graphicbutton->marginSize() == ISize(0,0) ))
                {
                   IRectangle rect1 =
                      graphicbutton->graphicWindow().nativeRect();
                   IRectangle rect2(rect1.left(),  rect1.top(),
                                    rect1.right()+1, rect1.top()+1);
                   graphicbutton->graphicWindow().refresh(
                      ICoordinateSystem::convertToApplication(
                         rect2, graphicbutton->size() ) );
                   rect2 = IRectangle(rect1.left(),  rect1.bottom(),
                                      rect1.left()+1, rect1.top()+1);
                   graphicbutton->graphicWindow().refresh(
                      ICoordinateSystem::convertToApplication(
                         rect2, graphicbutton->size() ) );
                }
#endif
              }
           else
              {
              // draw button down
              if (!graphicbutton->fGraphicPushButtonData->fDrawDown)
                 {
                 graphicbutton->fGraphicPushButtonData->fDrawDown = true;
                 bMovedImage = true;
                 location.setX(location.x()+1);
                 location.setY(location.y()+yDeltaDown);
                 } /* endif */
              }
           if (bMovedImage)
              graphicbutton->graphicWindow().moveTo(location);
         }

#ifdef IC_PM
       // intentional fall through to next case
      case WM_SETWINDOWPARAMS:
      case BM_SETDEFAULT:
      case WM_SYSVALUECHANGED:
         if (!graphicbutton)
           graphicbutton = (IGraphicPushButton*)evt.window();

         graphicbutton->setStyle(graphicbutton->style() |
                                 IWindow::clipChildren.asUnsignedLong());
         defaultProcedure(evt);
         graphicbutton->setStyle(graphicbutton->style() &
                                 ~IWindow::clipChildren.asUnsignedLong());
#else
         defaultProcedure( evt );
         if (bMovedImage)
           graphicbutton->graphicWindow().refresh( IWindow::paintAllImmediate);
#endif
         return true;

#ifdef IC_WIN
      case WM_ENABLE:
         graphicbutton = (IGraphicPushButton*)evt.window();
         if (graphicbutton)
            graphicbutton->graphicWindow().enable( evt.parameter1() );

         break;
#endif

      default:
         break;
   } /* endswitch */
   return false;
}



/*------------------------------------------------------------------------------
| IGraphicPushButtonData::IGraphicPushButtonData                               |
------------------------------------------------------------------------------*/
IGraphicPushButtonData::IGraphicPushButtonData() :
   fDrawDown( false )
{
}

/*------------------------------------------------------------------------------
| IGraphicPushButtonData::~IGraphicPushButtonData                              |
------------------------------------------------------------------------------*/
IGraphicPushButtonData::~IGraphicPushButtonData()
{
}
