/*******************************************************************************
* FILE NAME: ihelp1.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ihelp.hpp.  (platform dependent)                                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Note we depend on ireslib.cpp being first
#pragma priority( -2147482102 )
#define INCL_WINHELP
#define INCL_WININPUT
#define INCL_WINMESSAGEMGR
#define INCL_WINWINDOWMGR
extern "C" {
  #include <iwindefs.h>
#ifdef IC_WIN
  #include <ipfx.h>
#endif
}

#include <ihelp.hpp>
#include <ihelpprv.hpp>
#include <ihelpsta.hpp>
#include <itrace.hpp>
#include <iapp.hpp>
#if (!defined(IC_MOTIF) && !defined(_IWCNAME_))
  #include <iwcname.hpp>
#endif

#ifdef IC_WIN
  #include <iplatfrm.hpp>
  #include <ihelptbl.hpp>
  #include <isubmenu.hpp>
#endif

#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <iframe.hpp>
#include <ipoint.hpp>
#include <iprocadr.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <ithread.hpp>
#include <iwindow.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHELP1_CPP_
  #include <ipagetun.h>
#endif

//***************************************************************
// Public styles.
//***************************************************************
#pragma data_seg(ICLStaticConst)
const IHelpWindow::Style
  IHelpWindow::noStyle            = 0,
  IHelpWindow::classDefaultStyle  = 0,
  IHelpWindow::ipfCompatible      = 0x00000001;
#pragma data_seg()

#pragma data_seg(ICLNonConst)
#ifdef IC_WIN
  IHelpWindow::Style IHelpWindow::currentDefaultStyle = IHelpWindow::noStyle;
#else
  IHelpWindow::Style IHelpWindow::currentDefaultStyle = IHelpWindow::ipfCompatible;
#endif
#pragma data_seg()

/*------------------------------------------------------------------------------
| IHelpWindow::IHelpWindow                                                     |
|  Initialize the PM help structure and create an instance of the help         |
|  facility.                                                                   |
------------------------------------------------------------------------------*/
IHelpWindow::IHelpWindow ( IFrameWindow             *associatedWindow,
                           const IHelpWindow::Style &style )
            : fHelpWindowData ( new IHelpWindowData )
{
  if (style & ipfCompatible)
  {
    fHelpWindowData->isUsingIPF = true;

    HELPINIT   helpstruct;
#ifdef IC_PM
    helpstruct.cb              = sizeof(HELPINIT);
    helpstruct.ulReturnCode    = 0;
    helpstruct.pszTutorialName = 0;
    helpstruct.phtHelpTable    = 0;

    helpstruct.hmodHelpTableModule =
                  IApplication::current().userResourceLibrary().handle();
    helpstruct.idAccelTable        = 0;
    helpstruct.idActionBar         = 0;
    helpstruct.pszHelpWindowTitle  = 0;
    helpstruct.fShowPanelId        = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryName  = 0;
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance(IThread::current().anchorBlock(),
                                       &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("WinCreateHelpInstance");
    }
#endif
#ifdef IC_WIN
    helpstruct.hHelpTableModule = 0;
    helpstruct.hAccelMenuBarModule = 0;
    helpstruct.ulAccelTable = 0;
    helpstruct.menuBarHierarchy = 0;
    helpstruct.pszHelpWindowTitle = 0;
    helpstruct.usShowPanelId = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryNames = 0;
    helpstruct.pHelpTable = 0;
    helpstruct.usHelpTableID = 0;
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance( IWindow::objectWindow()->handle(),
                                        &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("XhCreateHelpInstance");
    }
    // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
    fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
      // handleEventsFor is called in setAssociatedWindow.
#endif

    if ( associatedWindow )
      setAssociatedWindow( associatedWindow );
#ifdef IC_PM
    startHandlingEventsFor(fHelpWindowData->helpHandle);
#endif
  } // end of IPF Compatibility

#ifdef IC_WIN
  else // Native windows Help
  {
    // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
    fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
      // handleEventsFor is called in setAssociatedWindow.
    if ( associatedWindow )
      setAssociatedWindow( associatedWindow );
  }
  // For win32s we have to add hooks for F1 help because there are no
  // WM_HELP messages on that environment.
  if (IPlatform::isWin32S())
    IHelpStatics::addHooks();
#endif

}


/*------------------------------------------------------------------------------
| IHelpWindow::IHelpWindow                                                     |
|  Initialize the PM help structure and create an instance of the help         |
|  facility.                                                                   |
------------------------------------------------------------------------------*/
IHelpWindow::IHelpWindow ( const IResourceId        &helpTable,
                           IFrameWindow             *associatedWindow,
                           const IHelpWindow::Style &style )
            : fHelpWindowData ( new IHelpWindowData )
{
  IASSERTPARM(associatedWindow!=0);
  if (style & ipfCompatible)
  {
    fHelpWindowData->isUsingIPF = true;

    HELPINIT   helpstruct;
#ifdef IC_PM
    helpstruct.cb              = sizeof(HELPINIT);
    helpstruct.ulReturnCode    = 0;
    helpstruct.pszTutorialName = 0;
    helpstruct.phtHelpTable    = (PHELPTABLE)MAKELONG(
                                          (unsigned long)(helpTable.id() ),
                                          0xffff);

    helpstruct.hmodHelpTableModule =
                  IApplication::current().userResourceLibrary().handle();
    helpstruct.idAccelTable        = 0;
    helpstruct.idActionBar         = 0;
    helpstruct.pszHelpWindowTitle  = 0;
    helpstruct.fShowPanelId        = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryName  = 0;
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance(IThread::current().anchorBlock(),
                                       &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("WinCreateHelpInstance");
    }
    setAssociatedWindow( associatedWindow );
    startHandlingEventsFor(fHelpWindowData->helpHandle);
#endif
#ifdef IC_WIN
    helpstruct.hHelpTableModule = 0;
    helpstruct.hAccelMenuBarModule = 0;
    helpstruct.ulAccelTable = 0;
    helpstruct.menuBarHierarchy = 0;
    helpstruct.pszHelpWindowTitle = 0;
    helpstruct.usShowPanelId = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryNames = 0;
    helpstruct.pHelpTable = 0;
    helpstruct.usHelpTableID = 0;
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance(IWindow::objectWindow()->handle(),
                                       &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("XhCreateHelpInstance");
    }
#endif

  } // end of IPF Compatibility

#ifdef IC_WIN
  // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
  fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
    // handleEventsFor is called in setAssociatedWindow.
  setAssociatedWindow( associatedWindow );
  if ( helpTable.id()!=0 )
    fHelpWindowData->pHlpTable = new IHelpTable(helpTable);
  // For win32s we have to add hooks for F1 help because there are no
  // WM_HELP messages on that environment.
  if (IPlatform::isWin32S())
    IHelpStatics::addHooks();
#endif

}

/*------------------------------------------------------------------------------
| IHelpWindow::IHelpWindow                                                     |
|  Initialize the PM help structure from the given Settings class and          |
|  create an instance of the help facility.                                    |
------------------------------------------------------------------------------*/
IHelpWindow::IHelpWindow( const Settings           &settings,
                          IFrameWindow             *associatedWindow,
                          const IHelpWindow::Style &style )
            : fHelpWindowData ( new IHelpWindowData )
{
  if ( settings.HTResLib.length() )
  {
     this->fHelpWindowData->pHTResLib =
              new IDynamicLinkLibrary( (char*)settings.HTResLib );
     this->fHelpWindowData->newHTResLib = true;
  }
  else
  {
     this->fHelpWindowData->pHTResLib =
             &( IApplication::current().userResourceLibrary() );
  }

  if ( settings.MAResLib.length() )
  {
     this->fHelpWindowData->pMAResLib =
              new IDynamicLinkLibrary( (char*)settings.MAResLib );
  }
  //Store the initial libraries so we can concatenate when using addLibraries()
  fHelpWindowData->helpLibraries = settings.helpLibraries;

  if (style & ipfCompatible)
  {
    fHelpWindowData->isUsingIPF = true;
    HELPINIT   helpstruct;
#ifdef IC_PM
    helpstruct.cb              = sizeof(HELPINIT);
    helpstruct.ulReturnCode    = 0;
    helpstruct.pszTutorialName = settings.tutorial.length() ?
                                   (char*)(settings.tutorial) : 0;
    helpstruct.phtHelpTable    = (PHELPTABLE)MAKELONG(
                                   (unsigned long)(settings.helpTableId),
                                   0xffff);
    if ( settings.HTResLib.length() )
       helpstruct.hmodHelpTableModule =
                    this->fHelpWindowData->pHTResLib->handle();
    else
       helpstruct.hmodHelpTableModule =
                    IApplication::current().userResourceLibrary().handle();
    if ( settings.MAResLib.length() )
       helpstruct.hmodAccelActionBarModule =
                    this->fHelpWindowData->pMAResLib->handle();
    else
       helpstruct.hmodAccelActionBarModule = 0;
    helpstruct.idAccelTable       = settings.acceleratorId;
    helpstruct.idActionBar        = settings.menuBarId;
    helpstruct.pszHelpWindowTitle = (char*)settings.title;
    helpstruct.fShowPanelId       = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryName = (char*)(settings.helpLibraries);
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance(IThread::current().anchorBlock(),
                                     &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("WinCreateHelpInstance");
    }
    if ( this->fHelpWindowData->pHTResLib )
       this->fHelpWindowData->pHTResLib->loadHelpTable( this,
                                               settings.helpTableId);
#endif
#ifdef IC_WIN
    helpstruct.pHelpTable = 0;
    helpstruct.usHelpTableID = 0;
    helpstruct.hHelpTableModule = 0;
    if ( settings.MAResLib.length() )
       helpstruct.hAccelMenuBarModule =
                    this->fHelpWindowData->pMAResLib->handle();
    else
       helpstruct.hAccelMenuBarModule = 0;
    helpstruct.ulAccelTable        = settings.acceleratorId;
    helpstruct.menuBarHierarchy    = settings.menuBarId;
    helpstruct.pszHelpWindowTitle  = (char*)settings.title;
    helpstruct.usShowPanelId       = CMIC_HIDE_PANEL_ID;
    helpstruct.pszHelpLibraryNames = (char*)(settings.helpLibraries);;
    fHelpWindowData->helpHandle =
      IHelpStatics::createHelpInstance(IWindow::objectWindow()->handle(),
                                       &helpstruct);
    if (!fHelpWindowData->helpHandle)
    {
       ITHROWGUIERROR("XhCreateHelpInstance");
    }
    // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
    fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
      // handleEventsFor is called in setAssociatedWindow.
#endif

    if ( associatedWindow )
       setAssociatedWindow( associatedWindow );
#ifdef IC_PM
    startHandlingEventsFor(fHelpWindowData->helpHandle);
#endif

    if (settings.usingHelpId)
       this->sendEvent(HM_REPLACE_USING_HELP,
                       IEventParameter1(settings.usingHelpId),
                       IEventParameter2(0) );

  } // end of IPF Compatibility

#ifdef IC_WIN
  else // Native windows help
  {
    // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
    fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
      // handleEventsFor is called in setAssociatedWindow.
    if ( associatedWindow )
       setAssociatedWindow( associatedWindow );
  }

  if ( settings.helpTableId != 0 )
    fHelpWindowData->pHlpTable =
         new IHelpTable( IResourceId(settings.helpTableId,
                                    *(this->fHelpWindowData->pHTResLib)));
  // For win32s we have to add hooks for F1 help because there are no
  // WM_HELP messages on that environment.
  if (IPlatform::isWin32S())
    IHelpStatics::addHooks();
#endif
}


/*------------------------------------------------------------------------------
| IHelpWindow::IHelpWindow                                                     |
|  Create an IHelpWindow object from an existing window handle.                |
------------------------------------------------------------------------------*/
IHelpWindow::IHelpWindow(const IWindowHandle& helpWindowHandle)
            : fHelpWindowData ( new IHelpWindowData )
{
  IASSERTPARM(helpWindowHandle!=0);

  fHelpWindowData->helpHandle = helpWindowHandle;
  fHelpWindowData->isUsingIPF = true;
#ifdef IC_PM
  startHandlingEventsFor(fHelpWindowData->helpHandle);
#endif
  setAutoDestroyWindow(false);
#ifdef IC_WIN
  // Handle WM_HELP messages for windows (VK_F1 messages for win32s)
  fHelpWindowData->winHelpHandler = new IWinHelpHandler(this,fHelpWindowData);
    // handleEventsFor is called in setAssociatedWindow.
#endif

}


/*------------------------------------------------------------------------------
| IHelpWindow::~IHelpWindow                                                    |
------------------------------------------------------------------------------*/
IHelpWindow::~IHelpWindow()
{
  if (this->isIPFCompatible())
  {
       // the isWin32S condition is added only as a temporary solution to
       // the problem of LIBIPF32.DLL hanging on win32s when the following
       // scenario is run:
       //   application 1 creates a help instance
       //   application 2 creates another help instance before application 1
       //     terminates
       //   application 2 destroys its help instance
       //   application 1 tries to destroy its help instance, but hangs.
       // Adding this condition circumvents the hang, but does cause a memory
       //   leak.  Any single application that creates and destroys many help
       //   instances may run into a problem.  (jks)
    if ( fHelpWindowData->helpHandle && !deleteIsInProcess()
           && !IPlatform::isWin32S() )
       IHelpStatics::destroyHelpInstance( fHelpWindowData->helpHandle );
  }
#ifdef IC_WIN
  else
  {
    if ( !deleteIsInProcess() )
      WinHelp((HWND)this->handle(),
              fHelpWindowData->helpLibraries,
              HELP_QUIT,
              0);
  }
  if (fHelpWindowData->winHelpHandler)
    delete fHelpWindowData->winHelpHandler;

  if (fHelpWindowData->pHlpTable)
    delete fHelpWindowData->pHlpTable;

  if (fHelpWindowData->pMAResLib)
    delete fHelpWindowData->pMAResLib;
#endif

  delete fHelpWindowData;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IWinHelpHandler::IWinHelpHandler                                             |
|  Save pointer to help window private data.                                   |
------------------------------------------------------------------------------*/
IWinHelpHandler::IWinHelpHandler(IHelpWindow *helpWindow,
                                 IHelpWindowData *helpWindowData)
           : pHelpWindow(helpWindow)
           , pHlpWndData(helpWindowData)
{

}

/*------------------------------------------------------------------------------
| IWinHelpHandler::~IWinHelpHandler                                            |
------------------------------------------------------------------------------*/
IWinHelpHandler::~IWinHelpHandler()
{

}

/*------------------------------------------------------------------------------
| IWinHelpHandler::dispatchHandlerEvent                                        |
|  Dispatch WM_HELP messages.                                                  |
------------------------------------------------------------------------------*/
Boolean IWinHelpHandler::dispatchHandlerEvent(IEvent &evt)
{

   // Handle WM_HELP messages, except for win32s, then VK_F1
   if ( evt.eventId() == WM_HELP )
   {
     return winHelpRequested(evt);
   }
   else if ( evt.eventId() == IC_UM_QUERY_HELP_WINDOW )
   {
     evt.setResult( (void *) (pHelpWindow) );
     return true;
   }
   // dismiss the helpwindow when the last frame that showed help dismisses.
   else if ( evt.eventId() == WM_DESTROY ||
            (evt.eventId() == WM_SHOWWINDOW && evt.parameter2() == SW_HIDE) )
   {
     if ( evt.handle()==pHlpWndData->ownerForIPF )
     {
       pHelpWindow->hide();
       return false;
     }
   }

 return false;
}

/*------------------------------------------------------------------------------
| IWinHelpHandler::winHelpRequested                                            |
|  Process WM_HELP messages by getting mapping control id to help id from the  |
|  resource file help table.  If no help table, display general help.          |
------------------------------------------------------------------------------*/
Boolean IWinHelpHandler::winHelpRequested(IEvent &evt)
{
  unsigned short frameID=0;
  unsigned short controlID=0;
  unsigned short helpID=0;
  unsigned short frameOrMenu=(unsigned short)HLPM_FRAME;
  HWND requestingHwnd=0;
  IFrameWindow* activeFrame=0;

  HELPINFO helpInfo=*((LPHELPINFO)(void *)(evt.parameter2()));
  requestingHwnd = helpInfo.hItemHandle;
  controlID   = (unsigned short)helpInfo.iCtrlId;
  helpID      = (unsigned short)helpInfo.dwContextId;

  if (helpInfo.iContextType == HELPINFO_MENUITEM)
  {
    frameOrMenu = (unsigned short)HLPM_MENU;
    if (helpID==0)
    {
      IMenu *tempMenu=new IMenu(requestingHwnd);
      if (tempMenu)
        helpID = (unsigned short)tempMenu->itemHelpId(controlID);
    }
    requestingHwnd = evt.window()->handle();
  }
  else
  {
    // see if a helpid is stored in the window.
    // if getwindowcontexthelpid doesnt find a helpid in the requestingHwnd,
    // then it searches the parent (for child window) or owner chain
    // so that it can inherit helpIds.

    if (helpID==0)
    {
      HWND wnd = requestingHwnd;
      while ( helpID==0 && wnd!=0 )
      {
        if ( IPlatform::isWin32S() )
        {
          IWindow *iwnd=0;
          iwnd = IWindow::windowWithHandle( wnd );
          if ( iwnd )
            helpID = (unsigned short)iwnd->helpId();
        }
        else
          helpID = (unsigned short)GetWindowContextHelpId( wnd );

        HWND parent = IPARENTOF( wnd );
        if ( parent!=0 && parent!=IWindow::desktopWindow() )
          wnd = parent;
        else
          wnd = IOWNEROF( wnd );
      }
    }
  }

  // If setActiveWindow has been called, then we have its handle saved in
  // private data. Use it for searching the help table and as the frame
  // requesting help for native windows help. If not, search the parent
  // and owner chain of the control for a frame to use.
  activeFrame = ihelpwindowSearchChain(requestingHwnd,
                                       pHlpWndData->lastActiveHandle);
  if ( activeFrame == 0 )
    return false;

  // If we don't have a helpId already, but we do have a helpTable and the
  // of the control for which help was requested, then search the helpTable.
  if (helpID == 0)
  {
    // look up controlID in helptable if we have one
    if (pHlpWndData->pHlpTable)
    {
      frameID = (unsigned short)activeFrame->id();
      helpID = pHlpWndData->pHlpTable->helpForWindowID(controlID, frameID);
    }

    if ( helpID == IHelpTable::noHelpFound || helpID == 0)
    {
      IEventParameter1 parm1( frameOrMenu, 0);
      IEventParameter2 parm2( frameID, controlID);
      Boolean processedByHandler=
         activeFrame->sendEvent(HM_HELPSUBITEM_NOT_FOUND, parm1, parm2);

      // if subitemnotfound was processed by the helphandler, return.
      if (processedByHandler)
      {
        evt.setResult( true );
        return true;
      }

      // look for extended help in helptable, if there is one
      if (pHlpWndData->pHlpTable )
        helpID = pHlpWndData->pHlpTable->extendedHelpForWindowID(frameID);

      // if there is no extended help, then return
      if (helpID == 0)
      {
        evt.setResult( true );
        return true;
      }

    }
  }

  // If we do have a helpId, then call the appropriate help manager.
  if (pHlpWndData->isUsingIPF)
  {
    IHelpStatics::callIPF(pHlpWndData->helpHandle, HM_DISPLAY_HELP,
                            (LPVOID)helpID, NULL);
  }
  else
  {
    WinHelp(pHelpWindow->handle(),
            pHlpWndData->helpLibraries,
            HELP_CONTEXT,
            helpID);
  }

  pHlpWndData->ownerForIPF = activeFrame->handle();

  evt.setResult( true );
  return true;
}
#endif  // IC_WIN

/*------------------------------------------------------------------------------
| IHelpWindow::sizeTo                                                          |
------------------------------------------------------------------------------*/
IHelpWindow& IHelpWindow::sizeTo(const ISize& newSize)
{
#ifdef IC_PM
    RECTL   rectl;
    IWindowHandle hwnd = coverPageWindow();
    if ( !hwnd )
    {
       ITHROWLIBRARYERROR(IC_NO_HELP_COVERPAGE, IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
    }
    WinQueryWindowRect(hwnd, &rectl);
    WinMapWindowPoints(hwnd, HWND_DESKTOP, (PPOINTL)&rectl, 2);

    IRectangle newRect =  ICoordinateSystem::convertToApplication(
                          IRectangle(rectl),
                          desktopWindow()->size() ) ;

    moveSizeTo( newRect.sizeTo( newSize ) );
#endif
#ifdef IC_WIN
  if (this->isIPFCompatible())
  {
    Inherited::sizeTo(newSize);
  } // end ipfCompatibility
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IHelpWindow::moveTo                                                          |
------------------------------------------------------------------------------*/
IHelpWindow& IHelpWindow::moveTo(const IPoint& newPos)
{
#ifdef IC_PM
    RECTL   rectl;
    IWindowHandle hwnd = coverPageWindow();
    if ( !hwnd )
    {
       ITHROWLIBRARYERROR(IC_NO_HELP_COVERPAGE, IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
    }
    WinQueryWindowRect(hwnd, &rectl);

  /* Note: WinQueryWindowRect always returns bottom left = (0,0)   */
    rectl.xLeft    = newPos.x();
    rectl.yBottom  = newPos.y();
    rectl.xRight  += newPos.x();
    rectl.yTop    += newPos.y();

    moveSizeTo( IRectangle( rectl ) );
#endif
#ifdef IC_WIN
  if (this->isIPFCompatible())
  {
    Inherited::moveTo(newPos);
  } // end ipfCompatibility
#endif

  return *this;
}


/*------------------------------------------------------------------------------
| IHelpWindow::moveSizeTo                                                      |
------------------------------------------------------------------------------*/
IHelpWindow& IHelpWindow::moveSizeTo(const IRectangle& newRect)
{
#ifdef IC_PM
  RECTL   rectl = ICoordinateSystem::convertToNative (
                  newRect,
                  desktopWindow()->size() ).asRECTL();
  this->sendEvent(HM_SET_COVERPAGE_SIZE,
                  IEventParameter1(&rectl),
                  IEventParameter2(0) );
#endif
#ifdef IC_WIN
  if (this->isIPFCompatible())
  {
    Inherited::moveSizeTo(newRect);
  }
  else
  {
    HELPWININFO helpWinInfo;
    LPHELPWININFO lpHelpWinInfo=&helpWinInfo;

    helpWinInfo.wStructSize = sizeof(HELPWININFO);
    helpWinInfo.x           = (int)newRect.left();
    helpWinInfo.y           = (int)newRect.bottom();
    helpWinInfo.dx          = (int)newRect.right();
    helpWinInfo.dy          = (int)newRect.top();
    helpWinInfo.wMax        = SW_RESTORE;
//  helpWinInfo.rgchMember  = 0;

    WinHelp((HWND)this->handle(),
            fHelpWindowData->helpLibraries,
            HELP_SETWINPOS,
            (DWORD)lpHelpWinInfo);
  }
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IHelpWindow::setActiveWindow                                                 |
------------------------------------------------------------------------------*/
IHelpWindow& IHelpWindow::setActiveWindow( IFrameWindow* activeWindow,
                                           IFrameWindow* relativeWindow )
{
  if (this->isIPFCompatible())
  {
#ifdef IC_PM
    HWND defaultRelWin = HWND_PARENT;
    if (activeWindow)
    {
      if (relativeWindow)
      {
        this->sendEvent(HM_SET_ACTIVE_WINDOW,
                        IEventParameter1( activeWindow->handle() ),
                        IEventParameter2( relativeWindow->handle() ) );
      }
      else
      {
        this->sendEvent(HM_SET_ACTIVE_WINDOW,
                        IEventParameter1( activeWindow->handle() ),
                        IEventParameter2( defaultRelWin ) );
      }
#endif
#ifdef IC_WIN
    if (activeWindow)
    {
      if (relativeWindow)
      {
        this->sendEvent(HM_SET_ACTIVE_WINDOW,
                        IEventParameter1( relativeWindow->handle() ),
                        IEventParameter2(0) );
      }
      else
      {
        this->sendEvent(HM_SET_ACTIVE_WINDOW,
                        IEventParameter1( activeWindow->handle() ),
                        IEventParameter2(0) );
      }
#endif
    }
    else
       this->sendEvent(HM_SET_ACTIVE_WINDOW,
                       IEventParameter1(0),
                       IEventParameter2(0) );
  } // NOP for Native windows help
#ifdef IC_WIN
  fHelpWindowData->lastActiveHandle =
    activeWindow ? activeWindow->handle() : (IWindowHandle)0;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IHelpWindow::setAssociatedWindow                                             |
------------------------------------------------------------------------------*/
IHelpWindow& IHelpWindow::setAssociatedWindow( IFrameWindow* associatedWindow )
{
  IASSERTPARM( associatedWindow != 0 );
  IASSERTPARM( fHelpWindowData->winHelpHandler != 0 );

#ifdef IC_PM
  if (this->isIPFCompatible())
  {

    IWindowHandle frameHandle = associatedWindow->handle();

    if (!IHelpStatics::associateHelpInstance(fHelpWindowData->helpHandle, frameHandle))
       ITHROWGUIERROR("WinAssociateHelpInstance");
  }
#endif

#ifdef IC_WIN
  fHelpWindowData->winHelpHandler->handleEventsFor( associatedWindow );
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IHelpWindow::sendEvent                                                       |
------------------------------------------------------------------------------*/
IEventResult IHelpWindow::sendEvent( unsigned long  eventId,
                                     const IEventParameter1 &parm1,
                                     const IEventParameter2 &parm2) const
{
#ifdef IC_WIN
  long rc=0;
  // should only be called for ipf compatibility
  if (this->isIPFCompatible())
  {
    rc = IHelpStatics::callIPF( fHelpWindowData->helpHandle,
                                (unsigned short)eventId, parm1, parm2 );
  }
  return IEventResult((void *)rc);
#else
  return Inherited::sendEvent( eventId, parm1, parm2 );
#endif
}

/*------------------------------------------------------------------------------
| IHelpWindow::sendEvent                                                       |
------------------------------------------------------------------------------*/
IEventResult IHelpWindow::sendEvent ( const IEvent& event ) const
{
  return Inherited::sendEvent(event);
}

/*------------------------------------------------------------------------------
| IHelpWindow::sendEvent                                                       |
------------------------------------------------------------------------------*/
IEventResult IHelpWindow::sendEvent ( EventType eventType,
                                      const IEventParameter1& parm1,
                                      const IEventParameter2& parm2 ) const
{
  return Inherited::sendEvent( eventType, parm1, parm2 );
}

/*------------------------------------------------------------------------------
| IHelpWindow::defaultStyle                                                    |
|                                                                              |
| Return the default style for new help window objects.                        |
------------------------------------------------------------------------------*/
IHelpWindow::Style  IHelpWindow::defaultStyle()
{
  return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IHelpWindow::setDefaultStyle                                                 |
|                                                                              |
| Set the default style for new help window objects.                           |
------------------------------------------------------------------------------*/
void IHelpWindow::setDefaultStyle( const IHelpWindow::Style& newDefault )
{
  currentDefaultStyle = newDefault;
}
#endif

