/*******************************************************************************
* FILE NAME: ihelpevt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the class(es):                    *
*     IHelpMenuBarEvent                                                        *
*     IHelpErrorEvent                                                          *
*     IHelpHyperlinkEvent                                                      *
*     IHelpSubitemNotFoundEvent                                                *
*     IHelpTutorialEvent                                                       *
*     IHelpNotifyEvent                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WINHELP
#define INCL_WINHOOKS        // For HLPM_xxx.
extern "C" {
  #include <iwindefs.h>
  #ifdef IC_MOTIF
     #include <ipfx.h>
     #include <ixconst.h>     // ICLUI additions to ipf/x
  #endif
  #ifdef IC_WIN
     #include <ipfx.h>
  #endif
  }

#ifndef _IHELPEVT_
  #include <ihelpevt.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHELPEVT_CPP_
  #include <ipagetun.h>
#endif


/*-----------------------------------------------------------------------------
      Help Events Constructors
------------------------------------------------------------------------------*/
IHelpMenuBarEvent         :: IHelpMenuBarEvent(const IEvent& event)
                           : IEvent(event) {}

IHelpErrorEvent           :: IHelpErrorEvent(const IEvent& event)
                           : IEvent(event) {}

IHelpHyperlinkEvent       :: IHelpHyperlinkEvent(const IEvent& event)
                           : IEvent(event) {}

IHelpSubitemNotFoundEvent :: IHelpSubitemNotFoundEvent(const IEvent& event)
                           : IEvent(event) {}

IHelpTutorialEvent        :: IHelpTutorialEvent(const IEvent& event)
                           : IEvent(event) {}

IHelpNotifyEvent          :: IHelpNotifyEvent(const IEvent& event)
                           : IEvent(event) {}


/*-----------------------------------------------------------------------------
      Help Events Destructors
------------------------------------------------------------------------------*/
IHelpMenuBarEvent::~IHelpMenuBarEvent()
{ }

IHelpErrorEvent::~IHelpErrorEvent()
{ }

IHelpHyperlinkEvent::~IHelpHyperlinkEvent()
{ }

IHelpNotifyEvent::~IHelpNotifyEvent()
{ }

IHelpTutorialEvent::~IHelpTutorialEvent()
{ }

IHelpSubitemNotFoundEvent::~IHelpSubitemNotFoundEvent()
{ }


/*------------------------------------------------------------------------------
| IHelpMenuBarEvent::commandId                                                 |
------------------------------------------------------------------------------*/
unsigned long IHelpMenuBarEvent :: commandId() const
{
  return (unsigned long)parameter1();
}


/*------------------------------------------------------------------------------
| IHelpErrorEvent::ErrorType IHelpErrorEvent :: error                          |
------------------------------------------------------------------------------*/
IHelpErrorEvent::ErrorType IHelpErrorEvent :: error() const
{
  ErrorType type = unknown;

  switch ((unsigned long)parameter1())
  {
#ifdef IC_PM
   case HMERR_LOAD_DLL:
      type = loadingDLL;
      break;

   case HMERR_NO_FRAME_WND_IN_CHAIN:
      type = noWindow;
      break;
#endif

#ifdef IC_PM
   case HMERR_INVALID_ASSOC_APP_WND:
      type = invalidAppHandle;
      break;
#endif

#ifdef IC_MOTIF
   case HMERR_INVALID_APPL_WIDGET:
      type = invalidAppHandle;
      break;
#endif

#ifdef IC_PM
   case HMERR_INVALID_ASSOC_HELP_INST:
   case HMERR_INVALID_DESTROY_HELP_INST:
      type = invalidInstanceHandle;
      break;
#endif
#ifdef IC_MOTIF
   case HMERR_INVALID_HELP_WIDGET:
      type = invalidInstanceHandle;
      break;
#endif

#ifdef IC_PM
   case HMERR_INVALID_QUERY_APP_WND:
      type = invalidQueryHandle;
      break;

   case HMERR_HELPTABLE_UNDEFINE:
      type = noHelpTable;
      break;
#endif

#ifdef IC_PM
   case HMERR_HELP_INSTANCE_UNDEFINE:
   case HMERR_INVALID_HELP_INSTANCE_HDL:
   case HMERR_HELP_INST_CALLED_INVALID:
      type = noHelpInstance;
      break;
#endif
#ifdef IC_MOTIF
   case HMERR_NO_HELP_INST_IN_CHAIN:
      type = noHelpInstance;
      break;
#endif

#ifdef IC_PM
   case HMERR_HELPSUBITEM_NOT_FOUND:
   case HMERR_INVALID_HELPSUBITEM_SIZE:
      type = invalidId;
      break;
#endif
#ifdef IC_MOTIF
   case HMERR_HELPITEM_NOT_FOUND:
      type = invalidId;
      break;
#endif
#ifdef IC_WIN
   case HMERR_HELPITEM_NOT_FOUND:
      type = invalidId;
      break;
#endif
   case HMERR_INDEX_NOT_FOUND:
      type = noIndex;
      break;

   case HMERR_CONTENT_NOT_FOUND:
      type = noContent;
      break;

#ifdef IC_MOTIF
   case HMERR_LIB_FILE_NOT_FOUND:
#endif
   case HMERR_OPEN_LIB_FILE:
      type = openHelpFile;
      break;

   case HMERR_READ_LIB_FILE:
      type = readHelpFile;
      break;

   case HMERR_CLOSE_LIB_FILE:
      type = closeHelpFile;
      break;

   case HMERR_INVALID_LIB_FILE:
      type = invalidHelpFile;
      break;

#ifdef IC_PM
   case HMERR_NO_MEMORY:
   case HMERR_ALLOCATE_SEGMENT:
      type = notEnoughMemory;
      break;
#endif
#ifdef IC_MOTIFWIN
   case HMERR_MEMORY_ERROR:
      type = notEnoughMemory;
      break;
#endif

   case HMERR_FREE_MEMORY:
      type = unableFreeMemory;
      break;

   case HMERR_PANEL_NOT_FOUND:
      type = invalidHelpWindow;
      break;

#ifdef IC_PM
   case HMERR_DATABASE_NOT_OPEN:
      type = unopenDatabase;
      break;
#endif

#ifdef IC_MOTIF
   case HMERR_FONTSET_CREATION_FAIL:
   case HMERR_NO_DEFAULT_FONT_FOUND:
      type = fontNotFound;
      break;
#endif

   // The following Motif unique errors are handled by the default case
   // case HMERR_INVALID_HELP_STRUCTURE:
   // case HMERR_INVALID_MESSAGE:
   // case HMERR_INVALID_PARAMETERS:
   // case HMERR_OPEN_COPY_FILE:
   // case HMERR_WRITE_COPY_FILE:

   default:
      break;
  } /* endswitch */

  return type;
}


/*------------------------------------------------------------------------------
| IHelpHyperlinkEvent::id                                                      |
------------------------------------------------------------------------------*/
unsigned long IHelpHyperlinkEvent :: id() const
{
  return (unsigned long)parameter1();
}


/*------------------------------------------------------------------------------
| IHelpSubitemNotFoundEvent::isWindow                                          |
------------------------------------------------------------------------------*/
IBase::Boolean IHelpSubitemNotFoundEvent :: isWindow() const
{
  return (parameter1().number1() == (unsigned short)HLPM_WINDOW)
            ? true : false;
}


/*------------------------------------------------------------------------------
| IHelpSubitemNotFoundEvent::isFrame                                           |
------------------------------------------------------------------------------*/
IBase::Boolean IHelpSubitemNotFoundEvent :: isFrame() const
{
  return (parameter1().number1() == (unsigned short)HLPM_FRAME)
            ? true : false;
}


/*------------------------------------------------------------------------------
| IHelpSubitemNotFoundEvent::isMenu                                            |
------------------------------------------------------------------------------*/
IBase::Boolean IHelpSubitemNotFoundEvent :: isMenu() const
{
  return (parameter1().number1() == (unsigned short)HLPM_MENU)
            ? true : false;
}


/*------------------------------------------------------------------------------
| IHelpSubitemNotFoundEvent::topicId                                           |
------------------------------------------------------------------------------*/
unsigned long IHelpSubitemNotFoundEvent :: topicId() const
{
  return (unsigned long)parameter2().number1();
}


/*------------------------------------------------------------------------------
| IHelpSubitemNotFoundEvent::subtopicId                                        |
------------------------------------------------------------------------------*/
unsigned long IHelpSubitemNotFoundEvent :: subtopicId() const
{
  return (unsigned long)parameter2().number2();
}


/*------------------------------------------------------------------------------
| IHelpTutorialEvent::tutorialName                                             |
------------------------------------------------------------------------------*/
IString  IHelpTutorialEvent :: tutorialName() const
{
  return IString((char*)parameter1());
}


/*------------------------------------------------------------------------------
| IHelpNotifyEvent::controlId                                                  |
------------------------------------------------------------------------------*/
unsigned long  IHelpNotifyEvent :: controlId() const
{
#ifdef IC_PM
  return parameter1().number2();
#endif
#ifdef IC_WIN
  return parameter2().number2();
#endif
}


/*------------------------------------------------------------------------------
| IHelpNotifyEvent::windowHandle                                               |
------------------------------------------------------------------------------*/
IWindowHandle  IHelpNotifyEvent :: windowHandle() const
{
#ifdef IC_PM
  return IWindowHandle(parameter2());
#endif
#ifdef IC_WIN
  return IWindowHandle(parameter1());
#endif
}
