/* NOSHIP */
#ifndef _IHELPSTA_
#define _IHELPSTA_
/*******************************************************************************
* FILE NAME: ihelpsta.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the declaration(s) of the IHelpStatics class, which     *
*   is used to dynamically load helpmgr.dll.                                   *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <ibase.hpp>
#include <ihandle.hpp>

#ifdef IC_WIN
extern "C"
  {
    #include <ipfx.h>
  }
#endif

#pragma pack(4)

#ifndef IC_MOTIF
class IDynamicLinkLibrary;
#else
class IResLibPrivateData;
class IWindow;
#endif

#ifdef IC_WIN
class IFrameWindow;
class IWindow;
#endif

#ifdef IC_PM
typedef unsigned long
     (_System  ICreateInstance)( IAnchorBlockHandle, void * );
typedef Boolean
     (_System  IAssocInstance)( IWindowHandle, IWindowHandle );
typedef unsigned long
     (_System  IQueryInstance)( IWindowHandle );
typedef Boolean
     (_System  IDestroyInstance)( IWindowHandle );
typedef Boolean
     (_System  ILoadHelpTable)( IWindowHandle, unsigned long, IModuleHandle );
#endif

#ifdef IC_WIN
typedef long
     (__stdcall ICreateInstance)( HWND, PHELPINIT, HWND * );
typedef long
     (__stdcall IAssocInstance)( HWND, HWND );
typedef long
     (__stdcall IQueryInstance)( HWND, HWND * );
typedef long
     (__stdcall IDestroyInstance)( HWND );
typedef long
     (__stdcall ILoadHelpTable)( HWND, unsigned long, HWND );
typedef long
     (__stdcall ICallIPF)( HWND , unsigned short , void *, void * );
#endif

class IHelpStatics : public IBase
{
public:
 ~IHelpStatics          ( );

#ifndef IC_MOTIF
static IDynamicLinkLibrary
 &helpDLL               ( );
#endif // MOTIF

static IWindowHandle
#ifdef IC_PM
  createHelpInstance    ( const IAnchorBlockHandle& hab,
                          void* helpstruct ),
  queryHelpInstance     ( const IWindowHandle& windowHandle );
#endif
#ifdef IC_WIN
  createHelpInstance    ( const IWindowHandle& windowHandle,
                          void* helpstruct ),
  queryHelpInstance     ( const IWindowHandle& windowHandle );
#endif
#ifdef IC_MOTIF
  createHelpInstance    ( IWindowHandle appshell,
                          void*         helpstruct),
  queryHelpInstance     ( IWindowHandle windowHandle ),
  lastAssociatedHandle  (  IWindowHandle helpInst );
#endif // MOTIF

static Boolean
#ifndef IC_MOTIF
  associateHelpInstance ( const IWindowHandle& helpHandle,
                          const IWindowHandle& activeWindowHandle ),
  destroyHelpInstance   ( const IWindowHandle& helpHandle ),
  loadHelpTable         ( const IWindowHandle& helpHandle,
                          unsigned long id,
                          const IModuleHandle& hmod );
#else
  associateHelpInstance ( IWindowHandle helpHandle,
                          IWindowHandle activeWindowHandle),
  destroyHelpInstance   ( IWindowHandle helpHandle),
  loadHelpTable         ( IWindowHandle helpHandle,
                          unsigned long id,
                          IResLibPrivateData *resLibData);
#endif // MOTIF
#ifdef IC_WIN
static long
  callIPF               ( const IWindowHandle& helpHandle,
                          unsigned short usMsgType,
                          void* param1, void* param2);
static void
  addHooks              ( );
#endif

#ifndef IC_MOTIF
static ICreateInstance
 *entryCreate;
static IAssocInstance
 *entryAssoc;
static IQueryInstance
 *entryQuery;
static IDestroyInstance
 *entryDestroy;
static ILoadHelpTable
 *entryLoadHelpTable;
#ifdef IC_WIN
static ICallIPF
 *entryCallIPF;
#endif

static IDynamicLinkLibrary
 *helpDynLinkLib;

#ifdef IC_WIN
static HHOOK
  msgFilterHook,
  keyboardHook;
static unsigned short
  activeMenuItem;
static HWND
  activeMenu;
#endif

#endif // NOT MOTIF
};

#ifdef IC_WIN
IFrameWindow* ihelpwindowSearchChain( HWND contextHwnd, HWND lastActiveFrame );
#endif

#ifdef IC_MOTIF
class IAssociatedWindowHandle : public IBase {
/*******************************************************************************
| The IAssociatedWindowHandle class represents an association between          |
| a frame window and a help instance.  An instance of this class is created    |
| and stored in an ordered collection when help is associated with a           |
| frame.                                                                       |
*******************************************************************************/
public:
  IAssociatedWindowHandle (IWindowHandle help, IWindowHandle frame) :
      helpHandle (help), windowHandle(frame)
      {};
  IAssociatedWindowHandle (const IAssociatedWindowHandle & another ) :
      helpHandle(another.helpHandle), windowHandle(another.windowHandle)
      {};
  ~IAssociatedWindowHandle()
      {}

IAssociatedWindowHandle& operator= (const IAssociatedWindowHandle & another )
      {
      helpHandle = another.helpHandle; windowHandle = another.windowHandle;
      return *this;
      };

// instance data
  IWindowHandle  helpHandle;
  IWindowHandle  windowHandle;
};

/*----------------------------------------------------------------------------*/
/* Callback routines used for IPF/X interface                                 */
/*----------------------------------------------------------------------------*/
extern "C" {
  // prototype for callback routine for XmNhelpCallback
  void ihelpwindowCallback( _WidgetRec   *handle,
                            void         *client_data,
                            void         *call_data);
  // IPF/X application notify proc prototype
  typedef int (*PAppNotifyProc)() ;
  int  ihelpwindowAppNotifyProc( _WidgetRec    *helpInstance,
                                 unsigned short messageId,
                                 void          *parm1,
                                 void          *parm2) ;
}

// Functions of type PCheckWindowFunction are passed to the ihelpwindowSearchChain
// function to validate an IWindow
typedef IWindowHandle (*PCheckWindowFunction)( IWindow *pwin );
IWindowHandle  checkForHelpInstance( IWindow *pwin );
IWindowHandle  checkForFrameWindow( IWindow *pwin );

// routine to search parent and owner chain for window validated by testfn
IWindowHandle ihelpwindowSearchChain( IWindowHandle        handle,
                                      PCheckWindowFunction testfn );
#endif // IC_MOTIF

#pragma pack()

#endif // _IHELPSTA_
