/*******************************************************************************
* FILE NAME: iinfoa.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iinfoa.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINWINDOWMGR
#define INCL_WINMENUS
#include <iwindefs.h>

#include <iinfoa.hpp>
#include <iexcept.hpp>
#include <itrace.hpp>
#include <iframe.hpp>
#include <imenu.hpp>
#include <imnitem.hpp>
#include <isubmenu.hpp>
#include <ifont.hpp>
#include <iapp.hpp>
#include <iwcname.hpp>
#include <inotifev.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IINFOA_CPP_
  #include <ipagetun.h>
#endif

#define DEFAULT_TEXT_LIMIT  80


/*------------------------------------------------------------------------------
| &IInfoArea::setStringTableOffset                                             |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setStringTableOffset(long newOffset)
{
  if (stringTableOffset() != newOffset)
  {
    this->strTblOffset = newOffset;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IInfoArea::stringTableOffset                                                 |
------------------------------------------------------------------------------*/
long IInfoArea :: stringTableOffset ( ) const
  {
  return this->strTblOffset;
  }


/*------------------------------------------------------------------------------
| IInfoArea::IInfoArea                                                         |
------------------------------------------------------------------------------*/
IInfoArea :: IInfoArea ( IFrameWindow        *frame,
                         unsigned long        id )
  : IStaticText( id,                   // Id as given.
                 frame,                // Parent is the frame.
                 frame,                // Owner is the frame.
                 IRectangle(),         // Doesn't matter.
                 IWindow::visible |    // Style.
                   IStaticText::left |
                   IStaticText::top |
                   IStaticText::fillBackground),
    IMenuHandler(),
    pResLib( 0 ),
    numberOfLines( 1 ),
    strTblOffset( 0 ),
    inactive(),
    disabled(),
    missing(),
    status( noneShowing ),
    preventFlyUpdate(false)
/*------------------------------------------------------------------------------
  Leave pResLib at zero to indicate to utilize user resource library.
------------------------------------------------------------------------------*/
  {
  IMODTRACE_DEVELOP( "IInfoArea::ctor 1" );
  this -> setLimit( DEFAULT_TEXT_LIMIT );
  this -> addTo( frame );
  }

/*------------------------------------------------------------------------------
| IInfoArea::IInfoArea                                                         |
------------------------------------------------------------------------------*/
IInfoArea :: IInfoArea ( IFrameWindow        *frame,
                         unsigned long        id,
                         const char          *resDLLName )
  : IStaticText( id,                   // Id as given.
                 frame,                // Parent is the frame.
                 frame,                // Owner is the frame.
                 IRectangle(),         // Doesn't matter.
                 IWindow::visible |    // Style.
                   IStaticText::left |
                   IStaticText::top |
                   IStaticText::fillBackground),
    IMenuHandler(),
    pResLib( 0 ),
    numberOfLines( 1 ),
    strTblOffset( 0 ),
    inactive(),
    disabled(),
    missing(),
    status( noneShowing ),
    preventFlyUpdate(false)
  {
  IMODTRACE_DEVELOP( "IInfoArea::ctor 2" );
  this -> setResourceLibrary( resDLLName );
  this -> setLimit( DEFAULT_TEXT_LIMIT );
  this -> addTo( frame );
  }

/*------------------------------------------------------------------------------
| IInfoArea::IInfoArea                                                         |
------------------------------------------------------------------------------*/
IInfoArea :: IInfoArea ( IFrameWindow        *frame,
                         const IModuleHandle &resMod,
                         unsigned long        id )
  : IStaticText( id,                   // Id as given.
                 frame,                // Parent is the frame.
                 frame,                // Owner is the frame.
                 IRectangle(),         // Doesn't matter.
                 IWindow::visible |    // Style.
                   IStaticText::left |
                   IStaticText::top |
                   IStaticText::fillBackground),
    IMenuHandler(),
    pResLib( 0 ),
    numberOfLines( 1 ),
    strTblOffset( 0 ),
    inactive(),
    disabled(),
    missing(),
    status( noneShowing ),
    preventFlyUpdate(false)
  {
  IMODTRACE_DEVELOP( "IInfoArea::ctor 3" );
  this -> setResourceLibrary( resMod );
  this -> setLimit( DEFAULT_TEXT_LIMIT );
  this -> addTo( frame );
  }

/*------------------------------------------------------------------------------
| IInfoArea::IInfoArea                                                         |
------------------------------------------------------------------------------*/
IInfoArea :: IInfoArea ( IFrameWindow        *frame,
                         const char          *resDLLName,
                         unsigned long        id )
  : IStaticText( id,                   // Id as given.
                 frame,                // Parent is the frame.
                 frame,                // Owner is the frame.
                 IRectangle(),         // Doesn't matter.
                 IWindow::visible |    // Style.
                   IStaticText::left |
                   IStaticText::top |
                   IStaticText::fillBackground),
    IMenuHandler(),
    pResLib( 0 ),
    numberOfLines( 1 ),
    strTblOffset( 0 ),
    inactive(),
    disabled(),
    missing(),
    status( noneShowing ),
    preventFlyUpdate(false)
  {
  IMODTRACE_DEVELOP( "IInfoArea::ctor 4" );
  this -> setResourceLibrary( resDLLName );
  this -> setLimit( DEFAULT_TEXT_LIMIT );
  this -> addTo( frame );
  }

/*------------------------------------------------------------------------------
| IInfoArea::~IInfoArea                                                        |
------------------------------------------------------------------------------*/
IInfoArea :: ~IInfoArea()
  {
  IMODTRACE_DEVELOP( "IInfoArea::dtor" );

  // Remove extension...
  if ( this->isValid() )
  {
    IFrameWindow* pfrmParent = (IFrameWindow*)(this->parent());
    if ( pfrmParent && pfrmParent->isFrameWindow() )
      {
      pfrmParent -> removeExtension( this );
      this->stopHandlingEventsFor( pfrmParent );
      IHandler::stopHandlingEventsFor( this );
      }
  }

  // Delete resource library if we created one...
  delete this->pResLib;
  }

/*------------------------------------------------------------------------------
| IInfoArea::dispatchHandlerEvent                                              |
------------------------------------------------------------------------------*/
Boolean IInfoArea::dispatchHandlerEvent(IEvent& evt)
{
   // Check for menu item is being de-selected (losing focus)
   if ((evt.eventId() == WM_MENUSELECT) &&
#ifdef IC_WIN
       (evt.parameter1().highNumber() == 0xFFFF))
#endif
#ifndef IC_WIN
       (evt.parameter1().lowNumber() == 0xFFFF))
#endif
     {

#ifdef IC_WIN
     // It seems that this MENUSELECT only occurs when the
     // menu is closed...not when a submenu is closed.
     this -> showText( inactiveShowing, this->inactiveText() );
#endif

#ifndef IC_WIN
     // Get owner window.  If the owner is another menu, then reset the
     // info area text to the menu text, otherwise set inactive text
     IWindowHandle whOwner = IOWNEROF(evt.parameter2());
     if (whOwner)
     {
       IWindowClassName  className(whOwner);
       if ( className == WC_MENU )
       {
         int sItem = whOwner.sendEvent(MM_QUERYSELITEMID).number1();
         if (( sItem == MIT_ERROR) || ( sItem == MIT_NONE ))
         {
           this -> showText( inactiveShowing, this->inactiveText() );
         }
         else
         {
           evt.window()->sendEvent( WM_MENUSELECT, sItem,
                                    (unsigned long) whOwner);
         }
       }
       else
         this -> showText( inactiveShowing, this->inactiveText() );
     }
     else
       this -> showText( inactiveShowing, this->inactiveText() );
#endif

   }
   else if (evt.eventId() == IC_UM_QRY_PREVENTUPDATE)
   {
     evt.setResult(preventFlyUpdate);
     return true;
   }

   return IMenuHandler::dispatchHandlerEvent(evt);
}

/*------------------------------------------------------------------------------
| &IInfoArea::resourceLibrary                                                  |
------------------------------------------------------------------------------*/
IResourceLibrary &IInfoArea :: resourceLibrary ( ) const
/*------------------------------------------------------------------------------
  If we've got our own, return it, else use current user default.
------------------------------------------------------------------------------*/
  {
  if ( this->pResLib )
    return *this->pResLib;
  else
    return IApplication::current().userResourceLibrary();
  }

/*------------------------------------------------------------------------------
| &IInfoArea::setResourceLibrary                                               |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea :: setResourceLibrary ( const IModuleHandle &resMod )
/*------------------------------------------------------------------------------
  Construct an DLL from the input handle.
------------------------------------------------------------------------------*/
{
  delete this->pResLib;
  this->pResLib = 0;
  if (resMod)
  {
    pResLib = new IDynamicLinkLibrary( resMod );
  }
  else
  {
    pResLib = new IResourceLibrary();
  }
  notifyObservers(INotificationEvent(IInfoArea::resourceLibraryId,
                                     *this, true, (void*)pResLib));
  return *this;
}

/*------------------------------------------------------------------------------
| &IInfoArea::setResourceLibrary                                               |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea :: setResourceLibrary ( const char *resDLLName )
/*------------------------------------------------------------------------------
  Construct an DLL from the input dll name.
------------------------------------------------------------------------------*/
{
  delete this->pResLib;
  this->pResLib = 0;
  if (resDLLName)
  {
    pResLib = new IDynamicLinkLibrary(resDLLName);
  }
  else
  {
    pResLib = new IResourceLibrary();
  }
  notifyObservers(INotificationEvent(IInfoArea::resourceLibraryId,
                                     *this, true, (void*)pResLib));
  return *this;
}

/*------------------------------------------------------------------------------
| IInfoArea::inactiveText                                                      |
------------------------------------------------------------------------------*/
IString IInfoArea :: inactiveText ( ) const
  {
  return this->inactive;
  }

/*------------------------------------------------------------------------------
| IInfoArea::disabledText                                                      |
------------------------------------------------------------------------------*/
IString IInfoArea :: disabledText ( ) const
  {
  return this->disabled;
  }

/*------------------------------------------------------------------------------
| IInfoArea::missingText                                                       |
------------------------------------------------------------------------------*/
IString IInfoArea :: missingText  ( ) const
  {
  return this->missing;
  }

/*------------------------------------------------------------------------------
| IInfoArea::setString                                                         |
------------------------------------------------------------------------------*/
void IInfoArea :: setString ( IString &aString, unsigned long id )
  {
  aString = this->resourceLibrary().loadString( id );
  }

/*------------------------------------------------------------------------------
|  IInfoArea::setInactiveText                                                  |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setInactiveText(unsigned long id)
{
  IString newString(this->resourceLibrary().loadString(id));
  if (newString != this->inactive)
  {
    this->inactive = newString;
    this->status = inactiveShowing;
    if (this->status)
    {
      this->setText(this->inactiveText());
    }
    notifyObservers(INotificationEvent(IInfoArea::inactiveTextId,
                                       *this, true, (char*)inactive));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setInactiveText                                                  |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setInactiveText(const IString &aString)
{
  if (this->inactive != aString)
  {
    this->inactive = aString;
    this->status = inactiveShowing;
    if (this->status)
    {
      this->setText(this->inactiveText());
    }
    notifyObservers(INotificationEvent(IInfoArea::inactiveTextId,
                                       *this, true, (char*)inactive));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setDisabledText                                                  |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setDisabledText(unsigned long id)
{
  IString newString(this->resourceLibrary().loadString(id));
  if (newString != this->disabled)
  {
    this->disabled = newString;
    this->status = disabledShowing;
    if (this->status)
    {
      this->setText(this->disabledText());
    }
    notifyObservers(INotificationEvent(IInfoArea::disabledTextId,
                                       *this, true, (char*)disabled));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setDisabledText                                                  |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setDisabledText(const IString &aString)
{
  if (this->disabled != aString)
  {
    this->disabled = aString;
    this->status = disabledShowing;
    if (this->status)
    {
      this->setText(this->disabledText());
    }
    notifyObservers(INotificationEvent(IInfoArea::disabledTextId,
                                       *this, true, (char*)disabled));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setMissingText                                                   |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setMissingText(unsigned long id)
{
  IString newString(this->resourceLibrary().loadString(id));
  if (newString != this->missing)
  {
    this->missing = newString;
    this->status = missingShowing;
    if (this->status)
    {
      this->setText(this->missingText());
    }
    notifyObservers(INotificationEvent(IInfoArea::missingTextId,
                                       *this, true, (char*)missing));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setMissingText                                                   |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setMissingText(const IString &aString)
{
  if (this->missing != aString)
  {
    this->missing = aString;
    this->status = missingShowing;
    if (this->status)
    {
      this->setText(this->missingText());
    }
    notifyObservers(INotificationEvent(IInfoArea::missingTextId,
                                       *this, true, (char*)missing));
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::lineCount                                                        |
------------------------------------------------------------------------------*/
unsigned long IInfoArea::lineCount ( ) const
{
  return this->numberOfLines;
}

/*------------------------------------------------------------------------------
|  IInfoArea::setLineCount                                                     |
------------------------------------------------------------------------------*/
IInfoArea &IInfoArea::setLineCount ( unsigned long lines )
{
  IASSERTPARM(lines != 0);
  if (this->lineCount() != lines)
  {
     this->numberOfLines = lines;

     if (lines > 1)
     {
        this->setAlignment( IStaticText::topLeftWrapped );
     }
     else if (this->alignment() == IStaticText::topLeftWrapped)
     {
        this->setAlignment( IStaticText::topLeft );
     }

     this->setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
  }
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::calcMinimumSize                                                  |
------------------------------------------------------------------------------*/
ISize IInfoArea::calcMinimumSize ( ) const
{
  ISize sizMin = Inherited::calcMinimumSize();
  unsigned long lines = this->lineCount();
  return ISize( sizMin.width() * 2 / (lines + 1),
                sizMin.height() * lines );
}

/*------------------------------------------------------------------------------
|  IInfoArea::handleEventsFor                                                  |
------------------------------------------------------------------------------*/
IInfoArea  &IInfoArea :: handleEventsFor ( IFrameWindow *frame )
{
  IASSERTPARM(frame != 0);
  IHandler::handleEventsFor(frame);
  return *this;
}

/*------------------------------------------------------------------------------
|  IInfoArea::stopHandlingEventsFor                                            |
------------------------------------------------------------------------------*/
IInfoArea  &IInfoArea :: stopHandlingEventsFor ( IFrameWindow *frame )
{
  IASSERTPARM(frame != 0);
  IHandler::stopHandlingEventsFor(frame);
  return *this;
}

/*------------------------------------------------------------------------------
| IInfoArea::informationFor                                                    |
------------------------------------------------------------------------------*/
IString IInfoArea :: informationFor( unsigned long menuId ) const
/*------------------------------------------------------------------------------
  Load the string with menuId + offset and return it.
  If it can't be loaded, throw an exception.
------------------------------------------------------------------------------*/
  {
  IString result =
    this->resourceLibrary().tryToLoadString( menuId +
                                             this->stringTableOffset() );
  return result;
  }

/*------------------------------------------------------------------------------
| IInfoArea::addTo                                                             |
------------------------------------------------------------------------------*/
void IInfoArea :: addTo( IFrameWindow *frame )
/*------------------------------------------------------------------------------
  Add this object as a frame extension at bottom of client area.
------------------------------------------------------------------------------*/
  {
  frame -> addExtension ( this,
                          IFrameWindow::belowClient );
  // The handler is attached to the info area as well as the frame
  // so that it can respond to a user message from fly over help.
  IHandler::handleEventsFor( this );
  this -> handleEventsFor( frame );
  }

/*------------------------------------------------------------------------------
| IInfoArea::showText                                                          |
------------------------------------------------------------------------------*/
void IInfoArea :: showText ( IAStatus newStatus, const IString &aString )
/*------------------------------------------------------------------------------
  Reset the status and show the argument text.
------------------------------------------------------------------------------*/
  {
  this->status = newStatus;
  if (newStatus == inactiveShowing)
    // Allow fly over help to set help text into the information area.
    preventFlyUpdate = false;
  else
    // Prevent fly over help from setting help text into the information area.
    preventFlyUpdate = true;

  this->setText( aString );
  }

/*------------------------------------------------------------------------------
| IInfoArea::menuSelected                                                      |
|  Called when a menu item is highlighted                                      |
------------------------------------------------------------------------------*/
Boolean IInfoArea :: menuSelected  ( IMenuEvent& mnevt )
{
  unsigned long ulStrId=mnevt.menuItemId();

  // Special routine to handle disabled menu items.
  // If user has called setDisabledText() then that text will be
  // displayed for any disabled menu item.
  // Otherwise just display the string from the table.

  if (disabledText().size())  {   /* any disabled text? */
    Boolean isDisabled;

    try {
      isDisabled = mnevt.menuItem().isDisabled();    //change in IMenuHandler
    }
    catch(...) {
      /* ulStrId not found in menu */
      return false;
    }/* end catch */

    if (isDisabled) {
      this -> showText( disabledShowing, this->disabledText() );
      return false;
    }

  }

  IString strInfoText=IString();
  IAStatus  newStatus=noneShowing;

  strInfoText=this -> informationFor( ulStrId );

  if (strInfoText.length() == 0)
  {                // string not found in resource
     strInfoText=missingText();
     newStatus = missingShowing;
  } /* end if */

  this -> showText( newStatus, strInfoText );

  return false;
}/* end IInfoArea :: menuSelected(...) */


/*------------------------------------------------------------------------------
| IInfoArea::menuEnded                                                         |
------------------------------------------------------------------------------*/
Boolean IInfoArea :: menuEnded     ( IMenuEvent& menuEvent )
{
  // Do nothing... this is now handled in dispatch processing
  // this -> showText( inactiveShowing, this->inactiveText() );
  return false;
}/* end IInfoArea :: menuEnded(...) */


/*------------------------------------------------------------------------------
| &IInfoArea::handleEventsFor                                                  |
------------------------------------------------------------------------------*/
IHandler  &IInfoArea :: handleEventsFor ( IWindow *window )
{                              // private to hide version in IHandler
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   return *this;
}


/*------------------------------------------------------------------------------
| &IInfoArea::stopHandlingEventsFor                                            |
------------------------------------------------------------------------------*/
IHandler  &IInfoArea :: stopHandlingEventsFor ( IWindow *window )
{                              // private to hide version in IHandler
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   return *this;
}
