/*******************************************************************************
* FILE NAME: ilbdihdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ilbdihdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define INCL_WINLISTBOXES
#define INCL_WINMENUS
#define INCL_WINMESSAGEMGR
#define INCL_WINWINDOWMGR
extern "C" {
  #include <iwindefs.h>
  #include <stdio.h>
}

#include <ilbdihdr.hpp>

#ifndef _ILISTBOX_
  #include <ilistbox.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IWCNAME_
  #include <iwcname.hpp>
#endif

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ILBDIHDR_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::IListBoxDrawItemHandler                             |
------------------------------------------------------------------------------*/
IListBoxDrawItemHandler::IListBoxDrawItemHandler ( )
  : IListBoxDrawItemHandler::Inherited( )
{ }

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::~IListBoxDrawItemHandler                            |
|                                                                              |
| Empty destructor here for page tuning                                        |
------------------------------------------------------------------------------*/
IListBoxDrawItemHandler::~IListBoxDrawItemHandler ( )
{ }

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::dispatchHandlerEvent                                |
|                                                                              |
| Dispatch command functions for this handler.                                 |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::dispatchHandlerEvent ( IEvent& event )
{
  Boolean bRc = false;
  switch ( event.eventId() )
  {
     case WM_MEASUREITEM:
     {
        IListBoxSizeItemEvent sizeEvent( event );
        IWindowHandle listBoxHwnd = sizeEvent.controlHandle();
        if ( listBoxHwnd  &&
             IWindowClassName( listBoxHwnd ) == WC_LISTBOX )
        {                                 // List box case.
           bRc = this->setItemSize( sizeEvent );

           ISize newSize = sizeEvent.itemSize();
           unsigned long newHeight = newSize.height();
#ifdef IC_WIN
           // Send message to listbox to tell it the width specified
           // by setItemSize.  Need this for horizontal scroll bars
           // to show up.
           listBoxHwnd.sendEvent( LB_SETHORIZONTALEXTENT,
                                  (unsigned long)sizeEvent.itemSize().width(),
                                  0 );
           // Fill in the MEASUREITEMSTRUCT with the size specified by
           // setItemSize.
           MEASUREITEMSTRUCT *mis = (MEASUREITEMSTRUCT*)(void*)(event.parameter2());
           mis->itemWidth  = newSize.width();
           mis->itemHeight = newHeight;
#endif

           if ( newHeight == 0 )
           {       // Don't leave height at 0 (can cause a trap).
              newHeight = 1;
              sizeEvent.setItemSize( newSize.setHeight( newHeight ));
           }
//#ifdef IC_PM
           else
           {  // Non-zero height specified.
              // Some major hacks:
              // - PM only honors the returned height when the item
              //   size is requested the first time, when the list box
              //   is created.  Force subsequent changes to be honored
              //   as well.
              // - The height is not honored from LM_SETITEMHEIGHT, if
              //   the list box is not as tall as the requested item
              //   height.  Force this, if necessary.

              // This handler is either attached to the list box
              // generating the WM_MEASUREITEM message, or to its
              // owner.  If IControlEvent::controlWindow returns 0,
              // this must mean we have a non-wrappered list box
              // control with a wrappered owner window, and the
              // handler is attached to the owner window.  This case
              // is supported only if the owner of the list box
              // control is also the control's parent (this is in
              // order for IWindow::handleWithParent to work).

              IWindow* listBox = sizeEvent.controlWindow();
              Boolean needToDelete = false;
              Boolean hackedHeight = false;
              Boolean origVisible = false;
              ISize origSize;

              if ( ! listBox )
              {
                 try
                 {
                    listBox = new IWindow( listBoxHwnd );
                    needToDelete = true;
                 }
                 catch ( IException& )
                 {
                    listBox = 0;
                 }
              }

              if ( listBox )
              {
                 origSize = listBox->size();
                 if ( origSize.height() < newHeight )
                 {  // LM_SETITEMHEIGHT fails if list box is too short.

                    // IWindow::isVisible doesn't report WS_VISIBLE for newly
                    // created, never-painted windows.
                    origVisible =
                      ISTYLEOF( listBox->handle() ) & WS_VISIBLE ? true : false;

                    listBox->disableUpdate();
                    listBox->sizeTo( ISize( origSize.width(), newHeight ));
                    hackedHeight = true;
                 }
              }

              // Force the height, in case PM ignores it (use the
              // window handle, since there may be no IListBox).
#ifdef IC_WIN
              listBoxHwnd.sendEvent( LM_SETITEMHEIGHT,
                                     IEventParameter1( (void*)(mis->itemID) ),
                                     IEventParameter2( newHeight ));
#else
              listBoxHwnd.sendEvent( LM_SETITEMHEIGHT, newHeight, 0 );
#endif

              if ( listBox )
              {         // Clean-up processing.
                 if ( hackedHeight )
                 {
                    listBox->sizeTo( origSize );
                    listBox->enableUpdate( origVisible );
                 }
                 if ( needToDelete )
                 {
                    delete listBox;
                 }
              }
           }
//#endif
        }

        event.setResult( sizeEvent.result() );
        break;
     }

     case WM_DRAWITEM:
     {
        IListBoxDrawItemEvent drawEvent( event );
        IWindowHandle listBoxHwnd = drawEvent.controlHandle();
        if ( listBoxHwnd  &&
             IWindowClassName( listBoxHwnd ) == WC_LISTBOX )
        {                                 // List box case.
           if ( drawEvent.isDrawItem() )
           {
              bRc = this->drawItem( drawEvent );
              if ( bRc )
              {
                 event.setResult( true );

                 // PM will draw the highlighting for the list box
                 // item, unless the fsState and fsStateOld flags are
                 // both set to 0.  In this case, it assumes that the
                 // application will manage the appearance of the
                 // item's selection state.

                 if ( drawEvent.isSelectionStateDrawn() )
                 {        // Don't let PM draw any highlighting.
#ifdef IC_WIN
                    ((LPDRAWITEMSTRUCT)(void*)event.parameter2())->itemState = 0;
#else
                    ((POWNERITEM)(void*)event.parameter2())->fsState = 0;
                    ((POWNERITEM)(void*)event.parameter2())->fsStateOld = 0;
#endif
                 }
              }
           }

           else if ( drawEvent.isSetHighlight() )
           {
              bRc = this->selectItem( drawEvent );
              if ( bRc )
              {
                 event.setResult( true );
                 if ( drawEvent.isSelectionStateDrawn() )
                 {
#ifdef IC_WIN
                    ((LPDRAWITEMSTRUCT)(void*)event.parameter2())->itemState = 0;
#else
                    ((POWNERITEM)(void*)event.parameter2())->fsState = 0;
                    ((POWNERITEM)(void*)event.parameter2())->fsStateOld = 0;
#endif
                 }
              }
           }

           else if ( drawEvent.isRemoveHighlight() )
           {
              bRc = this->deselectItem( drawEvent );
              if ( bRc )
              {
                 event.setResult( true );
                 if ( drawEvent.isSelectionStateDrawn() )
                 {
#ifdef IC_WIN
                    ((LPDRAWITEMSTRUCT)(void*)event.parameter2())->itemState = 0;
#else
                    ((POWNERITEM)(void*)event.parameter2())->fsState = 0;
                    ((POWNERITEM)(void*)event.parameter2())->fsStateOld = 0;
#endif
                 }
              }
           }
        }
        break;
     }
     default:
        break;
  } /* endswitch */

  return bRc;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::setItemSize                                         |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::setItemSize ( IListBoxSizeItemEvent& event )
{
  Boolean stopProcessingEvent = false;
#if (IC_OBSOLETE <= IC_OBSOLETE_1)
  ISize itemSize = event.itemSize();
  short itemHeight = (short)(itemSize.height());
  stopProcessingEvent = this->setHeight( itemHeight );
  event.setItemSize( itemSize.setHeight( itemHeight ));
#endif
  return stopProcessingEvent;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::drawItem                                            |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::drawItem ( IListBoxDrawItemEvent& event )
{
  Boolean stopProcessingEvent = false;
#if (IC_OBSOLETE <= IC_OBSOLETE_1)
  DrawFlag flag( 0 );
  stopProcessingEvent = this->draw( event, flag );
  if (flag & IListBoxDrawItemHandler::drewSelected)
  {
     event.setSelectionStateDrawn();
  }
#endif
  return stopProcessingEvent;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::selectItem                                          |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::selectItem ( IListBoxDrawItemEvent& event )
{
  Boolean stopProcessingEvent = false;
#if (IC_OBSOLETE <= IC_OBSOLETE_1)
  stopProcessingEvent = this->highlight( event );
  if (stopProcessingEvent)
  {           // Maintain compatibility with old version.
     event.setSelectionStateDrawn();
  }
#endif
  return stopProcessingEvent;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::deselectItem                                        |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::deselectItem ( IListBoxDrawItemEvent& event )
{
  Boolean stopProcessingEvent = false;
#if (IC_OBSOLETE <= IC_OBSOLETE_1)
  stopProcessingEvent = this->unhighlight( event );
  if (stopProcessingEvent)
  {           // Maintain compatibility with old version.
     event.setSelectionStateDrawn();
  }
#endif
  return stopProcessingEvent;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::handleEventsFor                                     |
|                                                                              |
| Attach handler to list box.                                                  |
------------------------------------------------------------------------------*/
IListBoxDrawItemHandler&
  IListBoxDrawItemHandler::handleEventsFor ( IListBox* listBox )
{
  Inherited::handleEventsFor( listBox );
  if ( listBox->isDrawItem() )
  {
     // The handler has already missed the initial WM_MEASUREITEM
     // message that PM uses to get the height of the list box items.
     // The handler may have also missed all other messages, if all
     // item have already been added to the list box.  Compensate for
     // this by generating a message that the handler will catch.

     IWindow* owner = listBox->owner();
     if ( owner )
     {
#ifdef IC_PM
        owner->sendEvent( WM_MEASUREITEM,
                          IEventParameter1( listBox->id() ),
                          0 );
#endif
#ifdef IC_WIN
        MEASUREITEMSTRUCT mis;
        mis.CtlType = ODT_LISTBOX;
        mis.CtlID   = listBox->id();
        mis.itemID  = 0;
        mis.itemWidth  = 0;
        mis.itemHeight = 0;
        // Get the value previously set, if any.
        mis.itemData =
           listBox->sendEvent( LB_GETITEMDATA,
                               IEventParameter1( (unsigned long)mis.itemID ),
                               IEventParameter2( 0 ) );
        owner->sendEvent( WM_MEASUREITEM,
                          IEventParameter1( (unsigned long)mis.CtlID ),
                          IEventParameter2( (void*) &mis ) );
#endif
     }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::handleEventsFor                                     |
|                                                                              |
| Attach handler to owner window of a list box.                                |
------------------------------------------------------------------------------*/
IListBoxDrawItemHandler&
  IListBoxDrawItemHandler::handleEventsFor ( IWindow* listBoxOwner )
{
  // The application must attach the handler to an owner window before
  // creating any owned, draw-item list boxes, to avoid missing the
  // initial WM_MEASUREITEM message that queries item heights.

  Inherited::handleEventsFor( listBoxOwner );
  return *this;
}


#if (IC_OBSOLETE <= IC_OBSOLETE_1)
//
// Block of obsolete code.  Obsolete level 1.
// These functions are here, to avoid creating a dependency on a separate
// DLL with the remainder of the obsolete functions, if they get packaged
// that way.
//

#pragma data_seg(ICLStaticConst)
  const IListBoxDrawItemHandler::DrawFlag
                    IListBoxDrawItemHandler::drewSelected ( 0, IDIF_SELECTED );
#pragma data_seg()

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::setHeight                                           |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::setHeight ( short& height )
{
  return false;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::draw                                                |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::draw ( IListBoxDrawItemEvent& event,
                                        DrawFlag& flag )
{
  return false;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::highlight                                           |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::highlight ( IListBoxDrawItemEvent& event )
{
  return false;
}

/*------------------------------------------------------------------------------
| IListBoxDrawItemHandler::unhighlight                                         |
------------------------------------------------------------------------------*/
Boolean IListBoxDrawItemHandler::unhighlight ( IListBoxDrawItemEvent& event )
{
  return false;
}
//
// End block of obsolete code.  Obsolete level 1.
//
#endif // IC_OBSOLETE
