/*******************************************************************************
* FILE NAME: imndievt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imndievt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <stdio.h>
  #define INCL_WINMENUS
  #include <iwindefs.h>
  }

#include <imndievt.hpp>
#include <irect.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMNDIEVT_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
  #define OIDATA ((LPDRAWITEMSTRUCT)ownerItemData())
#else
  #define OIDATA ((POWNERITEM)ownerItemData())
#endif

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::IMenuDrawItemEvent                                       |
------------------------------------------------------------------------------*/
IMenuDrawItemEvent::IMenuDrawItemEvent( IEvent& evt )
  : IMenuDrawItemEvent::Inherited(evt)
{ }

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::~IMenuDrawItemEvent                                      |
------------------------------------------------------------------------------*/
IMenuDrawItemEvent::~IMenuDrawItemEvent( )
{ }

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isDrawItem                                               |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isDrawItem() const
{
#ifdef IC_WIN
   return (OIDATA->itemAction & ODA_DRAWENTIRE) ? true : false;
#else
   return (OIDATA->fsAttribute ==
           OIDATA->fsAttributeOld) ? true : false;
#endif
}

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isSetHighlight                                           |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isSetHighlight() const
{
#ifdef IC_WIN
   return ((OIDATA->itemAction & ODA_SELECT) &&
           (OIDATA->itemState & ODS_SELECTED)) ? true : false;
#else
   return ( (OIDATA->fsAttribute != OIDATA->fsAttributeOld) &&
            (OIDATA->fsAttribute & MIA_HILITED) ) ?
      true : false;
#endif
}

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isRemoveHighlight                                        |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isRemoveHighlight() const
{
#ifdef IC_WIN
   return ((OIDATA->itemAction & ODA_SELECT) &&
           !(OIDATA->itemState & ODS_SELECTED)) ? true : false;
#else
   return ( (OIDATA->fsAttribute != OIDATA->fsAttributeOld) &&
            !(OIDATA->fsAttribute & MIA_HILITED) ) ?
      true : false;
#endif
}

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isSelected                                               |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isSelected() const
{
#ifdef IC_WIN
   return (OIDATA->itemState & ODS_SELECTED) ? true : false;
#else
   return (OIDATA->fsAttribute & MIA_FRAMED) ? true : false;
#endif
}

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isChecked                                                |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isChecked() const
{
#ifdef IC_WIN
   return (OIDATA->itemState & ODS_CHECKED) ? true : false;
#else
   return (OIDATA->fsAttribute & MIA_CHECKED) ? true : false;
#endif
}

/*------------------------------------------------------------------------------
| IMenuDrawItemEvent::isDisabled                                               |
------------------------------------------------------------------------------*/
Boolean IMenuDrawItemEvent::isDisabled() const
{
#ifdef IC_WIN
   return (OIDATA->itemState & ODS_DISABLED) ? true : false;
#else
   return (OIDATA->fsAttribute & MIA_DISABLED) ? true : false;
#endif
}
