/*******************************************************************************
* FILE NAME: inotebk.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IPageHandle                                                              *
*     INotebook                                                                *
*                                                                              *
*   See inotebk0.cpp for the implementation of the platform independent        *
*   nested classes:                                                            *
*     INotebook::Cursor                                                        *
*     INotebook::PageSettings                                                  *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_DOSMODULEMGR
  #define INCL_WINSYS
  #define INCL_WINENTRYFIELDS
  #define INCL_WINCLIPBOARD
  #define INCL_WININPUT             // WM_CHAR, etc.
  #define INCL_WINMESSAGEMGR        // for WNDPARAM struct
  #define INCL_NLS                  // ES_ANY, ES_SBCS, ES_DBCS, ES_MIXED
  #define INCL_WINSTDDRAG
  #define INCL_WINWINDOWMGR         // WinQueryWindowULong
  #define INCL_WINSTDBOOK
  #define INCL_WINDIALOGS           // Needed for PDLGTEMPLATE in BOOKPAGEINFO
  #define INCL_GPI
  #include <iwindefs.h>
}

#ifdef IC_WIN
  #ifndef _ICLNOTEBOOKW_
    #include <iclnbw.h>
  #endif
  #include <icctlsta.hpp>         // Statics to load dll.
#endif
#ifndef _INOTEBK_
  #include <inotebk.hpp>
#endif
#ifndef _INOTEBK0_
  #include <inotebk0.hpp>
#endif
#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _ICOLOR_
  #include <icolor.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFRAME_
  #include <iframe.hpp>
#endif
#ifndef _IMPHDR_
  #include <imphdr.hpp>
#endif
#ifndef _INOTEHDR_
  #include <inotehdr.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#ifdef IC_WIN
  #ifndef _ISIZEHDR_
    #include <isizehdr.hpp>
  #endif
  #ifndef _IPAGEHDR_
    #include <ipagehdr.hpp>
  #endif
  #ifndef _IPAINHDR_
    #include <ipainhdr.hpp>
  #endif
  #ifndef _IPAINEVT_
    #include <ipainevt.hpp>
  #endif
  #ifndef _IKEYEVT_
    #include <ikeyevt.hpp>
  #endif
  #ifndef _IGBITMAP_
    #include <igbitmap.hpp>
  #endif
  #ifndef _IWCNAME_
    #include <iwcname.hpp>
  #endif
  #ifndef _IES_
    #include <ies.h>
  #endif
#endif //IC_WIN


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _INOTEBK_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public notebook styles.                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const INotebook::Style
  INotebook::backPagesBottomRight = BKS_BACKPAGESBR,            // 0x00000001
  INotebook::backPagesBottomLeft  = BKS_BACKPAGESBL,            // 0x00000002
  INotebook::backPagesTopRight    = BKS_BACKPAGESTR,            // 0x00000004
  INotebook::backPagesTopLeft     = BKS_BACKPAGESTL,            // 0x00000008
  INotebook::majorTabsRight       = BKS_MAJORTABRIGHT,          // 0x00000010
  INotebook::majorTabsLeft        = BKS_MAJORTABLEFT,           // 0x00000020
  INotebook::majorTabsTop         = BKS_MAJORTABTOP,            // 0x00000040
  INotebook::majorTabsBottom      = BKS_MAJORTABBOTTOM,         // 0x00000080
  INotebook::squareTabs           ( 0, IBKS_SQUARETABS ),       // 0x00000001 *
  INotebook::roundedTabs          = BKS_ROUNDEDTABS,            // 0x00000100
  INotebook::polygonTabs          = BKS_POLYGONTABS,            // 0x00000200
  INotebook::solidBinding         ( 0, IBKS_SOLIDBIND ),        // 0x00000002 *
  INotebook::spiralBinding        = BKS_SPIRALBIND,             // 0x00000400
  INotebook::statusTextLeft       ( 0, IBKS_STATUSTEXTLEFT ),   // 0x00000004 *
  INotebook::statusTextRight      = BKS_STATUSTEXTRIGHT,        // 0x00001000
  INotebook::statusTextCenter     = BKS_STATUSTEXTCENTER,       // 0x00002000
  INotebook::tabTextLeft          ( 0, IBKS_TABTEXTLEFT ),      // 0x00000008 *
  INotebook::pmCompatible        ( 0, IBKS_PMCOMPATIBLE),       // 0x00000010 *
#ifdef IC_WIN
  INotebook::allTabsVisible      ( 0, IBKS_ALLTABSVISIBLE),     // 0x00000020 *
  INotebook::handleDrawTabs      ( 0, IBKS_HANDLEDRAWTABS),     // 0x00000040 *
#endif
  INotebook::tabTextRight         = BKS_TABTEXTRIGHT,           // 0x00004000
  INotebook::tabTextCenter        = BKS_TABTEXTCENTER,          // 0x00008000
  INotebook::classDefaultStyle  ( BKS_BACKPAGESBR |
                                  BKS_MAJORTABRIGHT |
                                  BKS_TABTEXTCENTER |
                                  WS_VISIBLE ,
                                  IBKS_SQUARETABS |
                                  IBKS_SOLIDBIND |
                                  IBKS_STATUSTEXTLEFT );
                               // 0x00000001 backPagesBottomRight
                               // 0x00000010 majorTabsRight
                               // 0x00008000 tabTextCenter
                               // 0x80000000 visible
                                             // 0x00000001 squareTabs
                                             // 0x00000002 solidBinding
                                             // 0x00000004 statusTextLeft

const INotebook::PageSettings::Attribute
  INotebook::PageSettings::noAttribute  = 0,                    // 0x00000000
  INotebook::PageSettings::statusTextOn = BKA_STATUSTEXTON,     // 0x00000001
  INotebook::PageSettings::majorTab     = BKA_MAJOR,            // 0x00000040
  INotebook::PageSettings::minorTab     = BKA_MINOR,            // 0x00000080
  INotebook::PageSettings::autoPageSize = BKA_AUTOPAGESIZE;     // 0x00000100

const INotebook::clrFlags
  INotebook::bgnPageColor  = IBKA_BACKGROUNDPAGECOLOR,          // 0x00000001 *
  INotebook::bgnMajorColor = IBKA_BACKGROUNDMAJORCOLOR,         // 0x00000002 *
  INotebook::bgnMinorColor = IBKA_BACKGROUNDMINORCOLOR,         // 0x00000004 *
  INotebook::fgnMajorColor = IBKA_FOREGROUNDMAJORCOLOR,         // 0x00000008 *
  INotebook::fgnMinorColor = IBKA_FOREGROUNDMINORCOLOR;         // 0x00000010 *
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new notebook objects (initial value).                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  INotebook::Style
       INotebook::currentDefaultStyle ( BKS_BACKPAGESBR |
                                        BKS_MAJORTABRIGHT |
                                        BKS_TABTEXTCENTER |
                                        WS_VISIBLE ,
                                        IBKS_SQUARETABS |
                                        IBKS_SOLIDBIND |
                                        IBKS_STATUSTEXTLEFT );
#ifdef IC_WIN
  Boolean  INotebook::hasBeenRegistered = false;
#endif
#pragma data_seg()

#ifdef IC_WIN
  #define UPAGE   (unsigned long)(void*)page
#else
  #define UPAGE   (unsigned long)page
#endif

#ifdef IC_WIN
  #define WC_PAGECLIP "IOC Page Clipping Window"
  #define ID_PAGECLIP 8000
#endif

/*------------------------------------------------------------------------------
| INotebookHandler                                                             |
|                                                                              |
| This class wrappers an object pointer and destroys the pointed to object     |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator INotebookHandler* ( );
INotebookHandler
 *operator -> ( )
  {
    return *this;
  }
 ~DefaultHandler ( )
  {
    if (ptr)
    {
      delete ptr;
      ptr = 0;
    }
  }
INotebookHandler
 *ptr;
} defaultHandler;

DefaultHandler :: operator INotebookHandler* ( )
{
  if ( !ptr )
    ptr = new INotebookHandler;

  return ptr;
}

/*------------------------------------------------------------------------------
| ITabBitmapMgr                                                                |
|                                                                              |
| This class is used to manage the tab bitmaps for a notebook's pages.         |
------------------------------------------------------------------------------*/
class ITabBitmapMgr {
public:
  ITabBitmapMgr ( const IBitmapHandle& tabBitmap,
                  const IPageHandle&   pgHandle );
virtual
 ~ITabBitmapMgr ( );
void
  setNext ( ITabBitmapMgr* );
ITabBitmapMgr
 *next    ( );
private:
IBitmapHandle
  bitmapHandle;
IPageHandle
  pageHandle;
ITabBitmapMgr
 *nextInList;
};

/*------------------------------------------------------------------------------
| ITabBitmapMgr::ITabBitmapMgr                                                 |
------------------------------------------------------------------------------*/
ITabBitmapMgr :: ITabBitmapMgr ( const IBitmapHandle& tabBitmap,
                                 const IPageHandle&   pgHandle )
{
  bitmapHandle = tabBitmap;
  pageHandle = pgHandle;
  nextInList = 0;
}

/*------------------------------------------------------------------------------
| ITabBitmapMgr::~ITabBitmapMgr                                                |
------------------------------------------------------------------------------*/
ITabBitmapMgr :: ~ITabBitmapMgr ( )
{ }

/*------------------------------------------------------------------------------
| ITabBitmapMgr::setNext                                                       |
------------------------------------------------------------------------------*/
void ITabBitmapMgr :: setNext ( ITabBitmapMgr* nextOne )
{
  nextInList = nextOne;
}

/*------------------------------------------------------------------------------
| ITabBitmapMgr::next                                                          |
------------------------------------------------------------------------------*/
ITabBitmapMgr* ITabBitmapMgr :: next ( )
{
  return nextInList;
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IPageClipPaintHandler                                                        |
|                                                                              |
| Handle paint requests for the page clipping window.                          |
------------------------------------------------------------------------------*/
class IPageClipPaintHandler : public IPaintHandler
{
typedef IPaintHandler
  Inherited;
public:
  IPageClipPaintHandler  ( );
virtual
  ~IPageClipPaintHandler ( );

static IPageClipPaintHandler
 *defaultHandler ( );

protected:
virtual Boolean
  paintWindow          ( IPaintEvent& event );
};

/*------------------------------------------------------------------------------
| IPageClipPaintHandler::IPageClipPaintHandler                                 |
| Constructor                                                                  |
------------------------------------------------------------------------------*/
IPageClipPaintHandler :: IPageClipPaintHandler( )
{}

/*------------------------------------------------------------------------------
| IPageClipPaintHandler::~IPageClipPaintHandler                                |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
IPageClipPaintHandler :: ~IPageClipPaintHandler( )
{}

/*------------------------------------------------------------------------------
| IPageClipPaintHandler::paintWindow                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IPageClipPaintHandler :: paintWindow ( IPaintEvent& event )
{
  IMODTRACE_DEVELOP( "IPageClipPaintHandler::paintWindow" );

  INotebook*
    pNotebook = (INotebook *)event.window()->parent();

  if (pNotebook)
  {
    IPresSpaceHandle hps = event.presSpaceHandle();
    event.clearBackground( pNotebook->pageBackgroundColor() );
    return( true );
  }

  return( false );
}

/*------------------------------------------------------------------------------
| IPageClipPaintHandlerStatics                                                 |
|                                                                              |
| Class to handle destruction of the static IPageClipPaintHandlerStatics       |
| object.                                                                      |
------------------------------------------------------------------------------*/
class IPageClipPaintHandlerStatics
{
public:
  ~IPageClipPaintHandlerStatics( )
  {
    if ( pDefHandler )
    {
      delete pDefHandler;
      pDefHandler = 0;
    }
  }

IPageClipPaintHandler
 *pDefHandler;
};

#pragma data_seg(ICLNonConst)
static IPageClipPaintHandlerStatics IDefPageClipStatics;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IPageClipPaintHandler::defaultHandler                                        |
|                                                                              |
------------------------------------------------------------------------------*/
IPageClipPaintHandler* IPageClipPaintHandler::defaultHandler( )
{
  if ( !IDefPageClipStatics.pDefHandler )
    IDefPageClipStatics.pDefHandler = new IPageClipPaintHandler();

  return( IDefPageClipStatics.pDefHandler );
}


/*------------------------------------------------------------------------------
| IPageClipEraseHandler                                                        |
|                                                                              |
| Handle background erase requests for the page clipping window.               |
------------------------------------------------------------------------------*/
class IPageClipEraseHandler : public IHandler
{
typedef IHandler
  Inherited;
public:
  IPageClipEraseHandler  ( );
virtual
  ~IPageClipEraseHandler ( );

static IPageClipEraseHandler
 *defaultHandler ( );

protected:
virtual Boolean
  dispatchHandlerEvent ( IEvent& event );
};

/*------------------------------------------------------------------------------
| IPageClipEraseHandler::IPageClipEraseHandler                                 |
| Constructor                                                                  |
------------------------------------------------------------------------------*/
IPageClipEraseHandler :: IPageClipEraseHandler( )
{}

/*------------------------------------------------------------------------------
| IPageClipEraseHandler::~IPageClipEraseHandler                                |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
IPageClipEraseHandler :: ~IPageClipEraseHandler( )
{}

/*------------------------------------------------------------------------------
| IPageClipEraseHandler::dispatchHandlerEvent                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IPageClipEraseHandler :: dispatchHandlerEvent( IEvent& event )
{
  /****************************************************************************/
  /* Only processes WM_ERASEBKGND to prevent default processing               */
  /****************************************************************************/
  switch( event.eventId() )
  {
    case WM_ERASEBKGND:
      event.setResult( true );
      return( true );

    default:
      break;
  }

  return( false );
}

/*------------------------------------------------------------------------------
| IPageClipEraseHandlerStatics                                                 |
|                                                                              |
| Class to handle destruction of the static IPageClipEraseHandlerStatics       |
| object.                                                                      |
------------------------------------------------------------------------------*/
class IPageClipEraseHandlerStatics
{
public:
  ~IPageClipEraseHandlerStatics( )
  {
    if ( pDefHandler )
    {
      delete pDefHandler;
      pDefHandler = 0;
    }
  }

IPageClipEraseHandler
 *pDefHandler;
};

#pragma data_seg(ICLNonConst)
static IPageClipEraseHandlerStatics IDefPageEraseStatics;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IPageClipEraseHandler::defaultHandler                                        |
|                                                                              |
------------------------------------------------------------------------------*/
IPageClipEraseHandler* IPageClipEraseHandler::defaultHandler( )
{
  if ( !IDefPageEraseStatics.pDefHandler )
    IDefPageEraseStatics.pDefHandler = new IPageClipEraseHandler();

  return( IDefPageEraseStatics.pDefHandler );
}


/*------------------------------------------------------------------------------
| IPageClipWindow                                                              |
|                                                                              |
| Page clipping window class for the Windows tab control implementation.       |
| Clips the application page window to the allowable display area within the   |
| tab control to prevent it from painting over areas of the control.           |
------------------------------------------------------------------------------*/
class IPageClipWindow : public IWindow
{
public:
  IPageClipWindow  ( INotebook*             parentNbk );

  IPageClipWindow  ( const IPageClipWindow& pageClipWindow );

IPageClipWindow
 &operator =       ( const IPageClipWindow& pageClipWindow );

virtual
  ~IPageClipWindow ( );

};

/*------------------------------------------------------------------------------
| IPageClipWindow::IPageClipWindow                                             |
| Constructor                                                                  |
------------------------------------------------------------------------------*/
IPageClipWindow :: IPageClipWindow( INotebook* parentNbk )
{
  if ( !parentNbk->hasBeenRegistered )
  {
    WNDCLASS wndClass;
    wndClass.style = CS_HREDRAW | CS_VREDRAW | CS_GLOBALCLASS;
    wndClass.lpfnWndProc = DefWindowProc;
    wndClass.cbClsExtra = 0;
    wndClass.cbWndExtra = 0;
    wndClass.hInstance = GetModuleHandle( 0 );
    wndClass.hIcon = 0;
    wndClass.hCursor = LoadCursor( NULL, IDC_ARROW );
    wndClass.hbrBackground = (HBRUSH)(COLOR_3DFACE+1);
    wndClass.lpszMenuName = 0;
    wndClass.lpszClassName = WC_PAGECLIP;

    if ( RegisterClass( &wndClass ) )
    {
      parentNbk->hasBeenRegistered = true;
    }
    else
    {
      ITHROWGUIERROR("RegisterClass");
    }
  }

  if ( parentNbk->hasBeenRegistered )
  {
    IWindowHandle
      handlePage = create( ID_PAGECLIP,
                           0,
                           WS_CHILD | WS_CLIPCHILDREN,
                           WC_PAGECLIP,
                           parentNbk->handle(),
                           IWindowHandle(0),
                           IRectangle(),
                           0,
                           0 );

    if ( handlePage.isValid() )
    {
      setAutoDestroyWindow( true );
      startHandlingEventsFor( handlePage );
      IPageClipPaintHandler::defaultHandler()->handleEventsFor( this );
      IPageClipEraseHandler::defaultHandler()->handleEventsFor( this );
    }
    else
    {
      ITHROWGUIERROR("CreateWindow");
    }
  }
}

/*------------------------------------------------------------------------------
| IPageClipWindow::IPageClipWindow                                             |
| Copy constructor                                                             |
------------------------------------------------------------------------------*/
IPageClipWindow :: IPageClipWindow( const IPageClipWindow& pageClipWindow )
{
}

/*------------------------------------------------------------------------------
| IPageClipWindow::operator =                                                  |
| Assignment                                                                   |
------------------------------------------------------------------------------*/
IPageClipWindow& IPageClipWindow ::
                      operator = ( const IPageClipWindow& pageClipWindow )
{
  return( *this );
}

/*------------------------------------------------------------------------------
| IPageClipWindow::~IPageClipWindow                                            |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
IPageClipWindow :: ~IPageClipWindow( )
{
  IPageClipPaintHandler::defaultHandler()->stopHandlingEventsFor( this );
  IPageClipEraseHandler::defaultHandler()->stopHandlingEventsFor( this );
}


/*------------------------------------------------------------------------------
| ITopPageKeyboardHandler                                                      |
|                                                                              |
| Process Alt+PageUp, Alt+PageDown, and Alt+UpArrow keys for the application   |
| page window that is on top.  This handler is only attached to the            |
| application page window that is on top.  When a new page becomes the top     |
| page this handler is removed from the current page and attached to the new   |
| top page.                                                                    |
------------------------------------------------------------------------------*/
class ITopPageKeyboardHandler : public IHandler
{
typedef IHandler
  Inherited;
public:
  ITopPageKeyboardHandler  ( );
virtual
  ~ITopPageKeyboardHandler ( );

static ITopPageKeyboardHandler
 *defaultHandler ( );

protected:
  dispatchHandlerEvent ( IEvent& event );
};

/*------------------------------------------------------------------------------
| ITopPageKeyboardHandler::ITopPageKeyboardHandler                             |
| Constructor                                                                  |
------------------------------------------------------------------------------*/
ITopPageKeyboardHandler :: ITopPageKeyboardHandler( )
{}

/*------------------------------------------------------------------------------
| ITopPageKeyboardHandler::~ITopPageKeyboardHandler                            |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
ITopPageKeyboardHandler :: ~ITopPageKeyboardHandler( )
{}

/*------------------------------------------------------------------------------
| ITopPageKeyboardHandler::dispatchHandlerEvent                                |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean ITopPageKeyboardHandler :: dispatchHandlerEvent( IEvent& event )
{
  Boolean stopProcessingEvent = false;

  /****************************************************************************/
  /* Only processes WM_SYSKEYDOWN to simulate CUA '91 notebook behavior.      */
  /****************************************************************************/
  switch( event.eventId() )
  {
    case WM_SYSKEYDOWN:
      {
        /**********************************************************************/
        /* Get a pointer to the notebook object.  The parent of the top       */
        /* application page window is always the page clipping window, and    */
        /* the parent of the page clipping window is always the notebook.     */
        /**********************************************************************/
        IWindowHandle
          hwndControl( event.window()->parent()->parent()->handle() );
        if ( !hwndControl )
          break;

        /**********************************************************************/
        /* Only process Windows tab control keyboard events of interest.      */
        /**********************************************************************/
        if ( IWindowClassName( hwndControl ) != WC_TABCONTROL )
          break;

        /**********************************************************************/
        /* Process key events for the Windows tab control.                    */
        /**********************************************************************/
        IKeyboardEvent keyevt( event );
        if ( !keyevt.isVirtual() || !keyevt.isAltDown() ||
             keyevt.isShiftDown() || keyevt.isCtrlDown() )
        {
          break;
        }

        /**********************************************************************/
        /* Obtain pointer to notebook object.                                 */
        /**********************************************************************/
        INotebook*
          pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

        /**********************************************************************/
        /* If Alt+PageUp or Alt+PageDown was pressed, send to the notebook    */
        /* for processing.                                                    */
        /**********************************************************************/
        if ( (keyevt.virtualKey() == IKeyboardEvent::pageUp)  ||
             (keyevt.virtualKey() == IKeyboardEvent::pageDown) )
        {
          HWND hwndOldFocus = IQUERYFOCUS(HWND_DESKTOP);

          pNotebk->sendEvent( event );

          /********************************************************************/
          /* If the focus hasn't changed after the page-turn, the application */
          /* left it on the old page, so put it on the tab control            */
          /********************************************************************/
          if ( IQUERYFOCUS(HWND_DESKTOP) == hwndOldFocus )
            pNotebk->setFocus();

          stopProcessingEvent = true;
        }
        /**********************************************************************/
        /* Alt+Up:  set the focus to the tab control                          */
        /**********************************************************************/
        else if ( keyevt.virtualKey() == IKeyboardEvent::up )
        {
          pNotebk->setFocus();
          stopProcessingEvent = true;
        }
      } //WM_SYSKEYDOWN
      break;

    default:
      break;
  }

  return( stopProcessingEvent );
}

/*------------------------------------------------------------------------------
| ITopPageKeyboardHandlerStatics                                               |
|                                                                              |
| Class to handle destruction of the static ITopPageKeyboardHandlerStatics     |
| object.                                                                      |
------------------------------------------------------------------------------*/
class ITopPageKeyboardHandlerStatics
{
public:
  ~ITopPageKeyboardHandlerStatics( )
  {
    if ( pDefHandler )
    {
      delete pDefHandler;
      pDefHandler = 0;
    }
  }

ITopPageKeyboardHandler
 *pDefHandler;
};

#pragma data_seg(ICLNonConst)
static ITopPageKeyboardHandlerStatics IDefTopPageKBStatics;
#pragma data_seg()

/*------------------------------------------------------------------------------
| ITopPageKeyboardHandler::defaultHandler                                      |
|                                                                              |
------------------------------------------------------------------------------*/
ITopPageKeyboardHandler* ITopPageKeyboardHandler::defaultHandler( )
{
  if ( !IDefTopPageKBStatics.pDefHandler )
    IDefTopPageKBStatics.pDefHandler = new ITopPageKeyboardHandler();

  return( IDefTopPageKBStatics.pDefHandler );
}


/*------------------------------------------------------------------------------
| INotebookData class definition                                               |
------------------------------------------------------------------------------*/
class INotebookData : public IBase
{
public:
  INotebookData  ( );

  ~INotebookData ( );

IPageClipWindow
 *pPageClipWindow;

HIMAGELIST
  hWndImageList;        // Image list window handle

long
  lMajorTabHeight;
};

/*------------------------------------------------------------------------------
| INotebookData::INotebookData                                                 |
|                                                                              |
| This class contains the                                                      |
------------------------------------------------------------------------------*/
INotebookData::INotebookData()
  : pPageClipWindow (0)
  , hWndImageList (0)
  , lMajorTabHeight (0)
{ }

/*------------------------------------------------------------------------------
| INotebookData::~INotebookData                                                |
|                                                                              |
| Destructor here for page tuning.                                             |
------------------------------------------------------------------------------*/
INotebookData::~INotebookData()
{
  // Destroy the page clipping window if it exists
  if ( this->pPageClipWindow )
    delete this->pPageClipWindow;

  // Destroy the image list if it exists
  if ( this->hWndImageList )
  {
    if ( !ImageList_Destroy( this->hWndImageList ) )
      ITHROWGUIERROR("ImageList_Destroy");
  }
}

/*------------------------------------------------------------------------------
| ITabCtrlPageSequence                                                         |
------------------------------------------------------------------------------*/
class ITabCtrlPageSequence : public IEqualitySequence< IPageHandle > {
public:
  ITabCtrlPageSequence ( );
 ~ITabCtrlPageSequence ( );
};

/*------------------------------------------------------------------------------
| ITabCtrlPageSequence::ITabCtrlPageSequence                                   |
| Constructor                                                                  |
------------------------------------------------------------------------------*/
ITabCtrlPageSequence :: ITabCtrlPageSequence( )
  : IEqualitySequence< IPageHandle >( 20 )
{
}

/*------------------------------------------------------------------------------
| ITabCtrlPageSequence::~ITabCtrlPageSequence                                  |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
ITabCtrlPageSequence :: ~ITabCtrlPageSequence( )
{
}
#endif //IC_WIN


/*------------------------------------------------------------------------------
| IPageHandle::IPageHandle                                                     |
|                                                                              |
| Class constructor                                                            |
------------------------------------------------------------------------------*/
IPageHandle :: IPageHandle ( Value pageId )
  : IHandle( pageId )
{ }


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Construct a notebook from a window id, parent, and owner window handles,     |
| and initial size rectangle, and a style.                                     |
------------------------------------------------------------------------------*/
INotebook :: INotebook ( unsigned long windowId,
                         IWindow* parent,
                         IWindow* pOwner,
                         const IRectangle& initial,
                         const Style& style )
#ifdef IC_WIN
  : fNotebookData( new INotebookData )
  , pTabCtrlPageSeqCl( 0 )
  , ulClPagesInserted( 0 )
#else
  : fNotebookData( 0 )
#endif
  , bmClTabBitmapMgr( 0 )
  , pnotebookColors( 0 )
  , colorFlags( 0 )
  , ulClValidate( 0 )
{
  // assertions on input parms
  IASSERTPARM(parent != 0);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

#ifdef __BORLANDC__
  IWindowHandle owner;
  if (pOwner == 0)
    owner = IWindowHandle(0);
  else
    owner = pOwner->handle();
#else
  IWindowHandle owner = (pOwner == 0) ? IWindowHandle(0) : pOwner->handle();
#endif

  // Default notebook window class name
  IString windowClass( WC_NOTEBOOK );

#ifdef IC_WIN
  if ( isPMCompatible() )
  {
    // Use CUA '91 notebook control.
    // If needed, load the DLL containing the control code.
    IControlStatics::loadControlDLL();
  }
  else
  {
    //Use native tab control
    InitCommonControls();
    windowClass = WC_TABCONTROL;
  }
#endif

  IWindowHandle whNB =
      this -> create( windowId,
                      0,
                      convertToGUIStyle( style ),
                      windowClass,
                      parent->handle(),
                      owner,
                      initial,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( style, true ) );

  startHandlingEventsFor( whNB );
  this -> addHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->handleEventsFor( this );

  /********************************************************************/
  /* pmCompatible style is set for CUA '91 notebook in both PM and    */
  /* Windows.  If this style is not set, treat the notebook as a      */
  /* Windows tab control.                                             */
  /********************************************************************/
  if ( isPMCompatible() )
  {
    // Set Default Major Tab Dimensions
    SHORT MajorTabWidth  = 50;
    SHORT MajorTabHeight = 30;
    setMajorTabSize( ISize(MajorTabWidth,MajorTabHeight) );

    // Set Default Minor Tab Dimensions
    SHORT MinorTabWidth  = 50;
    SHORT MinorTabHeight = 30;
    setMinorTabSize( ISize(MinorTabWidth,MinorTabHeight) );

    // Set Default DogEar Dimensions
    SHORT DogEarWidth   = 30;
    SHORT DogEarHeight  = 20;
    setPageButtonSize( ISize(DogEarWidth,DogEarHeight) );
  }
  else
  {
    //Create the page clipping window
    fNotebookData->pPageClipWindow = new IPageClipWindow( this );

    //Since we have defined our own tab control item structure, we must
    //indicate the number of extra bytes in addition to the TC_ITEMHEADER
    //structure.  The TC_ITEMHEADER structure must be the first field in
    //the new item structure.
    if ( !TabCtrl_SetItemExtra( whNB,
                                sizeof(IOC_ITEM) - sizeof(TC_ITEMHEADER) ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItemExtra");
    }

    //Create the tab page sequence collection
    pTabCtrlPageSeqCl = new ITabCtrlPageSequence( );

    //The tab control automatically sizes the tabs based upon the
    //tab text and/or bitmap size.  We do not need to set a default
    //size.
  }
}


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Construct a notebook from a window id and the window handle which serves     |
| as both parent and owner.                                                    |
------------------------------------------------------------------------------*/
INotebook :: INotebook( unsigned long windowId,
                        IWindow* parentAndOwner )
#ifdef IC_WIN
  : fNotebookData( new INotebookData )
  , pTabCtrlPageSeqCl( 0 )
  , ulClPagesInserted( 0 )
#else
  : fNotebookData( 0 )
#endif
  , bmClTabBitmapMgr( 0 )
  , pnotebookColors( 0 )
  , colorFlags( 0 )
  , ulClValidate( 0 )
{
  IASSERTPARM(parentAndOwner != 0);

  setAutoDestroyWindow(false);
  startHandlingEventsFor(windowId, parentAndOwner);
  this -> addHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

#ifdef IC_WIN
  /********************************************************************/
  /* Get the window class of the control window to determine if       */
  /* we're dealing with a tab control.                                */
  /********************************************************************/
  IWindowHandle
    hwndControl = IWindow::handleWithParent( windowId,
                                             parentAndOwner->handle() );
  if ( IWindowClassName( hwndControl ) == WC_TABCONTROL )
  {
    //Create the page clipping window
    fNotebookData->pPageClipWindow = new IPageClipWindow( this );

    //Since we have defined our own tab control item structure, we must
    //indicate the number of extra bytes in addition to the TC_ITEMHEADER
    //structure.  The TC_ITEMHEADER structure must be the first field in
    //the new item structure.
    if ( !TabCtrl_SetItemExtra( hwndControl,
                                sizeof(IOC_ITEM) - sizeof(TC_ITEMHEADER) ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItemExtra");
    }

    //Create the tab page sequence collection
    pTabCtrlPageSeqCl = new ITabCtrlPageSequence( );

    //The tab control automatically sizes the tabs based upon the
    //tab text and/or bitmap size.  We do not need to set a default
    //size, so we return.
    return;
  }
#endif

  // Set Default Major Tab Dimensions
  SHORT MajorTabWidth  = 50;
  SHORT MajorTabHeight = 30;
  setMajorTabSize( ISize(MajorTabWidth,MajorTabHeight) );

  // Set Default Minor Tab Dimensions
  SHORT MinorTabWidth  = 50;
  SHORT MinorTabHeight = 30;
  setMinorTabSize( ISize(MinorTabWidth,MinorTabHeight) );

  // Set Default DogEar Dimensions
  SHORT DogEarWidth   = 30;
  SHORT DogEarHeight  = 20;
  setPageButtonSize( ISize(DogEarWidth,DogEarHeight) );
}


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Constructor to instantiate from an existing push button control.             |
------------------------------------------------------------------------------*/
INotebook :: INotebook ( const IWindowHandle& handle )

#ifdef IC_WIN
  : fNotebookData( new INotebookData )
  , pTabCtrlPageSeqCl( 0 )
  , ulClPagesInserted( 0 )
#else
  : fNotebookData( 0 )
#endif
  , bmClTabBitmapMgr( 0 )
  , pnotebookColors( 0 )
  , colorFlags( 0 )
  , ulClValidate( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(handle);
  this -> addHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

#ifdef IC_WIN
  /********************************************************************/
  /* Get the window class of the control window to determine if       */
  /* we're dealing with a tab control.                                */
  /********************************************************************/
  if ( IWindowClassName( handle ) == WC_TABCONTROL )
  {
    //Create the page clipping window
    fNotebookData->pPageClipWindow = new IPageClipWindow( this );

    //Since we have defined our own tab control item structure, we must
    //indicate the number of extra bytes in addition to the TC_ITEMHEADER
    //structure.  The TC_ITEMHEADER structure must be the first field in
    //the new item structure.
    if ( !TabCtrl_SetItemExtra( handle,
                                sizeof(IOC_ITEM) - sizeof(TC_ITEMHEADER) ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItemExtra");
    }

    //Create the tab page sequence collection
    pTabCtrlPageSeqCl = new ITabCtrlPageSequence( );
  }
#endif
}


/*------------------------------------------------------------------------------
| INotebook::~INotebook                                                        |
|                                                                              |
| Constructor to instantiate from an existing notebook control.                |
------------------------------------------------------------------------------*/
INotebook :: ~INotebook ( )
{
  this -> removeHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->stopHandlingEventsFor(this);
  delete (pnotebookColors);

  while (bmClTabBitmapMgr)            // Remove the associated bitmap managers
  {
    ITabBitmapMgr *tmpTabBitmapMgr = bmClTabBitmapMgr;
    bmClTabBitmapMgr = bmClTabBitmapMgr->next();
    delete tmpTabBitmapMgr;
  }

#ifdef IC_WIN
  if ( !isPMCompatible() )
  {
    if (pTabCtrlPageSeqCl)
      delete pTabCtrlPageSeqCl;
  }

  if ( fNotebookData )
    delete fNotebookData;
#endif
}


/*------------------------------------------------------------------------------
| INotebook::defaultStyle                                                      |
|                                                                              |
| Return the default style for new notebook objects.                           |
------------------------------------------------------------------------------*/
INotebook::Style INotebook :: defaultStyle ( )
{
  return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| INotebook::setDefaultStyle                                                   |
|                                                                              |
| Sets the default style.                                                      |
|                                                                              |
| Note: This function is inlined on all platforms except Windows due to        |
|       variables it references not being exported.                            |
------------------------------------------------------------------------------*/
void INotebook::setDefaultStyle ( const INotebook::Style& style )
{
  currentDefaultStyle = style;
}
#endif //IC_WIN

/*------------------------------------------------------------------------------
| INotebook::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long INotebook::convertToGUIStyle ( const IBitFlag& guiStyle,
                                             Boolean bExtOnly ) const
{
  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
    // BKS_ styles use all but one of their available bits in the lower
    // word of the base GUI style.  Therefore, obtain that portion asis,
    // masking out the upper word. Four of the styles were denoted as
    // 0 (OS/2 PM), due to the lack of available bits.  We are representing
    // those as extended bits so we or the developer can accurately test
    // for them later.  Since their native values are 0, we do not have to
    // do any additional processing here.
    ulStyle |= guiStyle.asUnsignedLong() & IBKS_MASK;

#ifdef IC_WIN
    if ( !isPMCompatible() )
    {
      // Set the new styles required by the tab control
      ulStyle = TCS_TABS | TCS_SINGLELINE |
                WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE;

      // Set the multiple row tab style if requested
      ulStyle |= ( guiStyle.asExtendedUnsignedLong() &
                  allTabsVisible.asExtendedUnsignedLong() ) ? TCS_MULTILINE : 0;

      // Set the owner draw tab style if requested
      ulStyle |= ( guiStyle.asExtendedUnsignedLong() &
                handleDrawTabs.asExtendedUnsignedLong() ) ? TCS_OWNERDRAWFIXED
                                                          : 0;
    }

    ulStyle |= WS_CHILD;
#endif
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| INotebook::setBinding                                                        |
|                                                                              |
| Set the binding style of the notebook.                                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setBinding ( Binding binding )
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    unsigned long ulOldStyle = ulStyle;
    unsigned long ulExtStyle = extendedStyle();

    if (binding == spiral)
    {
      ulStyle |= spiralBinding.asUnsignedLong();
      ulExtStyle &= ~solidBinding.asExtendedUnsignedLong();
    }
    else if (binding == solid)
    {
      ulStyle &= ~spiralBinding.asUnsignedLong();
      ulExtStyle |= solidBinding.asExtendedUnsignedLong();
    }
    else
    {
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
    }

    if (ulStyle != ulOldStyle)
    {
      setStyle(ulStyle);
      refresh();

      // Update extended style ...
      setExtendedStyle(ulExtStyle);
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setOrientation                                                    |
|                                                                              |
| Set the orientation style of the notebook.                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setOrientation ( Orientation orientation )
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    unsigned long ulOldStyle = ulStyle;
    unsigned long ulMask = ~(backPagesBottomRight.asUnsignedLong()  |
                             backPagesBottomLeft.asUnsignedLong()   |
                             backPagesTopRight.asUnsignedLong()     |
                             backPagesTopLeft.asUnsignedLong()      |
                             majorTabsRight.asUnsignedLong()        |
                             majorTabsLeft.asUnsignedLong()         |
                             majorTabsTop.asUnsignedLong()          |
                             majorTabsBottom.asUnsignedLong());
    ulStyle &= ulMask;
    switch (orientation)
    {
      case backpagesBottomTabsRight:
        ulStyle |= (backPagesBottomRight.asUnsignedLong() |
                    majorTabsRight.asUnsignedLong());
        break;
      case backpagesTopTabsRight:
        ulStyle |= (backPagesTopRight.asUnsignedLong() |
                    majorTabsRight.asUnsignedLong());
        break;
      case backpagesBottomTabsLeft:
        ulStyle |= (backPagesBottomLeft.asUnsignedLong() |
                    majorTabsLeft.asUnsignedLong());
        break;
      case backpagesTopTabsLeft:
        ulStyle |= (backPagesTopLeft.asUnsignedLong() |
                    majorTabsLeft.asUnsignedLong());
        break;
      case backpagesRightTabsTop:
        ulStyle |= (backPagesTopRight.asUnsignedLong() |
                    majorTabsTop.asUnsignedLong());
        break;
      case backpagesLeftTabsTop:
        ulStyle |= (backPagesTopLeft.asUnsignedLong() |
                    majorTabsTop.asUnsignedLong());
        break;
      case backpagesRightTabsBottom:
        ulStyle |= (backPagesBottomRight.asUnsignedLong() |
                    majorTabsBottom.asUnsignedLong());
        break;
      case backpagesLeftTabsBottom:
        ulStyle |= (backPagesBottomLeft.asUnsignedLong() |
                    majorTabsBottom.asUnsignedLong());
        break;
      default:
        ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable);
        break;
    }

    if (ulStyle != ulOldStyle)
    {
      setStyle(ulStyle);
      refresh();
      notifyObservers(INotificationEvent(INotebook::orientationId,
                                         *this, true, (void*)orientation));
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setTabShape                                                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabShape ( TabShape tabShape )
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    unsigned long ulOldStyle = ulStyle;
    unsigned long ulExtStyle = extendedStyle();
    unsigned long ulMask = ~(roundedTabs.asUnsignedLong()  |
                             polygonTabs.asUnsignedLong());
    ulStyle &= ulMask;
    ulExtStyle &= ~squareTabs.asExtendedUnsignedLong();

    switch (tabShape)
    {
      case square:
        ulExtStyle |= squareTabs.asExtendedUnsignedLong();
        break;
      case rounded:
        ulStyle |= roundedTabs.asUnsignedLong();
        break;
      case polygon:
        ulStyle |= polygonTabs.asUnsignedLong();
        break;
      default:
        ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable);
        break;
    }

    if (ulStyle != ulOldStyle)
    {
      setStyle(ulStyle);
      refresh();

      // Update extended style ...
      setExtendedStyle(ulExtStyle);
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setStatusTextAlignment                                            |
|                                                                              |
| Set the alignment style of the text in the notebook's status line.           |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusTextAlignment ( TextAlignment align )
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    unsigned long ulOldStyle = ulStyle;
    unsigned long ulExtStyle = extendedStyle();
    unsigned long ulMask = ~(statusTextRight.asUnsignedLong()  |
                             statusTextCenter.asUnsignedLong());
    ulStyle &= ulMask;
    ulExtStyle &= ~statusTextLeft.asExtendedUnsignedLong();

    switch (align)
    {
      case left:
        ulExtStyle |= statusTextLeft.asExtendedUnsignedLong();
        break;
      case right:
        ulStyle |= statusTextRight.asUnsignedLong();
        break;
      case center:
        ulStyle |= statusTextCenter.asUnsignedLong();
        break;
      default:
        ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable);
        break;
    }

    if (ulStyle != ulOldStyle)
    {
      setStyle(ulStyle);
      refresh();

      // Update extended style ...
      setExtendedStyle(ulExtStyle);
    }
  }

  return( *this );
}

/*------------------------------------------------------------------------------
| INotebook::setTabTextAlignment                                               |
|                                                                              |
| Set the alignment of the text in the notebook's tabs.                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabTextAlignment ( TextAlignment align )
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    unsigned long ulOldStyle = ulStyle;
    unsigned long ulExtStyle = extendedStyle();
    unsigned long ulMask = ~(tabTextRight.asUnsignedLong()  |
                             tabTextCenter.asUnsignedLong());
    ulStyle &= ulMask;
    ulExtStyle &= ~statusTextLeft.asExtendedUnsignedLong();

    switch (align)
    {
      case left:
        ulExtStyle |= tabTextLeft.asExtendedUnsignedLong();
        break;
      case right:
        ulStyle |= tabTextRight.asUnsignedLong();
        break;
      case center:
        ulStyle |= tabTextCenter.asUnsignedLong();
        break;
      default:
        ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable);
        break;
    }

    if (ulStyle != ulOldStyle)
    {
      setStyle(ulStyle);
      refresh();

      // Update extended style ...
      setExtendedStyle(ulExtStyle);
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::binding                                                           |
------------------------------------------------------------------------------*/
INotebook::Binding INotebook :: binding ( ) const
{
  if ( isPMCompatible() )
  {
    return( (style() & spiralBinding.asUnsignedLong()) ?  INotebook::spiral
                                                       :  INotebook::solid );
  }
  else
  {
    return( INotebook::solid );                            //return default
  }
}


/*------------------------------------------------------------------------------
| INotebook::orientation                                                       |
------------------------------------------------------------------------------*/
INotebook::Orientation INotebook :: orientation ( ) const
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    INotebook::Orientation currentOrientation(backpagesRightTabsBottom);

    if (ulStyle & majorTabsRight.asUnsignedLong())
    {
      if (ulStyle & backPagesBottomRight.asUnsignedLong())
        currentOrientation = INotebook::backpagesBottomTabsRight;
      else
        currentOrientation = INotebook::backpagesTopTabsRight;
    }
    else if (ulStyle & majorTabsLeft.asUnsignedLong())
    {
      if (ulStyle & backPagesBottomLeft.asUnsignedLong())
        currentOrientation = INotebook::backpagesBottomTabsLeft;
      else
        currentOrientation = INotebook::backpagesTopTabsLeft;
    }
    else if (ulStyle & majorTabsTop.asUnsignedLong())
    {
      if (ulStyle & backPagesTopLeft.asUnsignedLong())
        currentOrientation = INotebook::backpagesLeftTabsTop;
      else
        currentOrientation = INotebook::backpagesRightTabsTop;
    }
    else if (ulStyle & majorTabsBottom.asUnsignedLong())
    {
      if (ulStyle & backPagesBottomLeft.asUnsignedLong())
        currentOrientation = INotebook::backpagesLeftTabsBottom;
      else
        currentOrientation = INotebook::backpagesRightTabsBottom;
    }
    return currentOrientation;
  }
  else
  {
    return( INotebook::backpagesRightTabsTop );
  }
}


/*------------------------------------------------------------------------------
| INotebook::tabShape                                                          |
------------------------------------------------------------------------------*/
INotebook::TabShape INotebook :: tabShape ( ) const
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    if (ulStyle & roundedTabs.asUnsignedLong())
      return( INotebook::rounded );
    else if (ulStyle & polygonTabs.asUnsignedLong())
      return( INotebook::polygon );
    else
      return( INotebook::square );
  }
  else
  {
    return( INotebook::rounded );
  }
}


/*------------------------------------------------------------------------------
| INotebook::statusTextAlignment                                               |
|                                                                              |
| Return the alignment of the status text of the notebook.                     |
------------------------------------------------------------------------------*/
INotebook::TextAlignment INotebook :: statusTextAlignment ( ) const
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    if (ulStyle & statusTextRight.asUnsignedLong())
      return( INotebook::right );
    else if (ulStyle & statusTextCenter.asUnsignedLong())
      return( INotebook::center );
    else
      return( INotebook::left );
  }
  else
  {
    return( INotebook::left );                             //return default
  }
}


/*------------------------------------------------------------------------------
| INotebook::tabTextAlignment                                                  |
|                                                                              |
| Return the alignment of the tab text of the notebook.                        |
------------------------------------------------------------------------------*/
INotebook::TextAlignment INotebook :: tabTextAlignment ( ) const
{
  if ( isPMCompatible() )
  {
    unsigned long ulStyle = style();
    if (ulStyle & tabTextRight.asUnsignedLong())
      return( INotebook::right );
    else if (ulStyle & tabTextCenter.asUnsignedLong())
      return( INotebook::center );
    else
      return( INotebook::left );
  }
  else
  {
    return( INotebook::center );
  }
}


/*------------------------------------------------------------------------------
| INotebook::setMajorTabSize                                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setMajorTabSize ( const ISize& sizeMajorTab )
{
  IMODTRACE_DEVELOP( "INotebook::setMajorTabSize" );

  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent( BKM_SETDIMENSIONS,
                                MPFROM2SHORT(sizeMajorTab.width(),
                                            sizeMajorTab.height()),
                                BKA_MAJORTAB );
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETDIMENSIONS");
    }
  }
  else
  {
    unsigned long ulStyle = style();
    if ( !(ulStyle & TCS_FIXEDWIDTH) )
    {
      setStyle( ulStyle | TCS_FIXEDWIDTH );
    }

    if ( isDrawTabsEnabled() )
    {
      fNotebookData->lMajorTabHeight = sizeMajorTab.height();
      tabControlResize( size() );
    }

    sendEvent( TCM_SETITEMSIZE,
               0,
               MPFROM2SHORT(sizeMajorTab.width(),
                            sizeMajorTab.height()) );

    /**************************************************************************/
    /* Simulate a resizing event to force repainting of the client area       */
    /**************************************************************************/
    if ( !isDrawTabsEnabled() )
      parent()->sizeTo( parent()->size() );
  }

  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setMinorTabSize                                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setMinorTabSize ( const ISize& sizeMinorTab )
{
  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent( BKM_SETDIMENSIONS,
                                MPFROM2SHORT(sizeMinorTab.width(),
                                            sizeMinorTab.height()),
                                BKA_MINORTAB );
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETDIMENSIONS");
    }
  }

  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setPageButtonSize                                                 |
|                                                                              |
| Set the size of the arrow buttons used to turn the notebook's pages, in      |
| pixels.                                                                      |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setPageButtonSize ( const ISize& sizePageButton )
{
  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent( BKM_SETDIMENSIONS,
                                MPFROM2SHORT(sizePageButton.width(),
                                             sizePageButton.height()),
                                BKA_PAGEBUTTON );
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETDIMENSIONS");
    }
  }

  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::refreshTabs                                                       |
|                                                                              |
| Causes all tabs in the notebook to be invalidated (and thus repainted).      |
------------------------------------------------------------------------------*/
INotebook& INotebook :: refreshTabs ( )
{
  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent(BKM_INVALIDATETABS, 0, 0);
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_INVALIDATETABS");
    }
  }
  else
  {
    RECT firstTab,
         lastTab,
         updateRect;

    if (( TabCtrl_GetItemRect( handle(), 0, &firstTab ) )  &&
        ( TabCtrl_GetItemRect( handle(), (totalPages() - 1), &lastTab ) ))
    {
      updateRect.left = firstTab.left;
      updateRect.top = firstTab.top;
      updateRect.right = lastTab.right;
      updateRect.bottom = lastTab.bottom;

      InvalidateRect( handle(), &updateRect, false );
    }
    else
    {
      ITHROWGUIERROR("TabCtrl_GetItemRect");
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| WIN32 Color Support                                                          |
|                                                                              |
| Windows has no API equivalent for Presentation Parameters.  The following    |
| notebook color functions are NOPed for Windows.  The query functions return  |
| an appropriate default IColor object.                                        |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| INoteBook::backgroundColor                                                   |
|                                                                              |
| Returns the background color of the notebook.                                |
------------------------------------------------------------------------------*/
IColor INotebook::backgroundColor ( ) const
{
#ifdef IC_PM
  return (IWindow::color(PP_BACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::defaultControl)));
#else
  return IGUIColor( IGUIColor::defaultControl );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| INoteBook::hiliteBackgroundColor                                             |
|                                                                              |
| Returns the hilite background color of the notebook.                         |
------------------------------------------------------------------------------*/
IColor INotebook::hiliteBackgroundColor ( ) const
{
#ifdef IC_PM
  return (IWindow::color(PP_HILITEBACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::defaultButton)));
#else
  return IGUIColor( IGUIColor::defaultButton );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| INoteBook::pageBackgroundColor                                               |
|                                                                              |
| Returns the background page color of the notebook.                           |
------------------------------------------------------------------------------*/
IColor INotebook::pageBackgroundColor ( ) const
{
  if ( isPMCompatible() )
  {
    /********************************************************************/
    /* If this color was set previously, retrieve it from where we saved*/
    /* it and return it.  Otherwise, if a color has not been previously */
    /* set, return the default.                                         */
    /********************************************************************/
    if (colorFlags & INotebook::bgnPageColor)
    {
      if (pnotebookColors)
      {
        return (UnsignedLongAsRGB(pnotebookColors->pageBackgroundColor));
      }
    }
  }

  return( IGUIColor(IGUIColor::notebookPageBgnd) );
}

/*------------------------------------------------------------------------------
| INoteBook::majorTabBackgroundColor                                           |
|                                                                              |
| Returns the major tab background color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::majorTabBackgroundColor ( ) const
{
  if ( isPMCompatible() )
  {
    /********************************************************************/
    /* If this color was set previously, retrieve it from where we saved*/
    /* it and return it.  Otherwise, if a color has not been previously */
    /* set, return the default.                                         */
    /********************************************************************/
    if (colorFlags & INotebook::bgnMajorColor)
    {
      if (pnotebookColors)
      {
        return (UnsignedLongAsRGB(pnotebookColors->majorTabBackgroundColor));
      }
    }
  }

  return( IGUIColor(IGUIColor::notebookPageBgnd) );
}

/*------------------------------------------------------------------------------
| INoteBook::minorTabBackgroundColor                                           |
|                                                                              |
| Returns the minor tab background color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::minorTabBackgroundColor ( ) const
{
  if ( isPMCompatible() )
  {
    /********************************************************************/
    /* If this color was set previously, retrieve it from where we saved*/
    /* it and return it.  Otherwise, if a color has not been previously */
    /* set, return the default.                                         */
    /********************************************************************/
    if (colorFlags & INotebook::bgnMinorColor)
    {
      if (pnotebookColors)
      {
        return (UnsignedLongAsRGB(pnotebookColors->minorTabBackgroundColor));
      }
    }
  }

  return( IGUIColor(IGUIColor::notebookPageBgnd) );
}

/*------------------------------------------------------------------------------
| INoteBook::majorTabForegroundColor                                           |
|                                                                              |
| Returns the major tab foreground color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::majorTabForegroundColor ( ) const
{
  if ( isPMCompatible() )
  {
    /********************************************************************/
    /* If this color was set previously, retrieve it from where we saved*/
    /* it and return it.  Otherwise, if a color has not been previously */
    /* set, return the default.                                         */
    /********************************************************************/
    if (colorFlags & INotebook::fgnMajorColor)
    {
      if (pnotebookColors)
      {
        return (UnsignedLongAsRGB(pnotebookColors->majorTabForegroundColor));
      }
    }
  }

  return( IGUIColor(IGUIColor::windowText) );
}

/*------------------------------------------------------------------------------
| INoteBook::minorTabForegroundColor                                           |
|                                                                              |
| Returns the minor tab foreground color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::minorTabForegroundColor ( ) const
{
  if ( isPMCompatible() )
  {
    /********************************************************************/
    /* If this color was set previously, retrieve it from where we saved*/
    /* it and return it.  Otherwise, if a color has not been previously */
    /* set, return the default.                                         */
    /********************************************************************/
    if (colorFlags & INotebook::fgnMinorColor)
    {
      if (pnotebookColors)
      {
        return (UnsignedLongAsRGB(pnotebookColors->minorTabForegroundColor));
      }
    }
  }

  return IGUIColor(IGUIColor::windowText);
}

/*------------------------------------------------------------------------------
| INoteBook::setPageBackgroundColor                                            |
|                                                                              |
| Sets the page background color in a notebook.                                |
------------------------------------------------------------------------------*/
INotebook& INotebook::setPageBackgroundColor ( const IColor &color )
{
  if ( isPMCompatible() )
  {
    setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDPAGECOLOR);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMajorTabBackgroundColor                                        |
|                                                                              |
| Sets the major tab background color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMajorTabBackgroundColor ( const IColor &color )
{
  if ( isPMCompatible() )
  {
    setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDMAJORCOLOR);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMinorTabBackgroundColor                                        |
|                                                                              |
| Sets the minor tab background color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMinorTabBackgroundColor ( const IColor &color )
{
  if ( isPMCompatible() )
  {
    setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDMINORCOLOR);
  }

  /********************************************************************/
  /* Windows tab control has no concept of minor tabs.  Therefore,    */
  /* this function is a no-op.                                        */
  /********************************************************************/
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMajorTabForegroundColor                                        |
|                                                                              |
| Sets the major tab foreground color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMajorTabForegroundColor ( const IColor &color )
{
  if ( isPMCompatible() )
  {
    setNotebookColors (color.asRGBLong(), BKA_FOREGROUNDMAJORCOLOR);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMinorTabForegroundColor                                        |
|                                                                              |
| Sets the minor tab foreground color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMinorTabForegroundColor ( const IColor &color )
{
  if ( isPMCompatible() )
  {
    setNotebookColors (color.asRGBLong(), BKA_FOREGROUNDMINORCOLOR);
  }

  /********************************************************************/
  /* Windows tab control has no concept of minor tabs.  Therefore,    */
  /* this function is a no-op.                                        */
  /********************************************************************/
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetPageBackgroundColor                                          |
|                                                                              |
| Resets the page background color by undoing a previous set.                  |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetPageBackgroundColor ( )
{
  colorFlags &= ~bgnPageColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMajorTabBackgroundColor                                      |
|                                                                              |
| Resets the major tab background color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMajorTabBackgroundColor ( )
{
  colorFlags &= ~bgnMajorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMinorTabBackgroundColor                                      |
|                                                                              |
| Resets the minor tab background color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMinorTabBackgroundColor ( )
{
  colorFlags &= ~bgnMinorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMajorTabForegroundColor                                      |
|                                                                              |
| Resets the major tab foreground color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMajorTabForegroundColor ( )
{
  colorFlags &= ~fgnMajorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMinorTabForegroundColor                                      |
|                                                                              |
| Resets the minor tab foreground color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMinorTabForegroundColor ( )
{
  colorFlags &= ~fgnMinorColor;
  return *this;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| INoteBook::areAllTabsVisible                                                 |
|                                                                              |
| Returns true if multiple rows of tabs can be displayed on the tab control.   |
------------------------------------------------------------------------------*/
Boolean INotebook::areAllTabsVisible ( ) const
{
  if ( isPMCompatible() )
  {
    return( false );
  }
  else
  {
    return( (style() & TCS_MULTILINE) ? true : false );
  }
}

/*------------------------------------------------------------------------------
| INoteBook::isDrawTabsEnabled                                                 |
|                                                                              |
| Returns true if owner draw tabs style is specified.                          |
------------------------------------------------------------------------------*/
Boolean INotebook::isDrawTabsEnabled ( ) const
{
  if ( isPMCompatible() )
  {
    return( false );
  }
  else
  {
    return( (style() & TCS_OWNERDRAWFIXED) ? true : false );
  }
}
#endif

/*------------------------------------------------------------------------------
| INoteBook::isPMCompatible                                                    |
|                                                                              |
| Returns true if the notebook is displayed as the CUA '91 notebook.           |
------------------------------------------------------------------------------*/
Boolean INotebook::isPMCompatible ( ) const
{
#ifdef IC_PM
  return( true );
#else
  return( (extendedStyle() &
           pmCompatible.asExtendedUnsignedLong()) ? true : false );
#endif
}

//
// PageSettings and Cursor classes now implemented in inotebk0.cpp
//

/*------------------------------------------------------------------------------
| INotebook::insertPageInfo                                                    |
|                                                                              |
| Set the values in the PageSettings into the notebook.                        |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: insertPageInfo ( const PageSettings& pageInfo,
                                          const IPageHandle&  referencePage,
                                          IWindow*            window,
                                          unsigned long       fPosition )
{
  IMODTRACE_DEVELOP( "INotebook::insertPageInfo" );

  IPageHandle pageHandle( 0 );

#ifdef IC_WIN
  if ( isPMCompatible() )
  {
    /**************************************************************************/
    /* Insert page in the CUA '91 notebook.                                   */
    /**************************************************************************/
    pageHandle =
        (void*)insertPage( MPFROMLONG( ((void*)referencePage) ),
                    MPFROM2SHORT( (USHORT)pageInfo.pageStyle.asUnsignedLong(),
                                  fPosition) );
  }
  else
  {
    /**************************************************************************/
    /* Handle the tab control page in a separate function to increase         */
    /* readability/maintainability of the code.                               */
    /**************************************************************************/
    return( insertPageInfo2( pageInfo, referencePage, window, fPosition ) );
  }
#else
  pageHandle =
        insertPage( MPFROMLONG( referencePage ),
                    MPFROM2SHORT( (USHORT)pageInfo.pageStyle.asUnsignedLong(),
                                  fPosition) );
#endif

  if (window && pageHandle)
  {
    IRectangle pageRect;

    /**************************************************************************/
    /* If the page window is a frame, we need to set the owner to NULL.  The  */
    /* notebook will set the parent to the page window (id 8006) and then we  */
    /* can set the owner to the notebook.  (We cannot set the owner to the    */
    /* notebook initially because this might cause the frame to have the same */
    /* parent and owner and cause problems).  If the owner of the frame is    */
    /* not the notebook, a focus loop problem will occur.                     */
    /**************************************************************************/
#ifndef IC_WIN
    if ( window->isFrameWindow() )
      window->setOwner( NULL );
#endif

    /**************************************************************************/
    /* save the page window rect                                              */
    /**************************************************************************/
    if ( pageInfo.isAutoSize() )
      pageRect = window->rect();

    /**************************************************************************/
    /* Set (associate) the application page window with the notebook page     */
    /**************************************************************************/
    Boolean
      returnBool = sendEvent( BKM_SETPAGEWINDOWHWND,
#ifdef IC_WIN
                              (unsigned long)(void*)pageHandle,
                              (unsigned long)(void*)(window->handle()) );
#else
                              (unsigned long)pageHandle,
                              (unsigned long)window->handle() );
#endif

    if (!returnBool)
    {
      ITHROWGUIERROR("BKM_SETPAGEWINDOWHWND");
    }
    else
    {
      /************************************************************************/
      /* restore the page window size                                         */
      /************************************************************************/
      if ( pageInfo.isAutoSize() )
        window->sizeTo( ISize( pageRect.width(), pageRect.height() ));

      /************************************************************************/
      /* If there is no owner for the page, set it to the notebook.           */
      /************************************************************************/
#ifndef IC_WIN
      if ( window->owner() == NULL )
        window->setOwner( this );
#endif
    }
  }

  if (pageHandle)
  {
    if (pageInfo.savedStatusText.size() != 0)
      setStatusText(pageHandle, pageInfo.savedStatusText);

    if (pageInfo.savedTabText.size() != 0)
      setTabText(pageHandle, pageInfo.savedTabText);

    if (pageInfo.savedTabBitmap != 0)
      setTabBitmap(pageHandle, pageInfo.savedTabBitmap);

    setUserData(pageHandle, pageInfo.savedUserData);
  }
  return pageHandle;
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| INotebook::insertPageInfo2                                                   |
|                                                                              |
| Set the values in the PageSettings into the notebook.                        |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: insertPageInfo2 ( const PageSettings& pageInfo,
                                           const IPageHandle&  referencePage,
                                           IWindow*            window,
                                           unsigned long       fPosition )
{
  IMODTRACE_DEVELOP( "INotebook::insertPageInfo2" );

  /****************************************************************************/
  /* Update the page styles so we can store them in our extended tab control  */
  /* item.                                                                    */
  /****************************************************************************/
  if ( pageInfo.isAutoSize() )
  {
    pageInfo.fPageSettingsData->tabCtrlItem.ulPageStyle |=
                                   PageSettings::autoPageSize.asUnsignedLong();
  }
  else
  {
    pageInfo.fPageSettingsData->tabCtrlItem.ulPageStyle &=
                                  ~PageSettings::autoPageSize.asUnsignedLong();
  }

  /****************************************************************************/
  /* Add the application page window handle and its parent's handle to our    */
  /* extended tab control item, as well as any application user-defined data  */
  /* to our extended tab control item.                                        */
  /****************************************************************************/
  if (window)
  {
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindow = window->handle();
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindowParent =
        ( window->parent() ? window->parent()->handle() : IWindowHandle(0) );
  }
  else
  {
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindow = 0;
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindowParent = 0;
  }
  pageInfo.fPageSettingsData->tabCtrlItem.ulUserData = pageInfo.userData();

  /****************************************************************************/
  /* Insert the tab page into the tab control, and associate the application  */
  /* page window to it if one is supplied.                                    */
  /****************************************************************************/
  IPageHandle
    pageHandle = insertTabPage( pageInfo, referencePage, fPosition );

  /****************************************************************************/
  /* If we have an application page window and if the tab page was inserted   */
  /* successfully then do the following:                                      */
  /****************************************************************************/
  if (window && pageHandle)
  {
    IRectangle pageRect;

    /**************************************************************************/
    /* Make the page clipping window the parent of the application page       */
    /* window.                                                                */
    /**************************************************************************/
    window->setParent( pageClippingWindow() );

    /**************************************************************************/
    /* Simulate CUA notebook's processing of the addition of the first page   */
    /* in the notebook.                                                       */
    /**************************************************************************/
    if ( pageHandle.asUnsigned() == 1 )
    {
      PAGESELECTNOTIFY pageSelectNotify;
      pageSelectNotify.hwndBook = handle();
      pageSelectNotify.ulPageIdNew = 1;
      pageSelectNotify.ulPageIdCur = 1;

      /************************************************************************/
      /* Set tab selection on the first tab in the tab control.               */
      /************************************************************************/
      processTabSelect( &pageSelectNotify, false );
    }
    else
      window->hide();
  }

  if (pageHandle)
  {
    if (pageInfo.savedTabBitmap != 0)
      setTabBitmap(pageHandle, pageInfo.savedTabBitmap);
  }

  return( pageHandle );
}
#endif


/*------------------------------------------------------------------------------
| INotebook::addFirstPage                                                      |
|                                                                              |
| Add a given window to the notebook as the first page, using the given        |
| page settings.                                                               |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addFirstPage ( const PageSettings& pageInfo,
                                        IWindow* window )
{
  return insertPageInfo(pageInfo, IPageHandle(0), window, BKA_FIRST);
}


/*------------------------------------------------------------------------------
| INotebook::addLastPage                                                       |
|                                                                              |
| Add a given window to the notebook as the last page, using the given         |
| page settings.                                                               |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addLastPage ( const PageSettings& pageInfo,
                                       IWindow* window )
{
  return insertPageInfo(pageInfo, IPageHandle(0), window, BKA_LAST);
}


/*------------------------------------------------------------------------------
| INotebook::addPageBefore                                                     |
|                                                                              |
| Add a given window to the notebook before the referenced page, using the     |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageBefore ( const PageSettings& newPageInfo,
                                         const IPageHandle& referencePage,
                                         IWindow* window )
{
  return insertPageInfo(newPageInfo, referencePage, window, BKA_PREV);
}


/*------------------------------------------------------------------------------
| INotebook::addPageBefore                                                     |
|                                                                              |
| Add a given window to the notebook before the referenced page, using the     |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageBefore ( const PageSettings& newPageInfo,
                                         const Cursor& cursor,
                                         IWindow* window )
{
  return insertPageInfo(newPageInfo, cursor.current(), window, BKA_PREV);
}


/*------------------------------------------------------------------------------
| INotebook::addPageAfter                                                      |
|                                                                              |
| Add a given window to the notebook after the referenced page, using the      |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageAfter ( const PageSettings& newPageInfo,
                                        const IPageHandle& referencePage,
                                        IWindow* window )
{
  return insertPageInfo(newPageInfo, referencePage, window, BKA_NEXT);
}


/*------------------------------------------------------------------------------
| INotebook::addPageAfter                                                      |
|                                                                              |
| Add a given window to the notebook after the referenced page, using the      |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageAfter ( const PageSettings& newPageInfo,
                                        const Cursor& cursor,
                                        IWindow* window )
{
  return insertPageInfo(newPageInfo, cursor.current(), window, BKA_NEXT);
}


/*------------------------------------------------------------------------------
| INotebook::insertPage                                                        |
------------------------------------------------------------------------------*/
unsigned long INotebook :: insertPage ( void* mp1, void* mp2 )
{
  unsigned long retVal = sendEvent(BKM_INSERTPAGE, mp1, mp2);
  if (!retVal)
  {
    ITHROWGUIERROR("BKM_INSERTPAGE");
  }
  ulClValidate++;
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::removePage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removePage ( const IPageHandle& page )
{
#ifdef IC_WIN
  deletePage( MPFROMLONG( ((void*)page)), MPFROMSHORT (BKA_SINGLE) );
#else
  deletePage (MPFROMLONG (page), MPFROMSHORT (BKA_SINGLE));
#endif

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::removePage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removePage ( const Cursor& cursor )
{
  return removePage(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::removeAllPages                                                    |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeAllPages ( )
{
  if ( isPMCompatible() )
  {
    deletePage (0, MPFROMSHORT (BKA_ALL));
  }
  else
  {
    /**************************************************************************/
    /* We cannot use the tab control's delete all functions if we want to     */
    /* send a page deleted notification.  Instead we have to loop thru        */
    /* every page to simulate the PM behavior, since the tab control does     */
    /* not have a TCN_ notification defined for page deletion.                */
    /**************************************************************************/
#if 0
    /**************************************************************************/
    /* If we have created an image list, then remove all of the images        */
    /* from it.                                                               */
    /**************************************************************************/
    if ( fNotebookData->hWndImageList )
    {
      TabCtrl_RemoveImage( handle(), (WPARAM)-1 );
    }

    /**************************************************************************/
    /* Delete all of the pages.                                               */
    /**************************************************************************/
    if ( !TabCtrl_DeleteAllItems( handle() ) )
    {
      ITHROWGUIERROR("TabCtrl_DeleteAllItems");
    }
#endif

    /**************************************************************************/
    /* Hide the page clipping window if we're deleting all of the pages.      */
    /**************************************************************************/
    pageClippingWindow()->hide();

    /**************************************************************************/
    /* This is the hack to simulate the PM page deleted notification that     */
    /* was mentioned above.                                                   */
    /**************************************************************************/
    unsigned long
      ulTotalPages = totalPages();
    for ( unsigned i = 1; i <= ulTotalPages; i++ )
    {
      deletePage( MPFROMLONG ((void*)tabPageHandle( 1 )),
                  MPFROMSHORT (BKA_ALL) );
    }
    /******************************************************************/
    /* Reset pages inserted count to 0, since we have just deleted    */
    /* all of the pages.                                              */
    /******************************************************************/
    ulClPagesInserted = 0;

    /******************************************************************/
    /* Also clear the equality sequence of all its entries            */
    /******************************************************************/
    tabPageHandleCollection()->removeAll();
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::removeTabSection                                                  |
|                                                                              |
| If the page specified is a major tab page, the specified page and all        |
| subsequent pages up to the next major tab page are removed.                  |
| If the page specified is a minor tab page, the specified page and all        |
| subsequent pages up to the next page with any tab are removed.               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeTabSection ( const IPageHandle& page )
{
  if ( isPMCompatible() )
  {
    unsigned short pageStyle = sendEvent( BKM_QUERYPAGESTYLE,
                                          MPFROMLONG( UPAGE ),
                                          0 );

    IASSERTSTATE( ( pageStyle & (BKA_MAJOR | BKA_MINOR) ) != 0 );

    deletePage (MPFROMLONG( UPAGE ), MPFROMSHORT( BKA_TAB ));
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::removeTabSection                                                  |
|                                                                              |
| If the page cursored is a major tab page, remove the cursored page and       |
| all subsequent pages up to the next major tab page.                          |
| If the page cursored is a minor tab page, remove the cursored page and       |
| all subsequent pages up to the next page with any tab.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeTabSection ( const Cursor& cursor )
{
  return removeTabSection(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::deletePage                                                        |
------------------------------------------------------------------------------*/
Boolean INotebook :: deletePage ( void* mp1, void* mp2 )
{
  if ( isPMCompatible() )
  {
    Boolean retVal = true;

    retVal = sendEvent(BKM_DELETEPAGE, mp1, mp2);
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_DELETEPAGE");
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /* ---------------------------------------------------------------------- */
    /* Query the tab item's information                                       */
    /**************************************************************************/
    IOC_ITEM tabCtrlItem;
    tabCtrlItem.itemHeader.mask = TCIF_PARAM | TCIF_IMAGE;

    long
      iPageIndex = tabPageIndex( mp1 );

    if ( !TabCtrl_GetItem( handle(), iPageIndex, &tabCtrlItem ) )
    {
      if ( (unsigned long)mp2 == BKA_SINGLE )
        ITHROWGUIERROR("TabCtrl_GetItem");
      else
        return( false );
    }

    /**************************************************************************/
    /* If we have created an image list, then check to see if an image exists */
    /* for the specified page.  If an image exists, then remove it from the   */
    /* image list.                                                            */
    /**************************************************************************/
    IMAGEINFO imageInfo;
    if ( fNotebookData->hWndImageList )
    {
      if ( tabCtrlItem.itemHeader.iImage != -1 )
      {
        /**********************************************************************/
        /* Get the bitmap's handle before we delete it from the image list,   */
        /* so we can pass it as part of the page deleted notification.        */
        /**********************************************************************/
        ImageList_GetImageInfo( fNotebookData->hWndImageList,
                                tabCtrlItem.itemHeader.iImage,
                                &imageInfo );

        TabCtrl_RemoveImage( handle(), tabCtrlItem.itemHeader.iImage );
      }
    }

    /**************************************************************************/
    /* If the delete request is for the current top page, we must select a    */
    /* new top page, unless the notebook is empty.                            */
    /**************************************************************************/
    if ( ((unsigned long)mp2 == BKA_SINGLE) &&
         (IPageHandle( mp1 ) == topPage()) )
    {
      /************************************************************************/
      /* Reset the top page to the next page if it exists, or the previous    */
      /* page if it exists.                                                   */
      /************************************************************************/
      long
        iNewPageIndex = iPageIndex + 1;
      long
        retVal = TabCtrl_SetCurSel( handle(), iNewPageIndex );
      if ( retVal == -1 )
      {
        if ( iPageIndex )
        {
          iNewPageIndex = iPageIndex - 1;
          retVal = TabCtrl_SetCurSel( handle(), iNewPageIndex );
        }
      }

      if ( retVal != -1 )
      {
        /**********************************************************************/
        /* Notify the parent (i.e. owner) that a new page is now the top      */
        /* page. because the current top page is being deleted.               */
        /**********************************************************************/
        PAGESELECTNOTIFY pageSelectNotify;
        pageSelectNotify.hwndBook = handle();
        pageSelectNotify.ulPageIdNew =
                           tabPageHandle( iNewPageIndex + 1 ).asUnsigned();
        pageSelectNotify.ulPageIdCur = (long)mp1;

        processTabSelect( &pageSelectNotify, false );
      }
    }

    /**************************************************************************/
    /* Delete the specified page.                                             */
    /**************************************************************************/
    if ( !TabCtrl_DeleteItem( handle(), iPageIndex ) )
    {
      ITHROWGUIERROR("TabCtrl_DeleteItem");
    }

    /**************************************************************************/
    /* Send the page deleted notification to simulate PM behavior             */
    /**************************************************************************/
    DELETENOTIFY deleteNotify;
    deleteNotify.hwndBook = handle();
    deleteNotify.hwndPage = tabCtrlItem.hPageWindow;
    deleteNotify.ulAppPageData = tabCtrlItem.ulUserData;
    if ( fNotebookData->hWndImageList )
      deleteNotify.hbmTab = imageInfo.hbmImage;
    else
      deleteNotify.hbmTab = NULL;

    parent()->handle().sendEvent( WM_COMMAND,
                                  IEventParameter1( (unsigned short)id(),
                                                    BKN_PAGEDELETED ),
                                  IEventParameter2( &deleteNotify ) );

    /**************************************************************************/
    /* Reset the parent of the application page window for the tab page we    */
    /* just deleted.  We only want to do this if this if the application page */
    /* isn't associated with any other tab page.                              */
    /**************************************************************************/
    Boolean bResetParent = true;
    if ( (unsigned long)mp2 == BKA_SINGLE )
    {
      IOC_ITEM tempTabCtrlItem;
      unsigned long
        ulTotalPages = totalPages();
      for ( unsigned i = 0; i < ulTotalPages; i++ )
      {
        if ( TabCtrl_GetItem( handle(), i, &tempTabCtrlItem ) )
        {
          if ( tempTabCtrlItem.hPageWindow == tabCtrlItem.hPageWindow )
          {
            bResetParent = false;
            break;
          }
        }
      }
    }

    /**************************************************************************/
    /* If we didn't find a page (other than the one being freed) that has the */
    /* same application page window as the one being freed, reset the         */
    /* application page window's parent to its original parent.               */
    /**************************************************************************/
    if ( bResetParent )
    {
      IWindow* pPageWindow = windowWithHandle( tabCtrlItem.hPageWindow );
      if ( pPageWindow )
      {
        SetParent( tabCtrlItem.hPageWindow, tabCtrlItem.hPageWindowParent );
        pPageWindow->hide();
      }
    }

    /**************************************************************************/
    /* Remove the tab control page handle from our tab page sequence          */
    /* collection.                                                            */
    /**************************************************************************/
    if ( (unsigned long)mp2 == BKA_SINGLE )
      removeTabPage( mp1 );
  } //tab control

  ulClValidate--;
  return( true );
}


/*------------------------------------------------------------------------------
| INotebook::turnToPage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: turnToPage ( const IPageHandle& page )
{
  IMODTRACE_DEVELOP( "INotebook::turnToPage" );

  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent(BKM_TURNTOPAGE, UPAGE, 0);
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_TURNTOPAGE");
    }
  }
  else
  {
    /**************************************************************************/
    /* Verify that pages exist in the Windows tab control.                    */
    /**************************************************************************/
    if ( !totalPages() )
    {
      return( *this );
    }
    else
    {
      /************************************************************************/
      /* Fill in the PAGESELECTNOTIFY structure that we use in Windows        */
      /************************************************************************/
      PAGESELECTNOTIFY pageSelectNotify;
      pageSelectNotify.hwndBook = handle();
      pageSelectNotify.ulPageIdNew =  page.asUnsigned();
      pageSelectNotify.ulPageIdCur = (DWORD)topPage().asUnsigned();

      /************************************************************************/
      /* The tab selection is now in progress (pending).  Do our required     */
      /* processing and issue the selection pending notification.  If the     */
      /* new page selection is rejected, cancel the page turn request.        */
      /************************************************************************/
      if ( processTabSelect( &pageSelectNotify, true ) )
        return( *this );

      /************************************************************************/
      /* Set the index of the tab page we're turning to.                      */
      /************************************************************************/
      long
        iPageIndex = tabPageIndex( page );
      long
        retVal = TabCtrl_SetCurSel( handle(), iPageIndex );
      if ( retVal == -1 )
      {
        ITHROWGUIERROR("TabCtrl_SetCurSel");
      }

      /************************************************************************/
      /* The tab selection has already occurred and the new tab is now the    */
      /* top page in the tab control.  Do our required processing and issue   */
      /* the selection notification.                                          */
      /************************************************************************/
      processTabSelect( &pageSelectNotify, false );

    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::turnToPage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: turnToPage ( const Cursor& cursor )
{
  return turnToPage(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::topPage                                                           |
|                                                                              |
| Returns an IPageHandle object that references the current top page of the    |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: topPage ( ) const
{
  IPageHandle topPageHandle (0);

  if ( isPMCompatible() )
  {
    /**************************************************************************/
    /* CUA '91 notebook                                                       */
    /**************************************************************************/
    unsigned long
      retVal = queryPage (0, MPFROM2SHORT (BKA_TOP, 0));
    if (retVal)
#ifdef IC_WIN
      topPageHandle = (void*)retVal;
#else
      topPageHandle = retVal;
#endif
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control                                                    */
    /**************************************************************************/
    long
      iPageIndex = TabCtrl_GetCurSel( handle() );
    if ( iPageIndex == -1 )
    {
      ITHROWGUIERROR("TabCtrl_GetCurSel");
    }

    /**************************************************************************/
    /* We must increment the index since the tab control uses a 0-based index */
    /* and sequence collection positions are 1-based.                         */
    /**************************************************************************/
    topPageHandle = tabPageHandle( iPageIndex + 1 );
  }

  return( topPageHandle );
}


/*------------------------------------------------------------------------------
| INotebook::firstPage                                                         |
|                                                                              |
| Returns an IPageHandle object that references the first page of the notebook.|
------------------------------------------------------------------------------*/
IPageHandle INotebook :: firstPage ( ) const
{
  unsigned long retVal;
  IPageHandle firstPageHandle (0);

  if ( isPMCompatible() )
  {
    retVal = queryPage (0, MPFROM2SHORT (BKA_FIRST, 0));
    if (retVal)
#ifdef IC_WIN
      firstPageHandle = (void*)retVal;
#else
      firstPageHandle = retVal;
#endif
  }
  else
  {
    /**************************************************************************/
    /* Value of the first page is always the first element in the tab page    */
    /* sequence collection.                                                   */
    /**************************************************************************/
    if ( totalPages() )
      firstPageHandle = tabPageHandle( 1 );
  }

  return( firstPageHandle );
}


/*------------------------------------------------------------------------------
| INotebook::lastPage                                                          |
|                                                                              |
| Returns an IPageHandle object that references the last page of the notebook. |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: lastPage ( ) const
{
  unsigned long retVal;
  IPageHandle lastPageHandle (0);

  if ( isPMCompatible() )
  {
    retVal = queryPage (0, MPFROM2SHORT (BKA_LAST, 0));
    if (retVal)
#ifdef IC_WIN
      lastPageHandle = (void*)retVal;
#else
      lastPageHandle = retVal;
#endif
  }
  else
  {
    /**************************************************************************/
    /* Value of the last page is always the last element in the tab page      */
    /* sequence collection.                                                   */
    /**************************************************************************/
    lastPageHandle = tabPageHandle( totalPages() );
  }

  return( lastPageHandle );
}


/*------------------------------------------------------------------------------
| INotebook::nextPage                                                          |
|                                                                              |
| Returns an IPageHandle object that references the next page of the notebook. |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: nextPage ( const IPageHandle& page ) const
{
  IPageHandle nextPageHandle (0);

  if (page)
  {
    if ( isPMCompatible() )
    {
      unsigned long retVal = queryPage( MPFROMLONG( UPAGE ),
                                        MPFROM2SHORT( BKA_NEXT, 0 ));
      if (retVal)
#ifdef IC_WIN
        nextPageHandle = (void*)retVal;
#else
        nextPageHandle = retVal;
#endif
    }
    else
    {
      /************************************************************************/
      /* Verify the reference page.  If it does not exist, throw an exception.*/
      /************************************************************************/
      ICursor *cursor = tabPageHandleCollection()->newCursor();
      if ( tabPageHandleCollection()->locate( page, *cursor ) )
      {
        /**********************************************************************/
        /* Value of the next page is always the position of the reference     */
        /* page's element + 1 (in the tab page sequence collection).          */
        /**********************************************************************/
        unsigned long
          ulPosition = tabPageHandleCollection()->position( *cursor ) + 1;
        if ( ulPosition <= tabPageHandleCollection()->numberOfElements() )
        {
          nextPageHandle = tabPageHandle( ulPosition );
        }
      }
      else
      {
        ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                            IBaseErrorInfo::invalidParameter,
                            IException::recoverable);
      }
    }
  }

  return( nextPageHandle );
}


/*------------------------------------------------------------------------------
| INotebook::previousPage                                                      |
|                                                                              |
| Returns an IPageHandle object that references the page preceding the         |
| referenced page.                                                             |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: previousPage ( const IPageHandle& page ) const
{
  IPageHandle prevPageHandle (0);

  if (page)
  {
    if ( isPMCompatible() )
    {
      unsigned long retVal = queryPage( MPFROMLONG( UPAGE ),
                                        MPFROM2SHORT( BKA_PREV, 0 ));
      if (retVal)
#ifdef IC_WIN
        prevPageHandle = (void*)retVal;
#else
        prevPageHandle = retVal;
#endif
    }
    else
    {
      /************************************************************************/
      /* Verify the reference page.  If it does not exist, throw an exception.*/
      /************************************************************************/
      ICursor *cursor = tabPageHandleCollection()->newCursor();
      if ( tabPageHandleCollection()->locate( page, *cursor ) )
      {
        /**********************************************************************/
        /* Value of the previous page is always the position of the reference */
        /* page's element - 1 (in the tab page sequence collection).          */
        /**********************************************************************/
        unsigned long
          ulPosition = tabPageHandleCollection()->position( *cursor ) - 1;
        if ( ulPosition >= 1 )
        {
          prevPageHandle = tabPageHandle( ulPosition );
        }
      }
      else
      {
        ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                            IBaseErrorInfo::invalidParameter,
                            IException::recoverable);
      }
    }
  }

  return( prevPageHandle );
}


/*------------------------------------------------------------------------------
| INotebook::queryPage                                                         |
------------------------------------------------------------------------------*/
unsigned long INotebook :: queryPage ( void* mp1, void* mp2 ) const
{
  long retVal = (long)sendEvent(BKM_QUERYPAGEID, mp1, mp2);

  if (retVal == BOOKERR_INVALID_PARAMETERS)
  {
    ITHROWGUIERROR("BKM_QUERYPAGEID:INVALID_PARAMETERS");
  }
  return (unsigned long)retVal;
}


/*------------------------------------------------------------------------------
| INotebook::window                                                            |
|                                                                              |
| Returns the window used by the notebook page.                                |
------------------------------------------------------------------------------*/
IWindow* INotebook :: window ( const INotebook::Cursor& cursor ) const
{
  return window(cursor.current());     // Return window for current page
}


/*------------------------------------------------------------------------------
| INotebook::window                                                            |
|                                                                              |
| Returns the window used by the notebook page.                                |
------------------------------------------------------------------------------*/
IWindow* INotebook :: window ( const IPageHandle& page ) const
{
  IWindow* pwinReturn = 0;             // Assume no window

  if ( isPMCompatible() )
  {
    IWindowHandle wh =
#ifdef IC_WIN
      (void*)sendEvent(BKM_QUERYPAGEWINDOWHWND, UPAGE, 0);
    if (((long)(unsigned long)(void*)wh) == BOOKERR_INVALID_PARAMETERS)
#else
      (unsigned long)sendEvent(BKM_QUERYPAGEWINDOWHWND, UPAGE, 0);
    if (((long)(unsigned long)wh) == BOOKERR_INVALID_PARAMETERS)
#endif
    {
      ITHROWGUIERROR("BKM_QUERYPAGEWINDOWHWND");
    }
    else if (wh != 0)
    {
      pwinReturn = windowWithHandle(wh);
    }
  }
  else
  {
    /**************************************************************************/
    /* Locate the notebook page.  If it does not exist, throw an exception.   */
    /**************************************************************************/
    ICursor *cursor = tabPageHandleCollection()->newCursor();
    if ( tabPageHandleCollection()->locate( page, *cursor ) )
    {
      /************************************************************************/
      /* Get the pointer to the application page window from the extended tab */
      /* control item.                                                        */
      /************************************************************************/
      IOC_ITEM tabCtrlItem;
      tabCtrlItem.itemHeader.mask = TCIF_PARAM;
      if ( TabCtrl_GetItem( handle(), tabPageIndex( page ), &tabCtrlItem ) )
      {
        pwinReturn = windowWithHandle( tabCtrlItem.hPageWindow );
      }
      else
      {
        ITHROWGUIERROR("TabCtrl_GetItem");
      }
    }
    else
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }
  }

  return( pwinReturn );
}


/*------------------------------------------------------------------------------
| INotebook::pageSettings                                                      |
|                                                                              |
| Create and return a PageSettings object representing the page.               |
------------------------------------------------------------------------------*/
INotebook::PageSettings INotebook :: pageSettings
                          ( const INotebook::Cursor& cursor ) const
{
   return pageSettings(cursor.current());
}

/*------------------------------------------------------------------------------
| INotebook::pageSettings                                                      |
|                                                                              |
| Create and return a PageSettings object representing the page.               |
------------------------------------------------------------------------------*/
INotebook::PageSettings INotebook :: pageSettings
                          ( const IPageHandle& page ) const
{
  INotebook::PageSettings newPageSettings (0);
  char bookBuffer [100];

  if ( isPMCompatible() )
  {
    /**************************************************************************/
    /* CUA '91 notebook processing                                            */
    /**************************************************************************/
    BOOKTEXT bookText;

    unsigned long ulStyle =
           (unsigned long)sendEvent(BKM_QUERYPAGESTYLE, UPAGE, 0);

    if (ulStyle & PageSettings::autoPageSize.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::autoPageSize);
    if (ulStyle & PageSettings::statusTextOn.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::statusTextOn);
    if (ulStyle & PageSettings::majorTab.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::majorTab);
    if (ulStyle & PageSettings::minorTab.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::minorTab);

    newPageSettings.savedUserData =
                  sendEvent(BKM_QUERYPAGEDATA, UPAGE, 0);

    bookText.textLen = sizeof (bookBuffer);
    bookText.pString = bookBuffer;
    short usRetVal   = sendEvent( BKM_QUERYSTATUSLINETEXT,
                                  UPAGE,
                                  &bookText );

    if (usRetVal && (usRetVal != (short)BOOKERR_INVALID_PARAMETERS))
      newPageSettings.savedStatusText = IString (bookBuffer);

    usRetVal = sendEvent(BKM_QUERYTABTEXT, UPAGE, &bookText);
    if (usRetVal && (usRetVal != (short)BOOKERR_INVALID_PARAMETERS))
      newPageSettings.savedTabText = IString (bookBuffer);

    HBITMAP hTabBitmap =
                 sendEvent(BKM_QUERYTABBITMAP, UPAGE, 0);
#ifdef IC_WIN
    if (hTabBitmap && (hTabBitmap != (void*)BOOKERR_INVALID_PARAMETERS))
#else
    if (hTabBitmap && ((long)hTabBitmap != BOOKERR_INVALID_PARAMETERS))
#endif
    {
      newPageSettings.savedTabBitmap = IBitmapHandle (hTabBitmap);
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /* ---------------------------------------------------------------------- */
    /* Locate the notebook page.  If it does not exist, throw an exception.   */
    /**************************************************************************/
    ICursor *cursor = tabPageHandleCollection()->newCursor();
    if ( !tabPageHandleCollection()->locate( page, *cursor ) )
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }

    /**************************************************************************/
    /* Get tab text for the tab page and any user-defined data.  Get page     */
    /* settings attributes from the extended tab control item.                */
    /**************************************************************************/
    IOC_ITEM tabCtrlItem;
    tabCtrlItem.itemHeader.mask = TCIF_TEXT | TCIF_IMAGE | TCIF_PARAM;
    tabCtrlItem.itemHeader.pszText = bookBuffer;
    tabCtrlItem.itemHeader.cchTextMax = sizeof(bookBuffer);
    if ( !TabCtrl_GetItem( handle(), tabPageIndex( page ), &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_GetItem");
    }

    if ( tabCtrlItem.ulPageStyle & PageSettings::autoPageSize.asUnsignedLong() )
      newPageSettings.pageStyle |= (PageSettings::autoPageSize);
    if ( tabCtrlItem.ulPageStyle & PageSettings::statusTextOn.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::statusTextOn);
    if ( tabCtrlItem.ulPageStyle & PageSettings::majorTab.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::majorTab);
    if ( tabCtrlItem.ulPageStyle & PageSettings::minorTab.asUnsignedLong())
      newPageSettings.pageStyle |= (PageSettings::minorTab);

    newPageSettings.savedUserData = tabCtrlItem.ulUserData;

    if ( tabCtrlItem.itemHeader.cchTextMax )
      newPageSettings.savedTabText = IString( tabCtrlItem.itemHeader.pszText );

    /**************************************************************************/
    /* Get tab bitmap for the tab page.                                       */
    /**************************************************************************/
    if ( fNotebookData->hWndImageList )
    {
      IMAGEINFO imageInfo;
      if ( ImageList_GetImageInfo( fNotebookData->hWndImageList,
                                   tabCtrlItem.itemHeader.iImage,
                                   &imageInfo ) )
      {
        newPageSettings.savedTabBitmap = IBitmapHandle( imageInfo.hbmImage );
      }
      else
      {
        ITHROWGUIERROR("ImageList_GetImageInfo");
      }
    }
  } //isPMCompatible()

  return newPageSettings;
}


/*------------------------------------------------------------------------------
| INotebook::setStatusText                                                     |
|                                                                              |
| Change the status text of the specified page in the notebook.                |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusText ( const IPageHandle& page,
                                        const char* text )
{
  PageSettings settings = this->pageSettings( page );

  if ( ! settings.isStatusTextOn() )
  {           // Ensure the page has a status text area.
     ITHROWLIBRARYERROR( IC_NOSTATUSTEXTLINE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }
  else if ( settings.statusText() != text )
  {
    if ( isPMCompatible() )
    {
      IString strText( text );         //Avoid compiler warning!
      Boolean retVal =
         this->sendEvent( BKM_SETSTATUSLINETEXT, UPAGE, (char*)strText );
      if ( ! retVal )
      {
         ITHROWGUIERROR( "BKM_SETSTATUSLINETEXT" );
      }
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setStatusText                                                     |
|                                                                              |
| Change the status text of the specified page in the notebook.                |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusText ( const IPageHandle& page,
                                          const IResourceId& resId )
{
  return setStatusText(page,
                       resId.resourceLibrary().loadString(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setTabText                                                        |
|                                                                              |
| Change the tab text of the specified notebook page.                          |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabText ( const IPageHandle& page,
                                     const char* text )
{
  IString strText( text );               //Avoid compiler warning!

  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent( BKM_SETTABTEXT, UPAGE, (char *)strText );
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETTABTEXT");
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /* ---------------------------------------------------------------------- */
    /* Locate the notebook page.  If it does not exist, throw an exception.   */
    /**************************************************************************/
    ICursor *cursor = tabPageHandleCollection()->newCursor();
    if ( !tabPageHandleCollection()->locate( page, *cursor ) )
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }

    /**************************************************************************/
    /* Set tab text for the tab page.                                         */
    /**************************************************************************/
    IOC_ITEM tabCtrlItem;
    tabCtrlItem.itemHeader.mask = TCIF_TEXT;
    tabCtrlItem.itemHeader.pszText = strText;
    if ( !TabCtrl_SetItem( handle(), tabPageIndex( page ), &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItem");
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setTabText                                                        |
|                                                                              |
| Change the tab text of the specified notebook page.                          |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabText ( const IPageHandle& page,
                                     const IResourceId& resId )
{
  return setTabText(page,
                    resId.resourceLibrary().loadString(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setTabBitmap                                                      |
|                                                                              |
| Change the tab bitmap for the specified notebook page.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabBitmap ( const IPageHandle& page,
                                       const IBitmapHandle& bitmap )
{
  Boolean retVal = true;

  if ( isPMCompatible() )
  {
    retVal = sendEvent( BKM_SETTABBITMAP, UPAGE,
#ifdef IC_WIN
                        (unsigned long)(void*)bitmap);
#else
                        (unsigned long)bitmap);
#endif

    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETTABBITMAP");
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /* ---------------------------------------------------------------------- */
    /* Locate the notebook page.  If it does not exist, throw an exception.   */
    /**************************************************************************/
    ICursor *cursor = tabPageHandleCollection()->newCursor();
    if ( !tabPageHandleCollection()->locate( page, *cursor ) )
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }

    /**************************************************************************/
    /* Get icon/bitmap dimensions.                                            */
    /**************************************************************************/
    long lCXIcon = GetSystemMetrics( SM_CXICON );
    long lCYIcon = GetSystemMetrics( SM_CYICON );

    /**************************************************************************/
    /* We must use an image list per the tab control's requirements.  Create  */
    /* it if it does not exist.                                               */
    /**************************************************************************/
    if ( !fNotebookData->hWndImageList )
    {
      /************************************************************************/
      /* Create the image list to initially contain 10 bitmaps and grow it    */
      /* by an increment of 10 whenever the current increment list size is    */
      /* exceeded.  Also, set the bitmap to the system icon size.             */
      /************************************************************************/
      fNotebookData->hWndImageList =
        ImageList_Create( (int)lCXIcon, (int)lCYIcon, ILC_COLOR, 10, 10 );

      if ( !fNotebookData->hWndImageList )
        ITHROWGUIERROR("ImageList_Create");

      /************************************************************************/
      /* Set the tab control's image list.                                    */
      /************************************************************************/
      TabCtrl_SetImageList( handle(), fNotebookData->hWndImageList );
    }

    /**************************************************************************/
    /* Set the image index in the tab control item that identifies the        */
    /* bitmap's index in the image list.  Query the tab control item for      */
    /* the specified page.                                                    */
    /**************************************************************************/
    long iPageIndex = tabPageIndex( page );
    IOC_ITEM tabCtrlItem;
    tabCtrlItem.itemHeader.mask = TCIF_PARAM | TCIF_IMAGE;
    if ( !TabCtrl_GetItem( handle(), iPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_GetItem");
    }

    /**************************************************************************/
    /* Use an IGBitmap object to size the bitmap to the system icon size.     */
    /**************************************************************************/
    IGBitmap bmpImage( bitmap );
    bmpImage.sizeTo( ISize( lCXIcon, lCYIcon ) );

    /**************************************************************************/
    /* If the image for the specified page does not exist, add the bitmap     */
    /* to the image list.                                                     */
    /**************************************************************************/
    if ( tabCtrlItem.itemHeader.iImage == -1 )
    {
      long
        retVal = ImageList_Add( fNotebookData->hWndImageList,
                                bmpImage.handle(),
                                NULL );

      //If add was not successful, throw an exception
      if ( retVal == -1 )
      {
        ITHROWGUIERROR("ImageList_Add");
      }

      /************************************************************************/
      /* Use the index returned from the add to update the tab control item.  */
      /************************************************************************/
      tabCtrlItem.itemHeader.iImage = (int)retVal;
    }
    else
    {
      /************************************************************************/
      /* If the image for the specified page exists,  replace it with the     */
      /* specified bitmap.                                                    */
      /************************************************************************/
      if ( !ImageList_Replace( fNotebookData->hWndImageList,
                               tabCtrlItem.itemHeader.iImage,
                               bmpImage.handle(),
                               NULL ) )
      {
        ITHROWGUIERROR("ImageList_Replace");
      }
    }

    /**************************************************************************/
    /* Set the image index in the tab control item that identifies the        */
    /* bitmap's index in the image list.                                      */
    /**************************************************************************/
    if ( !TabCtrl_SetItem( handle(), iPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItem");
    }
  }

  ITabBitmapMgr *tmpBitmapMgr = new ITabBitmapMgr(bitmap, page);
  tmpBitmapMgr->setNext(bmClTabBitmapMgr);
  bmClTabBitmapMgr = tmpBitmapMgr;
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setTabBitmap                                                      |
|                                                                              |
| Change the tab bitmap for the specified notebook page.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabBitmap ( const IPageHandle& page,
                                       const IResourceId& resId )
{
  return setTabBitmap(page,
                      resId.resourceLibrary().loadBitmap(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setUserData                                                       |
|                                                                              |
| Change the user data in the specified notebook page.                         |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setUserData ( const IPageHandle& page,
                                      unsigned long ulData )
{
  if ( isPMCompatible() )
  {
    Boolean retVal = sendEvent( BKM_SETPAGEDATA, UPAGE, ulData );
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETPAGEDATA");
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /* ---------------------------------------------------------------------- */
    /* Locate the notebook page.  If it does not exist, throw an exception.   */
    /**************************************************************************/
    ICursor *cursor = tabPageHandleCollection()->newCursor();
    if ( !tabPageHandleCollection()->locate( page, *cursor ) )
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }

    /**************************************************************************/
    /* Get the extended tab item data for the tab page.                       */
    /**************************************************************************/
    long iPageIndex = tabPageIndex( page );
    IOC_ITEM tabCtrlItem;
    tabCtrlItem.itemHeader.mask = TCIF_PARAM;
    if ( !TabCtrl_GetItem( handle(), iPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_GetItem");
    }

    /**************************************************************************/
    /* Set the user data for the tab page.                                    */
    /**************************************************************************/
    tabCtrlItem.ulUserData = ulData;
    if ( !TabCtrl_SetItem( handle(), iPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItem");
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setWindow                                                         |
|                                                                              |
| Associates the specified page with the window.                               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setWindow ( const IPageHandle& page,
                                    IWindow*           window )
{
  IMODTRACE_DEVELOP( "INotebook::setWindow" );

  /****************************************************************************/
  /* Throw assertion if window is 0                                           */
  /****************************************************************************/
  IASSERTPARM(window != 0);

  INotebook::PageSettings pageInfo = this->pageSettings( page );

  if ( isPMCompatible() )
  {
    /**************************************************************************/
    /* CUA '91 notebook processing                                            */
    /**************************************************************************/

#ifndef IC_WIN
    /****************************************************************************/
    /* If the page window is a frame, we need to set the owner to NULL.  The  */
    /* notebook will set the parent to the page window (id 8006) and then we  */
    /* can set the owner to the notebook.  (We cannot set the owner to the    */
    /* notebook initially because this might cause the frame to have the same */
    /* parent and owner and cause problems).  If the owner of the frame is    */
    /* not the notebook, a focus loop problem will occur.                     */
    /**************************************************************************/
    if ( window->isFrameWindow() )
      window->setOwner( NULL );
#endif

#ifdef IC_WIN
    unsigned long temphwnd = 0;
    if (window)
      temphwnd = (unsigned long)(void*)window->handle();
#else
    unsigned long temphwnd = (window ? (unsigned long)window->handle() : 0);
#endif
    Boolean returnBool = sendEvent(BKM_SETPAGEWINDOWHWND, UPAGE, temphwnd );

    if (!returnBool)
    {
      ITHROWGUIERROR("BKM_SETPAGEWINDOWHWND");
    }
    else
    {
#ifndef IC_WIN
       /***********************************************************************/
       /* If there is no owner for the page, set it to the notebook.          */
       /***********************************************************************/
       if ( window->owner() == NULL )
         window->setOwner( this );
#endif
    }
  }
  else
  {
    /**************************************************************************/
    /* Windows tab control processing                                         */
    /**************************************************************************/
    if ( pageInfo.isAutoSize() )
    {
      pageInfo.fPageSettingsData->tabCtrlItem.ulPageStyle |=
                                  PageSettings::autoPageSize.asUnsignedLong();
    }
    else
    {
      pageInfo.fPageSettingsData->tabCtrlItem.ulPageStyle &=
                                 ~PageSettings::autoPageSize.asUnsignedLong();
    }

    /**************************************************************************/
    /* Add the application page window handle and its parent's handle to our  */
    /* extended tab control item in addition to any user-defined data.        */
    /**************************************************************************/
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindow = window->handle();
    pageInfo.fPageSettingsData->tabCtrlItem.hPageWindowParent =
        ( window->parent() ? window->parent()->handle() : IWindowHandle(0) );
    pageInfo.fPageSettingsData->tabCtrlItem.ulUserData = pageInfo.userData();

    pageInfo.fPageSettingsData->tabCtrlItem.itemHeader.mask = TCIF_PARAM;
    if ( !TabCtrl_SetItem( handle(),
                           tabPageIndex( page ),
                           &pageInfo.fPageSettingsData->tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_SetItem");
    }

    /**************************************************************************/
    /* Make the page clipping window the parent of the application page       */
    /* window.                                                                */
    /**************************************************************************/
    window->setParent( pageClippingWindow() );

    /**************************************************************************/
    /* Size and show the application page window if it is on the top page.    */
    /**************************************************************************/
    if ( page == topPage() )
    {
      /************************************************************************/
      /* Get the tab control's client area                                    */
      /************************************************************************/
      RECTL tabCtrlRect = rect().asRECTL();

      /************************************************************************/
      /* Calculate the display rectangle based upon the tab                   */
      /* control's client area, adjusted                                      */
      /************************************************************************/
      adjustRect( false, &tabCtrlRect );

      /************************************************************************/
      /* Show the top page window.                                            */
      /************************************************************************/
      showTopPage( tabPageIndex( page ), &tabCtrlRect );
    }
    else
    {
      /************************************************************************/
      /* Hide the application window since it is not on the top page.          */
      /************************************************************************/
      window->hide();
    }
  }

  return( *this );
}


/*------------------------------------------------------------------------------
| INotebook::setWindow                                                         |
|                                                                              |
| Associates the specified page with the window.                               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setWindow ( const Cursor&  cursor,
                                    IWindow*       window )
{
  return setWindow(cursor.current(), window);
}


/*------------------------------------------------------------------------------
| INotebook::notebookSize                                                      |
|                                                                              |
| Return the size of the notebook needed to display a page.                    |
------------------------------------------------------------------------------*/
ISize INotebook :: notebookSize ( const IPageHandle& page ) const
{
  IWindow *pageWindow = this->window( page );
  RECTL pageSize = { 0, 0, 100, 100 };

  if ( pageWindow )
  {
    IRectangle pageRect = pageWindow->rect();
    if ( pageRect.size() == ISize() )
    {        // Use minimum size instead.
      pageRect.sizeTo( pageWindow->minimumSize() );
    }
    pageSize = pageRect.asRECTL();
  }

  if ( isPMCompatible() )
  {
    Boolean res = handle().sendEvent( BKM_CALCPAGERECT,
                                      &pageSize, false );
    if ( res == false )
    {
      ITHROWGUIERROR( "BKM_CALCPAGERECT" );
    }
  }
  else
  {
    adjustRect( true, &pageSize );
  }

  return( ISize( pageSize ) );
}

/*------------------------------------------------------------------------------
| INotebook::pageSize                                                          |
|                                                                              |
| Return the size of the "viewport" in which the notebook draws the pages      |
| it contains.                                                                 |
------------------------------------------------------------------------------*/
ISize INotebook :: pageSize ( ) const
{
  RECTL notebookSize = rect().asRECTL();

  if ( isPMCompatible() )
  {
    Boolean res = handle().sendEvent(BKM_CALCPAGERECT,
                                     &notebookSize, true);
    if (res == false)
    {
       ITHROWGUIERROR("BKM_CALCPAGERECT");
    }
  }
  else
  {
    adjustRect( false, &notebookSize );
  }

  return( ISize( notebookSize ) );
}


/*------------------------------------------------------------------------------
| INotebook::totalPages                                                        |
|                                                                              |
| Return the total number of pages in the notebook.                            |
------------------------------------------------------------------------------*/
unsigned long INotebook :: totalPages ( ) const
{
  unsigned long ulRetVal;

  if ( isPMCompatible() )
  {
    ulRetVal = (unsigned long)queryPageCount( 0, MPFROMSHORT (BKA_END) );
  }
  else
  {
    ulRetVal = (unsigned long)TabCtrl_GetItemCount( handle() );
  }

  return( ulRetVal );
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMajorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a major tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMajorTab ( const IPageHandle& page ) const
{
  unsigned long ulRetVal = 1;

  if ( isPMCompatible() )
  {
    ulRetVal = (unsigned long)queryPageCount( MPFROMLONG( UPAGE ),
                     MPFROMSHORT( PageSettings::majorTab.asUnsignedLong()) );
  }

  return( ulRetVal );
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMajorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a major tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMajorTab ( const Cursor& cursor) const
{
  return pagesToMajorTab(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMinorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a minor tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMinorTab ( const IPageHandle& page ) const
{
  unsigned long ulRetVal = 0;

  if ( isPMCompatible() )
  {
    ulRetVal = (unsigned long)queryPageCount( MPFROMLONG( UPAGE ),
                      MPFROMSHORT( PageSettings::minorTab.asUnsignedLong()) );
  }

  return( ulRetVal );
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMinorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a minor tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMinorTab ( const Cursor& cursor ) const
{
  return pagesToMinorTab(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::pagesToEnd                                                        |
|                                                                              |
| Return the number of pages between the specified page and the end of the     |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToEnd ( const IPageHandle& page ) const
{
  unsigned long ulRetVal = 0;

  if ( isPMCompatible() )
  {
    ulRetVal = (unsigned long)queryPageCount( MPFROMLONG (UPAGE),
                                              MPFROMSHORT (BKA_END) );
  }
  else
  {
    long lEnd = totalPages() - tabPageIndex( page );
    if ( lEnd >= 0 )
    {
      ulRetVal = lEnd;
    }
    else
    {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
    }
  }

  return( ulRetVal );
}


/*------------------------------------------------------------------------------
| INotebook::pagesToEnd                                                        |
|                                                                              |
| Return the number of pages between the specified page and the end of the     |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToEnd ( const Cursor& cursor ) const
{
  return pagesToEnd(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::isEmpty                                                           |
|                                                                              |
| Queries whether the notebook has any pages.                                  |
------------------------------------------------------------------------------*/
Boolean INotebook :: isEmpty ( ) const
{
  if ( totalPages() )
    return false;
  else
    return true;
}


/*------------------------------------------------------------------------------
| INotebook::queryPageCount                                                    |
|                                                                              |
| Queries whether the notebook has any pages.                                  |
------------------------------------------------------------------------------*/
short INotebook :: queryPageCount ( void* mp1, void* mp2 ) const
{
  short retVal = sendEvent(BKM_QUERYPAGECOUNT, mp1, mp2);

  if (retVal == BOOKERR_INVALID_PARAMETERS)
  {
     ITHROWGUIERROR("BKM_QUERYPAGECOUNT");
  }
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::calcMinimumSize                                                   |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The height is      |
| based on the minimum size of the superclass control, with the notebook       |
| drawing around it.                                                           |
------------------------------------------------------------------------------*/
ISize INotebook :: calcMinimumSize ( ) const
{
  /****************************************************************************/
  /* Get notebook rectangle                                                   */
  /****************************************************************************/
  RECTL pageRect = this->rect().asRECTL();
  ISize szLargestBook( 0, 0 );

  /****************************************************************************/
  /* Determine the largest notebook size which is determined by the largest   */
  /* application page window.                                                 */
  /****************************************************************************/
  for ( IPageHandle pageHandle = this->firstPage();
        pageHandle;
        pageHandle = this->nextPage( pageHandle ))
  {
    ISize
      szNoteBook = this->notebookSize( pageHandle );
    szLargestBook = szLargestBook.maximum( szNoteBook );
  }

  /****************************************************************************/
  /* If either the width or height is 0, then we must use the control's API   */
  /* to calculate the minimum size.                                           */
  /****************************************************************************/
  if ( (szLargestBook.height() == 0) ||
       (szLargestBook.width()  == 0) )
  {
    ISize
      pageSize = this->Inherited::calcMinimumSize();   // Default min size.

    Boolean bRetVal = true;

    if ( isPMCompatible() )
    {
      bRetVal = this->sendEvent( BKM_CALCPAGERECT, &pageRect, true );
    }
    else
    {
      adjustRect( true, &pageRect );
    }

    if ( bRetVal )
    {
      RECTL newPageRect =
#ifdef IC_WIN
               IRectangle( IPoint( pageRect.left, pageRect.bottom ),
                           pageSize ).asRECTL();
#else
               IRectangle( IPoint( pageRect.xLeft, pageRect.yBottom ),
                           pageSize ).asRECTL();
#endif

      bRetVal = true;

      if ( isPMCompatible() )
      {
        bRetVal = this->sendEvent( BKM_CALCPAGERECT, &newPageRect, false );
      }
      else
      {
        adjustRect( false, &newPageRect );
      }

      if ( bRetVal )
      {
        szLargestBook = ISize( newPageRect );
      }
      else
      {
        ITHROWGUIERROR( "BKM_CALCPAGERECT" );
      }
    }
    else
    {
      ITHROWGUIERROR( "BKM_CALCPAGERECT" );
    }
  }

  return( szLargestBook );
}

/*------------------------------------------------------------------------------
| INotebook::setNotebookColors                              *** private ***    |
|                                                                              |
| This function handles the setting of the notebook colors that are not        |
| identified by system defined PP_* values.  The notebook has some special     |
| color areas defined that can only be set using a specific notebook message.  |
| In addition, the presentation manager notebook does not provide a way to     |
| query these colors.                                                          |
------------------------------------------------------------------------------*/
void INotebook::setNotebookColors ( unsigned long color,
                                    unsigned long colorArea )
{
  Boolean colorChanged = false;

  /********************************************************************/
  /* Check to see if the current color is different from the color we */
  /* are trying to set.                                               */
  /********************************************************************/
  switch (colorArea)
  {
    case BKA_BACKGROUNDPAGECOLOR:
      if (pageBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_BACKGROUNDMAJORCOLOR:
      if (majorTabBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_BACKGROUNDMINORCOLOR:
      if (minorTabBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_FOREGROUNDMAJORCOLOR:
      if (majorTabForegroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_FOREGROUNDMINORCOLOR:
      if (minorTabForegroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;
  }

  if (colorChanged)
  {
    /* Set the color on the notebook. */
    Boolean retVal = sendEvent(BKM_SETNOTEBOOKCOLORS, color, colorArea);
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETNOTEBOOKCOLORS");
    }
  }
}

/*------------------------------------------------------------------------------
| INotebook::storeNotebookColors                              *** private ***  |
|                                                                              |
| This function handles the storing of the notebook colors that are not        |
| identified by system defined PP_* values.  This enables the colors to be     |
| querying later since the system notebook does not provide a query mechanism  |
| of its own.                                                                  |
------------------------------------------------------------------------------*/
void INotebook::storeNotebookColors ( unsigned long color,
                                      unsigned long colorArea )
{
  if (!pnotebookColors)
  {
    pnotebookColors = new NotebookColors;
  }

  /*******************************************************************/
  /* Set a flag to indicate which color we are setting and store the */
  /* color we set so it can be queried later.                        */
  /*******************************************************************/
  if (colorArea == BKA_BACKGROUNDPAGECOLOR)
  {
    pnotebookColors->pageBackgroundColor = color;
    colorFlags |= INotebook::bgnPageColor;
  }
  else if (colorArea == BKA_BACKGROUNDMAJORCOLOR)
  {
    pnotebookColors->majorTabBackgroundColor = color;
    colorFlags |= INotebook::bgnMajorColor;
  }
  else if (colorArea == BKA_BACKGROUNDMINORCOLOR)
  {
    pnotebookColors->minorTabBackgroundColor = color;
    colorFlags |= INotebook::bgnMinorColor;
  }
  else if (colorArea == BKA_FOREGROUNDMAJORCOLOR)
  {
    pnotebookColors->majorTabForegroundColor = color;
    colorFlags |= INotebook::fgnMajorColor;
  }
  else if (colorArea == BKA_FOREGROUNDMINORCOLOR)
  {
    pnotebookColors->minorTabForegroundColor = color;
    colorFlags |= INotebook::fgnMinorColor;
  }
}

/*------------------------------------------------------------------------------
| INotebook::UnsignedLongAsRGB                            *** private ***      |
|                                                                              |
| This function converts an unsigned long to an RGB value and returns it as    |
| an IColor object.                                                            |
------------------------------------------------------------------------------*/
IColor INotebook::UnsignedLongAsRGB ( unsigned long color ) const
{
  /****************************************************************************/
  /* Convert the long color to an RGB value then to an IColor object.         */
  /*                                                                          */
  /* Note: Remember that the calculations for red and blue are flipped in     */
  /*       Windows.                                                           */
  /****************************************************************************/
#ifdef IC_MOTIFPM
  unsigned char blue  = (unsigned char)(color);
  unsigned char red   = (unsigned char)(color >> 16);
#else
  unsigned char red   = (unsigned char)(color);
  unsigned char blue  = (unsigned char)(color >> 16);
#endif //IC_MOTIFPM

  unsigned char green = (unsigned char)(color >> 8);
  return IColor(red, green, blue);
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| INotebook::pageClippingWindow                           *** private ***      |
|                                                                              |
| This function returns the page clipping window that is used to prevent the   |
| application page window from drawing over the tab control.                   |
| using the autoPageSize attribute.                                            |
------------------------------------------------------------------------------*/
IWindow* INotebook::pageClippingWindow ( void ) const
{
  return( fNotebookData->pPageClipWindow );
}


/*------------------------------------------------------------------------------
| INotebook::processTabSelect                             *** private ***      |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean INotebook :: processTabSelect ( void* pPageSelNotify,
                                        Boolean bSelectPending ) const
{
  IMODTRACE_DEVELOP( "INotebook::processTabSelect" );

  PAGESELECTNOTIFY* pPSN = (PAGESELECTNOTIFY *)pPageSelNotify;
  long iNewPageIndex = tabPageIndex( (void*)pPSN->ulPageIdNew );
  long iCurrentPageIndex = tabPageIndex( (void*)pPSN->ulPageIdCur );

  if (bSelectPending)
  {
    /**************************************************************************/
   /* Simulate the CUA '91 notebook's select pending notification             */
    /**************************************************************************/
    return( (Boolean)parent()->handle().sendEvent(
                                  WM_COMMAND,
                                  IEventParameter1( (unsigned short)id(),
                                                    BKN_PAGESELECTEDPENDING ),
                                  IEventParameter2( pPSN ) ) );
  }
  else
  {
    /**************************************************************************/
    /* Simulate the CUA '91 notebook's page selection notification            */
    /**************************************************************************/
    parent()->handle().sendEvent( WM_COMMAND,
                                  IEventParameter1( (unsigned short)id(),
                                                    BKN_PAGESELECTED ),
                                  IEventParameter2( pPSN ) );
  }

  /****************************************************************************/
  /* Initialize variables we'll use later.                                    */
  /****************************************************************************/
  IOC_ITEM tabCtrlItem;
  tabCtrlItem.itemHeader.mask = TCIF_PARAM;
  Boolean bChangeFocus = false;

  /****************************************************************************/
  /* If the current page index equals the new page index, then we are         */
  /* processing the selection notification that occurs when the first page    */
  /* is added to the notebook.  If this is the case, do not hide the          */
  /* application window for the current page.                                 */
  /****************************************************************************/
  if ( iCurrentPageIndex != iNewPageIndex )
  {
    /**************************************************************************/
    /* Get the application page window's index for the current tab page.      */
    /**************************************************************************/
    if ( !TabCtrl_GetItem( handle(), iCurrentPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_GetItem");
    }

    /**************************************************************************/
    /* Hide the application page window for the current tab page if the       */
    /* application page window exists.  Check to see if it has the focus      */
    /* before hiding as we will use this indicator later.                     */
    /**************************************************************************/
    IWindow* pOldPageWindow = windowWithHandle( tabCtrlItem.hPageWindow );
    if ( pOldPageWindow )
    {
      bChangeFocus = pOldPageWindow->hasFocus();
      pOldPageWindow->hide();

      ITopPageKeyboardHandler::defaultHandler()->stopHandlingEventsFor(
                                                              pOldPageWindow );
    }
  }

  /****************************************************************************/
  /* Get the tab control's client area                                        */
  /****************************************************************************/
  RECTL tabCtrlRect = rect().asRECTL();

  /****************************************************************************/
  /* Calculate the display rectangle based upon the tab                       */
  /* control's client area, adjusted                                          */
  /****************************************************************************/
  adjustRect( false, &tabCtrlRect );

  /****************************************************************************/
  /* Show the new top page window.                                            */
  /****************************************************************************/
  showTopPage( iNewPageIndex, &tabCtrlRect );

  /****************************************************************************/
  /* If the focus was on the old application page window, set it to the new   */
  /* application page window if it exists.                                    */
  /****************************************************************************/
  if (bChangeFocus)
  {
    /**************************************************************************/
    /* Get the tab item for the newly selected tab page.                      */
    /**************************************************************************/
    tabCtrlItem.itemHeader.mask = TCIF_PARAM;
    if ( !TabCtrl_GetItem( handle(), iNewPageIndex, &tabCtrlItem ) )
    {
      ITHROWGUIERROR("TabCtrl_GetItem");
    }

    IWindow* pNewPageWindow = windowWithHandle( tabCtrlItem.hPageWindow );
    if ( pNewPageWindow )
    {
      pNewPageWindow->setFocus();
    }
  }

  return( true );
}


/*------------------------------------------------------------------------------
| INotebook::showTopPage                                  *** private ***      |
|                                                                              |
| Show the top page.                                                           |
------------------------------------------------------------------------------*/
void INotebook::showTopPage( long iPageIndex, void* pRect,
                             Boolean bNotResizing ) const
{
  IMODTRACE_DEVELOP( "INotebook::showTopPage" );

  /****************************************************************************/
  /* Get the tab control item for the currently selected page.                */
  /****************************************************************************/
  IOC_ITEM tabCtrlItem;
  tabCtrlItem.itemHeader.mask = TCIF_PARAM;
  if ( !TabCtrl_GetItem( handle(), iPageIndex, &tabCtrlItem ) )
  {
    ITHROWGUIERROR("TabCtrl_GetItem");
  }

  RECT tabCtrlRect;

  /****************************************************************************/
  /* Position, size, and show the page clipping window.  We position and      */
  /* size it here as well as in the private resize handler to handle the      */
  /* the tab control's placement on a canvas.  If the tab control is placed   */
  /* on a canvas, then there is a possibility that a resize event is not      */
  /* generated.  Hence, we resize here so both the page clipping window and   */
  /* the application page window are shown.  Note that we must re-calculate   */
  /* and re-adjust the display rectangle because of this dilemna.             */
  /****************************************************************************/
  if ( !pageClippingWindow()->isVisible() )
  {
    /**************************************************************************/
    /* Get the tab control's client area                                      */
    /**************************************************************************/
    RECTL tempRectl = rect().asRECTL();

    /**************************************************************************/
    /* Calculate the display rectangle based upon the tab                     */
    /* control's client area, adjusted                                        */
    /**************************************************************************/
    SetRect( &tabCtrlRect, 0, 0,
             (int)(tempRectl.right - tempRectl.left),
             (int)(tempRectl.bottom - tempRectl.top) );

    adjustRect( false, &tabCtrlRect );

    SetWindowPos( pageClippingWindow()->handle(),
                  NULL, (int)tabCtrlRect.left, (int)tabCtrlRect.top,
                  (int)(tabCtrlRect.right - tabCtrlRect.left),
                  (int)(tabCtrlRect.bottom - tabCtrlRect.top),
                  SWP_NOACTIVATE );

    pageClippingWindow()->show();
  }
  else
  {
    CopyRect( &tabCtrlRect, (LPRECT)pRect );
  }

  /****************************************************************************/
  /* If the application page window exists, position and size it to fit the   */
  /* tab control's display area, as long as the autoPageSize attribute is     */
  /* specified.  Also, show it as well.                                       */
  /****************************************************************************/
  IWindow* pPageWindow = windowWithHandle( tabCtrlItem.hPageWindow );
  if ( pPageWindow )
  {
    if ( tabCtrlItem.ulPageStyle &
                     INotebook::PageSettings::autoPageSize.asUnsignedLong() )
    {
      SetWindowPos( tabCtrlItem.hPageWindow,
                    NULL, 0, 0,
                    (int)(tabCtrlRect.right - tabCtrlRect.left),
                    (int)(tabCtrlRect.bottom - tabCtrlRect.top),
                    0 );
    }

    if ( !pPageWindow->isVisible() )
      pPageWindow->show();

    if ( bNotResizing )
      ITopPageKeyboardHandler::defaultHandler()->handleEventsFor( pPageWindow );
  }
  else
  {
    /**************************************************************************/
    /* Otherwise, invalidate the page clipping window so it is repainted.     */
    /**************************************************************************/
    pageClippingWindow()->refresh();
  }
}

/*------------------------------------------------------------------------------
| INotebook::adjustRect                                   *** private ***      |
|                                                                              |
| Adjust the display rectangle.                                                |
------------------------------------------------------------------------------*/
void INotebook::adjustRect ( Boolean bTabCtrlSize, void* pRect ) const
{
  IMODTRACE_DEVELOP( "INotebook::adjustRect" );

  if ( IRectangle( *(LPRECTL)pRect ) != IRectangle() )
  {
    if ( isDrawTabsEnabled()  &&  !bTabCtrlSize )
    {
      /************************************************************************/
      /* if owner draw tabs, then we must calculate the display area to       */
      /* avoid problems that occur when the tab is resized.                   */
      /************************************************************************/
      long lTabHeight;
      if ( fNotebookData->lMajorTabHeight )
        lTabHeight = fNotebookData->lMajorTabHeight + 4;
      else
      {
        RECT tabRect;
        TabCtrl_GetItemRect( handle(), 0, &tabRect );
        lTabHeight = tabRect.top + tabRect.bottom;
      }

      ((LPRECTL)pRect)->top += lTabHeight;
      ((LPRECTL)pRect)->left += 4;
      ((LPRECTL)pRect)->right -= 4;
      ((LPRECTL)pRect)->bottom -= 4;
    }
    else
    {
      TabCtrl_AdjustRect( handle(), bTabCtrlSize, (LPRECTL)pRect );
    }
  }
}

/*------------------------------------------------------------------------------
| INotebook::tabControlResize                             *** private ***      |
|                                                                              |
| Resize the tab control, page clipping, and application page windows.         |
------------------------------------------------------------------------------*/
void INotebook::tabControlResize ( const ISize& newSize ) const
{
  IMODTRACE_DEVELOP( "INotebook::tabControlResize" );

  /****************************************************************************/
  /* Calculate the display rectangle, assuming the tab control is the size    */
  /* of the client area                                                       */
  /****************************************************************************/
  RECT tabCtrlRect;
  SetRect( &tabCtrlRect, 0, 0,
           (int)newSize.width(),
           (int)newSize.height() );

  adjustRect( false, &tabCtrlRect );

  /****************************************************************************/
  /* Position and size the page clipping window, and show it if it is not     */
  /* visible.                                                                 */
  /****************************************************************************/
  SetWindowPos( pageClippingWindow()->handle(),
                NULL, (int)tabCtrlRect.left, (int)tabCtrlRect.top,
                (int)(tabCtrlRect.right - tabCtrlRect.left),
                (int)(tabCtrlRect.bottom - tabCtrlRect.top),
                SWP_NOACTIVATE );

  if ( !pageClippingWindow()->isVisible() )
    pageClippingWindow()->show();

  /****************************************************************************/
  /* We are finished if there are no pages in the notebook.                   */
  /****************************************************************************/
  if ( !totalPages() )
  {
    return;
  }

  /****************************************************************************/
  /* Get the index for the currently selected tab page.                       */
  /****************************************************************************/
  long
    iPageIndex = TabCtrl_GetCurSel( handle() );
  if ( iPageIndex == -1 )
  {
    ITHROWGUIERROR("TabCtrl_GetCurSel");
  }

  /****************************************************************************/
  /* Show the top page window.                                                */
  /****************************************************************************/
  showTopPage( iPageIndex, &tabCtrlRect, false );

  /****************************************************************************/
  /* Send the new page size notification to simulate PM behavior              */
  /****************************************************************************/
  parent()->handle().sendEvent( WM_COMMAND,
                                IEventParameter1( (unsigned short)id(),
                                                  BKN_NEWPAGESIZE ),
                                IEventParameter2( handle() ) );
}

/*------------------------------------------------------------------------------
| INotebook::tabPageHandleCollection                      *** private ***      |
|                                                                              |
| Returns a pointer to the collection of tab control page handles.             |
------------------------------------------------------------------------------*/
ITabCtrlPageSequence* INotebook :: tabPageHandleCollection( ) const
{
  return( this->pTabCtrlPageSeqCl );
}

/*------------------------------------------------------------------------------
| INotebook::tabPageIndex                                 *** private ***      |
|                                                                              |
| Returns the tab control page handle at the given position.                   |
------------------------------------------------------------------------------*/
long INotebook :: tabPageIndex( const IPageHandle& pageHandle ) const
{
  long iPageIndex = -1;

  ICursor *cursor = tabPageHandleCollection()->newCursor();
  if ( tabPageHandleCollection()->locate( pageHandle, *cursor ) )
  {
    /**************************************************************************/
    /* Position is 1-based, and tab page indicies are 0-based, so adjust.     */
    /**************************************************************************/
    iPageIndex = tabPageHandleCollection()->position( *cursor ) - 1;
  }
  return( iPageIndex );
}

/*------------------------------------------------------------------------------
| INotebook::tabPageHandle                                *** private ***      |
|                                                                              |
| Returns the tab control page handle at the given position.                   |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: tabPageHandle( unsigned long ulPosition ) const
{
  return( tabPageHandleCollection()->elementAtPosition( ulPosition ) );
}

/*------------------------------------------------------------------------------
| INotebook::insertTabPage                                *** private ***      |
|                                                                              |
| Adds the tab control page handle at the given position.  Store the page      |
| handle as the element.  The page handle is unique, as we assign a unique     |
| value to each page as it is added to the notebook.  Since we add the page    |
| in the same location in both the tab control and in this equality sequence,  |
| we can locate the page handle's position in the sequence and use the         |
| position as the index of the page to the actual tab control.  However,       |
| positions in the equality sequence are 1-based and the tab control's page    |
| index is 0-based, so we must remember to adjust the position and index       |
| where appropriate.                                                           |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: insertTabPage( const PageSettings& pageInfo,
                                        const IPageHandle& referencePage,
                                        unsigned long ulPosition )
{
  unsigned long ulIndex;

  /****************************************************************************/
  /* Create a unique page handle for each page that is inserted.  Never       */
  /* decrement ulClPagesInserted (unless an exception is thrown).             */
  /****************************************************************************/
  ulClPagesInserted++;
  IPageHandle pageHandle( (void*)ulClPagesInserted  );

  switch( ulPosition )
  {
    case BKA_FIRST:
    {
      tabPageHandleCollection()->addAsFirst( pageHandle );
      ulIndex = 0;
      break;
    }

    case BKA_PREV:
    {
      ICursor *cursor = tabPageHandleCollection()->newCursor();
      if ( tabPageHandleCollection()->locate( referencePage, *cursor ) )
      {
        tabPageHandleCollection()->addAsPrevious( pageHandle, *cursor );
        ulIndex = tabPageHandleCollection()->position( *cursor ) - 1;
      }
      else
      {
        ulClPagesInserted--;

        //Throw exception
        ITHROWLIBRARYERROR(IC_INDEX_OUT_OF_RANGE,
            IBaseErrorInfo::invalidRequest,
            IException::recoverable);
      }

      break;
    }

    case BKA_NEXT:
    {
      ICursor *cursor = tabPageHandleCollection()->newCursor();
      if ( tabPageHandleCollection()->locate( referencePage, *cursor ) )
      {
        tabPageHandleCollection()->addAsNext( pageHandle, *cursor );
        ulIndex = tabPageHandleCollection()->position( *cursor ) - 1;
      }
      else
      {
        ulClPagesInserted--;

        //Throw exception
        ITHROWLIBRARYERROR(IC_INDEX_OUT_OF_RANGE,
            IBaseErrorInfo::invalidRequest,
            IException::recoverable);
      }

      break;
    }

    case BKA_LAST:
    {
      tabPageHandleCollection()->addAsLast( pageHandle );
      ulIndex = totalPages();
      break;
    }

    default:
      ulClPagesInserted--;

      //Throw exception
      ITHROWLIBRARYERROR(IC_INDEX_OUT_OF_RANGE,
          IBaseErrorInfo::invalidRequest,
          IException::recoverable);
      break;
  } // end switch

  /****************************************************************************/
  /* Insert the tab control page at the given index.  Please note that since  */
  /* we have extended the tab control item, we must include the TCIF_PARAM    */
  /* style to indicate that additional information is contained within it.    */
  /****************************************************************************/
  pageInfo.fPageSettingsData->tabCtrlItem.itemHeader.mask |= TCIF_PARAM;

  if ( TabCtrl_InsertItem( handle(), ulIndex,
                           &pageInfo.fPageSettingsData->tabCtrlItem ) == -1 )
  {
    ulClPagesInserted--;

    //If insert was not successful, throw an exception
    ITHROWGUIERROR("TCM_INSERTITEM");
  }

  ulClValidate++;

  return( pageHandle );
}

/*------------------------------------------------------------------------------
| INotebook::removeTabPage                                *** private ***      |
|                                                                              |
| Removes the tab control page handle at the given position.                   |
------------------------------------------------------------------------------*/
Boolean INotebook :: removeTabPage( const IPageHandle& pageHandle )
{
  ICursor *cursor = tabPageHandleCollection()->newCursor();
  if ( tabPageHandleCollection()->locate( pageHandle, *cursor ) )
  {
    tabPageHandleCollection()->removeAt( *cursor );
  }
//  else
//  {
//    Throw exception
//  }
  return( true );
}

#endif  //IC_WIN
