#ifndef _ISCROLL_
  #define _ISCROLL_
/*******************************************************************************
* FILE NAME: iscroll.hpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*    IScrollBar                                                                *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <icontrol.hpp>
#include <irect.hpp>

class IRange;
class IWindowHandle;
class IScrollBarData;
class IColor;
class ISize;

#pragma pack(4)

class IScrollBar : public IControl {
typedef IControl
  Inherited;

public:
class Style;
/*------------------------------- Constructors -------------------------------*/
  IScrollBar ( unsigned long     id,
               IWindow*          parent,
               IWindow*          owner,
               const IRectangle& initial = IRectangle(),
               const Style&      style = defaultStyle() );

  IScrollBar ( unsigned long     id,
               IWindow*          parent,
               IWindow*          owner,
               const IRange&     scrollableItems,
               unsigned long     visibleItemCount,
               unsigned long     scrollBoxPosition = 1,
               const IRectangle& initial = IRectangle(),
               const Style&      style = defaultStyle() );

  IScrollBar ( unsigned long     id,
               IWindow*          parent );

  IScrollBar ( const IWindowHandle& handle );

virtual
 ~IScrollBar ( );

/*---------------------------------- Styles ----------------------------------*/
INESTEDBITFLAGCLASSDEF2(Style, IScrollBar, IWindow, IControl);
static const Style
  IC_IMPORTU horizontal,
  IC_IMPORTU vertical,
#ifndef IC_WIN
  IC_IMPORTU autoSize,
#endif
  IC_IMPORTU classDefaultStyle;

static Style
  defaultStyle      ( );

static void
  setDefaultStyle   ( const Style&    style );

#ifdef IC_PMWIN
virtual unsigned long
  convertToGUIStyle ( const IBitFlag& style,
                      Boolean         extendedOnly = false ) const;
#endif

/*------------------------------ System Values -------------------------------*/
static unsigned long
  systemScrollBarWidth     ( Boolean verticalScrollBar = true ),
  systemScrollBoxLength    ( Boolean verticalScrollBar = true ),
  systemScrollButtonLength ( Boolean verticalScrollBar = true );

/*--------------------------- Query Operations -------------------------------*/
Boolean
  isHorizontal ( ) const,
  isVertical   ( ) const;

/*------------------------- Manipulation Operations --------------------------*/
virtual IScrollBar
 &setScrollableRange ( const IRange& minMax ),
 &setVisibleCount    ( unsigned long scrollableRangeUnits ),
 &setScrollBar       ( const IRange& scrollableRange,
                       unsigned long visibleCount );
virtual IScrollBar
 &moveScrollBoxTo    ( unsigned long firstItem = 1 );

#ifdef IC_MOTIF
virtual IScrollBar
 &setPrevScrollBoxPosition ( unsigned long value );
#endif

virtual IRange
  scrollableRange     ( ) const,
  scrollBoxRange      ( ) const;
virtual unsigned long
  visibleCount        ( ) const,
  scrollBoxPosition   ( ) const;

#ifdef IC_MOTIF
virtual unsigned long
  prevScrollBoxPosition ( ) const;
#endif

/*--------------------------- Scrolling Increment ----------------------------*/
virtual unsigned long
  minScrollIncrement     ( ) const,
  pageScrollIncrement    ( ) const;
virtual IScrollBar
 &setMinScrollIncrement  ( unsigned long scrollableRangeUnits = 1 ),
 &setPageScrollIncrement ( unsigned long scrollableRangeUnits = 0 );

/*---------------------------------- Colors ----------------------------------*/
#ifndef IC_WIN_FLAGNOP
virtual IColor
  foregroundColor       () const,
  hiliteForegroundColor () const;
#endif

/*-------------------------------- Overrides ---------------------------------*/
#ifdef IC_MOTIF
virtual IScrollBar
  &setForegroundColor       ( const IColor &color),
  &setHiliteForegroundColor ( const IColor &color);

virtual IScrollBar
  &resetForegroundColor       (),
  &resetHiliteForegroundColor ();

virtual IScrollBar
  &moveSizeTo( const IRectangle &rect );

virtual IPoint
  position( ) const;

virtual ISize
  size( ) const;
#endif

/*----------------------- Notification Event Descriptions --------------------*/
static INotificationId const
  IC_IMPORTU scrollBoxPositionId;

/*----------------------- Observer Notification ------------------------------*/
virtual IScrollBar
 &enableNotification ( Boolean enable=true );


protected:
/*--------------------------- Protected Overrides ----------------------------*/
virtual ISize
  calcMinimumSize ( ) const;

/*---------------------- Event-Handling Implementation -----------------------*/
#ifdef IC_MOTIF
virtual void
  registerCallbacks   ( ),
  unregisterCallbacks ( );
virtual Boolean
  passEventToOwner    ( IEvent& event );
#endif


private:
/*------------------------------ Hidden Members ------------------------------*/
  IScrollBar            ( const IScrollBar& scrollBar );
IScrollBar
 &operator=             ( const IScrollBar& scrollBar );

/*--------------------------------- Private ----------------------------------*/
friend class IScrollBarData;
#ifdef IC_WIN
friend class IScrollBarHandler;
#endif //IC_WIN

#ifdef IC_PMWIN
void
  initialize( unsigned long     windowId,
              IWindow*          parent,
              IWindow*          owner,
              const IRectangle& initial,
              const IRange&     scrollableItems,
              unsigned long     visibleItemCount,
              unsigned long     scrollBoxPosition,
              const Style&      style );

unsigned long
  ulClMinScrollIncrement,
  ulClPageScrollIncrement;
#endif

#ifdef IC_WIN
void
  initializeControlData ( const IRange& scrollableItems,
                          unsigned long visibleItemCount,
                          struct tagSCROLLINFO* controlData );
#endif
#ifdef IC_PM
void
  initializeControlData ( const IRange& scrollableItems,
                          unsigned long visibleItemCount,
                          struct _SBCDATA* controlData ),
  windowData            ( struct _WNDPARAMS* windowData ) const;
#endif

static Style
  currentDefaultStyle;

IScrollBarData
 *fScrollBarData;

}; // IScrollBar

INESTEDBITFLAGCLASSFUNCS(Style, IScrollBar);

#pragma pack()

#ifndef I_NO_INLINES
  #include <iscroll.inl>
#endif

#endif /* _ISCROLL_ */
