/*******************************************************************************
* FILE NAME: isplitcv.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in isplitcv.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_GPILOGCOLORTABLE
  #define INCL_GPIPRIMITIVES
  #define INCL_WINFRAMEMGR
  #define INCL_WININPUT
  #define INCL_WINMESSAGEMGR
  #define INCL_WINPOINTERS
  #define INCL_WINSYS
  #define INCL_WINTRACKRECT
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <isplitch.hpp>
#include <isplitpc.hpp>
#include <isplitcv.hpp>

#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _ICOLOR_
  #include <icolor.hpp>
#endif
#ifndef _IEVENT_
  #include <ievent.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IHANDLER_
  #include <ihandler.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ISIZEEVT_
  #include <isizeevt.hpp>
#endif
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif
#ifndef _IWPOSBUF_
  #include <iwposbuf.hpp>
#endif
#ifndef _IKSSET_H
  #include <iksset.h>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ISPLITCV_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public canvas styles.                                                        |
|                                                                              |
| Notes: The below style bits must not conflict with those used by inherited   |
|          class.                                                              |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ISplitCanvas::Style
  ISplitCanvas::vertical          ( 0, ICNV_VERTICAL ),
  ISplitCanvas::horizontal        ( 0, ICNV_HORIZONTAL ),
  ISplitCanvas::noSplitBars       ( 0, ICNV_NOSPLITBARS ),
  ISplitCanvas::classDefaultStyle ( WS_VISIBLE, ICNV_VERTICAL );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
ISplitCanvas::Style
              ISplitCanvas::currentDefaultStyle ( WS_VISIBLE, ICNV_VERTICAL );
#pragma data_seg()

/*------------------------------------------------------------------------------
| ISplitPercentageSet                                                          |
|                                                                              |
| Declare a key sorted set for the split percentage objects.                   |
------------------------------------------------------------------------------*/
class ISplitPercentageSet : public IKeySortedSet<ISplitPercentage*, IWindow*> {
public:
  ISplitPercentageSet ( );
 ~ISplitPercentageSet ( );
};

/*------------------------------------------------------------------------------
| ISplitPercentageSet::ISplitPercentageSet                                     |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
ISplitPercentageSet::ISplitPercentageSet ( )
  : IKeySortedSet<ISplitPercentage*, IWindow*>()
{ }

/*------------------------------------------------------------------------------
| ISplitPercentageSet::~ISplitPercentageSet                                    |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
ISplitPercentageSet::~ISplitPercentageSet ( )
{
  this->removeAll( &ISplitPercentageDeleter );
}


/*------------------------------------------------------------------------------
| IPaneSeq                                                                     |
|                                                                              |
| Declare a linked list for the pane objects                                   |
------------------------------------------------------------------------------*/
class IPaneSeq {
public:
  IPaneSeq ( );
 ~IPaneSeq ( );
Boolean
  isEmpty  ( );
void
  addFirst ( IPaneObj * ),
  addLast  ( IPaneObj * );

private:
IPaneObj
 *pFirstCl,
 *pLastCl;

friend ISplitCanvas;
};

/*------------------------------------------------------------------------------
| IPaneSeq::IPaneSeq                                                           |
|                                                                              |
| Initialize the linked list.                                                  |
------------------------------------------------------------------------------*/
IPaneSeq::IPaneSeq ( )
  : pFirstCl( 0 ), pLastCl( 0 )
{ }

/*------------------------------------------------------------------------------
| IPaneSeq::~IPaneSeq                                                          |
|                                                                              |
| Delete all objects in the linked list.                                       |
------------------------------------------------------------------------------*/
IPaneSeq::~IPaneSeq ( )
{
  if ( pFirstCl )
  {
     IPaneObj *tmpPaneObj;
     while ( pFirstCl )
     {
       tmpPaneObj = pFirstCl->pNext;
       delete pFirstCl;
       pFirstCl = tmpPaneObj;
     }
  }
}

/*------------------------------------------------------------------------------
| IPaneSeq::addFirst                                                           |
|                                                                              |
| Add an item to the beginning of the linked list.                             |
------------------------------------------------------------------------------*/
void IPaneSeq::addFirst ( IPaneObj* paneObj )
{
  paneObj->pNext = pFirstCl;
  pFirstCl = paneObj;
}

/*------------------------------------------------------------------------------
| IPaneSeq::addLast                                                            |
|                                                                              |
| Add an item to the end of the linked list.                                   |
------------------------------------------------------------------------------*/
void IPaneSeq::addLast ( IPaneObj* paneObj )
{
  if ( ! pFirstCl )
  {
    pFirstCl = paneObj;
    pLastCl  = paneObj;
  }
  else
  {
    pLastCl->pNext = paneObj;
    pLastCl = paneObj;
  }
}

/*------------------------------------------------------------------------------
| IPaneSeq::isEmpty                                                            |
|                                                                              |
| See if the linked list is empty.                                             |
------------------------------------------------------------------------------*/
Boolean IPaneSeq::isEmpty ( )
{
  return (pFirstCl ? false : true);
}


/*------------------------------------------------------------------------------
| ISplitCanvas::ISplitCanvas                                                   |
|                                                                              |
| Constructor.  Window is created in ICanvas class.                            |
------------------------------------------------------------------------------*/
ISplitCanvas::ISplitCanvas( unsigned long windowIdentifier,
                            IWindow* parent,
                            IWindow* owner,
                            const IRectangle& initialSize,
                            const Style& style )
  : ISplitCanvas::Inherited( ),
    sizClCanvasSize( -1, -1 ),
    clrClMiddleColor( IColor::paleGray ),
    clrClEdgeColor( IColor::darkGray ),
    pClSplitBarSet( 0 ),
    fSplitCanvasData( 0 )
{
  // Check for valid style combination.
  if ( (style & horizontal)  &&  (style & vertical) )
  {                                   // Mutually exclusive styles.
     ITHROWLIBRARYERROR( IC_INVALIDSTYLE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable );
  }

  // If no split bar orientation was specified, assume horizontal
  // for backwards compatibility.
  Style createStyle = style;
  if ( !( style & horizontal )  &&  !( style & vertical ) )
  {                // No split bar orientation specified.
     createStyle |= ISplitCanvas::horizontal;
  }

  // Save the extended style to make sure we have a copy of it stored.
  setExtendedStyle( extendedStyle() | createStyle.asExtendedUnsignedLong() );

  bClRatiosNotResolved = true;

  pClSplitPercentageSet = new ISplitPercentageSet();
  pClSplitCanvasHandler = new ISplitCanvasHandler();

  // Get the border widths used to draw the split bar.
  ulClEdgeThickness   = IQUERYSYSVALUE( SV_CXBORDER );
  ulClMiddleThickness = IQUERYSYSVALUE( SV_CXSIZEBORDER ) -
                          (2 * ulClEdgeThickness);

  pClSplitCanvasHandler->handleEventsFor( this );

  // Create the presentation system window.
  this->initialize( windowIdentifier,
                    parent,
                    owner,
                    initialSize,
                    this->convertToGUIStyle( createStyle ),
                    this->convertToGUIStyle( createStyle, true ) );
}

/*------------------------------------------------------------------------------
| ISplitCanvas::~ISplitCanvas                                                  |
|                                                                              |
| Destructor.  Remove and delete collection objects.                           |
------------------------------------------------------------------------------*/
ISplitCanvas::~ISplitCanvas()
{
  if ( pClSplitPercentageSet )
  {
     delete pClSplitPercentageSet;
  }

  if ( pClSplitCanvasHandler )
  {
     pClSplitCanvasHandler->stopHandlingEventsFor( this );
     delete pClSplitCanvasHandler;
  }
  // delete pClSplitBarSet;
  // delete fSplitCanvasData;
}

/*------------------------------------------------------------------------------
| ISplitPercentageDeleter                                                      |
|                                                                              |
| This function is used as input to allElementsDo() and results in all members |
| of the collection being deleted.                                             |
------------------------------------------------------------------------------*/
Boolean ISplitPercentageDeleter ( ISplitPercentage* const&
                                  splitPercentageObj,
                                  void*   )
{
  delete (ISplitPercentage*)splitPercentageObj;
  return true;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::layout                                                         |
|                                                                              |
| Layout management routine.                                                   |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::layout()
{
  IMODTRACE_DEVELOP( "ISplitCanvas::layout()" );

  if ( this->isLayoutDistorted( IWindow::layoutChanged ))
  {
     ITRACE_DEVELOP( "isLayoutDistorted(layoutChanged)==true" );
     /**************************************************************/
     /* The layout is distorted so calculate the position and size */
     /* of each of the child controls based on the percentage of   */
     /* the window they occupied at time of window resizing.       */
     /**************************************************************/

     /**************************************************************/
     /* Build a list of the canvas' child controls, as enumerated  */
     /* from PM.  This list will be used as a workspace for        */
     /* positioning and sizing the child controls.                 */
     /*                                                            */
     /* Note that child controls are enumerated in the z-order,    */
     /* which is the reverse order that they are layed out (and    */
     /* tabbed through).                                           */
     /**************************************************************/
     ITRACE_DEVELOP( "Start getting all child controls" );
     ISize szCanvas;
     unsigned long ulCanvasWidth;
     unsigned long ulCanvasHeight;
     unsigned long ulRemainingWindowPixels;
     unsigned long ulTotalPixelsUsed = 0;
     unsigned long ulExtStyle = this->extendedStyle();

     /***************************************************************/
     /* if sizClCanvasSize.width == -1, the queried size of the     */
     /* canvas is valid, else we entered layout() from the WM_SIZE  */
     /* message and we should use the size passed into the WM_SIZE  */
     /* message which has been saved in the sizClCanvasSize object. */
     /***************************************************************/
     if ( sizClCanvasSize.width() == -1 )
     {
        szCanvas = this->size();
     }
     else
     {
        szCanvas = sizClCanvasSize;
        sizClCanvasSize.setWidth( -1 ).setHeight( -1 );
     }

     ulCanvasWidth  = szCanvas.width();
     ulCanvasHeight = szCanvas.height();

     /***************************************************************/
     /* Do NOT size controls if canvas is ZERO height or width      */
     /* since it is not really visible.                             */
     /***************************************************************/
     if ( (ulCanvasWidth == 0) || (ulCanvasHeight == 0) )
     {
        return *this;
     }

     /***************************************************************/
     /* Set the split bar width according to whether the canvas has */
     /* split-bars or not.                                          */
     /***************************************************************/
     unsigned long ulSplitBarWidth = 0;
     if ( !(ulExtStyle & noSplitBars.asExtendedUnsignedLong()) )
        ulSplitBarWidth = (ulClEdgeThickness * 2) + ulClMiddleThickness;

     if ( bClRatiosNotResolved )
        this->resolveRatios();

     ulClChildCount = 0;                 // Reset child count to zero.
     unsigned long ulNonZeroSizeChildCount = 0;
     unsigned long ulRatioTotal = 0;
     IWindowHandle whChild;
     IPaneSeq* pPaneSeq = new IPaneSeq();

     IWindow::ChildCursor cwCursor( *this );
     for ( cwCursor.setToFirst(); cwCursor.isValid(); cwCursor.setToNext() )
     {
        whChild = this->childAt( cwCursor );
        IWindow* pwinChild = IWindow::windowWithHandle( whChild );
        if ( pwinChild )
        {                         // Have child control object.
           IPaneObj* ppcvChild = new IPaneObj( pwinChild );

           /**********************************************************/
           /* If the control has a ratio other than the canvas       */
           /* default ratio, get the controls ratio.                 */
           /**********************************************************/
           if ( pClSplitPercentageSet->containsElementWithKey( pwinChild ))
           {
              ppcvChild->ulClPercentage = resolvedPercentage( pwinChild );
           }
           ulRatioTotal += ppcvChild->ulClPercentage;
           if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
           {
              if ( this->areChildrenReversed() )
                 pPaneSeq->addLast( ppcvChild );
              else
                 pPaneSeq->addFirst( ppcvChild );
           }
           else               // Horizontal splitbars.
           {
              if ( this->areChildrenReversed() )
                 pPaneSeq->addFirst( ppcvChild );
              else
                 pPaneSeq->addLast( ppcvChild );
           }

           if ( ppcvChild->ulClPercentage != 0 )
           {       // Control must stay zero size.
              ulNonZeroSizeChildCount++;
           }

           ulClChildCount++;
        }
     }
     ITRACE_DEVELOP( "End getting all child controls" );

     if ( pPaneSeq->isEmpty() )  // No controls are in the canvas.
     {
        bClRatiosNotResolved = true;
        delete pPaneSeq;
        return *this;
     }
     /***************************************************************/
     /* Remove the split-bar width from the ratio unit calculation  */
     /* and remember there are no split bars to the left of the     */
     /* first control and to the right of the last control.         */
     /***************************************************************/
     if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
        ulCanvasWidth -= (ulSplitBarWidth * (ulClChildCount - 1));
     else
        ulCanvasHeight -= (ulSplitBarWidth * (ulClChildCount - 1));

     IPaneObj* ppcvWinnerOfPixelsObj;
     IPaneObj* ppcvObj = pPaneSeq->pFirstCl;

     while ( ppcvObj )
     {                            // Process all contents.
       if ( ppcvObj->ulClPercentage != 0 )
          ppcvWinnerOfPixelsObj = ppcvObj;

       if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
       {
          ppcvObj->ulControlSize = (ulCanvasWidth *
                                     (ppcvObj->ulClPercentage)) /
                                       ulRatioTotal;
       }
       else
       {
          ppcvObj->ulControlSize = (ulCanvasHeight *
                                     (ppcvObj->ulClPercentage)) /
                                       ulRatioTotal;
       }
       ulTotalPixelsUsed += ppcvObj->ulControlSize;

       ppcvObj = ppcvObj->pNext;
     }

     /*****************************************************************/
     /* Due to rounding errors, there is still the possibility that   */
     /* a window unit (or a few window units) is unused, this unit    */
     /* will be added to the size of the first non-zero size control. */
     /*****************************************************************/
     if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
        ulRemainingWindowPixels = ulCanvasWidth - ulTotalPixelsUsed;
     else
        ulRemainingWindowPixels = ulCanvasHeight - ulTotalPixelsUsed;

     ITRACE_DEVELOP( IString( "Pixels = " ) +
                       IString( ulRemainingWindowPixels ));
     if ( ulRemainingWindowPixels > 4 )
     {
        ppcvObj = pPaneSeq->pFirstCl;
        while ( ppcvObj )
        {                            // Process all contents.
           unsigned long ulNonZeroSizeParts = 0;

           if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
              ulNonZeroSizeParts = (ulRemainingWindowPixels *
                                     (ppcvObj->ulClPercentage)) /
                                       ulRatioTotal;
           else
              ulNonZeroSizeParts = (ulRemainingWindowPixels *
                                     (ppcvObj->ulClPercentage)) /
                                       ulRatioTotal;

           ppcvObj->ulControlSize += ulNonZeroSizeParts;
           ulTotalPixelsUsed += ulNonZeroSizeParts;

           ppcvObj = ppcvObj->pNext;
        }
     }
     /***************************************************************/
     /* Add final pixels to last window that has a non-zero size.   */
     /***************************************************************/
     if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
        ulRemainingWindowPixels = ulCanvasWidth - ulTotalPixelsUsed;
     else
        ulRemainingWindowPixels = ulCanvasHeight - ulTotalPixelsUsed;

     ppcvWinnerOfPixelsObj->ulControlSize += ulRemainingWindowPixels;

     unsigned long ulCoord = 0;
     IWindowPosBuffer wposbuf = this->fixupChildren();

     // Size and position child windows in native coordinates, since
     // this is what IWindowPosBuffer expects.
     ppcvObj = pPaneSeq->pFirstCl;
     while ( ppcvObj )
     {
       if ( ulExtStyle & vertical.asExtendedUnsignedLong() )
       {
          wposbuf.moveSizeTo( ppcvObj->pwinCl,
                              IRectangle( IPair( ulCoord, 0 ),
                                          ISize( ppcvObj->ulControlSize,
                                                 ulCanvasHeight )));
          ulCoord += (ppcvObj->ulControlSize + ulSplitBarWidth);
       }
       else
       {
#ifdef IC_ORIGINLOWERLEFT
          wposbuf.moveSizeTo( ppcvObj->pwinCl,
                              IRectangle( IPair( 0, ulCoord ),
                                          ISize( ulCanvasWidth,
                                                 ppcvObj->ulControlSize )));
          ulCoord += (ppcvObj->ulControlSize + ulSplitBarWidth);
#else
          ulCoord += ppcvObj->ulControlSize;
          wposbuf.moveSizeTo( ppcvObj->pwinCl,
                              IRectangle( IPair( 0, szCanvas.height() - ulCoord ),
                                          ISize( ulCanvasWidth,
                                                 ppcvObj->ulControlSize )));
          ulCoord += ulSplitBarWidth;
#endif

       }

       ITRACE_DEVELOP( IString( "hwndCtrl = " ) +
                       IString( ppcvObj->pwinCl->handle().asUnsigned()) );
       ITRACE_DEVELOP( IString( "Ratio = " ) +
                       IString( ppcvObj->ulClPercentage ));
       ITRACE_DEVELOP( IString( "Size = " ) +
                       IString( (unsigned long)(ppcvObj->ulControlSize)) );
       ITRACE_DEVELOP( IString( "CVSize = " ) + szCanvas.asString() );

       ppcvObj = ppcvObj->pNext;
     }

     wposbuf.apply();           // Reposition/size child controls.

     if ( pPaneSeq )
     {
        delete pPaneSeq;
     }

    sizClCanvasSize.setWidth( -1 ).setHeight( -1 );

    this->setLayoutDistorted( 0, IWindow::layoutChanged );
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setSplitWindowPercentage                                       |
|                                                                              |
| Set the default split bar ratio (percentage of the window).                  |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setSplitWindowPercentage ( IWindow* pwin,
                                                  unsigned long ulPercentage )
{
  if ( pClSplitPercentageSet->containsElementWithKey( pwin ))
  {
     ISplitPercentage* obj = pClSplitPercentageSet->elementWithKey( pwin );
     obj->ulClPercentage = ulPercentage;
     obj->ulClResolvedPercentage = ulPercentage;
  }
  else
  {
     ISplitPercentage* obj = new ISplitPercentage( pwin, ulPercentage );
     pClSplitPercentageSet->add( obj );
     obj->ulClResolvedPercentage = ulPercentage;
  }

  bClRatiosNotResolved = true;
  this->setLayoutDistorted( IWindow::layoutChanged, 0 );

  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::splitWindowPercentage                                          |
|                                                                              |
| Get the default split bar ratio (percentage of the window).                  |
------------------------------------------------------------------------------*/
unsigned long ISplitCanvas::splitWindowPercentage ( IWindow* pwin )
{
  if ( bClRatiosNotResolved )
  {
     this->resolveRatios();
  }

  if ( pClSplitPercentageSet->containsElementWithKey( pwin ))
  {
     ISplitPercentage* obj = pClSplitPercentageSet->elementWithKey( pwin );
     return obj->ulClPercentage;
  }

  return 0;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setResolvedPercentage                                          |
|                                                                              |
| Set resolved percentage of the window.                                       |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setResolvedPercentage ( IWindow* pwin,
                                                  unsigned long ulPercentage )
{
  if ( pClSplitPercentageSet->containsElementWithKey( pwin ))
  {
     ISplitPercentage* obj = pClSplitPercentageSet->elementWithKey( pwin );
     obj->ulClResolvedPercentage = ulPercentage;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::resolvedPercentage                                             |
|                                                                              |
| Get the resolved percentage of the window.                                   |
------------------------------------------------------------------------------*/
unsigned long ISplitCanvas::resolvedPercentage ( IWindow* pwin )
{
  if ( pClSplitPercentageSet->containsElementWithKey( pwin ))
  {
     ISplitPercentage* obj = pClSplitPercentageSet->elementWithKey( pwin );
     return obj->ulClResolvedPercentage;
  }
  return 0;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::resolveRatios                                                  |
|                                                                              |
| Resolve percentages to a total of 128 (which is used to divide by).          |
------------------------------------------------------------------------------*/
void ISplitCanvas::resolveRatios ( )
{
  IMODTRACE_DEVELOP( "ISplitCanvas::resolveRatios()" );

  IPaneSeq* pPaneSeq  = new IPaneSeq();
  IWindowHandle whChild;

  Boolean              bNoRatioForWindow = false;
  unsigned long        ulWindowCount = 0;
  unsigned long        ulRatioTotal = 0;
  IWindow::ChildCursor cwCursor( *this );

  for ( cwCursor.setToFirst(); cwCursor.isValid(); cwCursor.setToNext() )
  {
     whChild = this->childAt( cwCursor );
     IWindow* pwinChild = IWindow::windowWithHandle( whChild );
     if ( pwinChild )
     {                         // Have child control object.
        IPaneObj* ppcvChild = new IPaneObj( pwinChild );

        /**********************************************************/
        /* If the control has a ratio other than the canvas       */
        /* default ratio, get the controls ratio.                 */
        /**********************************************************/
        if ( pClSplitPercentageSet->containsElementWithKey( pwinChild ))
        {
           ISplitPercentage* pdsrObject =
                       pClSplitPercentageSet->elementWithKey( pwinChild );
           ppcvChild->ulClPercentage = pdsrObject->ulClResolvedPercentage;
           ulWindowCount++;
           ulRatioTotal += ppcvChild->ulClPercentage;
        }
        else
        {
           bNoRatioForWindow = true;
        }
        pPaneSeq->addFirst( ppcvChild );
     }
  }

  /****************************************************************/
  /* If a window was added without setting a ratio for it, we     */
  /* need to give it a ratio value proportional to the ratios of  */
  /* the other windows (that already have a value).               */
  /****************************************************************/

  IPaneObj* ppcvObj;
  if ( bNoRatioForWindow )
  {
     ppcvObj = pPaneSeq->pFirstCl;
     while ( ppcvObj )
     {
        if ( !pClSplitPercentageSet->containsElementWithKey( ppcvObj->pwinCl ))
        {
           /*******************************************************/
           /* This window does not have a ratio, so create one.   */
           /*******************************************************/
           if ( (ulWindowCount != 0) && (ulRatioTotal != 0) )
              ppcvObj->ulClPercentage = ulRatioTotal / ulWindowCount;
           else
              ppcvObj->ulClPercentage = 10L;
           this->setSplitWindowPercentage( ppcvObj->pwinCl,
                                           ppcvObj->ulClPercentage );
           ulRatioTotal += ppcvObj->ulClPercentage;
           ulWindowCount++;
        }
        ppcvObj = ppcvObj->pNext;
     }
  }

  if ( pPaneSeq->isEmpty() )  // No controls are in the canvas.
  {                                    // Nothing more to do.
     delete pPaneSeq;
     return;
  }

  unsigned long ul100Ratio;
  unsigned long ulTotal100PercentUsed = 0;
  IPaneObj* ppcvWinnerOfPercents;

  ppcvObj = pPaneSeq->pFirstCl;
  while ( ppcvObj )
  {
     ul100Ratio = (ppcvObj->ulClPercentage * 100) / ulRatioTotal;
     this->setSplitWindowPercentage( ppcvObj->pwinCl, ul100Ratio );
     this->setResolvedPercentage( ppcvObj->pwinCl, ppcvObj->ulClPercentage );

     if ( ul100Ratio != 0 )
        ppcvWinnerOfPercents = ppcvObj;

     ulTotal100PercentUsed += ul100Ratio;
     ppcvObj = ppcvObj->pNext;
  }

  bClRatiosNotResolved = false;        // Clear flag.

  /*******************************************************************/
  /* Add the left over percentage to the last window with a non-zero */
  /* ratio.                                                          */
  /*******************************************************************/
  if ( ppcvWinnerOfPercents )
  {
     ul100Ratio = this->splitWindowPercentage( ppcvWinnerOfPercents->pwinCl );
     this->setSplitWindowPercentage( ppcvWinnerOfPercents->pwinCl,
                                     (100 - ulTotal100PercentUsed +
                                                          ul100Ratio) );
     /****************************************************************/
     /* Since setSplitWindowPercentage also changes the resolved     */
     /* value, we need to reset it back to the correct user setting. */
     /****************************************************************/
     this->setResolvedPercentage( ppcvWinnerOfPercents->pwinCl,
                                  ppcvWinnerOfPercents->ulClPercentage );
  }

  if ( pPaneSeq )
  {
     delete pPaneSeq;
  }

  return;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setOrientation                                                 |
|                                                                              |
| Change the panes orientation.                                                |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setOrientation ( Orientation value )
{
  Boolean attrChanged = false;
  unsigned long ulExtStyle = this->extendedStyle();

  if ( value == verticalSplit &&
       !(ulExtStyle & vertical.asExtendedUnsignedLong()) )
  {
     ulExtStyle |= vertical.asExtendedUnsignedLong();
     ulExtStyle &= ~horizontal.asExtendedUnsignedLong();
     this->setLayoutDistorted( IWindow::layoutChanged |
                                 IWindow::immediateUpdate, 0 );
     attrChanged = true;
  }
  else if ( value == horizontalSplit &&
            (ulExtStyle & vertical.asExtendedUnsignedLong()) )
  {
     ulExtStyle |= horizontal.asExtendedUnsignedLong();
     ulExtStyle &= ~vertical.asExtendedUnsignedLong();
     this->setLayoutDistorted( IWindow::layoutChanged |
                               IWindow::immediateUpdate, 0 );
     attrChanged = true;
  }

  if ( attrChanged )
  {
     this->setExtendedStyle( ulExtStyle );
     this->notifyObservers( INotificationEvent( ISplitCanvas::orientationId,
                                                *this, true, (void*)value ));
  }

  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::defaultStyle                                                   |
------------------------------------------------------------------------------*/
ISplitCanvas::Style ISplitCanvas::defaultStyle ( )
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::convertToGUIStyle                                              |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ISplitCanvas::convertToGUIStyle ( const IBitFlag& guiStyle,
                                                Boolean bExtOnly ) const
{
  // Obtain the style from the class (ICanvas) that we inherit from.
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if ( bExtOnly )
  {
     // Use mask to only return extended styles in the user defined range.
     ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
     // No styles are defined for the base canvas, so we do not mask.
     ulStyle |= guiStyle.asUnsignedLong();

     // No additional processing required since ICanvas::convertToGUIStyle is
     // doing all the required work ...
  }

  return ulStyle;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setLayoutDistorted                                             |
|                                                                              |
| Treat a size change like a layout change.                                    |
------------------------------------------------------------------------------*/
ISplitCanvas&
  ISplitCanvas::setLayoutDistorted ( unsigned long layoutAttributeOn,
                                     unsigned long layoutAttributeOff )
{
  unsigned long ulFlagsOn =
     (layoutAttributeOn & (unsigned long)(~IWindow::sizeChanged));

  if ( layoutAttributeOn & IWindow::sizeChanged )
  {                               // Need to resize/reposition panes.
     ulFlagsOn |= IWindow::layoutChanged;  // Flag for ICanvas.
  }
  Inherited::setLayoutDistorted( ulFlagsOn, layoutAttributeOff );

  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setSplitBarThickness                                           |
|                                                                              |
| Sets the thickness of the specified portion of the split bar.                |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setSplitBarThickness ( SplitBarArea value,
                                                   unsigned long thickness )
{
  if ( this->splitBarThickness( value ) != thickness )
  {
    if ( value == splitBarEdge )
    {
       ulClEdgeThickness = thickness;
    }
    else
    {
       ulClMiddleThickness = thickness;
    }
    this->setLayoutDistorted( IWindow::layoutChanged |
                                IWindow::immediateUpdate, 0 );
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::splitBarThickness                                              |
|                                                                              |
| Get the thickness of the specified portion of the split bar.                 |
------------------------------------------------------------------------------*/
unsigned long ISplitCanvas::splitBarThickness ( SplitBarArea value )
{
  return (value == splitBarEdge) ? ulClEdgeThickness : ulClMiddleThickness;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::splitBarEdgeColor                                              |
|                                                                              |
| Returns the color of the edge of the splitbar of an ISplitCanvas.            |
------------------------------------------------------------------------------*/
IColor ISplitCanvas::splitBarEdgeColor ( ) const
{
  return clrClEdgeColor;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::splitBarMiddleColor                                            |
|                                                                              |
| Returns the color of the middle of the splitbar of an ISplitCanvas.          |
------------------------------------------------------------------------------*/
IColor ISplitCanvas::splitBarMiddleColor ( ) const
{
  return clrClMiddleColor;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setSplitBarEdgeColor                                           |
|                                                                              |
| Sets the color of the edge of the splitbar of an ISplitCanvas.               |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setSplitBarEdgeColor ( const IColor& color )
{
  clrClEdgeColor = color;
  this->refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setSplitBarMiddleColor                                         |
|                                                                              |
| Sets the color of the middle of the splitbar of an ISplitCanvas.             |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::setSplitBarMiddleColor ( const IColor &color )
{
  clrClMiddleColor = color;
  this->refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::resetSplitBarEdgeColor                                         |
|                                                                              |
| Resets the color of the edge of the splitbar of an ISplitCanvas.             |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::resetSplitBarEdgeColor ( )
{
  clrClEdgeColor = IColor( IColor::darkGray );
  this->refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::resetSplitBarMiddleColor                                       |
|                                                                              |
| Resets the color of the middle of the splitbar of an ISplitCanvas.           |
------------------------------------------------------------------------------*/
ISplitCanvas& ISplitCanvas::resetSplitBarMiddleColor ( )
{
  clrClMiddleColor = IColor(IColor::paleGray);
  this->refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ISplitCanvas::setDefaultStyle                                                |
|                                                                              |
| Sets the default style.                                                      |
|                                                                              |
| Note: This function is inlined on all platforms except Windows due to        |
|       variables it references not being exported                             |
------------------------------------------------------------------------------*/
#ifdef IC_WIN
void  ISplitCanvas::setDefaultStyle ( const ISplitCanvas::Style& style )
{
  currentDefaultStyle = style;
}
#endif //IC_WIN
