/*******************************************************************************
* FILE NAME: ivporhdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ivporhdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WINBUTTONS           // BS_xxx, BM_xxx
#define INCL_WINDIALOGS           // WinEnumDlgItem, etc.
#define INCL_WINFRAMEMGR          // WM_FOCUSCHANGE
#define INCL_WININPUT             // WM_CHAR, WinGetKeyState, WinQueryFocus
#define INCL_WINMESSAGEMGR        // WM_HSCROLL, WM_VSCROLL
#define INCL_WINSCROLLBARS        // SB_xxx
#define INCL_WINWINDOWMGR         // WinDefWindowProc, etc.
#include <iwindefs.h>

#include <ivporhdr.hpp>

#include <ivport.hpp>

#include <icconst.h>
#include <icoordsy.hpp>
#include <ievent.hpp>
#include <iexcept.hpp>
#include <ikeyevt.hpp>
#include <ipainevt.hpp>
#include <iscroll.hpp>
#include <itrace.hpp>
#include <ivrect.hpp>

#define MAX_SCROLL_LIMIT  0x8000

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IVPORHDR_CPP_
  #include <ipagetun.h>
#endif


IViewPortResizeHandler* IViewPortStatics :: pClResizeHandler = 0;
IViewPortScrollHandler* IViewPortStatics :: pClScrollHandler = 0;
IViewPortKeyboardHandler* IViewPortStatics :: pClKeyboardHandler = 0;
IViewPortPaintHandler* IViewPortStatics :: pClPaintHandler = 0;
IViewRectanglePaintHandler* IViewPortStatics :: pClRectPaintHandler = 0;


/*------------------------------------------------------------------------------
| IViewPortStatics::~IViewPortStatics                                          |
|                                                                              |
| Destructor to clean up.                                                      |
------------------------------------------------------------------------------*/
IViewPortStatics :: ~IViewPortStatics ( )
{
  if ( pClResizeHandler )
  {
     delete pClResizeHandler;
  }
  if ( pClScrollHandler )
  {
     delete pClScrollHandler;
  }
  if ( pClKeyboardHandler )
  {
     delete pClKeyboardHandler;
  }
  if ( pClPaintHandler )
  {
     delete pClPaintHandler;
  }
  if ( pClRectPaintHandler )
  {
     delete pClRectPaintHandler;
  }
}

/*------------------------------------------------------------------------------
| IViewPortStatics::resizeHandler                                              |
|                                                                              |
| Returns the view port resize handler.                                        |
------------------------------------------------------------------------------*/
IViewPortResizeHandler& IViewPortStatics :: resizeHandler ( )
{
  if ( pClResizeHandler == 0 )
  {
     pClResizeHandler = new IViewPortResizeHandler;
  }
  return *pClResizeHandler;
}

/*------------------------------------------------------------------------------
| IViewPortStatics::scrollHandler                                              |
|                                                                              |
| Returns the view port scroll handler.                                        |
------------------------------------------------------------------------------*/
IViewPortScrollHandler& IViewPortStatics :: scrollHandler ( )
{
  if ( pClScrollHandler == 0 )
  {
     pClScrollHandler = new IViewPortScrollHandler;
  }
  return *pClScrollHandler;
}

/*------------------------------------------------------------------------------
| IViewPortStatics::keyboardHandler                                            |
|                                                                              |
| Returns the view port keyboard handler.                                      |
------------------------------------------------------------------------------*/
IViewPortKeyboardHandler& IViewPortStatics :: keyboardHandler ( )
{
  if ( pClKeyboardHandler == 0 )
  {
     pClKeyboardHandler = new IViewPortKeyboardHandler;
  }
  return *pClKeyboardHandler;
}

/*------------------------------------------------------------------------------
| IViewPortStatics::paintHandler                                               |
|                                                                              |
| Returns the view port paint handler.                                         |
------------------------------------------------------------------------------*/
IViewPortPaintHandler& IViewPortStatics :: paintHandler ( )
{
  if ( pClPaintHandler == 0 )
  {
     pClPaintHandler = new IViewPortPaintHandler;
  }
  return *pClPaintHandler;
}

/*------------------------------------------------------------------------------
| IViewPortStatics::rectanglePaintHandler                                      |
|                                                                              |
| Returns the paint handler for the view rectangle object.                     |
------------------------------------------------------------------------------*/
IViewRectanglePaintHandler& IViewPortStatics :: rectanglePaintHandler ( )
{
  if ( pClRectPaintHandler == 0 )
  {
     pClRectPaintHandler = new IViewRectanglePaintHandler;
  }
  return *pClRectPaintHandler;
}


/*------------------------------------------------------------------------------
| IViewPortResizeHandler::IViewPortResizeHandler                               |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IViewPortResizeHandler :: IViewPortResizeHandler ( )
  : IViewPortResizeHandler::Inherited ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::~IViewPortResizeHandler                              |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IViewPortResizeHandler :: ~IViewPortResizeHandler ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::windowResize                                         |
|                                                                              |
| Resizes the view port window and adjust the scroll bars if necessary.        |
------------------------------------------------------------------------------*/
Boolean IViewPortResizeHandler :: windowResize ( IResizeEvent& event )
{
  IMODTRACE_DEVELOP( "IViewPortResizeHandler::windowResize()" );

  if ( event.oldSize() != event.newSize() )
  {                               // Size change occurred.
     /**************************************************************/
     /* Cause the viewport to redraw scroll bars, view window.     */
     /**************************************************************/
     IViewPort *pvport = (IViewPort*)event.window();
     pvport->setLayoutDistorted( IWindow::sizeChanged, 0 );
                                  // Flag change occurred.
     if ( pvport->isVisible() )
     {                            // Delay if not visible.
        if ( event.newSize().width() > event.oldSize().width() )
        {                         // Viewport width has grown.
           pvport->refresh();     // Process the change.
        }
     }
  }                               // End size change occurred.
  return true;
}

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::handleEventsFor                                      |
|                                                                              |
| Attaches the resize handler to the specified view port object.               |
------------------------------------------------------------------------------*/
IViewPortResizeHandler&
  IViewPortResizeHandler :: handleEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::handleEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::stopHandlingEventsFor                                |
|                                                                              |
| Detach the resize handler from the specified view port object.               |
------------------------------------------------------------------------------*/
IViewPortResizeHandler&
  IViewPortResizeHandler :: stopHandlingEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::stopHandlingEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::handleEventsFor                                      |
------------------------------------------------------------------------------*/
IHandler& IViewPortResizeHandler :: handleEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortResizeHandler::stopHandlingEventsFor                                |
------------------------------------------------------------------------------*/
IHandler& IViewPortResizeHandler :: stopHandlingEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}


/*------------------------------------------------------------------------------
| IViewPortScrollHandler::IViewPortScrollHandler                               |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IViewPortScrollHandler :: IViewPortScrollHandler ( )
  : IViewPortScrollHandler::Inherited ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::~IViewPortScrollHandler                              |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IViewPortScrollHandler :: ~IViewPortScrollHandler ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::lineUp                                               |
|                                                                              |
| Scrolls the view window upward when the up scroll button of the vertical     |
| scroll bar is clicked.                                                       |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: lineUp ( IScrollEvent& event )
{
  return this->verticalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::lineDown                                             |
|                                                                              |
| Scrolls the view window downward when the down scroll button of the vertical |
| scroll bar is clicked.                                                       |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: lineDown ( IScrollEvent& event )
{
  return this->verticalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::lineLeft                                             |
|                                                                              |
| Scrolls the view window left when the left scroll button of the horizontal   |
| scroll bar is clicked.                                                       |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: lineLeft ( IScrollEvent& event )
{
  return this->horizontalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::lineRight                                            |
|                                                                              |
| Scrolls the view window right when the right scroll button of the horizontal |
| scroll bar is clicked.                                                       |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: lineRight ( IScrollEvent& event )
{
  return this->horizontalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::pageUp                                               |
|                                                                              |
| Scrolls the view window up a page when the scroll shaft above the scroll box |
| of the vertical scroll bar is clicked.                                       |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: pageUp ( IScrollEvent& event )
{
  return this->verticalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::pageDown                                             |
|                                                                              |
| Scrolls the view window down a page when the scroll shaft below the scroll   |
| box of the vertical scroll bar is clicked.                                   |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: pageDown ( IScrollEvent& event )
{
  return this->verticalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::pageLeft                                             |
|                                                                              |
| Scrolls the view window left a page when the scroll shaft to the left of the |
| scroll box of the horizontal scroll bar is clicked.                          |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: pageLeft ( IScrollEvent& event )
{
  return this->horizontalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::pageRight                                            |
|                                                                              |
| Scrolls the view window right a page when the scroll shaft to the right of   |
| the scroll box of the horizontal scroll bar is clicked.                      |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: pageRight ( IScrollEvent& event )
{
  return this->horizontalScroll( event );
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::verticalScroll                                       |
|                                                                              |
| Scroll the view window vertically and move the scroll box of the vertical    |
| scroll bar as appropriate for the scroll message.                            |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: verticalScroll ( IScrollEvent& event )
{
  Boolean bProcessed = false;
  IViewPort *pvportOwner = (IViewPort*)event.window();
  IScrollBar* scrollBar = pvportOwner->verticalScrollBar();
  if ( event.scrollBarWindow() == scrollBar )
  {                               // Viewport's vertical scroll bar.
     unsigned long ulPos;         // Position for scroll box.

     // Adjust for huge logical height.
     unsigned long viewWindowHeight =
                     pvportOwner->viewWindowSize().height();
     if ( viewWindowHeight < MAX_SCROLL_LIMIT )
     {                  // View port with normal <32K range.
        this->moveScrollBox( event );
        ulPos = scrollBar->scrollBoxPosition();
     }
     else
     {           // View port gave scaled down values to scroll bar.
        // Get the current scroll box position in view port units
        // directly from the view port (don't trust using the real
        // scroll box position and converting it to view port units
        // because of potential rounding errors).
        ulPos = viewWindowHeight -
                  pvportOwner->viewWindowDrawRectangle().top() + 1;
        long lDelta = event.scrollAmount();  // In view port units.

        // Re-adjust the scroll amount for the case of a partial
        // scroll to the top or bottom of the view window.  The
        // adjustments made by IScrollEvent::scrollAmount were using
        // scroll bar units.  The below adjustments use view port
        // units.
#ifdef IC_WIN
        unsigned long type = event.parameter1().lowNumber();
#else
        unsigned long type = event.parameter2().highNumber();
#endif
        unsigned long ulNormalScroll;
        if ( type == SB_PAGEUP  ||  type == SB_PAGEDOWN )
        {
           ulNormalScroll = scrollBar->pageScrollIncrement();
        }
        else if ( type == SB_LINEUP  ||  type == SB_LINEDOWN )
        {
           ulNormalScroll = scrollBar->minScrollIncrement();
        }

        Boolean bAtMaxEnd = false;
        if ( type == SB_PAGEUP  ||  type == SB_LINEUP )
        {
           if ( 0 - lDelta < ulNormalScroll )
           {            // Need to adjust the adjusted value.
              lDelta = 1 - ulPos;
              if ( ulNormalScroll < 0 - lDelta )
              {
                 lDelta = 0 - ulNormalScroll;
              }
           }
        }
        else if ( (type == SB_PAGEDOWN  ||  type == SB_LINEDOWN)  &&
                  (lDelta < ulNormalScroll) )
        {               // Need to adjust the adjusted value.
           lDelta = pvportOwner->viewWindowDrawRectangle().bottom();
           if (ulNormalScroll < lDelta)
           {
              lDelta = ulNormalScroll;
           }
           else
           {
              bAtMaxEnd = true;
           }
        }

        ulPos += lDelta;   // New scroll box position.
        unsigned long ulNewPos;
        if ( ! bAtMaxEnd )
        {
           unsigned long yRatio =
                        (viewWindowHeight / MAX_SCROLL_LIMIT) + 1;
           ulNewPos = ulPos / yRatio;  // Remap to scroll bar units.
        }
        else
        {
           ulNewPos = scrollBar->scrollBoxRange().upperBound();
        }
        scrollBar->moveScrollBoxTo( ulNewPos );
     }                  // End huge logical height.

     pvportOwner->scrollViewVerticallyTo( ulPos - 1 );
     bProcessed = true;
  }
  return bProcessed;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::horizontalScroll                                     |
|                                                                              |
| Scroll the view window horizontally and move the scroll box of the horizontal|
| scroll bar as needed for the scroll message.                                 |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: horizontalScroll ( IScrollEvent& event )
{
  Boolean bProcessed = false;
  IViewPort *pvportOwner = (IViewPort*)event.window();
  IScrollBar* scrollBar = pvportOwner->horizontalScrollBar();
  if ( event.scrollBarWindow() == scrollBar )
  {                               // Viewport's horizontal scroll bar.
     unsigned long ulPos;         // Position for scroll box.

     // Adjust for huge logical width.
     unsigned long viewWindowWidth =
                     pvportOwner->viewWindowSize().width();
     if ( viewWindowWidth < MAX_SCROLL_LIMIT )
     {                  // View port with normal <32K range.
        this->moveScrollBox( event );
        ulPos = scrollBar->scrollBoxPosition();
     }
     else
     {           // View port gave scaled down values to scroll bar.
        // Get the current scroll box position in view port units
        // directly from the view port (don't trust using the real
        // scroll box position and converting it to view port units
        // because of potential rounding errors).
        ulPos = pvportOwner->viewWindowDrawRectangle().left() + 1;
        long lDelta = event.scrollAmount();  // In view port units.

        // Re-adjust the scroll amount for the case of a partial
        // scroll to the left or right edge of the view window.  The
        // adjustments made by IScrollEvent::scrollAmount were using
        // scroll bar units.  The below adjustments use view port
        // units.
#ifdef IC_WIN
        unsigned long type = event.parameter1().lowNumber();
#else
        unsigned long type = event.parameter2().highNumber();
#endif
        unsigned long ulNormalScroll;
        if ( type == SB_PAGELEFT  ||  type == SB_PAGERIGHT )
        {
           ulNormalScroll = scrollBar->pageScrollIncrement();
        }
        else if ( type == SB_LINELEFT  ||  type == SB_LINERIGHT )
        {
           ulNormalScroll = scrollBar->minScrollIncrement();
        }

        Boolean bAtMaxEnd = false;
        if ( type == SB_PAGELEFT  ||  type == SB_LINELEFT )
        {
           if ( 0 - lDelta < ulNormalScroll )
           {            // Need to adjust the adjusted value.
              lDelta = 1 - ulPos;
              if ( ulNormalScroll < 0 - lDelta )
              {
                 lDelta = 0 - ulNormalScroll;
              }
           }
        }
        else if ( (type == SB_PAGERIGHT  ||  type == SB_LINERIGHT)  &&
                  (lDelta < ulNormalScroll) )
        {               // Need to adjust the adjusted value.
           lDelta = viewWindowWidth -
                      pvportOwner->viewWindowDrawRectangle().right();
           if ( ulNormalScroll < lDelta )
           {
              lDelta = ulNormalScroll;
           }
           else
           {
              bAtMaxEnd = true;
           }
        }

        ulPos += lDelta;   // New scroll box position.
        unsigned long ulNewPos;
        if ( ! bAtMaxEnd )
        {
           unsigned long xRatio =
                        (viewWindowWidth / MAX_SCROLL_LIMIT) + 1;
           ulNewPos = ulPos / xRatio;  // Remap to scroll bar units.
        }
        else
        {
           ulNewPos = scrollBar->scrollBoxRange().upperBound();
        }
        scrollBar->moveScrollBoxTo( ulNewPos );
     }                  // End huge logical height.

     pvportOwner->scrollViewHorizontallyTo( ulPos - 1 );
     bProcessed = true;
  }
  return bProcessed;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::scrollBoxTrack                                       |
|                                                                              |
| View window has been scrolled by dragging the scroll box of either the       |
| vertical or horizontal scroll bar.                                           |
------------------------------------------------------------------------------*/
Boolean IViewPortScrollHandler :: scrollBoxTrack ( IScrollEvent& event )
{
  Boolean bProcessed = false;
  IViewPort *pvportOwner = (IViewPort*)event.window();

  IScrollBar* scrollBar = event.scrollBarWindow();
  if ( scrollBar == pvportOwner->horizontalScrollBar() )
  {                               // Horizontal scroll bar.
     unsigned long ul0BasedScrollTo = event.newScrollBoxPosition() - 1;

     // Adjust for huge logical width.
     unsigned long viewWindowWidth =
                       pvportOwner->viewWindowSize().width();
     if ( viewWindowWidth >= MAX_SCROLL_LIMIT )
     {           // View port gave scaled down values to scroll bar.
        unsigned long scrollBoxMax = scrollBar->scrollBoxRange().upperBound();
        unsigned long effectiveScrollBoxMax =
                    viewWindowWidth + 1 -
                      pvportOwner->viewWindowDrawRectangle().width();
        if ( ul0BasedScrollTo + 1 == scrollBoxMax )
        {               // Scrolled to end.
           ul0BasedScrollTo = effectiveScrollBoxMax - 1;
        }
        else
        {               // Extrapolate where we are.
           double xRatio = (double)(effectiveScrollBoxMax - 1) /
                                               (scrollBoxMax - 1);
           ul0BasedScrollTo *= xRatio;
                        // Remap it back to view port's units.
        }
     }                  // End huge logical width.

     pvportOwner->scrollViewHorizontallyTo( ul0BasedScrollTo );

     bProcessed = true;
  }
  else if ( scrollBar == pvportOwner->verticalScrollBar() )
  {                               // Vertical scroll bar
     unsigned long ul0BasedScrollTo = event.newScrollBoxPosition() - 1;

     // Adjust for huge logical height.
     unsigned long viewWindowHeight =
                       pvportOwner->viewWindowSize().height();
     if ( viewWindowHeight >= MAX_SCROLL_LIMIT )
     {           // View port gave scaled down values to scroll bar.
        unsigned long scrollBoxMax = scrollBar->scrollBoxRange().upperBound();
        unsigned long effectiveScrollBoxMax =
                   viewWindowHeight + 1 -
                     pvportOwner->viewWindowDrawRectangle().height();
        if ( ul0BasedScrollTo + 1 == scrollBoxMax )
        {               // Scrolled to end.
           ul0BasedScrollTo = effectiveScrollBoxMax - 1;
        }
        else
        {               // Extrapolate where we are.
           double yRatio = (double)(effectiveScrollBoxMax - 1) /
                                               (scrollBoxMax - 1);
           ul0BasedScrollTo *= yRatio;
                        // Remap it back to view port's units.
        }
     }                  // End huge logical height.

     pvportOwner->scrollViewVerticallyTo( ul0BasedScrollTo );

     bProcessed = true;
  }                               // Else user's scroll bar.

  return bProcessed;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::handleEventsFor                                      |
|                                                                              |
| Attach the scroll handler to the specified view port object.                 |
------------------------------------------------------------------------------*/
IViewPortScrollHandler&
  IViewPortScrollHandler :: handleEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::handleEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::stopHandlingEventsFor                                |
|                                                                              |
| Detach the scroll handler from the specified view port object.               |
------------------------------------------------------------------------------*/
IViewPortScrollHandler&
  IViewPortScrollHandler :: stopHandlingEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::stopHandlingEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::handleEventsFor                                      |
------------------------------------------------------------------------------*/
IHandler& IViewPortScrollHandler :: handleEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortScrollHandler::stopHandlingEventsFor                                |
------------------------------------------------------------------------------*/
IHandler& IViewPortScrollHandler :: stopHandlingEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}


/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::IViewPortKeyboardHandler                           |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IViewPortKeyboardHandler :: IViewPortKeyboardHandler ( )
  : IViewPortKeyboardHandler::Inherited ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::~IViewPortKeyboardHandler                          |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IViewPortKeyboardHandler :: ~IViewPortKeyboardHandler ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::virtualKeyPress                                    |
|                                                                              |
| Process keys used to scroll the client window.                               |
------------------------------------------------------------------------------*/
Boolean IViewPortKeyboardHandler :: virtualKeyPress ( IKeyboardEvent& event )
{
  Boolean bProcessed = false;

  IKeyboardEvent::VirtualKey vkey = event.virtualKey();

  // Page scrolling.
  if (vkey == IKeyboardEvent::pageUp)
  {                               // PageUp/PgUp key.
     if ( event.isCtrlDown() )
     {                            // Ctrl+PageUp: scroll left.
        IViewPort *pvport = (IViewPort*)event.window();

        pvport->sendEvent( WM_HSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT( SB_PAGELEFT, 0 ),
                           IEventData(
                              pvport->horizontalScrollBar()->handle()) );
#else
                           pvport->horizontalScrollBar()->id(),
                           MPFROM2SHORT( 0, SB_PAGELEFT ));
#endif
        bProcessed = true;
     }
     else if ( event.isShiftDown() == false  &&  event.isAltDown() == false )
     {                            // PageUp alone: scroll up.
        IViewPort *pvport = (IViewPort*)event.window();

        pvport->sendEvent( WM_VSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT(SB_PAGEUP, 0 ) ,
                           IEventData(
                              pvport->verticalScrollBar()->handle()) );
#else
                           pvport->verticalScrollBar()->id(),
                           MPFROM2SHORT( 0, SB_PAGEUP ));
#endif
        bProcessed = true;
     }
  }                               // End pageUp key.
  else if ( vkey == IKeyboardEvent::pageDown )
  {                               // PageDown/PgDn key.
     if ( event.isCtrlDown() )
     {                            // Ctrl+PageDown: scroll right.
        IViewPort *pvport = (IViewPort*)event.window();
        pvport->sendEvent( WM_HSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT( SB_PAGERIGHT, 0 ),
                           IEventData(
                              pvport->horizontalScrollBar()->handle()) );
#else
                           pvport->horizontalScrollBar()->id(),
                           MPFROM2SHORT( 0, SB_PAGERIGHT ));
#endif
        bProcessed = true;
     }
     else if ( event.isShiftDown() == false  &&  event.isAltDown() == false )
     {                            // PageDown alone: scroll down.
        IViewPort *pvport = (IViewPort*)event.window();
        pvport->sendEvent( WM_VSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT( SB_PAGEDOWN, 0 ),
                           IEventData(
                              pvport->verticalScrollBar()->handle()) );
#else
                           pvport->verticalScrollBar()->id(),
                           MPFROM2SHORT( 0, SB_PAGEDOWN ));
#endif
        bProcessed = true;
     }
  }                               // End pageDown key.

  // Line scrolling.
  else if ( vkey == IKeyboardEvent::up  ||
            vkey == IKeyboardEvent::down )
  {                               // Up or down arrow key.
     if ( event.isCtrlDown() == false  &&
          event.isShiftDown() == false  &&  event.isAltDown() == false )
     {                            // Arrow key alone: line scroll.
        IViewPort *pvport = (IViewPort*)event.window();
        pvport->sendEvent( WM_VSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT( (vkey == IKeyboardEvent::up) ?
                                            SB_LINEUP : SB_LINEDOWN ,
                                         0 ) ,
                           IEventData(
                              pvport->verticalScrollBar()->handle()) );
#else
                           pvport->verticalScrollBar()->id(),
                           MPFROM2SHORT( 0,
                                         (vkey == IKeyboardEvent::up) ?
                                            SB_LINEUP : SB_LINEDOWN) );
#endif
        bProcessed = true;
     }
  }                               // End up or down arrow key.
  else if ( vkey == IKeyboardEvent::left  ||
            vkey == IKeyboardEvent::right )
  {                               // Left or right arrow key.
     if ( event.isCtrlDown() == false  &&
          event.isShiftDown() == false  &&  event.isAltDown() == false )
     {                            // Arrow key alone: line scroll.
        IViewPort *pvport = (IViewPort*)event.window();
        pvport->sendEvent( WM_HSCROLL,
#ifdef  IC_WIN
                           MPFROM2SHORT( (vkey == IKeyboardEvent::left) ?
                                            SB_LINELEFT : SB_LINERIGHT ,
                                         0 ) ,
                           IEventData(
                              pvport->horizontalScrollBar()->handle()) );
#else
                           pvport->horizontalScrollBar()->id(),
                           MPFROM2SHORT( 0,
                                         (vkey == IKeyboardEvent::left) ?
                                            SB_LINELEFT : SB_LINERIGHT) );
#endif
        bProcessed = true;
     }
  }                               // End left or right arrow key.

  /*****************************************************************/
  /* Also need to cause window scrolling to keep the cursor        */
  /* visible during tabbing/cursor movement.                       */
  /*****************************************************************/

  return bProcessed;
}

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::handleEventsFor                                    |
|                                                                              |
| Attach the keyboard handler to the specified view port object.               |
------------------------------------------------------------------------------*/
IViewPortKeyboardHandler&
  IViewPortKeyboardHandler :: handleEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::handleEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::stopHandlingEventsFor                              |
|                                                                              |
| Detach the keyboard handler from the specified view port object.             |
------------------------------------------------------------------------------*/
IViewPortKeyboardHandler&
  IViewPortKeyboardHandler :: stopHandlingEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::stopHandlingEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::handleEventsFor                                    |
------------------------------------------------------------------------------*/
IHandler& IViewPortKeyboardHandler :: handleEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortKeyboardHandler::stopHandlingEventsFor                              |
------------------------------------------------------------------------------*/
IHandler& IViewPortKeyboardHandler :: stopHandlingEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}


/*------------------------------------------------------------------------------
| IViewPortPaintHandler::IViewPortPaintHandler                                 |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IViewPortPaintHandler :: IViewPortPaintHandler ( )
  : IViewPortPaintHandler::Inherited ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::~IViewPortPaintHandler                                |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IViewPortPaintHandler :: ~IViewPortPaintHandler ( )
{ }

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::paintWindow                                           |
|                                                                              |
| Ignore paint events, since child windows will do painting.                   |
------------------------------------------------------------------------------*/
Boolean IViewPortPaintHandler :: paintWindow ( IPaintEvent& event )
{
  IViewPort* pvpViewPort = (IViewPort*)event.window();
  ISize dummy = pvpViewPort->viewWindowSize();
                  // Force view port to update child window positions.

  // Don't need to paint where horizontal scroll bar will.
  IScrollBar* horizontal = pvpViewPort->horizontalScrollBar();
#ifdef IC_PM
  SWP swp;
  WinQueryWindowPos( horizontal->handle(), &swp );
  if ( swp.fl & SWP_SHOW )   // (IWindow::isShowing invalidates too much.)
#endif
#ifndef IC_PM
  if ( IISWINDOWVISIBLE( horizontal->handle() ))
#endif
  {                                    // Scroll bar being used.
     IRectangle rectChild = horizontal->nativeRect();
     RECTL rectlChild = rectChild.asRECTL();
#ifdef IC_PM
     IVALIDATERECT( event.handle(), &rectlChild );
#endif
#ifndef IC_PM
     IVALIDATERECT( event.handle(), (RECT*)&rectlChild );
#endif
  }

  // Don't need to paint where vertical scroll bar will.
  IScrollBar* vertical = pvpViewPort->verticalScrollBar();
#ifdef IC_PM
  WinQueryWindowPos( vertical->handle(), &swp );
  if ( swp.fl & SWP_SHOW )  // (IWindow::isShowing invalidates too much.)
#endif
#ifndef IC_PM
  if ( IISWINDOWVISIBLE( horizontal->handle() ))
#endif
  {                                    // Scroll bar being used.
     IRectangle rectChild = vertical->nativeRect();
     RECTL rectlChild = rectChild.asRECTL();
#ifdef IC_PM
     IVALIDATERECT( event.handle(), &rectlChild );
#endif
#ifndef IC_PM
     IVALIDATERECT( event.handle(), (RECT*)&rectlChild );
#endif
  }

  // Don't need to paint where view rectangle will.
  IWindow* viewRect = IWindow::windowWithOwner( IC_VIEWPORT_VIEWRECTANGLE,
                                                pvpViewPort );
  IRectangle rectChild = viewRect->nativeRect();
  RECTL rectlChild = rectChild.asRECTL();
#ifdef IC_PM
  IVALIDATERECT( event.handle(), &rectlChild );
#endif
#ifndef IC_PM
  IVALIDATERECT( event.handle(), (RECT*)&rectlChild );
#endif

  return false;                        // Always pass event on.
}

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::handleEventsFor                                       |
|                                                                              |
| Attach the paint handler to the specified view port object.                  |
------------------------------------------------------------------------------*/
IViewPortPaintHandler&
  IViewPortPaintHandler :: handleEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::handleEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::stopHandlingEventsFor                                 |
|                                                                              |
| Detach the paint handler from the specified view port object.                |
------------------------------------------------------------------------------*/
IViewPortPaintHandler&
  IViewPortPaintHandler :: stopHandlingEventsFor ( IViewPort* viewport )
{
  IASSERTPARM( viewport != 0 );
  Inherited::stopHandlingEventsFor( viewport );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::handleEventsFor                                       |
------------------------------------------------------------------------------*/
IHandler& IViewPortPaintHandler :: handleEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewPortPaintHandler::stopHandlingEventsFor                                 |
------------------------------------------------------------------------------*/
IHandler& IViewPortPaintHandler :: stopHandlingEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}


/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::IViewRectanglePaintHandler                       |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IViewRectanglePaintHandler :: IViewRectanglePaintHandler ( )
  : IViewRectanglePaintHandler::Inherited ( )
{ }

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::~IViewRectanglePaintHandler                      |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IViewRectanglePaintHandler :: ~IViewRectanglePaintHandler ( )
{ }

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::paintWindow                                      |
|                                                                              |
| Ignore paint events, since child windows will do painting.                   |
------------------------------------------------------------------------------*/
Boolean IViewRectanglePaintHandler :: paintWindow ( IPaintEvent& event )
{
  IViewRectangle* pvrViewRect = (IViewRectangle*)event.window();
  IViewPort* pvpViewPort = (IViewPort*)pvrViewRect->parent();
  if ( pvpViewPort  &&  pvpViewPort->viewWindow() )
  {                                  // Validate where child will paint.
     IRectangle rectViewWindow( IPoint(), pvpViewPort->viewWindowSize() );
                                       // Size of the view window.
     IRectangle rectVisibleView =
         ICoordinateSystem::convertToNative(
            pvpViewPort->viewWindowDrawRectangle(),
            pvpViewPort->size() );
                                       // Portion in view rectangle.
     IPoint visibleOrigin = rectVisibleView.bottomLeft();
     long coord = visibleOrigin.x();
     visibleOrigin.setX( coord >= 0 ? 0 : 0 - coord );
                                    // (Should always end up being 0.)
     coord = visibleOrigin.y();
     visibleOrigin.setY( coord >= 0 ? 0 : 0 - coord );
                          // Adjust for white space below view window.
     IRectangle rectActualVisible = rectViewWindow & rectVisibleView;
     rectActualVisible.moveTo( visibleOrigin );
     RECTL rectlActualVisible = rectActualVisible.asRECTL();
#ifdef IC_PM
     IVALIDATERECT( event.handle(), &rectlActualVisible );
#endif
#ifndef IC_PM
     IVALIDATERECT( event.handle(), (RECT*)&rectlActualVisible );
#endif
  }
  return false;                        // Always pass event on.
}

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::handleEventsFor                                  |
|                                                                              |
| Attach the paint handler to the specified view port object.                  |
------------------------------------------------------------------------------*/
IViewRectanglePaintHandler&
  IViewRectanglePaintHandler :: handleEventsFor ( IViewRectangle* viewRect )
{
  IASSERTPARM( viewRect != 0 );
  Inherited::handleEventsFor( viewRect );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::stopHandlingEventsFor                            |
|                                                                              |
| Detach the paint handler from the specified view port object.                |
------------------------------------------------------------------------------*/
IViewRectanglePaintHandler&
  IViewRectanglePaintHandler :: stopHandlingEventsFor
                                                 ( IViewRectangle* viewRect )
{
  IASSERTPARM( viewRect != 0 );
  Inherited::stopHandlingEventsFor( viewRect );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::handleEventsFor                                  |
------------------------------------------------------------------------------*/
IHandler& IViewRectanglePaintHandler :: handleEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*------------------------------------------------------------------------------
| IViewRectanglePaintHandler::stopHandlingEventsFor                            |
------------------------------------------------------------------------------*/
IHandler& IViewRectanglePaintHandler :: stopHandlingEventsFor ( IWindow* )
{                              // Private to hide version in IHandler.
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}
