/* NOSHIP */
#ifndef _IVRECT_
  #define _IVRECT_
/*******************************************************************************
* FILE NAME: ivrect.hpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*     IViewRectangle                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <icanvas.hpp>
#include <irect.hpp>

#pragma pack(4)

// Forward declarations for other classes:
class IWindow;

class IViewRectangle : public ICanvas {
typedef ICanvas
  Inherited;
/*******************************************************************************
* Objects of the IViewRectangle class are used as an implementation feature of *
* an IViewPort, to provide a view rectangle.  The view rectangle is the        *
* bounding window for the view window, and so is used by the view port to      *
* ensure the view window gets clipped properly (and so doesn't overwrite the   *
* scroll bars of the view port).  The view rectangle is similar to the         *
* intermediate window added by the WC_NOTEBOOK class between the notebook      *
* window itself and a page window.                                             *
*******************************************************************************/
public:
/*-------------------------- Constructor/Destructor ----------------------------
| You can construct instances of this class by specifying a window ID, parent  |
| and owner windows, a position and size, and style.                           |
------------------------------------------------------------------------------*/
  IViewRectangle     ( unsigned long windowId,
                       IWindow* parent,
                       IWindow* owner,
                       const IRectangle& initial,
                       const Inherited::Style& style );
virtual
 ~IViewRectangle     ( );

/*-------------------------------- Overrides -----------------------------------
| This class overrides the following inherited function:                       |
|   setLayoutDistorted - Forwards child minimum size changes to the view port. |
------------------------------------------------------------------------------*/
virtual IViewRectangle
 &setLayoutDistorted ( unsigned long layoutAttributeOn,
                       unsigned long layoutAttributeOff );

private:
/*--------------------------------- Private ----------------------------------*/
  IViewRectangle     ( const IViewRectangle& );
IViewRectangle
 &operator=          ( const IViewRectangle& );
}; // IViewRectangle

#pragma pack()

#endif /* _IVRECT_ */
