/*******************************************************************************
* FILE NAME: imetatyp.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*     Definition of the functions found in imetatyp.hpp                        *                                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*******************************************************************************/
#include <imetatyp.hpp>
#include <icontext.hpp>
#include <ibasstrm.hpp>
#include <itypeint.hpp>
#include <iexcept.hpp>
#include <icconsta.h>
#include <itrace.hpp>

#define PVMARK(str)

/*------------------------------------------------------------------------------
| IMetaYpeInfo:IMetaTypeInfo                                                   |
|                                                                              |
| Constructor from a character string.                                         |
------------------------------------------------------------------------------*/
IMetaTypeInfo::IMetaTypeInfo(const char* typeName)
    :   theName(typeName),
        fInternal(0)
{
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:IMetaTypeInfo                                                   |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IMetaTypeInfo::IMetaTypeInfo()
{
    // should not be called
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:IMetaTypeInfo                                                   |
|                                                                              |
| Copy Constructor.                                                            |
------------------------------------------------------------------------------*/
IMetaTypeInfo::IMetaTypeInfo(const IMetaTypeInfo& other)
{
    // should not be called (why protected?)
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:operator=                                                       |
|                                                                              |
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
IMetaTypeInfo& IMetaTypeInfo::operator=(const IMetaTypeInfo& other)
{
    // should not be called (
    return *this;
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:~IMetaTypeInfo                                                  |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IMetaTypeInfo::~IMetaTypeInfo()
{
    delete [] (char*)theName;
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:GetTypeInternal                                                 |
|                                                                              |
| Return the type.                                                             |
------------------------------------------------------------------------------*/
ITypeInternal* IMetaTypeInfo::GetTypeInternal() const
{
    return fInternal;
}

/*------------------------------------------------------------------------------
| IMetaYpeInfo:SetTypeInternal                                                 |
|                                                                              |
| Set the type.                                                                |
------------------------------------------------------------------------------*/
void IMetaTypeInfo::SetTypeInternal(ITypeInternal* typeInternal)
{
    fInternal = typeInternal;
}

//
// IType class
//
TypeExtensionMacro(IMetaType)

/*------------------------------------------------------------------------------
| IMetaYpe::IMetaType                                                          |
|                                                                              |
| Constructor from IMetaTypeInfo                                               |
------------------------------------------------------------------------------*/
IMetaType::IMetaType(const IMetaTypeInfo& t)
    :   fInternalObject(t.GetTypeInternal())
{
}

/*------------------------------------------------------------------------------
| IMetaYpe::IMetaType                                                          |
|                                                                              |
| Copy Constructor.                                                            |
------------------------------------------------------------------------------*/
IMetaType::IMetaType(const IMetaType& other)
    :   fInternalObject(other.fInternalObject)
{
}

/*------------------------------------------------------------------------------
| IMetaYpe::IMetaType                                                          |
|                                                                              |
| Default Constructor.                                                         |
------------------------------------------------------------------------------*/
IMetaType::IMetaType()
    :   fInternalObject(0)
{
}

/*------------------------------------------------------------------------------
| IMetaYpe::IMetaType                                                          |
|                                                                              |
| Constructor from character class name.                                       |
------------------------------------------------------------------------------*/
IMetaType::IMetaType(const char* theClassName)
    :   fInternalObject(0)
{
    fInternalObject = ITypeInternal::GetType(theClassName);
}

/*------------------------------------------------------------------------------
| IMetaYpe::~IMetaType                                                         |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IMetaType::~IMetaType()
{
}

/*------------------------------------------------------------------------------
| IMetaYpe::operator=                                                          |
|                                                                              |
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
IMetaType& IMetaType::operator=(const IMetaType& other)
{
    if (this != &other) {
        fInternalObject = other.fInternalObject;
    }
    return *this;
}

/*------------------------------------------------------------------------------
| IMetaYpe::operator==                                                         |
|                                                                              |
| Equality operator.                                                           |
------------------------------------------------------------------------------*/
IBoolean IMetaType::operator==(const IMetaType& other) const
{
    return (*fInternalObject == *other.fInternalObject);
}

/*------------------------------------------------------------------------------
| IMetaYpe::operator==                                                         |
|                                                                              |
| Inequality operator.                                                         |
------------------------------------------------------------------------------*/
IBoolean IMetaType::operator!=(const IMetaType& other) const
{
    return (*fInternalObject != *other.fInternalObject);
}


// =======================================================================================

/*------------------------------------------------------------------------------
| IMetaYpe::typeInfo                                                           |
|                                                                              |
| Return a constant reference to the stored IMetaTypeInfo.                     |
------------------------------------------------------------------------------*/
const IMetaTypeInfo& IMetaType::typeInfo() const
{
    return *fInternalObject->GetTypeInfo();
}

/*------------------------------------------------------------------------------
| IMetaYpe::GetTypeInternal                                                    |
|                                                                              |
| Return a pointer to the internal type.                                       |
------------------------------------------------------------------------------*/
ITypeInternal* IMetaType::GetTypeInternal() const
{
    return fInternalObject;
}

/*------------------------------------------------------------------------------
| IMetaYpe::operator>>=                                                        |
|                                                                              |
| Stream out operator.                                                         |
------------------------------------------------------------------------------*/
IBaseStream& IMetaType::operator>>= (IBaseStream& toWhere) const
{
    IASSERTSTATE(fInternalObject != 0);

    fInternalObject->GetClassName() >>= toWhere;
    return toWhere;
}

/*------------------------------------------------------------------------------
| IMetaYpe::operator<<=                                                        |
|                                                                              |
| Stream in operator.                                                          |
------------------------------------------------------------------------------*/
IBaseStream& IMetaType::operator<<= (IBaseStream& fromWhere)
{
    IString className;
    className <<= fromWhere;
    fInternalObject = ITypeInternal::GetType(className);
    return fromWhere;
}

/*------------------------------------------------------------------------------
| IMetaYpe::name                                                               |
|                                                                              |
| Store the type name.                                                         |
------------------------------------------------------------------------------*/
void IMetaType::name(IString& name) const
{
    name = fInternalObject->GetClassName();
}

//
// class ITypeInstaller
//

ITypeInstaller::ITypeInstaller(const char* className,
                               const IMetaTypeInfo* (*tgbFunction)(ITypeHelper*))
{
    ITypeInternal::AddType(className, tgbFunction);
}

//
// IType helper class
//

ITypeHelper::ITypeHelper()
    :   fObject(0),
        fObjectType(0),
        fBaseType(0),
        fStream(0),
        fCreatedContext(false)
{
}

ITypeHelper::ITypeHelper(IBaseStream& s)
    :   fObject(0),
        fObjectType(0),
        fBaseType(0),
        fStream(&s),
        fCreatedContext(false)
{
}

ITypeHelper::~ITypeHelper()
{
    if (fStream && fCreatedContext) {
        fStream->setContext(0);
    }
}

void* ITypeHelper::CantConstructAbstractClass()
{
   ITHROWLIBRARYERROR(IC_STREAM_ABSTRACT_TYPE,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
   return 0;
}

void ITypeHelper::PrepareContext()
{
PVMARK("ITypeHelper::PrepareContext");

    if (fStream) {
        IContext* context = fStream->context();
        if (context == 0) {
            fCreatedContext = true;
            context = fStream->defaultContext();
            context->reset();
            fStream->setContext(context);
        }
    }
}

void ITypeHelper::SetObject(const void* obj, const ITypeInternal* t)
{
    if (obj != 0) {
        fObject = obj;
        fObjectType = t;
    } else
        fBaseType = t;
}

const void* ITypeHelper::GetObject() const
{
    return fObject;
}

void ITypeHelper::ThrowObject() const
{
    fObjectType->CallTGDFunction(fObject);
}

void ITypeHelper::CaughtObject(const void* obj)
{
    // At this point we may know some of the following:
    //
    //      fObjectType is the ITypeInternal for the derived class of the object, if not 0
    //      fBaseType is the ITypeInternal for the base class of the object, if not 0
    //

    ((ITypeInternal*)fObjectType)->SetLastBaseClass((ITypeInternal*)fBaseType, (long)(((char*)obj) - ((char*)fObject)));
    fObject = obj;
}

IContext::ObjectNumber ITypeHelper::UpdateContext(const void* newObj,
                                IBoolean& update) const
{
PVMARK("ITypeHelper::UpdateContext");
    IContext::ObjectNumber ret;
    IContext* context = fStream->context();  // won't be 0, caller assures that there's a context
    ret = context->add(newObj, update);
//qprintf("IType: 0x%x: %s %d\n", newObj, update ? "added" : "found", ret);
    return ret;
}

const void* ITypeHelper::LookupContext(IContext::ObjectNumber refNum) const
{
PVMARK("ITypeHelper::LookupContext");
    IContext* context = fStream->context();      // non-0 because of PrepareContext() by caller...
    const void* ret = context->find(refNum);
//qprintf("IType: 0x%x: looking %d\n", ret, refNum);
    return ret;
}

IBaseStream &operator <<=(ITypeHelper::EPreType& theValue, IBaseStream& fromwhere)
{
    unsigned char temp;
    temp <<= fromwhere;
    theValue = (ITypeHelper::EPreType) temp;
    return fromwhere;
}

IBaseStream &operator >>=(ITypeHelper::EPreType theValue, IBaseStream& towhere)
{
    unsigned char temp = theValue;
    temp >>= towhere;
    return towhere;
}

const IMetaTypeInfo* ITypeHelper::StaticCreate(
        const IString& className,
        const IMetaTypeInfo* (*tgbFunction)(ITypeHelper* helper),
        void* (*tgdFunction)(const void* obj))
{
    ITypeInternal* t = new ITypeInternal(className, tgbFunction, tgdFunction);
    return t->GetTypeInfo();
}

IString ITypeHelper::MakeClassName(
        const char* tName, const char* ptype1, const char* ptype2)
{
    IString name = tName;
    name += '<';
    name += ptype1;
    if (ptype2) {
        name += ',';
        name += ptype2;
    }
    name += '>';
    return name;
}

//
// dynamic_cast() implementation
//
const void* ITypeHelper::DynamicCastImplementation(
        const IMetaTypeInfo& derivedTypeInfo,
        const IMetaTypeInfo& castToTypeInfo,
        const void* obj)
{
PVMARK("ITypeHelper::DynamicCastImplementation");
    const void* ret = 0;

    ITypeInternal* t = derivedTypeInfo.GetTypeInternal();
    const void* mostDerived = obj;

    ITypeInternal* castToTypeInternal = castToTypeInfo.GetTypeInternal();
    long adjustToBase = 0;
    if (castToTypeInternal == t || castToTypeInternal == t->GetLastBaseClass(adjustToBase)) {
        ret = ((char*)mostDerived) + adjustToBase;
    } else {
        ITypeHelper helper;
        helper.SetObject(mostDerived, t);
        helper.SetObject(0, castToTypeInternal);
        try {
            (void) castToTypeInternal->CallTGBFunction(&helper);
            ret = helper.GetObject();
        } catch (void* ) {
            ret = 0;  // the dynamic cast failed
        }

    }
    return ret;
}

//
// ::Flatten(const AType* object) implementation
//
IBoolean ITypeHelper::MustStreamOutObject(const IMetaTypeInfo& baseTypeInfo, const IMetaTypeInfo* typeInfo, const void* obj)
{
PVMARK("ITypeHelper::MustStreamOutObject");

    IBoolean ret = false;

    if (obj == 0) {

        // Stream out a representation of 0...
        //
        ITypeHelper::EPreType header = ITypeHelper::kNil;
        header >>= *fStream;

        // ret = false;     // Obviously, we don't try to stream out 0 pointer...

    } else {

        // We have a non-0 object...
        //
        ITypeInternal* t = typeInfo->GetTypeInternal();
        const void* mostDerivedObj = obj;
        PrepareContext();

        IBoolean firstSeen = false;
        IContext::ObjectNumber refNum = UpdateContext(mostDerivedObj, firstSeen);
        if (firstSeen) {
        PVMARK("ITypeHelper::MustStreamOutObject - first");

            // Since this object isn't in our context, that means that this is
            // the first time we're flattening this object.  Obtain the top
            // of the object, and squirrel that along with the type of the object,
            // in our context...

            //
            // With the bookkeeping done, stream out a representation of the type
            // of this object.
            //
            ITypeHelper::EPreType header = ITypeHelper::kObject;
            header >>= *fStream;
            t->GetClassName() >>= *fStream;

            ret = true;     // On first encounter of object, stream it out...

//qprintf("(%d) ::Flatten(): %s\n", getpid(), t->GetTypeInfo()->name());
            #ifdef IC_TRACE_DEVELOP
                printf("ITypeHelper::MustStreamOutObject: userobj 0x%x adjobj 0x%x ref 0x%x\n",
                    obj, mostDerivedObj, refNum);
            #endif

        } else {
        PVMARK("ITypeHelper::MustStreamOutObject - !first");

            // We've previously flattened out this object.  We just
            // send its reference number, which ought to correspond to that
            // entry in the stream-in side's context...
            //
            ITypeHelper::EPreType header = ITypeHelper::kReference;
            header >>= *fStream;
            t->GetClassName() >>= *fStream;
            refNum >>= *fStream;

            #ifdef IC_TRACE_DEVELOP
                printf("ITypeHelper::MustStreamOutObject: userobj 0x%x ref 0x%x\n",
                    obj, refNum);
            #endif

            // ret = false; // Since hit detected in context, don't stream this out again...
//qprintf("(%d) ::Flatten(): %s (ref)\n", getpid(), t->GetTypeInfo()->name());
        }
    }
    return ret;
}

//
// ::Resurrect(AType*& objPointer, IBaseStream& stream) implementation
//
const void* ITypeHelper::ConstructObject(const IMetaTypeInfo& baseTypeInfo, IBoolean& mustStreamInObject)
{
PVMARK("ITypeHelper::ConstructObject");

    const void* ret;
    ITypeInternal* t;

    mustStreamInObject = false;

    ITypeHelper::EPreType header;

    // Figure out what kind of thing we have on the stream...
    //
    header <<= *fStream;
    switch (header) {

    case ITypeHelper::kNil:
        //
        // To indicate 0, we just return 0...
        //
        return 0;

    case ITypeHelper::kObject: {
    PVMARK("ITypeHelper::kObject");
        //
        // A first-time object is constructed, put into the context
        // entry, and returned...
        //
        PrepareContext();
        IString className;
        className <<= *fStream;
        t = ITypeInternal::GetType(className);
        void* newObj = t->CallTGDFunction((const void*)0);
//qprintf("::Resurrect(): 0x%x: %s\n", newObj, t->GetTypeInfo()->name());

        // At this point, enough is known to cache the address point of the top
        // of the object...
        //
//      TAddressPoint::SetType(newObj, t);

        // We must update our context, so that we'll properly handle repeated
        // references to this object and this type...  It is important to update the
        // context with the object first, followed by the type, since this is the same order
        // that it occurs on the flatten side.
        //
        IBoolean force = true;

        #ifdef IC_TRACE_DEVELOP
           IContext::ObjectNumber ref = UpdateContext(newObj, force);
           printf("ITypeHelper::MustCatchConstructedObject: kObject 0x%x ref 0x%x\n",
                newObj, ref);
        #else
           UpdateContext(newObj, force);
        #endif

        ret = newObj;
        mustStreamInObject = true;

        break;
    }

    case ITypeHelper::kReference: {
    PVMARK("ITypeHelper::kReference");

        IString className;
        className <<= *fStream;
        t = ITypeInternal::GetType(className);

        //
        // A repeated object is just pulled out of the context, and it is
        // also thrown so that our caller can catch it as the desired type...
        //
        PrepareContext();

        IContext::ObjectNumber refNum;
        refNum <<= *fStream;
        const void* obj = LookupContext(refNum);
        if (obj == 0) {
           ITHROWLIBRARYERROR(IC_STREAM_NO_OBJECT_IN_CONTEXT,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
        }

        #ifdef IC_TRACE_DEVELOP
// e undefined
//            printf("ITypeHelper::MustCatchConstructedObject: kReference obj 0x%x ref 0x%x\n",
//                e->GetObject(), refNum);
//            printf("  (cont...) type 0x%x %s\n", e->GetObjectType(), e->GetObjectType()->GetTypeInfo()->name());
        #endif

        ret = obj;
        mustStreamInObject = false;

        break;
    }

    default:
        ITHROWLIBRARYERROR(IC_STREAM_INVALID,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
    }

    ITypeInternal* baseTypeInternal = baseTypeInfo.GetTypeInternal();
    long adjustToBase = 0;
    if (baseTypeInternal == t || baseTypeInternal == t->GetLastBaseClass(adjustToBase)) {
        ret = ((char*)ret) + adjustToBase;
    } else {
        SetObject(ret, t);
        SetObject(0, baseTypeInternal);
        IBoolean mismatch = false;
        try {
            (void) baseTypeInternal->CallTGBFunction(this);
        } catch (void* ) {
            mismatch = true;
        }
        if (mismatch)
           ITHROWLIBRARYERROR(IC_STREAM_OBJECT_BASE_MISMATCH,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
        ret = GetObject();
    }

    return ret;
}

//
// ::CreateNewObject(AType*&, const IType& theType) implementation
//
const void* ITypeHelper::ConstructObject(const IMetaTypeInfo& baseTypeInfo,
                                         const IMetaType& theType)
{
    ITypeInternal* t = theType.GetTypeInternal();
//qprintf("(%d) ::CreateNewObject(): %s\n", getpid(), t->GetTypeInfo()->name());
    const void* ret = t->CallTGDFunction((const void*)0);
    ITypeInternal* baseTypeInternal = baseTypeInfo.GetTypeInternal();
    long adjustToBase = 0;
    if (baseTypeInternal == t || baseTypeInternal == t->GetLastBaseClass(adjustToBase)) {
        ret = ((char*)ret) + adjustToBase;
    } else {
        ITypeHelper helper;
        helper.SetObject(ret, t);
        helper.SetObject(0, baseTypeInternal);
        IBoolean mismatch = false;
        try {
            (void) baseTypeInternal->CallTGBFunction(&helper);
        } catch (void* ) {
            mismatch = true;
        }
        if (mismatch)
           ITHROWLIBRARYERROR(IC_STREAM_OBJECT_BASE_MISMATCH,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
        ret = helper.GetObject();
    }
    return ret;
}


