// *******************************************************************************
// * FILE NAME: icompimp.hpp                                                      *
// *                                                                              *
// * DESCRIPTION:                                                                 *
// *   Implementation of the classes:                                             *
// *     IAswServSpace                                                            *
// *     IComponentImp                                                            *
// *                                                                              *
// * COPYRIGHT:                                                                   *
// *   IBM Open Class Library                                                     *
// *   (C) Copyright International Business Machines Corporation 1992, 1996       *
// *   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
// *   US Government Users Restricted Rights - Use, duplication, or disclosure    *
// *   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
// *                                                                              *
// *   Portions or this component are based on work originating from Lotus.       *
// *   (C) Copyright Lotus Development Corporation. 1996 - All Rights Reserved.   *
// *                                                                              *
// *******************************************************************************/
#include <precomp.hpp>

#include "icompimp.hpp"

#include <icompnen.hpp>
#include <iembmod.hpp>
#include <iview.hpp>
#include <iguibndl.hpp>
#include <istatnry.hpp>
#include <icompfac.hpp>
#include <idocstor.hpp>
#include <imetatyp.hpp>
#include <icmpfram.hpp>
#include <imemstrm.hpp>

#include <ititle.hpp>
#include <imenubar.hpp>
#include <itbarfrm.hpp>

// have our component output to the Aswan CLASS debug channel
#define CH_DOCFWK_COMPONENT CH_CLASS_OBJINST


// *******************************************************************************
// *                               IAswServSpace                                 *
// *******************************************************************************

IAswServSpace::IAswServSpace( PASWCONTROL pControl )
   : CAswServSpace( pControl ),
     fpBundle( NULL )
{
}


IAswServSpace::~IAswServSpace(void)
{
   // Shutdown the base and leave
   Shutdown(TRUE);
}


void
IAswServSpace::Setup( IGUIBundle& bundle )
{
    CAswServSpace::Setup(TRUE); // TRUE: we ARE a frame space
    fpBundle = &bundle;
}


CAswResult IAswServSpace::AddTools( LPRECT prectBorder,
                                    BOOL fIsMDIContainer )
{
    CAswResult ar;
    do
    {
        IToolBar& toolBar = fpBundle->objectToolBar();
        IFrameWindow* toolFrame = toolBar.floatingFrame();
        ISize toolSize = toolBar.size();

        // reparent the toolbar to the container's frame window
        // so that it will minimize with the container
        HWND hwndParent;
        GetWindow(&hwndParent);
        SetParent((HWND)(void*)toolFrame->handle(), hwndParent);
        // SetParent seems to resize the tool bar, so reset it
        toolFrame->sizeTo( toolSize );

        ar = ResizeBorder(prectBorder);
        if (ASWFAILED(ar))
        {
            SetParent( (HWND)(void*)toolFrame->handle(),
                       (HWND)(void*)fpBundle->inPlaceFrame().handle() );
            break;
        }

        toolFrame->show();

    }
    while (FALSE);

    return ar;
}


void IAswServSpace::RemoveTools( void )
{

    IFrameWindow* toolFrame = fpBundle->objectToolBar().floatingFrame();

    toolFrame->hide();

    SetParent( (HWND)(void*)toolFrame->handle(),
               (HWND)(void*)fpBundle->inPlaceFrame().handle() );

}


CAswResult IAswServSpace::ResizeBorder( LPCRECT prcBorder )
{
    BORDERWIDTHS   bw;
    CAswResult     ar;

    do
    {
        // Make sure we can reserve space for what we need.  If this
        // fails then we just do without because the menu has
        // everything we really need.  A more willing server could
        // put tools in popup windows as well.

        IToolBar& toolBar = fpBundle->objectToolBar();
        ISize toolSize = toolBar.size();
        IPair::Coord toolHeight = toolSize.height();

        SetRect( &bw, 0, toolHeight, 0, 0 );

        ar = RequestBorderSpace(&bw);
        if( ASWFAILED(ar) )
        {
//          IASWSHOWERROR( ar );
            break;
        }

        //Safety net:  RequestBorderSpace may modify values in bw
        SetRect(&bw, 0, toolHeight, 0, 0);

        ar = SetBorderSpace(&bw);
        if( ASWFAILED(ar) )
        {
//          IASWSHOWERROR( ar );
            break;
        }

        // move it to 0,0 in the client area of the container's frame window
        toolSize.setWidth( prcBorder->right - prcBorder->left );
        toolBar.floatingFrame()->moveSizeTo( IRectangle( IPoint(0,0), toolSize ) );

    }
    while( FALSE );

    return ar;
}

// *******************************************************************************
// *                               IComponentImp                                 *
// *******************************************************************************


// ***************************************************************************
//                          Initialization and termination
// ***************************************************************************

// -----------------------------------------------------------------------------
// IComponentImp
// -----------------------------------------------------------------------------
IComponentImp::IComponentImp(IComponent* comp) :
    CAswObjectInstance( IComponentStationery::stationery(), NULL ),
    fpComponent( comp ),
    fpBundle( NULL ),
    fpStorage( NULL ),
    fServMenu( IComponentStationery::stationery() ),
    fSharedMenu( NULL ),
    fServFrameSpace( IComponentStationery::stationery() ),
    fShuttingDown( FALSE ),
    fIsStreamShutdown( FALSE ),
    fEmbedded( FALSE ),
    fpAswFormatSet( NULL ),
    fAswStream( IComponentStationery::stationery() )
{
}

// -----------------------------------------------------------------------------
// ~IComponentImp
// -----------------------------------------------------------------------------
IComponentImp::~IComponentImp()
{

//    SHUTDOWNANDDELETE( fpAswFormatSet );    // bbb

    if ( fSharedMenu )
        delete fSharedMenu;

    Shutdown( TRUE );
}

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------
CAswResult
IComponentImp::Setup( IGUIBundle* bundle, IModel* model )
{
    ASWMETHOD("Setup", (CH_DOCFWK_COMPONENT));
    ASWASSERT(bundle);
    ASWSETUPENTRYF("bundle=%p",(ASWSTR,bundle));

    CAswResult  ar;

    do
    {
        // Setup the base class
        ar = CAswObjectInstance::Setup( &IComponentStationery::stationery().componentFactory(),
                                        CAswObjectInstance::oitImplementedStorage |
                                        CAswObjectInstance::oitImplementedFile );
        if( ASWFAILED(ar) )
        {
            IASWSHOWERROR( ar );
            break;
        }

        // set up a stream inside the storage
        ar = fAswStream.Setup();
        if( ASWFAILED(ar) )
        {
            IASWSHOWERROR( ar );
            break;
        }

        // Set the bundle pointer
        fpBundle = bundle;

        // ------------------------------
        // supported clipboard formats
        // ------------------------------

        // register clipboard formats that we can serve up
        const int numFormats = 7;
        FORMATETC aFormats[ numFormats ];

        // format 0 - our model as an entity (can be absorbed by containers that "know" us)
        // must first register it as a valid clipboard format
        // using the model name as the format name
        fcfNative = (CLIPFORMAT) RegisterClipboardFormat(
                                           dynamicTypeInfo( *model ).name() );
        aFormats[0].cfFormat = fcfNative;
        aFormats[0].ptd = NULL;
        aFormats[0].dwAspect = DVASPECT_CONTENT;
        aFormats[0].lindex = -1;
        aFormats[0].tymed = TYMED_HGLOBAL;

        // format 1 - our model as an embed source (cannot be absorbed)
        fcfEmbedSource = (CLIPFORMAT)RegisterClipboardFormat( CF_EMBEDSOURCE );
        aFormats[1].cfFormat = fcfEmbedSource;
        aFormats[1].ptd = NULL;
        aFormats[1].dwAspect = DVASPECT_CONTENT;
        aFormats[1].lindex = -1;
        aFormats[1].tymed = TYMED_ISTORAGE;

        // format 2 - METAFILE
        // this is how we want to look in cached presentations
        aFormats[2].cfFormat = CF_METAFILEPICT;
        aFormats[2].ptd = NULL;
        aFormats[2].dwAspect = DVASPECT_CONTENT;
        aFormats[2].lindex = -1;
        aFormats[2].tymed = TYMED_MFPICT;

        // format 3 - Icon
        // called for if the user says "display as icon" during embedding or
        // if s/he converts representation with convert dialog
        aFormats[3].cfFormat = CF_METAFILEPICT;
        aFormats[3].ptd = NULL;
        aFormats[3].dwAspect = DVASPECT_ICON;
        aFormats[3].lindex = -1;
        aFormats[3].tymed = TYMED_MFPICT;

        // object descriptor
        aFormats[4].cfFormat = IComponentStationery::stationery().cfObjectDescriptor();
        aFormats[4].dwAspect = DVASPECT_CONTENT;
        aFormats[4].ptd      = NULL;
        aFormats[4].tymed    = TYMED_HGLOBAL;
        aFormats[4].lindex   = -1;

        // link source
        aFormats[5].cfFormat = IComponentStationery::stationery().cfLinkSource();
        aFormats[5].ptd      = NULL;
        aFormats[5].dwAspect = DVASPECT_CONTENT;
        aFormats[5].lindex   = -1;
        aFormats[5].tymed    = TYMED_ISTREAM;

        // link source descriptor
        aFormats[6].cfFormat = IComponentStationery::stationery().cfLinkSrcDescriptor();
        aFormats[6].dwAspect = DVASPECT_CONTENT;
        aFormats[6].ptd      = NULL;
        aFormats[6].tymed    = TYMED_HGLOBAL;
        aFormats[6].lindex   = -1;

        // Setup a new Aswan Source format set
        fpAswFormatSet = new CAswSourceFormatSet( IComponentStationery::stationery() );
        ar = fpAswFormatSet->Setup( aFormats, numFormats );
        if( ASWFAILED(ar) )
        {
            IASWSHOWERROR( ar );
            break;
        }

        // Give the format set to the object instance
        SetFormatSet( fpAswFormatSet );

    }
    while( FALSE );

    // If we did not succeed, then make sure we are shutdown
    if( ASWFAILED(ar) )
        Shutdown(TRUE);

    ASWEXITR(ar);

}


// -----------------------------------------------------------------------------
// Shutdown
// -----------------------------------------------------------------------------
CAswResult
IComponentImp::Shutdown( BOOL fEmergency )
{
    ASWMETHOD("Shutdown", (CH_DOCFWK_COMPONENT));
    ASWSHUTDOWNENTRY(fEmergency);

    CAswResult  ar;

    // check to see if we are already shutting down
    if( !fShuttingDown )
    {
        fShuttingDown = TRUE;

        // check to see if shutdown is OK (call base class method IsShutdownOK)
        // or shutdown unconditionally if this is an emergency shutdown
        if( fEmergency || IsShutdownOK() )
        {
            // Shutdown the members
            ar = fAswStream.Shutdown(fEmergency);
            ASWASSERT(ASWSUCCEEDED(ar));

            ar = fpAswFormatSet->Shutdown(fEmergency);
            ASWASSERT(ASWSUCCEEDED(ar));

            // Shutdown the base class
            ar = CAswObjectInstance::Shutdown(fEmergency);
            ASWASSERT(ASWSUCCEEDED(ar));
        }
        else
        {
            // not OK to shutdown now
            ar.Set(ASW_E_SHUTDOWNNOTOK, ASD_UNCHANGED);
        }

        fShuttingDown = FALSE;
    }

    else
    {
        // we are already shutting down
        ar.Set(S_OK, ASD_OK);
    }

    ASWSHUTDOWNEXITR(ar);

}


// ***************************************************************************
//                              Standard Server Methods
// ***************************************************************************


// -----------------------------------------------------------------------------
// OnClose
// -----------------------------------------------------------------------------
// Method called to notify an application that an object is closing.
// Based on dwSaveOption, applications should make sure their changes
// are saved and prepare to be shutdown.
CAswResult
IComponentImp::OnClose( DWORD dwSaveOption )
{
    ASWMETHOD("OnClose", (CH_DOCFWK_COMPONENT));
    ASWENTRYF("dwSaveOption=%lx", (ASWSTR, dwSaveOption));

    CAswResult ar(S_OK, ASD_OK);
    BOOL fSave = FALSE;

    // If object is dirty and we're asked to save, save it and close.
    if( dwSaveOption == OLECLOSE_SAVEIFDIRTY && GetIsDirty() )
        fSave = TRUE;

    // If asked to prompt, just save; this precludes the case where all
    // embedded components ask the user to save
    if( dwSaveOption == OLECLOSE_PROMPTSAVE && GetIsDirty() )
            fSave = TRUE;

    if ( fSave )
        // If asked, then tell the container to save us
        SaveObject();

    // Once we have saved, we should release our hold on any sub-storages
    // or streams.   Once close is finished, your object should not
    // access the sub-storages or streams again.
    fAswStream.Shutdown();
    fIsStreamShutdown = TRUE;

    ASWEXITR(ar);
}

// -----------------------------------------------------------------------------
// ObjectReleased
// -----------------------------------------------------------------------------
void
IComponentImp::ObjectReleased()
{

    // if we are NOT embedded (ie standalone) we dont do this
    // as it will be done in the File-Exit menu option handler
    if( fpComponent->isEmbedded() )
    {
        // shutdown if in-place windows exist.  If not, shutdown has
        // already been done in IGUIBundle::handleExit
        if ( fpBundle->destroyInPlaceWindows() )
        {
            // close our self and our embedded components (if any)
            CloseObjectInstance( OLECLOSE_NOSAVE );

            // finalize on view required before frame window is closed
            fpBundle->view().finalize();

            // send a close message to the frame window
            fpBundle->frameWindow().close();
        }

    }
}

// -----------------------------------------------------------------------------
// Launch
// -----------------------------------------------------------------------------
// informs the application that an object is being activated for
// out-of-place editing.
CAswResult
IComponentImp::Launch( LONG iVerb )
{
    ASWMETHOD("Launch", (CH_DOCFWK_COMPONENT));
    ASWENTRYF("iVerb=%ld", (ASWSTR, iVerb));

    // show our main window
    HWND hWnd = frameWindowHandle();
    ShowWindow( hWnd, SW_SHOWNORMAL );
    SetFocus( hWnd );

    ::PostMessage( hWnd, WM_COMMAND, kViewToolbar, 0 );

    // inform the world that we have shown our window
    AdviseOnShowWindow();

    ASWEXITR(CAswResult(S_OK, ASD_OK));
}

// -----------------------------------------------------------------------------
// Hide
// -----------------------------------------------------------------------------
// inform an application that an object is being hidden.
CAswResult
IComponentImp::Hide(LONG iVerb)
{
    ASWMETHOD("Hide", (CH_DOCFWK_COMPONENT));
    ASWENTRYF("iVerb=%ld", (ASWSTR, iVerb));

    // Hide the window and send the advise
    HWND hwndServer = frameWindowHandle();
    ShowWindow( hwndServer, SW_HIDE);

    // send out a notification that the window was sucessfully hidden
    AdviseOnHideWindow();

    ASWEXITR(CAswResult(S_OK, ASD_OK));
}



// -----------------------------------------------------------------------------
// SetHostNames
// -----------------------------------------------------------------------------
// method called to provide us with the container name and document name
// as a result we modify the title bar to show: <our app name> in <container obj>
CAswResult
IComponentImp::SetHostNames( LPASWCOLESTR pszContainerApp,
                             LPASWCOLESTR pszContainerObj )
{

    // convert from OLE strings to ANSI
    CAswANSIString AnsiContainerApp( pszContainerApp );
    CAswANSIString AnsiContainerObj( pszContainerObj ); // could be NULL

    // store in member vars
    // hostName() method will try to return Obj unless it is blank
    fstrHostAppName = IString( AnsiContainerApp );
    fstrHostObjName = IString( AnsiContainerObj );

    // set the title in the frame window
    ITitle title( &fpBundle->frameWindow() );

    // "%1 in %2"
    const char* strTitle = IBase::messageText(IC_CDF_EMBEDDED_TITLE,
                                         IComponentStationery::stationery().appName(),
                                         hostName() );
    title.setTitleText( strTitle );

    return ( CAswResult( S_OK, ASD_OK ) );
}



// ***************************************************************************
//                              In-Place Activation
// ***************************************************************************


// -----------------------------------------------------------------------------
// OnInPlaceActivate
// -----------------------------------------------------------------------------
// called by the framework when we are about to go in place
// Call the bundle to create a view to be used as the in-place view
void
IComponentImp::OnInPlaceActivate(void)
{
    // switch to in-place view, etc.
    fpBundle->useInPlaceWindows();

    // get the default menus from the bundle; the client view will have also
    // automatically inserted its menus during initialization
    IMenuBar& menuBar = fpBundle->objectMenuBar();

    if ( fSharedMenu == NULL )
        fSharedMenu = new ASWSHAREDMENU[menuBar.numberOfItems()];

    IMenu::Cursor menuCursor( fpBundle->objectMenuBar() );
    unsigned menuPos = 0;

    for ( menuCursor.setToFirst(); menuCursor.isValid(); menuCursor.setToNext() )
    {
        IMenuItem menuItem = menuBar.elementAt( menuCursor );
        switch ( menuItem.id() ) {
        case kEditMenu:
            fSharedMenu[menuPos].nGroup = ASW_MENUGROUP_EDIT;
            break;
        case kHelpMenu:
            fSharedMenu[menuPos].nGroup = ASW_MENUGROUP_HELP;
            break;
        default:
            fSharedMenu[menuPos].nGroup = ASW_MENUGROUP_OBJECT;
            break;
        }

        fSharedMenu[menuPos].nPosition = menuPos;
        menuPos++;
    }

//FB shouldn't we do setup and shutdown once (in constructor/destructor)
    fServMenu.Shutdown();
    fServFrameSpace.Shutdown();

    HMENU hmenuBar = (HMENU)(void*)menuBar.menuHandle();
    fServMenu.Setup( inPlaceViewHandle(), hmenuBar, menuPos, fSharedMenu );
//    fServMenu.Setup( (HWND)(void*)fpBundle->view().handle(), hmenuBar, menuPos, fSharedMenu );
    fServFrameSpace.Setup( *fpBundle );

    PASWSERVSPACE  pServFrameSpace = NULL;
    PASWSERVSPACE  pServDocSpace   = NULL;
    PASWSERVACCEL  pServAccel      = NULL;
    PASWSERVUNDO   pServUndo       = NULL;

    CAswOLEString szObjName( IApplication::current().userResourceLibrary()
                                               .loadString(IC_CDF_CLSID_AUX_2) );

    SetInPlaceOptions( &fServMenu,
                       &fServFrameSpace,
                        pServDocSpace,
                        pServAccel,
                        pServUndo,
                        szObjName.GetOLEPtr() );

}

// -----------------------------------------------------------------------------
// AdjustObjectRect
// -----------------------------------------------------------------------------
// Adjust the object rect during resize operations.
void
IComponentImp::AdjustObjectRect( LPRECT prcPos )
{

    SIZEL sizeHi, sizePix;
    sizePix.cx = ((*prcPos).right - (*prcPos).left) * fdZoomX;
    sizePix.cy = ((*prcPos).bottom - (*prcPos).top) * fdZoomY;
    XformSizeInPixelsToHimetric( NULL, &sizePix, &sizeHi );
    SetExtent( IComponent::kContent, &sizeHi );

}

// -----------------------------------------------------------------------------
// OnInPlaceDeactivate
// -----------------------------------------------------------------------------
// tell the server object that it's doing and inplace deactivate. Call the bundle to
// destroy the in-place view and return back to the out of place view
void
IComponentImp::OnInPlaceDeactivate(void)
{
    IEmbedderModel* embModel = NULL;
    ::dynamicCastTo( embModel, fpComponent->model() );

    // unselect any selected component
    if ( embModel != NULL && embModel->selectedComponent() != NULL )
        embModel->selectComponent( NULL );

    // switch back to out-of-place view, etc.
    fpBundle->useInPlaceWindows( false );

}


// -----------------------------------------------------------------------------
// GetInPlaceServerWindow
// -----------------------------------------------------------------------------
// This method is used by containers for a number of things:
//  - asks us to return our in-place server window handle
//  - when we go inplace active we must reparent to the provided parent window
//    ( the aswan hatch border )
//  - when we go out of in-place activation (NULL given for parent) we reparent back
//    to out internal frame window
//  - we get a position rect which identifies the current rectangle (in client pixels)
//    the first time through we compute the zoom factor, subsequent times we change
//    our extents to the new size (but factored by the zoom factor)
HWND
IComponentImp::GetInPlaceServerWindow( HWND hwndParent, RECT rcClip, LPRECT prcPos )
{


    HWND hwndObject = inPlaceViewHandle();
    HWND hwndCurrentParent = GetParent( hwndObject );

    // see if a parent window was provided
    if( hwndParent )
    {
        // position rect size
        long lWidth = (*prcPos).right - (*prcPos).left;
        long lHeight = (*prcPos).bottom - (*prcPos).top;
        ISize sizeCurrent = fpComponent->contentSize();

        // is this the first time?
        if( hwndParent != hwndCurrentParent )
        {

            // reparent to the container's frame window (aswan hatch border)
            ::SetParent( hwndObject, hwndParent );

            // get the ratio of the container rect to our current extents
            // this is the "zoom factor"
            fdZoomX = (double)sizeCurrent.width() / (double)lWidth;
            fdZoomY = (double)sizeCurrent.height() / (double)lHeight;

        }

    }

    else
    {
        // no parent window provided so reparent back to our frame window
        HWND hwndFrame = (HWND)(void*)fpBundle->inPlaceFrame().handle();
        SetParent( hwndObject, hwndFrame );
    }

    return hwndObject;

}

// -----------------------------------------------------------------------------
// HideInPlaceWindow
// -----------------------------------------------------------------------------
// tell the server object to hide its window (InPlace editing only)
BOOL
IComponentImp::HideInPlaceWindow(void)
{

    BOOL bOK = FALSE;

    // Hide the window and send the advise
    HWND hWnd = inPlaceViewHandle();

    if( hWnd )
    {
        ShowWindow( hWnd, SW_HIDE );
        // send out a notification that the window was sucessfully hidden
        AdviseOnHideWindow();
        bOK = TRUE;
    }

    return bOK;

}

// -----------------------------------------------------------------------------
// ShowInPlaceWindow
// -----------------------------------------------------------------------------
// tell the server object to show its window (InPlace editing only)
BOOL
IComponentImp::ShowInPlaceWindow(void)
{

    BOOL bOK = FALSE;

    // show our inplace window
    HWND hWnd = inPlaceViewHandle();

    if( hWnd )
    {
        ShowWindow( hWnd, SW_SHOWNORMAL );
        SetFocus( hWnd );

        // inform the world that we have shown our window
        AdviseOnShowWindow();
        bOK = TRUE;
    }

    return bOK;
}


// ***************************************************************************
//                              Resize Methods
// ***************************************************************************


// -----------------------------------------------------------------------------
// SetExtent
// -----------------------------------------------------------------------------
// called by the framework to reset the size of the object. This method delegates
// to the handleSetArea method.
CAswResult
IComponentImp::SetExtent( DWORD dwDrawAspect, LPSIZEL lpsizel )
{
    ASWMETHOD("SetExtent", (CH_DOCFWK_COMPONENT));
    ASWASSERT(lpsizel);
    ASWENTRYF("dwDrawAspect=0x%lx,lpsizel->cx=%ld,lpsizel->cy=%ld",
                    (ASWSTR,dwDrawAspect,lpsizel->cx,lpsizel->cy));

    // delegate to handleSetArea
    fpComponent->handleSetArea( (IComponent::EAspect)dwDrawAspect, ISize(*lpsizel) );

    ASWEXITR(CAswResult(S_OK, ASD_OK));
}



// -----------------------------------------------------------------------------
// GetExtent
// -----------------------------------------------------------------------------
// called by the framework to get our size. We delegate to handleGetArea.
CAswResult
IComponentImp::GetExtent( DWORD dwDrawAspect, LPSIZEL lpsizel )
{
    ASWMETHOD("GetExtent", (CH_DOCFWK_COMPONENT));
    ASWASSERT(lpsizel);
    ASWENTRYF("dwDrawAspect=0x%lx,lpsizel=%p",(ASWSTR,dwDrawAspect,lpsizel));

    // get the size
    *lpsizel = fpComponent->handleGetArea
                                ( (IComponent::EAspect)dwDrawAspect ).asSIZEL();

    ASWEXITR(CAswResult(S_OK, ASD_OK));
}


// ***************************************************************************
//                          Server Storage Methods
// ***************************************************************************
// Note: the server storage methods that used to be here have been moved
// to the file IDocStor.cpp since this file was too big for the compiler.



// ***************************************************************************
//                          Server/Container Methods
// ***************************************************************************

// no overrides of the base class methods required



// ***************************************************************************
//                              DATA METHODS
// ***************************************************************************

// -----------------------------------------------------------------------------
// GetNativeClipFormat
// -----------------------------------------------------------------------------
// Returns our native clipboard format (into the out paramater)
CAswResult
IComponentImp::GetNativeClipFormat( LPCLIPFORMAT pFormat )
{
    ASWMETHOD("GetNativeClipFormat", (CH_CLASS_OBJINST));
    ASWASSERT(pFormat);
    ASWENTRYF("pFormat=%p",(ASWSTR,pFormat));

    *pFormat = fcfNative;

    ASWEXITR(CAswResult(S_OK,ASD_OK));
}


// -----------------------------------------------------------------------------
// GetData
// -----------------------------------------------------------------------------
// Called when we are required to render our data

// These are the formats we promised (in Setup) to support:
//   0. Native,             CONTENT,    TYMED_HGLOBAL
//   1. EmbedSource         CONTENT,    TYMED_ISTORAGE
//   2. METAFILEPICT,       CONTENT,    TYMED_MFPICT
//   3. METAFILEPICT,       ICON,       TYMED_MFPICT
//   4. ObjectDescriptor    CONTENT,    TYMED_HGLOBAL   * not implemented here
//   5. LinkSource          CONTENT,    TYMED_ISTREAM
//   6. LinkSrcDescriptor   CONTENT,    TYMED_HGLOBAL   * not implemented here

// * - NOTE: the object descriptors are implemented in the idataex.cpp versions of
//           GetData. This should be the only place that these are required

CAswResult
IComponentImp::GetData( LPFORMATETC lpFormatEtc, PASWSTGMEDUNION pAswStgMedium )
{

    ASWMETHOD("GetData", (CH_DOCFWK_COMPONENT));
    ASWENTRYF("lpFormatEtc=%p,pAswStgMedium=%p", (ASWSTR,lpFormatEtc,pAswStgMedium) );

    CAswResult ar(DATA_E_FORMATETC, ASD_UNCHANGED);

    // did the calling program allocate the space? (if it did it would provide a "tymed")
    // if not, then we will render it into the medium we want and tell the calling program
    if( pAswStgMedium->tymed == TYMED_NULL )
    {

        HGLOBAL hGlobal = NULL;

        // check for the aspect requested - CONTENT and ICON supported

        // DVASPECT_CONTENT
        if ( lpFormatEtc->dwAspect & DVASPECT_CONTENT )
        {

            // NOTE: Object descriptors are NOT provided here, even though
            // we promised apps querying our format set that we would.
            // Typically the only time object descriptors are required is for
            // data exchange (ie. cut/copy/paste & drag/drop) so the object
            // descriptors are implemented there.

            // render the data into the format specified
            hGlobal = RenderFormat( lpFormatEtc->cfFormat );
            if (hGlobal)
            {

                // tell the calling container which storage medium we used
                switch (lpFormatEtc->cfFormat)
                {
                    case CF_METAFILEPICT:
                        pAswStgMedium->tymed = TYMED_MFPICT;
                        break;

                    default:
                        pAswStgMedium->tymed = TYMED_HGLOBAL;
                        break;
                }
                pAswStgMedium->hGlobal = hGlobal;
                ar.Set(S_OK, ASD_OK);
            }

            else // RenderFormat failed
            {
                // we did not handle the get
                #ifdef IC_SCREEN_DEBUG
                    cout << "*** ERROR Render format failed" << endl;
                #endif
                assert( false );
            }


        } // DVASPECT_CONTENT


        // see if this is requesting our ICON
        else
            if( ( lpFormatEtc->dwAspect & DVASPECT_ICON  ) &&
                    ( lpFormatEtc->cfFormat == CF_METAFILEPICT ) &&
                    ( lpFormatEtc->tymed & TYMED_MFPICT ) )
        {

            CAswOLEString strClsId = (char*)IApplication::current().userResourceLibrary().
                                             loadString(IC_CDF_CLSID);
            CLSID clsId;
            CLSIDFromString( strClsId.GetOLEPtr(), &clsId );
            // call OLE to get the icon that is registered for this class in the system registry
            hGlobal = OleGetIconOfClass( clsId, NULL, FALSE );

            if( hGlobal )
            {
                pAswStgMedium->tymed = TYMED_MFPICT;
                pAswStgMedium->hGlobal = hGlobal;
                ar.Set(S_OK, ASD_OK);
            }

            else
            {
                // no default icon registered for this CLSID - generate some default one
                #ifdef IC_SCREEN_DEBUG
                    cout << "*** ERROR could not generate icon" << endl;
                #endif
                assert( false );
            }

        } // DVASPECT_ICON


    } // we pick the storage medium to use


    // The calling program allocated the space
    else
    {

        // ---------------------------------------------------
        // This is mainly to process a CF_EMBEDSOURCE format
        // ---------------------------------------------------
        // the only case we handle is embedded source
        if( (lpFormatEtc->cfFormat == fcfEmbedSource ) )
        {
            // Use the object instance to get the embed source data
            ar = GetEmbedSourceData( pAswStgMedium->pAswStorage );
        }

        // ---------------------------------------------------
        // This is mainly to process a CF_LINKSOURCE format
        // ---------------------------------------------------
        // are they asking for a link?
        if( pAswStgMedium->tymed == TYMED_ISTREAM &&
            lpFormatEtc->cfFormat == IComponentStationery::stationery().cfLinkSource() )
        {

            // Render an embedded object into a storage
            ar = GetLinkSourceData( pAswStgMedium->pAswStream );
        }


    } // the calling program allocated the space


    ASWEXITR( ar );


} // GetData




// -----------------------------------------------------------------------------
// RenderFormat
// -----------------------------------------------------------------------------
// Called from GetData (above) to render the data in the specified clipboard format
HGLOBAL
IComponentImp::RenderFormat( CLIPFORMAT cf )
{


    HGLOBAL hMem = NULL;

    // our entire model
    if( cf == fcfNative )
        hMem = GetModelData();

    // metafile presentation (for cache)
    if( cf == CF_METAFILEPICT )
        hMem = RenderMetafilePict();

    // return the memory pointer or NULL if an error occurred
    return hMem;
}



// -----------------------------------------------------------------------------
// RenderMetafilePict
// -----------------------------------------------------------------------------
// This method is called from RenderFormat when the framework calls for a
// metafile representation of the data. It creates an empty metafile, then calls
// the view's draw contents method to draw into the metafile
HGLOBAL
IComponentImp::RenderMetafilePict( Boolean inPlaceView )
{

    // Create an empty memory metafile and return its handle.
//    HDC hDC = (HDC)CreateEnhMetaFile( NULL, NULL, NULL, NULL );
    HDC hDC = (HDC)CreateMetaFile( NULL );
    if( hDC == NULL )
        return NULL;

    // Set the mapping mode so that the metafile is "scalable"
    // then set the size (f into the metafile
    SetMapMode( hDC, MM_ANISOTROPIC );

    // get the current extent of our window in pixels
    RECT rc;
    rc.left = 0;
    rc.top = 0;
    rc.right = fpComponent->contentSize().width();
    rc.bottom = fpComponent->contentSize().height();

    // set origin of bitmap to 0,0
    SetWindowOrgEx( hDC, 0, 0, NULL );
    SetWindowExtEx( hDC, rc.right, rc.bottom, NULL );

    //Erase the background for bitmaps and metafiles.
    HPEN hPen = (HPEN)SelectObject( hDC, GetStockObject(NULL_PEN) );
    Rectangle( hDC, rc.left, rc.top, rc.right+1, rc.bottom+1 );
    SelectObject( hDC, hPen );

    // call the view to draw into the supplied DC
    // (hDC must be wrapped in an IPresSpaceHandle)
    IPresSpaceHandle ps( hDC );
    IRectangle irc( rc );
    IView& theView = fpBundle->view();
    theView.draw( ps, irc, TRUE );

//    HENHMETAFILE hMF = CloseEnhMetaFile( hDC );
    HMETAFILE hMF = CloseMetaFile( hDC );
    if( hMF == NULL )
        return NULL;

    // now that we have a handle to the metafile we wrap it into a METAFILEPICT structure
    // in global memory
    HGLOBAL hMem = GlobalAlloc( GMEM_DDESHARE | GMEM_MOVEABLE, sizeof(METAFILEPICT) );
    if( hMem == NULL )
    {
//        DeleteEnhMetaFile(hMF);
        DeleteMetaFile(hMF);
        return NULL;
    }

    LPMETAFILEPICT pMF = (LPMETAFILEPICT)GlobalLock( hMem );
    pMF->hMF = hMF;
    pMF->mm = MM_ANISOTROPIC;

    // Insert the extents in MM_HIMETRIC units.
    pMF->xExt = fpComponent->fExtent.width();
    pMF->yExt = fpComponent->fExtent.height();

    GlobalUnlock(hMem);
    return hMem;

}

// -----------------------------------------------------------------------------
// GetModelData
// -----------------------------------------------------------------------------
// Call out to model to provide the data
HGLOBAL
IComponentImp::GetModelData()
{

//FB do we need to copy the buffer length into the buffer as well ???
    IMemoryStream stream;
    ::flatten( fpComponent->model(), stream );
//FB what about embedded components ?????

    // allocate global memory
    HGLOBAL hMem = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE, stream.position() );
    if( hMem )
    {
        // call on the model to stream out the data
        void* lpDest=(void*)GlobalLock(hMem);

        memcpy( lpDest, stream.buffer(), stream.position() );

        GlobalUnlock(hMem);

    }

    return hMem;

}


//--------------------------------------------------
// SERVER STORAGE METHODS
//--------------------------------------------------

// ***************************************************************************
// *
// * IComponentImp::LoadFile()
// *
// * Description:
// *    Loads the document that is contained in the given file name. The loaded
// *    document is not to be shown to the user.
// *
// * Arguments:
// *    LPCSTR pszFileName
// *       Points to the absolute path of the file to open
// *
// *    DWORD dwReserved
// *       Reserved
// *
// * Returns:
// *    S_OK           Document was successfully loaded
// *    E_FAIL         Document could not be loaded
// *    E_OUTOFMEMORY  Out of memory
// *    Any STG_E_* value
// *
// ****************************************************************************/
CAswResult
IComponentImp::LoadFile( LPASWCOLESTR pszFileName,
                         DWORD )
{
   ASWMETHOD("LoadFile", (CH_DOCFWK_COMPONENT));
   ASWENTRY();
   ASWSHOWOLESTR( pszFileName );

   CAswResult  ar;


   // *** warning ***
   // The Aswan string classes are not public
   CAswANSIString astr( pszFileName );

   IString fileName = astr.GetANSIPtr();
   try
   {
       fpComponent->storage().loadFile( fileName );
       ar.Set( S_OK, ASD_OK );
   }

   catch ( IAswException& e )
   {
       ar = e.aswResult();
   }

   catch ( IException& )
   {
       ar.Set( STG_E_READFAULT, ASD_UNCHANGED );
   }

   ASWEXITR(ar);
}

// ***************************************************************************
// *
// * IComponentImp::SaveFile()
// *
// * Description:
// *    Saves a copy of the object to the indicated filename. The pszFilename
// *    can be NULL, indicating that a simple "File Save" operation is requested
// *    to the now current filename. A non-NULL value indicates either a
// *    "File Save As" or a "File Save Copy As" operation. In this case, the
// *    full path of the destination file is specified in pszFilename.
// *
// *    The fRemember parameter determines if a "Save As" or "Save Copy As"
// *    operation is requested by the user. If fRemember is TRUE, then the
// *    file pointed to by pszFilename should become the current working
// *    copy of the document ("Save As"); FALSE indicates that after completing
// *    the Save call, the working copy should remain whatever it was before
// *    the call ("Save Copy As").
// *
// *    If, upon exiting, the saved file is the current working copy, the
// *    internal dirty flag maintained by the object should be cleared and the
// *    caller must call SaveFileCompleted as some later time when he or she
// *    is done with the data in the file.
// *
// * Arguments:
// *    LPCSTR pszFileName
// *       Absolute path name of the file to be saved (can be NULL)
// *
// *    BOOL fRemember
// *       Valid only if pszFilename is NULL. Determines which disk file the
// *       document is to be logically associated with after the call completes.
// *
// ****************************************************************************/
CAswResult
IComponentImp::SaveFile( LPASWCOLESTR pszFileName,
                         BOOL         remember )
{
   ASWMETHOD("SaveFile", (CH_DOCFWK_COMPONENT));
   ASWENTRYF("remember=%s",(ASWSTR,ASWSHOWTF(remember)));
   ASWSHOWOLESTR(pszFileName);

   CAswResult  ar;

   // *** warning ***
   // The Aswan string classes are not public
   CAswANSIString astr(pszFileName);

   IString fileName = fpStorage->fileToUse( (LPSTR) astr.GetANSIPtr() );
   Boolean sameAsLoad = ( pszFileName == NULL );

   try {
       // save the data ("Save", "Save as", or "Save copy as")
       fpStorage->saveFile( fileName, sameAsLoad, remember );

       ar.Set(S_OK, ASD_OK);
   }
   catch ( IException& ) {
       ar.Set(STG_E_WRITEFAULT, ASD_UNCHANGED);
   }
//FB should we catch IAswException also??

   ASWEXITR(ar);
}

/*+**************************************************************************
 *
 * IComponentImp::SaveFileCompleted()
 *
 * Description:
 *    Indicates that the caller has saved the file with a call to SaveFile
 *    and is finished working.
 *
 * Arguments:
 *    LPCSTR pszFileName
         Pointer to the file in which the document was previously saved.
 *
 * Returns:
 *    S_OK     In all cases
 *
 ****************************************************************************/
CAswResult
IComponentImp::SaveFileCompleted( LPASWCOLESTR pszFileName )
{
   ASWMETHOD("SaveFileCompleted", (CH_DOCFWK_COMPONENT));
   ASWENTRY();
   ASWSHOWOLESTR(pszFileName);
   ASWEXITR(CAswResult(S_OK,ASD_OK));
}

/***************************************************************************
 *
 * IComponentImp::GetCurFile()
 *
 * Description:
 *    Returns either the absulute path name of the document's currently-
 *    associated file or the default filename prompt, if there is no
 *    currently-associated file.
 *
 * Arguments:
 *    LPSTR pszFileName
 *       Points to an allocated buffer to return the current path or the
 *       default save prompt of the document, as appropriote.
 *
 *    UINT cchMax
 *       Allocated length of the pszFileName buffer.
 *
 * Returns:
 *    S_OK           Return value is a path name
 *    S_FALSE        Return value is the save prompt
 *    E_FAIL         Undefined failure
 *
 * Warnings:
 *    The pszFileName buffer is ALREADY allocated. This method only needs
 *    to fill in the information. The only restriction is that the filename
 *    can not be longer than cchMax.
 *
 ****************************************************************************/
CAswResult
IComponentImp::GetCurFile( LPOLESTR pszFileName,
                           UINT     cchMax )
{
   ASWMETHOD("GetCurFile", (CH_DOCFWK_COMPONENT));
   ASWENTRYF("pszFileName=%p,cchMax=%d",(ASWSTR,pszFileName,cchMax));

   CAswResult  ar;
   CAswOLEString oleStr;

   IString name = fpStorage->fileName();

   //If we have no filename, return the prompt for File Open/Save.
   if ( 0 == name.length() ) {
       IString prompt = IString( '*' ) + IComponentStationery::stationery().fileExt();
       oleStr.TranslateA2W( prompt );
       ar.Set(S_FALSE,ASD_OK);
   }
   else {
       oleStr.TranslateA2W( name );
       ar.Set(S_OK,ASD_OK);
   }

   ASWLSTRCPYNW( pszFileName, oleStr, cchMax );

   ASWEXITR(ar);
}

/*+**************************************************************************
 *
 * IComponentImp::InitNewStorage()
 *
 ****************************************************************************/
CAswResult
IComponentImp::InitNewStorage( PASWSTORAGE pAswStg )
{
   ASWMETHOD("InitNewStorage", (CH_DOCFWK_COMPONENT));
   ASWENTRYF("pAswStg=%s", (ASWSTR, ASWCLASSID(pAswStg)));

   // if this method gets called, we are embedded.
   fEmbedded = TRUE;

   /*
    * The rules of IPersistStorage mean we pre-create anything
    * we'd need in SaveStorage (...,TRUE) for low-memory situations.
    * For us this means creating our "CONTENTS" stream and holding
    * onto that AswStream here.
    */

   CAswResult ar = pAswStg->CreateChildStream( &fAswStream, OLESTR("CONTENTS"),
                                STGM_DIRECT | STGM_CREATE | STGM_READWRITE |
                                STGM_SHARE_EXCLUSIVE );

   // set the new storage to be the one to use
   fpStorage->setStorage( pAswStg );

   ASWEXITR(ar);
}

/*+**************************************************************************
 *
 * IComponentImp::LoadStorage()
 *
 ****************************************************************************/
CAswResult
IComponentImp::LoadStorage( PASWSTORAGE pAswStg )
{
   ASWMETHOD("LoadStorage", (CH_DOCFWK_COMPONENT));
   ASWASSERT(pAswStg);
   ASWENTRYF("pAswStg=%s", (ASWSTR, ASWCLASSID(pAswStg)));

   CAswResult  ar (S_OK, ASD_OK);

   // open the correct child stream where the data ought to be
   ar = pAswStg->OpenChildStream(&fAswStream, OLESTR("CONTENTS"), STGM_DIRECT |
                                    STGM_READWRITE | STGM_SHARE_EXCLUSIVE);

   // if wo could open the stream, try to get the data from it
   if (ASWSUCCEEDED(ar))
   {
       // orphan and delete the old model
       delete fpComponent->orphanModel();

       try {
           // loadStorage, if successful, will adopt a new model
           fpStorage->loadStorage( pAswStg, &fAswStream );
       }

       //FB should we catch IAswException also??
       catch ( IException& ) {
           // if load failed, use an empty model
           fpComponent->adoptModel( IComponentStationery::stationery().createModel() );
           ar.Set(STG_E_READFAULT, ASD_UNCHANGED);
       }

   }

   // set this storage as the one to use for setup of new embedded components
   fpStorage->setStorage( pAswStg );

   ASWEXITR(ar);
}

/*+**************************************************************************
 *
 * IComponentImp::SaveStorage()
 *
 ****************************************************************************/
CAswResult
IComponentImp::SaveStorage( PASWSTORAGE pAswStgSave,
                            BOOL        sameAsLoad )
{
    ASWMETHOD("SaveStorage", (CH_DOCFWK_COMPONENT));
    ASWASSERT(pAswStgSave);
    ASWENTRYF("pAswStgSave=%s,sameAsLoad=%s",
       (ASWSTR, ASWCLASSID(pAswStgSave), ASWSHOWTF(sameAsLoad)));

    CAswResult  ar(S_OK, ASD_OK);
    CAswStream* strm = NULL;

    if (sameAsLoad )
    {
        if ( fIsStreamShutdown )
        {
            ar.Set( E_FAIL, ASD_UNCHANGED );
            ASWEXITR( ar );
        }

        LARGE_INTEGER   li;
        LISet32(li, 0);

        // Go to the beginning in the old stream
        ar = fAswStream.Seek(li, STREAM_SEEK_SET, NULL);
        if ( ASWFAILED( ar ) )
        {
            ASWEXITR( ar );
        }

        strm = &fAswStream;
    }

    try
    {
        fpStorage->saveStorage( pAswStgSave, strm, sameAsLoad, FALSE );
    }

    //FB should we catch IAswException also??
    catch ( IException& )
    {
        ar.Set(STG_E_WRITEFAULT, ASD_UNCHANGED);
    }


    ASWEXITR(ar);
}

// ***************************************************************************
// *
// * IComponentImp::SaveStorageCompleted()
// *
// ***************************************************************************
CAswResult
IComponentImp::SaveStorageCompleted( PASWSTORAGE pAswStgNew )
{
   ASWMETHOD("SaveStorageCompleted", (CH_DOCFWK_COMPONENT));
   ASWENTRYF("pAswStgNew=%s", (ASWSTR, ASWCLASSID(pAswStgNew)));

   CAswResult ar( S_OK, ASD_OK );

   // If pIStorage is NULL, then we don't need to do anything
   // since we already have all the pointers we need for Save.
   // Otherwise we have to release any held pointers and
   // reinitialize them from pIStorage.
    if( pAswStgNew )
    {

        do
        {
            PASWCONTROL pCtl = GetControl();
            if( !pCtl )
            {
                // Major catastrophe - Aswan Control gone!
                ASWASSERT( pCtl );
                ar.Set( E_FAIL, ASD_UNCHANGED );
                break;
            }

            CAswStream AswStream( pCtl );
            ar = AswStream.Setup ();
            if( ASWFAILED( ar ) )
            {
                break;
            }

            ar = pAswStgNew->OpenChildStream( &AswStream,
                                              OLESTR("CONTENTS"),
                                              STGM_DIRECT | STGM_READWRITE | STGM_SHARE_EXCLUSIVE );
            if( ASWFAILED( ar ) )
            {
                break;
            }

            fAswStream.Shutdown( TRUE );
            fAswStream.Setup();
            fAswStream.SetGivenStream( AswStream.GetRealStream() );


        } while( FALSE );

    }

    ASWEXITR( ar );
}

// ***************************************************************************
// *
// * IComponentImp::HandsOffStorage()
// *
// ***************************************************************************

CAswResult
IComponentImp::HandsOffStorage( void )
{
   ASWMETHOD("HandsOffStorage", (CH_DOCFWK_COMPONENT));
   ASWENTRY();

   CAswResult ar (S_OK, ASD_OK);

   // Re-initialize our held streams, if any
   if (NULL != fAswStream.GetRealStream ())
   {
      CAswResult arLocal = fAswStream.Shutdown (TRUE);
      // Shouldn't fail
      ASWASSERT(ASWSUCCEEDED(arLocal));
      ar = fAswStream.Setup ();
   }

   ASWEXITR(ar);
}
