/*******************************************************************************
* FILE NAME: immtime.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immtime.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #define INCL_MCIOS2
   #define IC_MULTIMEDIA
   #include <immdefs.h>
   }

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif

#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif

#include <immtime.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFRAME_CPP_
  #include <ipagetun.h>
#endif

#pragma data_seg(ICLStaticConst)
const unsigned long
  IMMTime::defaultTime = 0xFFFFFFFFL;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMMTime::~IMMTime                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::~IMMTime()
{
   IMODTRACE_DEVELOP("IMMTime::~IMMTime") ;
}

/*------------------------------------------------------------------------------
| IMMTime::IMMTime                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::IMMTime(unsigned long time)
       : fValue(time),
         fValid(true)
{
  IMODTRACE_DEVELOP("IMMTime::IMMTime(unsigned long)") ;
  if (time == defaultTime)
  {
     fValue = 0;
     fValid = false;
  }
}

/*------------------------------------------------------------------------------
| IMMTime::IMMTime                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::IMMTime(const IMMTime &time)
       : fValue(time.fValue),
         fValid(time.fValid)
{
   IMODTRACE_DEVELOP("IMMTime::IMMTime(IMMTime)") ;
}

/*------------------------------------------------------------------------------
| IMMTime::asMMTime                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::asMMTime() const
{
  IMODTRACE_DEVELOP("IMMTime::asMMTime") ;
  return fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::isValid                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::isValid() const
{
  IMODTRACE_DEVELOP("IMMTime::isValid") ;
  return fValid;
}

/*------------------------------------------------------------------------------
| IMMTime::setMMTime                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::setMMTime(unsigned long newTime)
{
  IMODTRACE_DEVELOP("IMMTime::setMMTime") ;
  fValue = newTime;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::setValid                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::setValid(Boolean isValid)
{
  IMODTRACE_DEVELOP("IMMTime::isValid") ;
  fValid = isValid;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::asString                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMTime::asString() const
{
   IMODTRACE_DEVELOP("IMMTime::asString") ;
   unsigned long msec(MSECFROMMM(fValue));
   unsigned long hours   = 0;
   unsigned long minutes = 0;
   unsigned long seconds = 0;

   if (msec)
   {
      hours = msec / (1000 * 60 * 60);
      msec %= 1000 * 60 * 60;
   }
   IString hoursStr(hours);
   if (msec)
   {
      minutes = msec / (1000 * 60);
      msec   %= 1000 * 60;
   }
   IString minutesStr(minutes);
   if (msec)
   {
      seconds = msec / 1000;
      msec   %= 1000;
   }
   IString secondsStr(seconds);
   IString msecStr(msec);
   minutesStr.rightJustify(2, '0');
   secondsStr.rightJustify(2, '0');
   msecStr.rightJustify(3, '0');
   return  IString(hoursStr   + ":" + minutesStr + ":" +
                   secondsStr + "." + msecStr);
}

/*------------------------------------------------------------------------------
| IMMTime::hours                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::hours() const
{
  IMODTRACE_DEVELOP("IMMTime::hours") ;
  unsigned long msec(MSECFROMMM(fValue));
  return msec / (1000 * 60 * 60);
}

/*------------------------------------------------------------------------------
| IMMTime::minutes                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::minutes() const
{
   IMODTRACE_DEVELOP("IMMTime::minutes") ;
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  return msec / (1000 * 60);
}

/*------------------------------------------------------------------------------
| IMMTime::seconds                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::seconds() const
{
  IMODTRACE_DEVELOP("IMMTime::seconds") ;
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  msec %= 1000 * 60;         // strip minutes component
  return msec / 1000;
}

/*------------------------------------------------------------------------------
| IMMTime::hundredths                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::hundredths() const
{
  IMODTRACE_DEVELOP("IMMTime::hundredths") ;
  unsigned long temp(thousandths());
  Boolean roundUp(temp % 10 > 4);
  temp /= 10;
  if (roundUp)
     ++temp;
  return temp;
}

/*------------------------------------------------------------------------------
| IMMTime::thousandths                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::thousandths() const
{
  IMODTRACE_DEVELOP("IMMTime::thousandths") ;
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  msec %= 1000 * 60;         // strip minutes component
  msec %= 1000;              // strip seconds component
  return msec;
}

/*------------------------------------------------------------------------------
| IMMTime::operator unsigned long                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMMTime::operator(unsigned long)") ;
  return fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator =                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator = (const IMMTime &time)
{
  IMODTRACE_DEVELOP("IMMTime::operator=") ;
  if (this != &time)   // check for assignment to self
  {
     fValue = time.asMMTime();
     fValid = time.isValid();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator +                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime IMMTime::operator + (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator+") ;
  IASSERTSTATE(isValid() && time.isValid());
  return IMMTime(fValue + time.asMMTime());
}

/*------------------------------------------------------------------------------
| IMMTime::operator -                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime IMMTime::operator - (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator-") ;
  IASSERTSTATE(isValid() && time.isValid());
  return IMMTime(fValue - time.asMMTime());
}

/*------------------------------------------------------------------------------
| IMMTime::operator +=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator += (const IMMTime &time)
{
  IMODTRACE_DEVELOP("IMMTime::operator+=") ;
  IASSERTSTATE(isValid() && time.isValid());
  fValue += time.asMMTime();
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator -=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator -= (const IMMTime &time)
{
  IMODTRACE_DEVELOP("IMMTime::operator-=") ;
  IASSERTSTATE(isValid() && time.isValid());
  fValue -= time.asMMTime();
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator <                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator < (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator<") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue < time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator >                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator > (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator>") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue > time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator ==                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator == (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator==") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue == time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator !=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator != (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator!=") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue != time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator <=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator <= (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator<=") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue <= time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator >=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator >= (const IMMTime &time) const
{
  IMODTRACE_DEVELOP("IMMTime::operator>=") ;
  IASSERTSTATE(isValid() && time.isValid());
  return fValue >= time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::ordinal                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::ordinal() const
{
  IMODTRACE_DEVELOP("IMMTime::ordinal") ;
  return MSECFROMMM(fValue);
}

/*------------------------------------------------------------------------------
| IMMTime::setTimeToOrdinal                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime& IMMTime::setTimeToOrdinal(unsigned long ordinal)
{
  IMODTRACE_DEVELOP("IMMTime::setTimeToOrdinal") ;
  setMMTime(MSECTOMM(ordinal));
  return *this;
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(const IMMMillisecondTime &time)
                  : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMMillisecondTime::IMMMillisecondTime(IMMMillisecondTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(const IMMTime &time)
                  : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMMillisecondTime::IMMMillisecondTime(IMMTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(unsigned long time)
                  : IMMTime(MSECTOMM(time))
{
  IMODTRACE_DEVELOP("IMMMillisecondTime::IMMMillisecondTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::operator unsigned long                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMMMillisecondTime::operator(unsigned long)") ;
  return MSECFROMMM(asMMTime());
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::~IMMMillisecondTime                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::~IMMMillisecondTime()
{
   IMODTRACE_DEVELOP("IMMMillisecondTime::~IMMMillisecondTime") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::IMMHourMinSecFrameTime                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(unsigned long val, unsigned long fps)
                      : IMMTime(HourMinSecFrameToMM(val, fps)),
                        fFramesPerSec(fps),
                        fOriginal(val)
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(value, fps)") ;
  if (val == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
     fOriginal = 0;
  }
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::IMMHourMinSecFrameTime                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(const IMMHourMinSecFrameTime& time)
                      : IMMTime(time),
                        fFramesPerSec(time.fFramesPerSec),
                        fOriginal(time.fOriginal)
{
   IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(IMMHourMinSecFrameTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::~IMMHourMinSecFrameTime                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::~IMMHourMinSecFrameTime()
{
   IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::~IMMHourMinSecFrameTime()") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::asString                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMHourMinSecFrameTime::asString() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::asString") ;
  return IString(hours()) + ":" +
         IString(minutes()).rightJustify(2, '0') + ":" +
         IString(seconds()).rightJustify(2, '0') + "." +
         IString(frames()).rightJustify(2, '0');
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::framesPerSecond                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::framesPerSecond() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::framesPerSecond") ;
  return fFramesPerSec;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::operator unsigned long                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::operator unsigned long() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::operator(unsigned long)") ;
  return ((IMMHourMinSecFrameTime*)this)->HourMinSecFrameFromMM(asMMTime(),fFramesPerSec);
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::ordinal                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::ordinal() const
{
   IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::ordinal") ;
   // change internal HourMinSecF value to frames
   // this will allow us to convert back to HourMinSecFrame from ordinal
  return (seconds() + minutes() * 60 + hours() * 60 * 60) *
         framesPerSecond() + frames();
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::setTimeToOrdinal                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime& IMMHourMinSecFrameTime::setTimeToOrdinal(unsigned long ordinal)
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::setTimeToOrdinal") ;
  unsigned long h, m, s, f;

  // first get odd frames (not on a second boundary)
  f = ordinal % framesPerSecond();
  ordinal -= f;
  ordinal /= framesPerSecond();

  // now ordinal is in seconds
  h = ordinal / (60 * 60);
  ordinal %= 60 * 60;

  m = ordinal / 60;
  s = ordinal % 60;

  fOriginal = h | (m << 8) | (s << 16) | (f << 24);
  setMMTime(HourMinSecFrameToMM(fOriginal, framesPerSecond()));
  setValid(true);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::frames                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::frames() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::frames") ;
  return (((unsigned long) (*this)) >> 24) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::hours                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::hours() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::hours") ;
  return ((unsigned long) (*this)) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::minutes                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::minutes() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::minutes") ;
  return (((unsigned long) (*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::seconds                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::seconds() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecFrameTime::seconds") ;
  return (((unsigned long) (*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::IMM24FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 24)
{
  IMODTRACE_DEVELOP("IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::IMM24FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(const IMM24FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{
   IMODTRACE_DEVELOP("IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(IMM24FramesPerSecondTime&)") ;
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::operator unsigned long                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMM24FramesPerSecondTime::operator(unsigned long)") ;
  return (unsigned long)(FPS24FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::~IMM24FramesPerSecondTime                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::~IMM24FramesPerSecondTime()
{
   IMODTRACE_DEVELOP("IMM24FramesPerSecondTime::~IMM24FramesPerSecondTime") ;
}


/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::IMM25FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(const IMM25FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{
   IMODTRACE_DEVELOP("IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(IMM25FramesPerSecondTime&)") ;
}

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::IMM25FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 25)
{
  IMODTRACE_DEVELOP("IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::operator unsigned long                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMM25FramesPerSecondTime::operator(unsigned long)") ;
  return (unsigned long)(FPS25FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::~IMM25FramesPerSecondTime                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::~IMM25FramesPerSecondTime()
{
   IMODTRACE_DEVELOP("IMM25FramesPerSecondTime::~IMM25FramesPerSecondTime") ;
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::IMM30FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(const IMM30FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{
   IMODTRACE_DEVELOP("IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(IMM30FramesPerSecondTime)") ;
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::IMM30FramesPerSecondTime                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 30)
{
  IMODTRACE_DEVELOP("IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::operator unsigned long                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMM30FramesPerSecondTime::operator(unsigned long)") ;
  return (unsigned long)(FPS30FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::~IMM30FramesPerSecondTime                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::~IMM30FramesPerSecondTime()
{
   IMODTRACE_DEVELOP("IMM30FramesPerSecondTime::~IMM30FramesPerSecondTime") ;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(unsigned long time)
                  : IMMTime(REDBOOKTOMM(time))
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::IMMMinSecFrameTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(const IMMMinSecFrameTime &time)
                  : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMMinSecFrameTime::IMMMinSecFrameTime(IMMMinSecFrameTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(const IMMTime &time)
                  : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMMinSecFrameTime::IMMMinSecFrameTime(IMMTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::asString                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMMinSecFrameTime::asString() const
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::asString") ;
  return IString(minutes()) + ":" +
         IString(seconds()).rightJustify(2, '0') + "." +
         IString(frames()).rightJustify(3, '0');
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::operator unsigned long                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::operator(unsigned long)") ;
  return REDBOOKFROMMM(asMMTime());
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::~IMMMinSecFrameTime                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::~IMMMinSecFrameTime()
{
   IMODTRACE_DEVELOP("IMMMinSecFrameTime::~IMMMinSecFrameTime") ;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::frames                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::frames() const
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::frames") ;
  return (((unsigned long) (*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::minutes                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::minutes() const
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::minutes") ;
  return (((unsigned long) (*this))) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::minutes                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::seconds() const
{
  IMODTRACE_DEVELOP("IMMMinSecFrameTime::seconds") ;
  return (((unsigned long) (*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(unsigned long time)
                 : IMMTime(HMSTOMM(time))
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::IMMHourMinSecTime(unsigned long)") ;
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(const IMMHourMinSecTime &time)
                 : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMHourMinSecTime::IMMHourMinSecTime(IMMHourMinSecTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(const IMMTime &time)
                 : IMMTime(time)
{
   IMODTRACE_DEVELOP("IMMHourMinSecTime::IMMHourMinSecTime(IMMTime&)") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::hours                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::hours() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::hours") ;
  return ((unsigned long)(*this)) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::minutes                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::minutes() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::minutes") ;
  return (((unsigned long)(*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::seconds                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::seconds() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::seconds") ;
  return (((unsigned long)(*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::asString                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMHourMinSecTime::asString() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::asString") ;
  return IString(hours()) + ":" +
         IString(minutes()).rightJustify(2, '0') + ":" +
         IString(seconds()).rightJustify(2, '0');
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::operator unsigned long                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::operator unsigned long () const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::operator(unsigned long)") ;
  return (unsigned long)(HMSFROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::~IMMHourMinSecTime                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::~IMMHourMinSecTime()
{
   IMODTRACE_DEVELOP("IMMHourMinSecTime::~IMMHourMinSecTime") ;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::ordinal                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::ordinal() const
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::ordinal") ;
  // return total seconds
  return seconds() + (minutes() * 60) + (hours() * 60 * 60);
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::setTimeToOrdinal                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime& IMMHourMinSecTime::setTimeToOrdinal(unsigned long ordinal)
{
  IMODTRACE_DEVELOP("IMMHourMinSecTime::setTimeToOrdinal") ;
  setMMTime(MSECTOMM(ordinal * 1000));
  return *this;
}

