/*******************************************************************************
* FILE NAME: ibmphdr.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ibmphdr.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_GPIBITMAPS
  #include <iwindefs.h>
  }

#include <ibmphdr.hpp>

#include <ipainevt.hpp>
#include <iswp.hpp>
#include <itrace.hpp>
#include <icontrol.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <ibmpctl.hpp>
#include <igrafctx.hpp>
#include <igbitmap.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBMPHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IBitmapHandler::IBitmapHandler                                               |
------------------------------------------------------------------------------*/
IBitmapHandler :: IBitmapHandler() {;}

/*------------------------------------------------------------------------------
| IBitmapHandler::~IBitmapHandler                                              |
------------------------------------------------------------------------------*/
IBitmapHandler :: ~IBitmapHandler() {;}

/*------------------------------------------------------------------------------
| IBitmapHandler::paintWindow                                                  |
| This inherited function is overridden:                                       |
|   paintWindow - Process the paint events for a bitmap control window.        |
------------------------------------------------------------------------------*/
Boolean IBitmapHandler :: paintWindow          ( IPaintEvent& evt )
{
  IBitmapControl* pBitmap = (IBitmapControl*)evt.window();
  ISize winsize = pBitmap->size();
#ifdef IC_PM
  POINTL aptlPoints[4];
  if (pBitmap->bitmap())
  {
     BITMAPINFOHEADER   pbmpData;
     pbmpData.cbFix = sizeof(BITMAPINFOHEADER);
     GpiQueryBitmapParameters(pBitmap->bitmap(),&pbmpData);
     aptlPoints[0].x = 0;
     aptlPoints[0].y = 0;
     aptlPoints[1].x = winsize.width()-1;
     aptlPoints[1].y = winsize.height()-1;
     aptlPoints[2].x = 0L;
     aptlPoints[2].y = 0L;
     aptlPoints[3].x = pbmpData.cx;
     aptlPoints[3].y = pbmpData.cy;
     GpiWCBitBlt(evt.presSpaceHandle(),pBitmap->bitmap(),
                 4,aptlPoints,ROP_SRCCOPY,BBO_IGNORE);
  }
  else
  {
     GpiErase(evt.presSpaceHandle());
  } /* endif */
#endif
#ifdef IC_WIN
  // Get a graphic context
  IGraphicContext gc(evt.presSpaceHandle());

  // draw the bitmap if we have one
  if ( pBitmap->bitmap() )
    {
    // Get a pointer to the current bitmap
    IGBitmap bitmap( pBitmap->bitmap() );
    ISize    bitmapSize = bitmap.size();
    // draw the bitmap
    bitmap.drawOn(gc,
                  IPoint(0,0),
                  IPoint( winsize.width()-1, winsize.height()-1 ),
                  IPoint(0,0),
                  IPoint( bitmapSize.width(), bitmapSize.height() ) );
    }
  else
    {
    // paint the current background color
    evt.clearBackground();
    }

#endif
  unsigned long length   = pBitmap->textLength();
  if (length)
     return false;
  return true;
}
