/*******************************************************************************
* FILE NAME: ibutton.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ibutton.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C"
{
  #define INCL_WINBUTTONS
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}
#include <ibase.hpp>
#include <ibutton.hpp>
#include <icconst.h>
#include <icmnfun.hpp>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <istring.hpp>
#include <inotifev.hpp>
#include <ihandler.hpp>
#include <ievent.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBUTTON_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public button styles                                                         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IButton::Style
#ifdef IC_PM
  IButton::noPointerFocus = BS_NOPOINTERFOCUS;
#endif
#ifdef IC_WIN
  IButton::noPointerFocus (0, IBS_NOPOINTERFOCUS);
#endif
#pragma data_seg()

#ifdef IC_WIN
class IButtonData : public IBase {
/*------------------------------------------------------------------------------
| The IButtonData class encapsulates private data and functions                |
| used by the IButton class.  An object of this class is created in            |
| the IButton constructors.  This is the os2/windows version.                  |
------------------------------------------------------------------------------*/
public:
  IButtonData ( );
 ~IButtonData ( );

Boolean
  fCheckingFocus;
HWND
  fOldFocusHandle;
}; // IButtonData

/*------------------------------------------------------------------------------
| IButtonData::IButtonData                                                     |
------------------------------------------------------------------------------*/
IButtonData::IButtonData ( )
  : fCheckingFocus( false ),
    fOldFocusHandle( 0 )
{ }

/*------------------------------------------------------------------------------
| IButtonData::~IButtonData                                                    |
------------------------------------------------------------------------------*/
IButtonData::~IButtonData ( )
{ }

class IButtonHandler : public IHandler {
typedef IHandler
  Inherited;
/*------------------------------------------------------------------------------
| This handler implements the BS_NOPOINTERFOCUS style for buttons.             |
------------------------------------------------------------------------------*/
public:

/*------------------------------- Constructors -------------------------------*/
  IButtonHandler ( );
virtual
  ~IButtonHandler ( );

/*----------------------------- Event Processing -----------------------------*/
virtual Boolean
  dispatchHandlerEvent ( IEvent& event );

}; // class IButtonHandler

/*------------------------------------------------------------------------------
| IButtonHandler::IButtonHandler                                               |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IButtonHandler::IButtonHandler()
{ }

/*------------------------------------------------------------------------------
| IButtonHandler::~IButtonHandler                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IButtonHandler::~IButtonHandler()
{ }

/*------------------------------------------------------------------------------
| IButtonHandler::dispatchHandlerEvent                                         |
------------------------------------------------------------------------------*/
IBase::Boolean IButtonHandler::dispatchHandlerEvent( IEvent& event )
{
  Boolean
    result(false);
  IButton*
    button(0);
  switch( event.eventId() )
     {

     case WM_LBUTTONDOWN:
        button = (IButton*)event.window();
        if ( (button) && !button->allowsMouseClickFocus()  )
           {
           button->fButtonData->fCheckingFocus = true;
           button->fButtonData->fOldFocusHandle = 0;
           }
        break;
     case WM_LBUTTONUP:
        button = (IButton*)event.window();
        if ( (button) && (button->fButtonData->fCheckingFocus) )
           {
           button->postEvent( IC_UM_CANVAS_SETFOCUS );
           }
        break;
     case WM_SETFOCUS:
        button = (IButton*)event.window();
        if ( (button) && (button->fButtonData->fCheckingFocus) )
           {
           button->fButtonData->fOldFocusHandle = (HWND)event.parameter1();
           }
        break;
     case WM_KILLFOCUS:
        button = (IButton*)event.window();
        if (button)
           {
           button->fButtonData->fCheckingFocus = false;
           }
        break;
     case IC_UM_CANVAS_SETFOCUS:
        button = (IButton*)event.window();
        if ( (button) && (button->fButtonData->fCheckingFocus) )
           {
           // if we get here we need to give focus back to fOldFocusHandle
           if ( button->fButtonData->fOldFocusHandle )
              {
              // have a handle to give focus back to
              SetFocus( button->fButtonData->fOldFocusHandle );
              }
           else
              {
              // give focus to top level window for this thread
              SetFocus( GetActiveWindow() );
              }

           button->fButtonData->fCheckingFocus = false;
           }
        result = true;
        break;

     default:
       break;
     }
  return result;
}

/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointer to the object |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator IButtonHandler* ();
IButtonHandler
 *operator -> ()
    {
    return *this;
    }
 ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
IButtonHandler
 *ptr;
} defaultHandler;

DefaultHandler::operator IButtonHandler* ( )
  {
  if ( !ptr )
    ptr = new IButtonHandler;
  return ptr;
  }
#endif //IC_WIN


/*------------------------------------------------------------------------------
| IButton::IButton                                                             |
|                                                                              |
------------------------------------------------------------------------------*/
IButton :: IButton ( )
#ifdef IC_WIN
                  : fButtonData( new IButtonData )
{
   defaultHandler->handleEventsFor( this );
}
#else
                  : fButtonData(0)
{ }
#endif

/*------------------------------------------------------------------------------
| IButton::~IButton                                                            |
------------------------------------------------------------------------------*/
IButton :: ~IButton ( )
{
#ifdef IC_WIN
   defaultHandler->stopHandlingEventsFor( this );
   delete fButtonData;
#endif
}

/*------------------------------------------------------------------------------
| IButton::enableMouseClickFocus                                               |
|                                                                              |
| Enables the button to receive the focus when the user clicks on the button   |
| using the mouse.                                                             |
------------------------------------------------------------------------------*/
IButton&  IButton :: enableMouseClickFocus ( Boolean bTurnOn )
{
#ifdef IC_PM
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle &= ~noPointerFocus.asUnsignedLong();
  }
  else
  {
    ulStyle |= noPointerFocus.asUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
  }
#endif
#ifdef IC_WIN
  unsigned long ulStyle = extendedStyle();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle &= ~noPointerFocus.asExtendedUnsignedLong();
  }
  else
  {
    ulStyle |= noPointerFocus.asExtendedUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setExtendedStyle(ulStyle);
  }
#endif
  return( *this );
}

/*------------------------------------------------------------------------------
| IButton::disableMouseClickFocus                                              |
------------------------------------------------------------------------------*/
IButton&  IButton :: disableMouseClickFocus ( )
{
  enableMouseClickFocus(false);
  return *this;
}


/*------------------------------------------------------------------------------
| IButton::allowsMouseClickFocus                                               |
------------------------------------------------------------------------------*/
Boolean  IButton :: allowsMouseClickFocus ( ) const
{
#ifdef IC_WIN
  if (extendedStyle() & noPointerFocus.asExtendedUnsignedLong())
#else
  if (style() & noPointerFocus.asUnsignedLong())
#endif
    return( false );
  else
    return( true );
}


/*------------------------------------------------------------------------------
| IButton::highlight                                                           |
|                                                                              |
| Highlight the button.  A highlighted button appears as if the mouse          |
| selection button were pressed with the mouse pointer over the button.        |
------------------------------------------------------------------------------*/
IButton&  IButton :: highlight ( Boolean highlight )
{
  handle().sendEvent(BM_SETHILITE, IEventParameter1(highlight),
                                   IEventParameter2(0));
  return *this;
}


/*------------------------------------------------------------------------------
| IButton::unhighlight                                                         |
------------------------------------------------------------------------------*/
IButton&  IButton :: unhighlight ( )
{
  return highlight(false);
}


/*------------------------------------------------------------------------------
| IButton::isHighlighted                                                       |
------------------------------------------------------------------------------*/
Boolean  IButton :: isHighlighted ( ) const
{
  IEventResult evt = handle().sendEvent(BM_QUERYHILITE,
                                        IEventParameter1(0),
                                        IEventParameter2(0));
#ifdef IC_WIN
  if (evt.asUnsignedLong() & 0x00000004)
#else
  if (evt.asUnsignedLong())
#endif
     return true;
  else
     return false;
}


/*------------------------------------------------------------------------------
| IButton::click                                                               |
------------------------------------------------------------------------------*/
IButton&  IButton :: click ( )
{
  clickButton(handle());
  return *this;
}

/*------------------------------------------------------------------------------
| IButton::foregroundColor                                                     |
|                                                                              |
| Returns the foreground color of the button.                                  |
------------------------------------------------------------------------------*/
IColor IButton::foregroundColor () const
{
  return (IWindow::color(PP_FOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::menuText)));
}

/*------------------------------------------------------------------------------
| IButton::backgroundColor                                                     |
|                                                                              |
| Returns the background color of the button.                                  |
------------------------------------------------------------------------------*/
IColor IButton::backgroundColor () const
{
  //Return the correct background for a pushbutton.
#ifdef IC_PM
  if (style() & BS_PRIMARYSTYLES)
#else
  if (!(((style() & BS_PRIMARYSTYLES) == BS_PUSHBUTTON) ||
        ((style() & BS_PRIMARYSTYLES) == BS_DEFPUSHBUTTON)))
#endif
  {
    return (IWindow::color(PP_BACKGROUNDCOLOR,
                           defaultbackgroundcolor(handle())));
  }
  return (IWindow::color(PP_BACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::buttonMiddle)));
}

/*------------------------------------------------------------------------------
| IButton::disabledForegroundColor                                             |
|                                                                              |
| Returns the disabled foreground color of the button.                         |
------------------------------------------------------------------------------*/
IColor IButton::disabledForegroundColor () const
{
#ifdef IC_PM
  return (IWindow::color(PP_DISABLEDFOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::menuText)));
#else
  return IGUIColor( IGUIColor::menuText );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| IButton::hiliteForegroundColor                                               |
|                                                                              |
| Returns the hilite foreground color of the button.                           |
------------------------------------------------------------------------------*/
IColor IButton::hiliteForegroundColor () const
{
#ifdef IC_PM
  return (IWindow::color(PP_HILITEFOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::menuText)));
#else
  return IGUIColor( IGUIColor::menuText );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| IButton::hiliteBackgroundColor                                               |
|                                                                              |
| Returns the hilite background color of the button.                           |
------------------------------------------------------------------------------*/
IColor IButton::hiliteBackgroundColor () const
{
#ifdef IC_PM
  return (IWindow::color(PP_HILITEBACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::menuText)));
#else
  return IGUIColor( IGUIColor::menuText );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| IButton::setText                                                             |
|                                                                              |
| Sets the text used by the button, and notifies the parent canvas             |
------------------------------------------------------------------------------*/
IButton &  IButton :: setText ( const char* pszText )
{
  Inherited::setText(pszText);
  setLayoutDistorted(IWindow::minimumSizeChanged, 0);
  return *this;
}


/*------------------------------------------------------------------------------
| IButton::setText                                                             |
|                                                                              |
| Sets the button text from a string table.                                    |
------------------------------------------------------------------------------*/
IButton &  IButton :: setText ( const IResourceId & residText )
{
  Inherited::setText(residText);
  return *this;
}
