/*******************************************************************************
* FILE NAME: icombob1.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icombonh.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define INCL_WINENTRYFIELDS

extern "C" {
   #include <iwindefs.h>
}

#ifndef _IWINDOW_
  #include <iwindow.hpp>
#endif

#ifndef _ICOMBOBS_
  #include <icombobs.hpp>
#endif

#ifndef _ICOMBOBX_
  #include <icombobx.hpp>
#endif

#ifndef _ICOMBONH_
  #include <icombonh.hpp>
#endif

#ifndef _IEVENT_
  #include <ievent.hpp>
#endif

#ifndef _ICTLEVT_
  #include <ictlevt.hpp>
#endif

#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif

#if (!defined(IC_MOTIF) && !defined(_IWCNAME_))
  #include <iwcname.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICOMBOB1_CPP_
  #include <ipagetun.h>
#endif

// Class to clean up static objects
class IComboBoxNotifyStatics
{
public:

~IComboBoxNotifyStatics ()
{
  delete comboBoxNotifyHandler;
  comboBoxNotifyHandler = 0;
}
static IComboBoxNotifyHandler
  &notificationHandler ( );

static IComboBoxNotifyHandler
 *comboBoxNotifyHandler;
};

// Initialize static objects and notification strings.
#pragma data_seg(ICLStaticConst)
static IComboBoxNotifyStatics comboBoxStatics;
INotificationId const IComboBox::addId    = "IComboBox::add";
INotificationId const IComboBox::removeId = "IComboBox::remove";
INotificationId const IBaseComboBox::enterId  = "IBaseComboBox::enter";
INotificationId const IBaseComboBox::selectId = "IBaseComboBox::select";
#pragma data_seg()

#pragma data_seg(ICLNonConst)
IComboBoxNotifyHandler* IComboBoxNotifyStatics::comboBoxNotifyHandler = 0;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IComboBoxNotifyStatics::notificationHandler                                  |
|                                                                              |
| Static function to create and return a notification handler for IBaseComboBox.   |
------------------------------------------------------------------------------*/
IComboBoxNotifyHandler& IComboBoxNotifyStatics::notificationHandler ()
{
  if (!(IComboBoxNotifyStatics::comboBoxNotifyHandler))
  {
    IComboBoxNotifyStatics::comboBoxNotifyHandler =
                                   new IComboBoxNotifyHandler();
  }
  return (*IComboBoxNotifyStatics::comboBoxNotifyHandler);
}

/*------------------------------------------------------------------------------
| IBaseComboBox::enableNotification                                                |
|                                                                              |
| Flags this part as being ready to accept notifications.  A Notifier, like    |
| IBaseComboBox, does not send notifications to observer objects until it is       |
| enabled.                                                                     |
|                                                                              |
| In the window hierarchy we override this function to allocate the correct    |
| notification handler.                                                        |
------------------------------------------------------------------------------*/
IBaseComboBox& IBaseComboBox::enableNotification (Boolean enabled)
{
  if (!notificationHandler())
  {
    setNotificationHandler(
               &(IComboBoxNotifyStatics::notificationHandler()));
  }
  IWindow::enableNotification(enabled);
  return *this;
}

/*------------------------------------------------------------------------------
| IComboBoxNotifyHandler::IComboBoxNotifyHandler                               |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IComboBoxNotifyHandler::IComboBoxNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IComboBoxNotifyHandler::~IComboBoxNotifyHandler                              |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IComboBoxNotifyHandler::~IComboBoxNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IComboBoxNotifyHandler::dispatchHandlerEvent                                 |
|                                                                              |
| This notifies IBaseComboBox Windows of the following events:                     |
|                                                                              |
| select(Id)                                                                   |
------------------------------------------------------------------------------*/
#ifdef IC_PMWIN
Boolean IComboBoxNotifyHandler::dispatchHandlerEvent (IEvent& anEvent)
{
  IWindow* thisWindow = anEvent.window();

  switch (anEvent.eventId())
  {
    case WM_CONTROL:
    {
      IControlEvent ctlevt(anEvent);
      IWindow* pwndControl = ctlevt.controlWindow();
      unsigned long ulMsg  = ctlevt.parameter1().highNumber();
      IWindowHandle hwndControl;

      if (pwndControl)
      {
        hwndControl = pwndControl->handle();
        thisWindow = pwndControl;
      }
      else
      {
        hwndControl = IWindow::handleWithParent(ctlevt.controlId(),
                                                ctlevt.handle());
        thisWindow = IWindow::windowWithOwner(ctlevt.controlId(),
                                              ctlevt.window());
      }

      if (hwndControl && thisWindow)
      {
        IWindowClassName className(hwndControl);
        if (className == WC_COMBOBOX)
        {
          if (ulMsg == CBN_LBSELECT)
          {
            thisWindow->notifyObservers(INotificationEvent(
                                        IBaseComboBox::selectId,
                                        *thisWindow));
          }
#ifdef IC_WIN
          else if (ulMsg == CBN_ENTER || ulMsg == IC_UM_CBN_ENTER)
#else
          else if (ulMsg == CBN_ENTER)
#endif
          {
            thisWindow->notifyObservers(INotificationEvent(
                                        IBaseComboBox::enterId,
                                        *thisWindow));
          }
        }
      }
    break;
    }

#ifdef IC_WIN
    case WM_CHAR:
    {
      if ((unsigned long)anEvent.parameter1() == VK_RETURN)
      {
        thisWindow->notifyObservers(INotificationEvent(
                                    IBaseComboBox::enterId,
                                    *thisWindow));
      }
    break;
    }
#endif

    default:
      break;
  }  /* end switch */

  return (Inherited::dispatchHandlerEvent(anEvent));
}

#endif // IC_PMWIN

