/*******************************************************************************
* FILE NAME: icombobs.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icombobs.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

/**********************************************************************/
/* Need IC_USE_CB defined before iwindefs since Windows macros are    */
/* different if processing listboxes that are part of comboboxes      */
/**********************************************************************/
extern "C" {
  #define IC_USE_CB
  #define INCL_WINENTRYFIELDS       // ES_xxx, CBS_xxx
  #define INCL_WINERRORS            // PMERR_HEAP_MAX_SIZE_REACHED
  #define INCL_WINLISTBOXES         // LS_HORZSCROLL
  #define INCL_NLS                  // ES_ANY, ES_SBCS, ES_DBCS, ES_MIXED
  #define INCL_WINWINDOWMGR         // WinQueryClassName
  #define INCL_WINSYS               // SV_CYBORDER
  #define INCL_GPICONTROL           // From ILISTBX2.HPP
  #define INCL_GPILCIDS             //
  #define INCL_GPILOGCOLORTABLE     //
  #define INCL_GPIPRIMITIVES        //

  #include <iwindefs.h>
  #include <stdlib.h>            // atoi
  #include <string.h>
}

#include <inotifev.hpp>
#include <icombobs.hpp>

#include <iapp.hpp>
#include <icconst.h>
#include <icolor.hpp>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <ihandle.hpp>
#include <ilistbox.hpp>
#include <imphdr.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iwcname.hpp>
#include <iplatfrm.hpp>
#include <ictlevt.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICOMBOBS_CPP_
  #include <ipagetun.h>
#endif

/******************************************************************************/
/* Public combobox styles.                                                    */
/******************************************************************************/
#pragma data_seg(ICLStaticConst)
const IBaseComboBox::Style
  IBaseComboBox::simpleType           = CBS_SIMPLE,
  IBaseComboBox::dropDownType         = CBS_DROPDOWN,
  IBaseComboBox::horizontalScroll     = LS_HORZSCROLL,
#ifdef IC_WIN
  IBaseComboBox::readOnlyDropDownType ( 0, ICBS_DROPDOWNLIST ),
  IBaseComboBox::oemData              = CBS_OEMCONVERT,
  IBaseComboBox::autoScroll           = CBS_AUTOHSCROLL,
  IBaseComboBox::border3D             ( 0, WS_EX_CLIENTEDGE ),
#else
  IBaseComboBox::readOnlyDropDownType = CBS_DROPDOWNLIST,
  IBaseComboBox::anyData              ( 0, IES_ANYDATA ),
  IBaseComboBox::sbcsData             = ES_SBCS,
  IBaseComboBox::dbcsData             = ES_DBCS,
  IBaseComboBox::mixedData            ( 0, IES_MIXEDDATA ),
  IBaseComboBox::autoScroll           = ES_AUTOSCROLL,
  IBaseComboBox::border3D             ( 0, 0 ),
#endif
#ifdef IC_WIN
  IBaseComboBox::classDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE | CBS_AUTOHSCROLL),
                                     IES_ANYDATA | WS_EX_CLIENTEDGE );
#else
  IBaseComboBox::classDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE | ES_AUTOSCROLL),
                                     IES_ANYDATA | WS_EX_CLIENTEDGE );
#endif

const unsigned long    IBaseComboBox::notFound = (unsigned long)ULONG_MAX;
const unsigned long    IBaseComboBox::first    = (unsigned long)ULONG_MAX;
#pragma data_seg()

/******************************************************************************/
/* Default style for new objects (initial value)                              */
/******************************************************************************/
#pragma data_seg(ICLNonConst)
#ifdef IC_WIN
IBaseComboBox::Style
  IBaseComboBox::currentDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE | CBS_AUTOHSCROLL),
                                       IES_ANYDATA | WS_EX_CLIENTEDGE );
#else
IBaseComboBox::Style
  IBaseComboBox::currentDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE | ES_AUTOSCROLL),
                                       IES_ANYDATA | WS_EX_CLIENTEDGE );
#endif
#pragma data_seg()


class IBaseComboBoxData : public IBase {
/*******************************************************************************
* The IBaseComboBoxData class encapsulates private data and functions used by the  *
* IBaseComboBox class.  An object of this class is created in the IBaseComboBox        *
* constructors.                                                                *
*******************************************************************************/
public:
  IBaseComboBoxData();

virtual
 ~IBaseComboBoxData();

  unsigned long    minimumRows,
                   minimumCharacters;

#ifdef IC_WIN

  unsigned long    charLimit;
  HBRUSH           bkgrndbrush;

//  Returns the handle of the entry field part of a combo box
static IWindowHandle
   handleOfEntryField(const IWindowHandle& combobox );

WNDPROC
  fdefaultEntryProc; // Default procedure for Entry field part of combo
IWindowHandle
  fentryHandle;      // handle of the entry field
#endif

private:
  IBaseComboBoxData::IBaseComboBoxData (const IBaseComboBoxData&);
  IBaseComboBoxData& operator= (const IBaseComboBoxData&);
};

/*------------------------------------------------------------------------------
| IBaseComboBoxData::IBaseComboBoxData                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IBaseComboBoxData::IBaseComboBoxData ( )
 : minimumRows (4),
   minimumCharacters (25)
{
#ifdef IC_WIN
  charLimit = 30000;
  bkgrndbrush = 0;
#endif
}

IBaseComboBoxData::~IBaseComboBoxData ( )
{
#ifdef IC_WIN
  if ( bkgrndbrush != 0 )
    DeleteObject(bkgrndbrush);
#endif
}

#ifdef IC_WIN
class IBaseComboHandler : public IHandler {
typedef IHandler
  Inherited;
public:

/*------------------------------- Constructors -------------------------------*/
  IBaseComboHandler ( );

virtual
  ~IBaseComboHandler ( );

/*----------------------------- Event Processing -----------------------------*/
virtual Boolean
  dispatchHandlerEvent ( IEvent& event );

}; // class IBaseComboHandler

/*------------------------------------------------------------------------------
| IBaseComboHandler::IBaseComboHandler                                         |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IBaseComboHandler::IBaseComboHandler()
{ }

/*------------------------------------------------------------------------------
| IBaseComboHandler::~IBaseComboHandler                                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IBaseComboHandler::~IBaseComboHandler()
{ }

/*------------------------------------------------------------------------------
| IBaseComboHandler::dispatchHandlerEvent                                      |
|                                                                              |
| Intercepts control color event                                               |
|    This handler returns false for all other events.                          |
------------------------------------------------------------------------------*/
IBase::Boolean IBaseComboHandler::dispatchHandlerEvent( IEvent& event )
{
  IBoolean
    result = false;

  switch( event.eventId() )
  {
    case WM_CONTROL:
    {
      IControlEvent ctlevt( event );
      IBaseComboBox *comboBox =
            (IBaseComboBox*)(ctlevt.controlWindow());
      if (comboBox)
      {
         unsigned long subMsg = ctlevt.parameter1().number2();
         IWindowHandle hwndControl = comboBox->handle();
         IWindowHandle hwndParent = IPARENTOF( hwndControl );
         if (subMsg == CBN_SELENDOK)
         {
            // Windows generates a CBN_SELENDOK notification in response to
            // a selection change in the combobox list.  It will not generate
            // a CBN_EDITCHANGE following the entry field update.  In order for
            // IEditHandler::dispatchHandlerEvents to call the edit function
            // following the entry field update, we post a notification event.
            hwndParent.postEvent( WM_COMMAND,
                  IEventParameter1( (short)IIDOF( hwndControl ), IC_UM_CBN_EFCHANGE),
                  IEventParameter2( hwndControl.asUnsigned()) );
            result = true;
         }
         else if (subMsg == CBN_CLOSEUP)
         {
            // Ensure ISelectHandler::enter processed after edit field updated.
            hwndParent.postEvent( WM_COMMAND,
                  IEventParameter1( (short)IIDOF( hwndControl ), IC_UM_CBN_ENTER),
                  IEventParameter2( hwndControl.asUnsigned()) );
            result = true;
         }
      }
      break;
    }

    case WM_PAINT:
         {
         // The following corrects a Dos/Windows problem that is
         // documented in MSDN Article #Q128110 - CBS_SIMPLE ComboBox
         // Repainting Problem
         if (IPlatform::isWin32S())
         {
           HPEN hOldPen;
           HBRUSH hOldBrush;
           RECT comboRect, entryfdRect, listboxRect;
           HWND hEntryfd, hListbox;

           HWND hWnd = event.handle();
           if (hWnd)
           {
             HDC comboDC = GetDC(hWnd);
             IBaseComboBox* combo = (IBaseComboBox*)event.window();

             if ( (combo->fBaseComboBoxData->fentryHandle) &&
                (combo->type() == IBaseComboBox::simple) )
             {
               HBRUSH hBrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
               HPEN hPen = CreatePen (PS_SOLID, 1, GetSysColor(COLOR_WINDOW));
               hOldBrush = SelectObject(comboDC, hBrush);
               hOldPen = SelectObject(comboDC, hPen);

               hEntryfd = combo->fBaseComboBoxData->fentryHandle;
               hListbox = GetWindow(hWnd, GW_CHILD);
               if (hListbox == hEntryfd)
                 hListbox = GetWindow(hWnd, GW_HWNDNEXT);
               GetClientRect(hWnd, &comboRect);
               GetClientRect(hEntryfd, &entryfdRect);
               GetClientRect(hListbox, &listboxRect);

#ifdef IC_TRACE_ALL
               ITRACE_ALL(IString("IBaseComboHandler::dispatch / LEFT")
               ITRACE_ALL(IString("rect=(") +
                          IString(comboRect.left) + IString(",") +
                          IString(entryfdRect.bottom) + IString(",") +
                          IString((comboRect.right-listboxRect.right)) + IString(",") +
                          IString(comboRect.bottom) + IString(")") );
#endif // IC_TRACE_ALL

               Rectangle (comboDC,
                          comboRect.left,
                          entryfdRect.bottom,
                          comboRect.right - listboxRect.right,
                          comboRect.bottom);

#ifdef IC_TRACE_ALL
               ITRACE_ALL(IString("IBaseComboHandler::dispatch / LEFT")
               ITRACE_ALL(IString("rect=(") +
                          IString((comboRect.right-listboxRect.right)) + IString(",") +
                          IString((entryfdRect.bottom+listboxRect.bottom)) + IString(",") +
                          IString(comboRect.right) + IString(",") +
                          IString(comboRect.bottom) + IString(")") );
#endif // IC_TRACE_ALL

               Rectangle (comboDC,
                          comboRect.right - listboxRect.right,
                          entryfdRect.bottom + listboxRect.bottom,
                          comboRect.right,
                          comboRect.bottom);
               DeleteObject(SelectObject(comboDC, hOldBrush));
               DeleteObject(SelectObject(comboDC, hOldPen));
             }  //
             ReleaseDC(hWnd, comboDC);
             // This handler does not do the complete paint for the
             // IComboBox; therefore, we must return false so that
             // the remaining painting gets done.
             result = false;
           }  // if (hWnd)
         }  // if (IPlatform::isWin32s())
         }  // case
      break;

    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORLISTBOX:
         {
           IWindow* ctlwin = 0;
           IWindowClassName className(IWindowHandle(event.parameter2()));
           if (className == WC_COMBOLBOX)
           {
             // Retrieve ComboBox hwnd that was tucked away during construction
             HWND combo = (HWND)GetWindowLong(event.parameter2(), GWL_USERDATA);
             ctlwin = IWindow::windowWithHandle(IWindowHandle(combo));
           }
           else if (className == WC_COMBOBOX)
           {
             ctlwin =
               IWindow::windowWithHandle(IWindowHandle(event.parameter2()));
           }
           else if (className == WC_ENTRYFIELD)
           {
             ctlwin = event.controlWindow();
           }
           if (ctlwin)
           {
             IColor clrFgnd =
               ((IBaseComboBox *) ctlwin)->foregroundColor();
             IColor clrBgnd =
               ((IBaseComboBox *) ctlwin)->backgroundColor();

             // Set foreground and background text colors in device context
             SetBkColor( (HDC)event.parameter1(), clrBgnd.asRGBLong() );
             SetTextColor( (HDC)event.parameter1(), clrFgnd.asRGBLong() );

             // Return HBRUSH for painting controls background
             if ( ((IBaseComboBox *) ctlwin)->fBaseComboBoxData->bkgrndbrush != 0 )
               DeleteObject( ((IBaseComboBox *) ctlwin)->fBaseComboBoxData->bkgrndbrush);
             ((IBaseComboBox *) ctlwin)->fBaseComboBoxData->bkgrndbrush =
               CreateSolidBrush (NEARESTPALETTECOLOR(clrBgnd.asRGBLong()));
             event.setResult( ((IBaseComboBox *) ctlwin)->fBaseComboBoxData->bkgrndbrush );

             result = true;
           }
         }
      break;

    default:
      break;
  }
  return result;
}

/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointer to the object |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator IBaseComboHandler* ();
IBaseComboHandler
 *operator -> ()
    {
    return *this;
    }
 ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
IBaseComboHandler
 *ptr;
} defaultHandler;

DefaultHandler::operator IBaseComboHandler* ( )
  {
  if ( !ptr )
    ptr = new IBaseComboHandler;
  return ptr;
  }
#endif //IC_WIN

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                 |
|                                                                              |
| Construct an IBaseComboBox on an IWindow.                                    |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox(unsigned long id,
                     IWindow* pParent,
                     IWindow* pOwner,
                     const IRectangle& rectInit,
                     const Style& style)
                   : fBaseComboBoxData( 0 )
{
  // assertions on input parms
  IASSERTPARM(pParent!=0);
  IASSERTPARM((style & (simpleType | dropDownType | readOnlyDropDownType))
               !=0);
  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  // Create a private data object.
  fBaseComboBoxData = new IBaseComboBoxData;

#ifdef __BORLANDC__
  IWindowHandle owner;
  if (pOwner == 0)
    owner = IWindowHandle(0);
  else
    owner = pOwner->handle();
#else
  IWindowHandle owner = (pOwner == 0) ? IWindowHandle(0) : pOwner->handle();
#endif

  IWindowHandle whComboBox =
      this -> create( id,
                      0,
                      convertToGUIStyle( style ),
                      WC_COMBOBOX,
                      pParent->handle(),
                      owner,
                      rectInit,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( style, true ) );

#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  startHandlingEventsFor(whComboBox);
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);
#ifdef IC_WIN
  // Subclass the entry field in the combo box so that we can get
  // navigation keys.
  fBaseComboBoxData->fentryHandle =
     IBaseComboBoxData::handleOfEntryField( whComboBox );
  if (fBaseComboBoxData->fentryHandle)
     {
     // setup our procedure in the window.
     fBaseComboBoxData->fdefaultEntryProc = (WNDPROC)
        ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                        icomboboxEntryWinProc);
     if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
        {
        // throw now rather than trap later.
        ITHROWGUIERROR( "ISUBCLASSWINDOW" );
        }
     }

     // IC_NOTYET...  The horizontal extend must be set for Windows to
     // display the horizontal scroll bar.  Temp fix is to set it to the
     // screen width.  Real fix could be to monitor adds/deletes and set
     // it to the widest string in current font.
     // Note:  Due to a Windows bug, horizontal scrollbars are only supported
     //        for simple comboboxes.
     if ((style & IBaseComboBox::horizontalScroll)
        && (IPlatform::isWin95()
        || (IPlatform::isNTNewShell()
            && !(style & IBaseComboBox::dropDownType)
            && !(style & IBaseComboBox::readOnlyDropDownType))))
     {
        whComboBox.sendEvent(CB_SETHORIZONTALEXTENT,
                             (unsigned long)IWindow::desktopWindow()->size().width(),
                             0);

        // Windows doesn't keep the WS_VSCROLL style with the combobox so store
        // it as an extended style for querying later.
        setExtendedStyle( extendedStyle() | LS_HORZSCROLL );
     }

  // Store whComboBox in the ComboListBox (COMBOLBOX) windows user data.
  // It is  used to retrieve colors when painting ComboListBox.  This is
  // necessary because the ComboListBox window is a child of the desktop
  // not the combo box.  Se we tuck away the ComboBox hwnd for each
  // combo list box created if nothing is already stored in GWL_USERDATA.
  HWND deskTop = desktopWindow()->handle();
  HWND comboBox = whComboBox;
  IWindowHandle comboLBox = GetWindow(desktopWindow()->handle(), GW_CHILD);
  while (comboLBox)
  {
    if (IWindowClassName(comboLBox) == WC_COMBOLBOX)
    {
      HWND curHwnd = (HWND) GetWindowLong(comboLBox, GWL_USERDATA);
      if ( !(curHwnd) )
      {
         SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
      }
    }
    comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
  }
  // Take care of CBS_SIMPLE too
  comboLBox = GetWindow(whComboBox, GW_CHILD);
  while ( (comboLBox) && (IWindowClassName(comboLBox) != WC_COMBOLBOX))
    comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
  if ( (comboLBox) && (IWindowClassName(comboLBox) == WC_COMBOLBOX) )
    SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
#endif
}

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                 |
|                                                                              |
| Construct an IBaseComboBox on a dialog.                                      |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox( unsigned long id, IWindow* parent)
                            : fBaseComboBoxData( 0 )
{
   // Create a private data object.
   fBaseComboBoxData = new IBaseComboBoxData;

   setAutoDestroyWindow(false);
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
   startHandlingEventsFor(id, parent);
   IMousePointerHandler::defaultHandler()->handleEventsFor(this);
#ifdef IC_WIN
   // Subclass the entry field in the combo box so that we can get
   // navigation keys.
   fBaseComboBoxData->fentryHandle =
      IBaseComboBoxData::handleOfEntryField( this->handle() );
   if (fBaseComboBoxData->fentryHandle)
      {
      // setup our procedure in the window.
      fBaseComboBoxData->fdefaultEntryProc = (WNDPROC)
         ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                         icomboboxEntryWinProc);
      if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
         {
         // throw now rather than trap later.
         ITHROWGUIERROR( "ISUBCLASSWINDOW" );
         }
      }

   // Store whComboBox in the ComboListBox (COMBOLBOX) windows user data.
   // It is  used to retrieve colors when painting ComboListBox.  This is
   // necessary because the ComboListBox window is a child of the desktop
   // not the combo box.  Se we tuck away the ComboBox hwnd for each
   // combo list box created if nothing is already stored in GWL_USERDATA.
   HWND deskTop = desktopWindow()->handle();
   HWND comboBox = this->handle();
   IWindowHandle comboLBox = GetWindow(desktopWindow()->handle(), GW_CHILD);
   while (comboLBox)
   {
     if (IWindowClassName(comboLBox) == WC_COMBOLBOX)
     {
       HWND curHwnd = (HWND) GetWindowLong(comboLBox, GWL_USERDATA);
       if ( !(curHwnd) )
       {
         SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
       }
     }
     comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
   }
   // Take care of CBS_SIMPLE too
   comboLBox = GetWindow(this->handle(), GW_CHILD);
   while ( (comboLBox) && (IWindowClassName(comboLBox) != WC_COMBOLBOX))
     comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
   if ( (comboLBox) && (IWindowClassName(comboLBox) == WC_COMBOLBOX) )
     SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
#endif
}

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                 |
|                                                                              |
| Construct an IBaseComboBox from an existing IBaseComboBox.                   |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox( const IWindowHandle& handle )
                            : fBaseComboBoxData( 0 )
{
   // Create a private data object.
   fBaseComboBoxData = new IBaseComboBoxData;

   setAutoDestroyWindow(false);
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
   startHandlingEventsFor(handle);
   IMousePointerHandler::defaultHandler()->handleEventsFor(this);
#ifdef IC_WIN
   // Subclass the entry field in the combo box so that we can get
   // navigation keys.
   fBaseComboBoxData->fentryHandle =
      IBaseComboBoxData::handleOfEntryField( handle );
   if (fBaseComboBoxData->fentryHandle)
      {
      // setup our procedure in the window.
      fBaseComboBoxData->fdefaultEntryProc = (WNDPROC)
         ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                         icomboboxEntryWinProc);
      if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
         {
         // throw now rather than trap later.
         ITHROWGUIERROR( "ISUBCLASSWINDOW" );
         }
      }

   // Store whComboBox in the ComboListBox (COMBOLBOX) windows user data.
   // It is  used to retrieve colors when painting ComboListBox.  This is
   // necessary because the ComboListBox window is a child of the desktop
   // not the combo box.  Se we tuck away the ComboBox hwnd for each
   // combo list box created if nothing is already stored in GWL_USERDATA.
   HWND deskTop = desktopWindow()->handle();
   HWND comboBox = handle;
   IWindowHandle comboLBox = GetWindow(desktopWindow()->handle(), GW_CHILD);
   while (comboLBox)
   {
     if (IWindowClassName(comboLBox) == WC_COMBOLBOX)
     {
       HWND curHwnd = (HWND) GetWindowLong(comboLBox, GWL_USERDATA);
       if ( !(curHwnd) )
       {
         SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
       }
     }
     comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
   }
   // Take care of CBS_SIMPLE too
   comboLBox = GetWindow(handle, GW_CHILD);
   while ( (comboLBox) && (IWindowClassName(comboLBox) != WC_COMBOLBOX))
     comboLBox = GetWindow(comboLBox, GW_HWNDNEXT);
   if ( (comboLBox) && (IWindowClassName(comboLBox) == WC_COMBOLBOX) )
     SetWindowLong(comboLBox, GWL_USERDATA, (LONG)comboBox);
#endif
}

/*------------------------------------------------------------------------------
| IBaseComboBox::~IBaseComboBox                                                |
|                                                                              |
| Reomve the default mouse pointer handler.                                    |
------------------------------------------------------------------------------*/
IBaseComboBox :: ~IBaseComboBox()
{
#ifdef IC_WIN
   // Remove the default Handler
   removeHandler(defaultHandler);
   // Remove the subclass of the entry field
   if ( (fBaseComboBoxData->fentryHandle) &&
        IISWINDOW(0, fBaseComboBoxData->fentryHandle) )
      {
      ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                      fBaseComboBoxData->fdefaultEntryProc);
      }
#endif
  IMousePointerHandler::defaultHandler()->stopHandlingEventsFor(this);
  if ( fBaseComboBoxData != 0 )
    delete fBaseComboBoxData;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::type                                                          |
------------------------------------------------------------------------------*/
IBaseComboBox::ControlType  IBaseComboBox::type() const
{
  unsigned long ulType =
           style() & (simpleType.asUnsignedLong()  |
                      dropDownType.asUnsignedLong() |
                      readOnlyDropDownType.asUnsignedLong());
  ControlType eReturn = simple;

#ifdef IC_WIN
  if ((ulType & simpleType.asUnsignedLong())  &&
      (ulType & dropDownType.asUnsignedLong()))
  {
    return( dropDownList );
  }
#endif

  switch (ulType)
  {
     case CBS_SIMPLE:
        break;
     case CBS_DROPDOWN:
        eReturn = dropDown;
        break;
     case CBS_DROPDOWNLIST:
        eReturn = dropDownList;
        break;
     default:
        break;
  }

  return eReturn;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::defaultStyle                                                      |
------------------------------------------------------------------------------*/
IBaseComboBox::Style  IBaseComboBox::defaultStyle()
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::setDefaultStyle                                               |
------------------------------------------------------------------------------*/
void IBaseComboBox::setDefaultStyle(const Style& style)
{
  currentDefaultStyle = style;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::convertToGUIStyle                                             |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IBaseComboBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // We don't get the IEntryField GUI styles since IEntryField styles are not
  // used by combo boxes.  Instead, get the ITextControl style.
  unsigned long ulStyle = ITextControl::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles that we don't overlap with our
    // extended styles.
    ulStyle |= guiStyle.asExtendedUnsignedLong() & ICBS_EXTGUIMASK;
  }
  else
  {
    // Let the clean CBS_ styles, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & ICBS_MASK;

#ifdef IC_WIN
    unsigned long ulTmp = guiStyle.asExtendedUnsignedLong() & ICBS_EXTMASK;

    // Add CBS_DROPDOWNLIST style if the corresponding extended bit is set
    if (ulTmp & readOnlyDropDownType.asExtendedUnsignedLong())
    {
       ulStyle |= CBS_DROPDOWNLIST;
    }

    ulStyle |= WS_CHILD | CBS_HASSTRINGS
               | WS_VSCROLL | CBS_NOINTEGRALHEIGHT;
#endif
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| IBaseComboBox::isListShowing                                                 |
------------------------------------------------------------------------------*/
Boolean   IBaseComboBox::isListShowing() const
{
  if (type() != simple)
  {
     IEventResult evt = handle().sendEvent(CBM_ISLISTSHOWING,
                                IEventParameter1(0),
                                IEventParameter2(0));
     if (evt.asUnsignedLong())
        return true;
     else
        return false;
  }
  else
     return true;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::showList                                                          |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox::showList(Boolean show)
{
  Boolean stateChanged = false;

  if (type() != simple)
  {
    if (show)
    {
      if (!isListShowing())
      {
        stateChanged = true;
      }
    }
    else if (isListShowing())
    {
      stateChanged = true;
    }

    if (stateChanged)
    {
      handle().sendEvent(CBM_SHOWLIST, IEventParameter1(show),
                                       IEventParameter2(0));
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::hideList                                                      |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox::hideList()
{
  showList(false);
  return *this;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IBaseComboBox::hasFocus                                                      |
------------------------------------------------------------------------------*/
Boolean         IBaseComboBox::hasFocus() const
{
   if ( extendedStyle() & readOnlyDropDownType.asExtendedUnsignedLong())
     return Inherited::hasFocus();      // static text type combobox
   else
     {                                  // has an Edit window
     char       className[8];
     HWND hwndEdit = GetWindow( handle(), GW_CHILD );
     GetClassName( hwndEdit, className, sizeof(className));
     if ( lstrcmp( className, "Edit" ))
       hwndEdit = GetWindow( hwndEdit, GW_HWNDNEXT );
     return hwndEdit == IQUERYFOCUS( HWND_DESKTOP );
     }
}
#endif

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IBaseComboBox::setLimit                                                      |
|                                                                              |
| Sets the maximum number of characters allowed in the entry field portion of  |
| the combo box.                                                               |
------------------------------------------------------------------------------*/
IBaseComboBox& IBaseComboBox::setLimit( unsigned long ulLimit )
{
  IASSERTSTATE(text().length() <= ulLimit);

  if (this->limit() != ulLimit)
  {
    /*************************************************************/
    /* Now set the new edit limit for the entry field.           */
    /*************************************************************/
    handle().sendEvent(CB_LIMITTEXT,
                       IEventParameter1(ulLimit),
                       IEventParameter2(0));
    fBaseComboBoxData->charLimit = ulLimit;
    setLayoutDistorted(IWindow::minimumSizeChanged, 0);
  }

  return( *this );
}
#endif


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IBaseComboBox::setLimit                                                      |
|                                                                              |
| Sets the maximum number of characters allowed in the entry field given a     |
| string resource-id that defines the limit.                                   |
------------------------------------------------------------------------------*/
IBaseComboBox& IBaseComboBox::setLimit( const IResourceId& residTextLimit )
{
  IString strTextLimit =
     residTextLimit.resourceLibrary().loadString(residTextLimit);
  setLimit(strTextLimit.asInt());
  return( *this );
}
#endif


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IBaseComboBox::limit                                                         |
|                                                                              |
| Return the maximum number of characters allowed in the entry field portion   |
| of the combo box.                                                            |
------------------------------------------------------------------------------*/
unsigned long IBaseComboBox::limit( ) const
{
  return( fBaseComboBoxData->charLimit );
}
#endif

/*------------------------------------------------------------------------------
| IBaseComboBox::calcMinimumSize                                               |
|                                                                              |
| Calculate the minimum size needed by the control.  The height is based on    |
| the font height and the width is based on the average font width multiplied  |
| by the text limit.                                                           |
------------------------------------------------------------------------------*/
ISize  IBaseComboBox::calcMinimumSize() const
{
  ISize sizMin = Inherited::calcMinimumSize();

  IBaseComboBox::ControlType comboType = type();

  // If we aren't a drop-down, add room for minimum rows
  if ( comboType != dropDown && comboType != dropDownList)
  {
    sizMin.setHeight ( sizMin.height() +
                       IQUERYSYSVALUE(SV_CYBORDER) +
                       minimumRows() * IFont(this).maxCharHeight());
     // Note: Only the bottom border is added to the height since the
     // listbox top border overlaps the entry field bottom border
     if ( ( extendedStyle() & border3D.asExtendedUnsignedLong() ) &&
          ( IPlatform::isWin95() || IPlatform::isNTNewShell() ))
     {
        sizMin += ISize(0,2);
     }
  }

  // Add size for the entry field border (doesn't seem to happen
  // via entry field calculation)
  sizMin += ISize((3 * IQUERYSYSVALUE(SV_CXBORDER) + 2),
#ifdef IC_WIN
                     (2 * IQUERYSYSVALUE(SV_CYBORDER)));
#else
                     (3 * IQUERYSYSVALUE(SV_CYBORDER) + 2));
#endif

#ifdef IC_WIN
  // If this is a drop down, add room for button.
  if (comboType == dropDown || comboType == dropDownList)
  {
     // Use the width of a vertical scrollbar arrow button for
     // button width.
     sizMin.setWidth( sizMin.width() +
                      GetSystemMetrics( SM_CXVSCROLL ));
  }
#endif

  if (isHorizontalScroll() && comboType != dropDown &&
      comboType != dropDownList)
  {
     sizMin.setHeight (sizMin.height() + IQUERYSYSVALUE(SV_CYHSCROLL));
  }

  return sizMin;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::layoutAdjustment                                              |
|                                                                              |
| Calculate the size and distance a dropDown Combination box should be         |
| adjusted after layout.                                                       |
------------------------------------------------------------------------------*/
IRectangle  IBaseComboBox::layoutAdjustment() const
{
  IBaseComboBox::ControlType comboType = type();

  if ( comboType == dropDown || comboType == dropDownList)
  {
     // Adjust position so that minimum rows of text will fit in the
     // drop down listbox
     unsigned long height = minimumRows() * IFont( this ).maxCharHeight() +
                            IQUERYSYSVALUE( SV_CYBORDER );
     if (isHorizontalScroll())
        height += IQUERYSYSVALUE( SV_CYHSCROLL );
     return IRectangle( 0, -height, 0, 0);
  }
  else
    return Inherited::layoutAdjustment();
}

/*------------------------------------------------------------------------------
| IBaseComboBox::visibleRectangle                                              |
|                                                                              |
| Calculate the actual displayed (painted) ares of the combo box.              |
------------------------------------------------------------------------------*/
IRectangle  IBaseComboBox::visibleRectangle() const
{
#ifndef IC_WIN
  IRectangle    trueRect = nativeRect();
  SWP           swp;

  IBaseComboBox::ControlType comboType = type();
  if ( comboType == dropDown || comboType == dropDownList)
  {
    IQUERYWINDOWPOS(IWINDOWFROMID(handle(),CBID_EDIT),&swp);
    unsigned long cyBorder = IQUERYSYSVALUE(SV_CYBORDER);
    trueRect = IRectangle ( IPoint(trueRect.left(),
                                   trueRect.bottom()+swp.y-(2*cyBorder)),
                            trueRect.topRight() );
  }
  else
  {
    IQUERYWINDOWPOS(IWINDOWFROMID(handle(),CBID_LIST),&swp);
    trueRect = IRectangle ( IPoint(trueRect.left(),
                            trueRect.bottom()+swp.y),
                            trueRect.topRight() );
  }
  return ICoordinateSystem::convertToApplication( trueRect, parentSize() );
#else
  RECT   rect;
  IQUERYWINDOWRECT(handle(), &rect);
  MapWindowPoints( HWND_DESKTOP, IPARENTOF( handle() ), (POINT*)&rect, 2);
  return ICoordinateSystem::convertToApplication( IRectangle( rect ),
                                                  parentSize() );
#endif
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IBaseComboBox::nativeRect                                                    |
------------------------------------------------------------------------------*/
IRectangle IBaseComboBox::nativeRect ( ) const
{
   if (type() == simple)
      return Inherited::nativeRect();

   RECT rect;
   handle().sendEvent( CB_GETDROPPEDCONTROLRECT,
                       IEventParameter1( 0 ),
                       IEventParameter2( &rect ));

   MapWindowPoints( HWND_DESKTOP, IPARENTOF( handle() ), (POINT*)&rect, 2);
   return IRectangle( rect );
}
#endif

/*------------------------------------------------------------------------------
| IBaseComboBox::setMinimumRows                                                |
|                                                                              |
| Sets the number of rows in a minimum size combobox.                          |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox :: setMinimumRows (unsigned long minimumRows)
{
  fBaseComboBoxData->minimumRows = minimumRows;
  setLayoutDistorted (minimumSizeChanged, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::minimumRows                                                   |
|                                                                              |
| Returns the number of rows in a minimum size combobox.                       |
------------------------------------------------------------------------------*/
unsigned long  IBaseComboBox :: minimumRows () const
{
  return fBaseComboBoxData->minimumRows;
}



#ifdef IC_PMWIN
IBaseComboBox& IBaseComboBox::setAlignment(IEntryField::Alignment alignment)
{
   ITHROWLIBRARYERROR1(IC_MEMBER_ACCESS_ERROR,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable,
                       "IBaseComboBox");
   return *this;
}
#endif

#ifdef IC_WIN

/*------------------------------------------------------------------------------
| IBaseComboBoxData::handleOfEntryField                                        |
------------------------------------------------------------------------------*/
IWindowHandle IBaseComboBoxData::handleOfEntryField(
                                         const IWindowHandle& combobox )
{
   // Find the edit control child of the combobox.
   IWindowHandle hwndChild = GetWindow(combobox, GW_CHILD);
   while ( (hwndChild) && (IWindowClassName(hwndChild) != WC_ENTRYFIELD))
      hwndChild = GetWindow(hwndChild, GW_HWNDNEXT);
   return hwndChild;
}

/*------------------------------------------------------------------------------
| icomboboxEntryWinProc                                                        |
| Window procedure for the entryfield part of a combo box.  This procedure     |
| intercepts the messages used for keyboard navigation and sends them to       |
| the combobox owner.  Other messages are send on to the original procedure    |
| for the combobox entryfield.                                                 |
------------------------------------------------------------------------------*/
long  __stdcall icomboboxEntryWinProc(void*         hwnd,
                                      unsigned int  msg,
                                      unsigned int  mp1,
                                      long          mp2)
{
   IWindowHandle  efParent   = IPARENTOF(hwnd);
   IBaseComboBox* combobox = (IBaseComboBox*)
      IWindow::windowWithHandle( efParent, false );
   if (combobox)
      {
      switch( msg )
         {
         case WM_CHAR:
            if (mp1 == VK_ENTER)
            {
               if (combobox->type() == IBaseComboBox::simple ||
                   (combobox->type() != IBaseComboBox::simple && !combobox->isListShowing()))
               {
                  IWindowHandle hwndParent = IPARENTOF( efParent );
                  hwndParent.sendEvent( WM_COMMAND,
                     IEventParameter1( (short)combobox->id(), IC_UM_CBN_ENTER),
                     IEventParameter2( efParent.asUnsigned() ));
               }
            }
            if (mp1 == VK_TAB ||
                mp1 == VK_ESC ||
                mp1 == VK_ENTER )
            {
               // send these to the combobox owner or parent.  We need
               // VK_TAB to handle keyboard navigation.  We need to pass on
               // VK_ESC and VK_ENTER for correct dialog behavior.
               IWindow* owner = combobox->owner();
               if (!owner)
                  owner = combobox->parent();
               owner->sendEvent((unsigned long)msg,
                                (unsigned long)mp1,
                                (unsigned long)mp2);
               return false;
            }
            break;

         case WM_SYSKEYUP:
         case WM_SYSKEYDOWN:
         {
           // Need to send Alt-mnemonic events to combobox owner or parent.
           IWindow *owner = combobox->owner();
           if (!owner)
              owner = combobox->parent();
           owner->sendEvent((unsigned long)msg,
                            (unsigned long)mp1,
                            (unsigned long)mp2);
           return false;
         }
         break;

         default:
            break;
         }

      return CallWindowProc(
                (FARPROC)combobox->fBaseComboBoxData->fdefaultEntryProc,
                hwnd, msg, mp1, mp2 );
      }
   else
      {
      // We are basically busted if we can't find our parent IWindow.  We do
      // what is probably the most reasonable thing and call the
      // default procedure.
      return IDEFWINDOWPROC( hwnd, msg, mp1, mp2 );
      }
}


#endif

#include <ilistbx3.hpp>

