/*******************************************************************************
* FILE NAME: iflyhhdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iflyhhdr.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WIN
  #define INCL_WINMESSAGEMGR
  #define INCL_WINWINDOWMGR
  #define INCL_WINPOINTERS
  #define INCL_WININPUT
  #define WINVER 0x0400
  #include <iwindefs.h>
  #ifdef IC_WIN
    #include <commctrl.h>
  #endif
  #include <string.h>
}

#include <iflyhhdr.hpp>
#include <iflytext.hpp>
#include <iflypriv.hpp>
#include <iapp.hpp>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <ictlevt.hpp>
#include <ithread.hpp>
#include <itrace.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IFLYHHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IHandleStringElement::IHandleStringElement                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IHandleStringElement::IHandleStringElement( unsigned long      handle,
                                            const IString&     flyText,
                                            const IString&     longText )
:  handleKey(handle)
 , flyStr(flyText)
 , longStr(longText)
 , flyResId(0)
 , longResId(0)
 , fStrings(true)
#ifdef IC_WIN
 , subclassProc(0)
 , flyData(0)
#endif
{}

/*------------------------------------------------------------------------------
| IHandleStringElement::IHandleStringElement                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IHandleStringElement::IHandleStringElement( unsigned long      handle,
                                            const IResourceId& flyText,
                                            const IResourceId& longText )
:  handleKey(handle)
 , flyStr()
 , longStr()
 , flyResId(flyText)
 , longResId(longText)
 , fStrings(false)
#ifdef IC_WIN
 , subclassProc(0)
 , flyData(0)
#endif
{}

/*------------------------------------------------------------------------------
| IHandleStringElement::~IHandleStringElement                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IHandleStringElement::~IHandleStringElement( )
{
}

/*------------------------------------------------------------------------------
| IHandleStringSet::IHandleStringSet                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IHandleStringSet::IHandleStringSet()
: IKeySortedSetOnSortedTabularSequence < IHandleStringElement*, unsigned long > (10)
{}

/*------------------------------------------------------------------------------
| IHandleStringSet::~IHandleStringSet                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IHandleStringSet::~IHandleStringSet()
{
  // Delete all the elements in the set and remove them.
  IHandleStringSet::Cursor cursor(*this);
  forCursor (cursor)
  {
    delete this->elementAt( cursor );
  }
  this->removeAll();
}

/*------------------------------------------------------------------------------
| Static Data.                                                                 |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
static IHandleStringSet fgHandleStringSet;
#pragma data_seg()

#ifdef IC_WIN

// The following window proc is used to subclass child windows of
// aggregate controls such as combo boxes and spin buttons.  These
// child windows are subclassed so that we can catch mouse messages
// and relay them onto the tool tip control.  This allows tool
// tip messages for all immediate child windows in an aggregate control.
/*------------------------------------------------------------------------------
| fFlyHelpSubclassProc                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long __stdcall fFlyHelpSubclassProc(HWND hwnd, unsigned long ulMsg,
                                             WPARAM mp1, LPARAM mp2)
{
  IHandleStringElement *element(0);
  // Get the handle string element from the hwnd.
  if ( fgHandleStringSet.containsElementWithKey( (unsigned long)hwnd ))
  {
    element = (IHandleStringElement*)
        fgHandleStringSet.elementWithKey( (unsigned long)hwnd );
  }
  else
  {
    // We should never get here!
    return 0;
  }

  switch (ulMsg)
  {
    case WM_MOUSEMOVE       :
    case WM_LBUTTONDOWN     :
    case WM_LBUTTONUP       :
    case WM_LBUTTONDBLCLK   :
    case WM_RBUTTONDOWN     :
    case WM_RBUTTONUP       :
    case WM_RBUTTONDBLCLK   :
    case WM_MBUTTONDOWN     :
    case WM_MBUTTONUP       :
    case WM_MBUTTONDBLCLK   :
    {
      // relay mouse messages to the tooltip control
      if ( element->flyData->fFlytxt->isValid() )
          {
          MSG msg;
          msg.hwnd    = hwnd;
          msg.message = (UINT)ulMsg;
          msg.wParam  = mp1;
          msg.lParam  = mp2;
          element->flyData->fFlytxt->sendEvent( TTM_RELAYEVENT,
                                                0,
                                                IEventData(&msg) );
          }
       }
       break;

    case WM_NOTIFY         :
       element->flyData->notifyEvent((NMHDR*)mp2);
       break;

    case WM_DESTROY:
       // Un-subclass the window and remove the window from the
       // handle string set.
       SetWindowLong( hwnd, GWL_WNDPROC, (long)element->subclassProc );
       fgHandleStringSet.removeElementWithKey((unsigned long)hwnd);
    break;

  }
  return (unsigned long)
           CallWindowProc((FARPROC)element->subclassProc, hwnd, ulMsg, mp1, mp2);
}

/*------------------------------------------------------------------------------
| IFlyOverData::setHelpText                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverData& IFlyOverData::setHelpText( const IWindowHandle& handle,
                                         IHandleStringElement& element )
{
  TOOLINFO  toolinfo;
  toolinfo.cbSize   = sizeof(TOOLINFO);
  toolinfo.uFlags   = TTF_IDISHWND;
  toolinfo.hwnd     = handle;
  toolinfo.uId      = (UINT)handle.asUnsigned();
  toolinfo.lpszText = LPSTR_TEXTCALLBACK;    // want TTN_NEEDTEXT message

  IWindow* child(IWindow::windowWithHandle( handle ));

  if (!child)
  {
    element.subclassProc = (unsigned long (* __stdcall)
                            (HWND hwnd, unsigned long ulMsg,
                             WPARAM mp1, LPARAM mp2))

                            SetWindowLong( handle, GWL_WNDPROC,
                                           (long)fFlyHelpSubclassProc );
    element.flyData = this;

    fFlytxt->sendEvent( TTM_ADDTOOL,
                        0,
                        IEventData( &toolinfo) );
  }
  else
  {
    if (child->sendEvent( IC_UM_IS_AGGREGATE_CTRL ).asUnsignedLong())
    {
      IWindow::ChildCursor aggregateCursor( *child );
      forCursor (aggregateCursor)
      {
        IWindowHandle childHandle(child->childAt( aggregateCursor ));
        if ( fgHandleStringSet.containsElementWithKey( childHandle.asUnsigned() ))
        {
          IHandleStringElement*
            childElement(fgHandleStringSet.elementWithKey( childHandle.asUnsigned()));

          if (element.fStrings)
          {
            childElement->flyStr    = element.flyStr;
            childElement->longStr   = element.longStr;
            childElement->flyResId  = 0;
            childElement->longResId = 0;
            childElement->fStrings  = true;
          }
          else
          {
            childElement->flyStr    = IString();
            childElement->longStr   = IString();
            childElement->flyResId  = element.flyResId;
            childElement->longResId = element.longResId;
            childElement->fStrings  = false;
          }
        }
      }
    }
  }
  return *this;
}

#endif

/*------------------------------------------------------------------------------
| IFlyOverData::IFlyOverData                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverData::IFlyOverData( IFlyText* flyText, ITextControl* longText,
                            unsigned long delay, unsigned long initialDelay,
                            IFlyOverHelpHandler* flyHelp )
: fFlytxt(flyText), fLongtxt(longText),
  fLongOffset(0), fFlyOffset(0),
#ifndef IC_WIN
  fTimer(),
  fPrevPos(),
#endif
  fDefault(),
  fSetPrevToNull(false),
#ifndef IC_WIN
  fHandle(0),
  fPrevHandle(0),
#endif
  fControlHandle(0),
#ifndef IC_WIN
  fControlWindow(0),
#endif
  fFly(flyHelp),
  fResLib(0),
  fTimerDelay(delay),            // Save the original timer value.
  fInitialTimerDelay(initialDelay)
{
#ifdef IC_WIN
  // Initialize the delays in the tooltips control
  if (fFlytxt)
     {
     fFlytxt->sendEvent( TTM_SETDELAYTIME,
                         IEventData( TTDT_INITIAL),
                         IEventData( fInitialTimerDelay ) );
     fFlytxt->sendEvent( TTM_SETDELAYTIME,
                         IEventData( TTDT_RESHOW ),
                         IEventData( fTimerDelay ) );
     // This is supposed to keep the help displayed.  Does not
     // seem to work.   Goes away after maybe 10 seconds.
     fFlytxt->sendEvent( TTM_SETDELAYTIME,              //stays as long as
                         IEventData( TTDT_AUTOPOP),     //mouse does
                         IEventData( INT_MAX ) );
     }
#endif

}

/*------------------------------------------------------------------------------
| IFlyOverData::~IFlyOverData                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverData::~IFlyOverData()
{
  delete fResLib;
#ifndef IC_WIN
  fTimer.stop();
#endif
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::IFlyOverHelpHandler                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler::IFlyOverHelpHandler( IFlyText* flyText,
                                          unsigned long initialDelay,
                                          unsigned long delay )
: fFlyOverData( new IFlyOverData(flyText, 0, delay, initialDelay, this))
{
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::IFlyOverHelpHandler                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler::IFlyOverHelpHandler( ITextControl* longText,
                                          unsigned long initialDelay,
                                          unsigned long delay )
: fFlyOverData( new IFlyOverData(0, longText, delay, initialDelay, this))
{
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::IFlyOverHelpHandler                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler::IFlyOverHelpHandler( IFlyText* flyText,
                                          ITextControl* longText,
                                          unsigned long initialDelay,
                                          unsigned long delay )
: fFlyOverData( new IFlyOverData(flyText, longText, delay, initialDelay, this))
{
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::~IFlyOverHelpHandler                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler::~IFlyOverHelpHandler()
{
  delete fFlyOverData;
  IHandleStringSet::Cursor cursor(fgHandleStringSet);
  for (cursor.setToFirst(); cursor.isValid();)
  {
    IHandleStringElement* element((IHandleStringElement*)fgHandleStringSet.elementAt( cursor ));

    if (IISWINDOW(IThread::current().anchorBlock(),
                  (HWND)element->handleKey ))
      cursor.setToNext();
    else
    {
      fgHandleStringSet.removeAt( cursor );
      delete element;
      cursor.setToFirst();
    }
  }
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::flyTextControl                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyText* IFlyOverHelpHandler::flyTextControl( ) const
{
  return fFlyOverData->fFlytxt;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::longTextControl                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ITextControl* IFlyOverHelpHandler::longTextControl( ) const
{
  return fFlyOverData->fLongtxt;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setFlyTextControl                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setFlyTextControl(IFlyText* flyText)
{
  fFlyOverData->fFlytxt = flyText;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setLongTextControl                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setLongTextControl(
                                                        ITextControl* longText )
{
  fFlyOverData->fLongtxt = longText;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::resourceLibrary                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IResourceLibrary& IFlyOverHelpHandler::resourceLibrary( ) const
{
  // If we've got our own, return it, else use the current user default.
  if ( fFlyOverData->fResLib )
    return *fFlyOverData->fResLib;
  else
    return IApplication::current().userResourceLibrary();
}

#ifndef IC_MOTIF
/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setResourceLibrary                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setResourceLibrary(
                                                   const IModuleHandle &resMod )
{
  // Construct a DLL from the input handle.

  delete fFlyOverData->fResLib;
  fFlyOverData->fResLib = 0;
  if ( resMod )
    fFlyOverData->fResLib = new IDynamicLinkLibrary( resMod );
  else
    fFlyOverData->fResLib = new IResourceLibrary();

  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setResourceLibrary                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setResourceLibrary(
                                                        const char *resDLLName )
{
  // Construct a DLL from the input dll name.

  delete fFlyOverData->fResLib;
  fFlyOverData->fResLib = 0;
  if ( resDLLName )
    fFlyOverData->fResLib = new IDynamicLinkLibrary( resDLLName );
  else
    fFlyOverData->fResLib = new IResourceLibrary();

  return *this;
}
#endif

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setFlyStringTableOffset                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setFlyTextStringTableOffset(
                                                                long newOffset )
{
  fFlyOverData->fFlyOffset = newOffset;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setLongStringTableOffset                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setLongStringTableOffset(
                                                                long newOffset )
{
  fFlyOverData->fLongOffset = newOffset;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::flyTextStringTableOffset                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
long IFlyOverHelpHandler::flyTextStringTableOffset ( ) const
{
  return fFlyOverData->fFlyOffset;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::longStringTableOffset                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
long IFlyOverHelpHandler::longStringTableOffset  ( ) const
{
  return fFlyOverData->fLongOffset;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::defaultText                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IFlyOverHelpHandler::defaultText ( ) const
{
  return fFlyOverData->fDefault;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setDefaultText                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setDefaultText( unsigned long id )
{
  fFlyOverData->fDefault = this->resourceLibrary().loadString( id );
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setDefaultText                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setDefaultText( const IString &id )
{
  fFlyOverData->fDefault = id;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::flyHelpText                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IFlyOverHelpHandler::flyHelpText(const IWindowHandle& handle ) const
{
  if ( fgHandleStringSet.containsElementWithKey( handle.asUnsigned() ))
  {
    IHandleStringElement*
      elementWithHandle(
         fgHandleStringSet.elementWithKey( handle.asUnsigned() ));
    return elementWithHandle->flyStr;
  }
  else
  {
    return IString();
  }
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::longHelpText                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IFlyOverHelpHandler::longHelpText(const IWindowHandle& handle ) const
{
  if ( fgHandleStringSet.containsElementWithKey( handle.asUnsigned() ))
  {
    IHandleStringElement*
      elementWithHandle(fgHandleStringSet.elementWithKey(
         handle.asUnsigned() ));
    return elementWithHandle->longStr;
  }
  else
  {
    return IString();
  }
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setHelpText                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setHelpText
                                               ( const IWindowHandle& handle,
                                                 const IString&       flyText,
                                                 const IString&       longText )
{
  IHandleStringElement* element(0);
  if ( fgHandleStringSet.containsElementWithKey( handle.asUnsigned() ))
  {
    element = (IHandleStringElement*)fgHandleStringSet.elementWithKey( handle.asUnsigned() );
    element->flyStr    = flyText;
    element->longStr   = longText;
    element->flyResId  = 0;
    element->longResId = 0;
    element->fStrings  = true;
  }
  else
  {
    element = new IHandleStringElement(handle.asUnsigned(),flyText,longText);
    fgHandleStringSet.add(element);
  }

#ifdef IC_WIN
  fFlyOverData->setHelpText( handle, *element );
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setHelpText                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setHelpText
                                               ( const IWindowHandle& handle,
                                                 const IResourceId&   flyText,
                                                 const IResourceId&   longText )
{
  IHandleStringElement* element(0);
  if ( fgHandleStringSet.containsElementWithKey( handle.asUnsigned() ))
  {
    element = (IHandleStringElement*)fgHandleStringSet.elementWithKey( handle.asUnsigned() );
    if (element->fStrings)
    {
      element->flyStr  = IString();
      element->longStr = IString();
    }
    element->flyResId  = flyText;
    element->longResId = longText;
    element->fStrings  = false;
  }
  else
  {
    element = new IHandleStringElement(handle.asUnsigned(),flyText,longText);
    fgHandleStringSet.add(element);
  }

#ifdef IC_WIN
  fFlyOverData->setHelpText( handle, *element );
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::removeHelpText                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::removeHelpText
                                                 ( const IWindowHandle& handle )
{
  if ( fgHandleStringSet.containsElementWithKey( handle.asUnsigned() ))
  {
    IHandleStringElement *oldElement((IHandleStringElement*)
       fgHandleStringSet.elementWithKey( handle.asUnsigned() ));
    fgHandleStringSet.removeElementWithKey( handle.asUnsigned() );

#ifdef IC_WIN
    IWindow* oldWindow(IWindow::windowWithHandle( handle ));
    if (!oldWindow && oldElement->subclassProc != 0)
    {
      SetWindowLong( handle, GWL_WNDPROC, (long)oldElement->subclassProc );

      TOOLINFO  toolinfo;
      toolinfo.cbSize = sizeof(TOOLINFO);
      toolinfo.hwnd   = handle;
      toolinfo.uId    = (UINT)handle.asUnsigned();

      fFlyOverData->fFlytxt->sendEvent( TTM_DELTOOL,
                                        0,
                                        IEventData( &toolinfo) );
    }
#endif
    delete oldElement;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setDelayTime                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setDelayTime( unsigned long milliseconds )
{
#ifdef IC_WIN
  IFlyText* flytext = this->flyTextControl();
  if (flytext)
     {
     flytext->sendEvent( TTM_SETDELAYTIME,
                         IEventData( TTDT_RESHOW ),
                         IEventData( milliseconds ) );
     }
#else
  fFlyOverData->fTimer.setInterval( milliseconds/2 );
#endif
  // Save the new timer delay.
  fFlyOverData->fTimerDelay = milliseconds;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::delayTime                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IFlyOverHelpHandler::delayTime( ) const
{
  // Return the last set time delay.
  // return fFlyOverData->fTimer.interval();
  return fFlyOverData->fTimerDelay;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setInitialDelayTime                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::setInitialDelayTime( unsigned long milliseconds )
{
#ifdef IC_WIN
  IFlyText* flytext = this->flyTextControl();
  if (flytext)
     {
     flytext->sendEvent( TTM_SETDELAYTIME,
                         IEventData( TTDT_INITIAL ),
                         IEventData( milliseconds ) );
     }
#endif
  // Save the new timer delay.
  fFlyOverData->fInitialTimerDelay = milliseconds;
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::initialDelayTime                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IFlyOverHelpHandler::initialDelayTime( ) const
{
  // Return the last set time delay.
  // return fFlyOverData->fTimer.interval();
  return fFlyOverData->fInitialTimerDelay;
}


/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::startHandlingEventsFor                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::handleEventsFor( IWindow* window )
{
  IASSERTPARM(window != 0);
  IFlyOverHelpHandler::Inherited::handleEventsFor( window );
#ifdef IC_WIN
  // Register the new control with the tooltips control.  Arrange
  // to have WM_NOTIFY (TTN_NEEDTEXT) sent to the window we are handling
  // events for.  We deal with this message in dispatchHandlerEvents.
  // If IWindow is changed to do WM_NOTIFY like WM_CONTROL, we
  // will need to change the hwnd value to be the owner.
  IFlyText* flytext = this->flyTextControl();
  if (flytext)
     {
     TOOLINFO  toolinfo;
     toolinfo.cbSize = sizeof(TOOLINFO);
     toolinfo.uFlags = TTF_IDISHWND;
     toolinfo.hwnd   = window->handle();  // sends WM_NOTIFY to this window
     toolinfo.uId    = (UINT)window->handle().asUnsigned();
                                         // window for which help applies
     toolinfo.lpszText = LPSTR_TEXTCALLBACK;    // want TTN_NEEDTEXT message
     // TOOLINFO fields rect and hinst not needed
     IMODTRACE_ALL
       ( IString("TTM_ADDTOOL") +
         IString(" uFlags="  ) +
         IString((unsigned long)toolinfo.uFlags).d2x() +
         IString(" hwnd="    ) +
         IString((unsigned long)toolinfo.hwnd).d2x() +
         IString(" uId="     ) +
         IString((unsigned long)toolinfo.uId).d2x() +
         IString(" lpszText=") +
         IString((unsigned long)toolinfo.lpszText).d2x() );
     flytext->sendEvent( TTM_ADDTOOL,
                         0,
                         IEventData( &toolinfo) );

     Boolean windowIsAggregate((Boolean)window->sendEvent( IC_UM_IS_AGGREGATE_CTRL ).asUnsignedLong());

     //
     // Perform a similar registration of child IWindows
     IWindow::ChildCursor cursor( *window );
     for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {
        IWindow* child = IWindow::windowWithHandle( window->childAt( cursor ));
        if (child)
           {
           // Send the IC_UM_FLY_PAINT message with lParam == 0 to check
           // if this handler is already installed on the child.  If so,
           // don't install it again.
           if (!child->sendEvent( IC_UM_FLY_PAINT,
                                  0,
                                  0 ).asUnsignedLong() )
              {
              if (!child->sendEvent( IC_UM_IS_AGGREGATE_CTRL ).asUnsignedLong())
                 {
                 this->handleEventsFor( child );
                 }
              else
                 {
                 Inherited::handleEventsFor( child );

                 toolinfo.hwnd   = child->handle();  // sends WM_NOTIFY to this window
                 toolinfo.uId    = (UINT)child->handle().asUnsigned();
                 flytext->sendEvent( TTM_ADDTOOL,
                                     0,
                                     IEventData( &toolinfo) );
                 IWindow::ChildCursor aggregateCursor( *child );
                 forCursor (aggregateCursor)
                    {
                    setHelpText( child->childAt( aggregateCursor ),
                                 IResourceId( child->id() + flyTextStringTableOffset() ),
                                 IResourceId( child->id() + longStringTableOffset() ));
                    }
                 }
              }
           }
        else if (windowIsAggregate)
           {
           setHelpText( window->childAt( cursor ),
                        IResourceId( window->id() + flyTextStringTableOffset() ),
                        IResourceId( window->id() + longStringTableOffset() ));
           }
        }
     }

#else
  if (!fFlyOverData->fTimer.isStarted())
  {
    fFlyOverData->fTimer.setInterval( fFlyOverData->fInitialTimerDelay/2 );
    fFlyOverData->fTimer.start(
      new ITimerMemberFn<IFlyOverData>(*fFlyOverData, IFlyOverData::timerpop));

    fFlyOverData->fPrevPos = IWindow::pointerPosition();
  }
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::stopHandlingEventsFor                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverHelpHandler& IFlyOverHelpHandler::stopHandlingEventsFor( IWindow* window )
{
  IASSERTPARM(window != 0);
  IFlyOverHelpHandler::Inherited::stopHandlingEventsFor( window );
#ifdef IC_WIN
  IFlyText* flytext = this->flyTextControl();
  if ( flytext && flytext->isValid() )
     {
     TOOLINFO  toolinfo;
     toolinfo.cbSize = sizeof(TOOLINFO);
     toolinfo.hwnd   = window->handle();  // sends WM_NOTIFY to this window
     toolinfo.uId    = (UINT)window->handle().asUnsigned();
                                         // window for which help applies
     // other TOOLINFO fields not needed
     flytext->sendEvent( TTM_DELTOOL,
                         0,
                         IEventData( &toolinfo) );
     //
     // Perform a similar deregistration of child IWindows
     IWindow::ChildCursor cursor( *window );
     for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {
        IWindow* child = IWindow::windowWithHandle( window->childAt( cursor ));
        if (child)
           {
           // Send the IC_UM_FLY_PAINT message with lParam == 0 to check
           // if this handler is already installed on the child.  If so,
           // then remove it.
           if (child->sendEvent( IC_UM_FLY_PAINT,
                                 0,
                                 0 ).asUnsignedLong() )
              {
              if (!child->sendEvent( IC_UM_IS_AGGREGATE_CTRL ).asUnsignedLong())
                 {
                 this->stopHandlingEventsFor( child );
                 }
              else
                 {
                 Inherited::stopHandlingEventsFor( child );

                 toolinfo.hwnd   = child->handle();  // sends WM_NOTIFY to this window
                 toolinfo.uId    = (UINT)child->handle().asUnsigned();
                 flytext->sendEvent( TTM_DELTOOL,
                                     0,
                                     IEventData( &toolinfo) );
                 IWindow::ChildCursor aggregateCursor( *child );
                 forCursor (aggregateCursor)
                    {
                    removeHelpText( child->childAt( aggregateCursor ));
                    }
                 }
              }
           }
        }
     }
#else
  fFlyOverData->fTimer.stop();
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::dispatchHandlerEvent                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFlyOverHelpHandler::dispatchHandlerEvent( IEvent& event )
{
#ifdef IC_WIN
  ITRACE_ALL(IString("FlyHelpHandler: HWND=") + IString(event.handle().asUnsigned()).d2x()+
             IString(", Msg=") + IString(event.eventId()).d2x()+
             IString(", MP1=") + IString((unsigned long)event.parameter1()).d2x()+
             IString(", MP2=") + IString((unsigned long)event.parameter2()).d2x());

  switch (event.eventId())
     {
     case WM_MOUSEMOVE       :
     case WM_LBUTTONDOWN     :
     case WM_LBUTTONUP       :
     case WM_LBUTTONDBLCLK   :
     case WM_RBUTTONDOWN     :
     case WM_RBUTTONUP       :
     case WM_RBUTTONDBLCLK   :
     case WM_MBUTTONDOWN     :
     case WM_MBUTTONUP       :
     case WM_MBUTTONDBLCLK   :
        {
        // relay mouse messages to the tooltip control
        IFlyText* flytext = this->flyTextControl();
        if ( flytext && flytext->isValid() )
           {
           MSG msg;
           msg.hwnd    = event.handle();
           msg.message = (UINT)event.eventId();
           msg.wParam  = event.parameter1();
           msg.lParam  = event.parameter2();
           flytext->sendEvent( TTM_RELAYEVENT,
                               0,
                               IEventData(&msg) );
           }
        }
        break;

     case WM_NOTIFY         :
        {
        NMHDR*  nmhdr = (NMHDR*)(event.parameter2().asUnsignedLong());
        fFlyOverData->notifyEvent( nmhdr );
        }  // case WM_NOTIFY

        break;

     case WM_PARENTNOTIFY   :
        // Detect the creation of child windows.   Deletion of child
        // windows is handled by stopHandlingEvents.
        {
        if ( event.parameter1().lowNumber() == WM_CREATE )
           {
           // We use the user message IC_UM_FLY_PAINT to notify ourselves
           // of the creation of a child window.  We need to post it
           // to ourselves so that IWindow will have done its creation
           // logic when we process it.  WM_PARENTNOTIFY is sent
           // by CreateWindowEx, and therefore windowWithHandle will
           // always return 0.
           event.handle().postEvent(
               IC_UM_FLY_PAINT,
               IEventData( event.parameter1().highNumber() ),
               event.parameter2() );
           event.setResult(0);   // required by Win32
           return true;          // stop handlers
           }
        }  // case WM_PARENTNOTIFY

        break;

     case IC_UM_FLY_PAINT   :
        // The message name is a misnomer in Windows.   We don't need
        // this message for painting in Windows, so it is reused.
        // A more appropriate name is IC_UM_FLY_HANDLER
        //
        // IC_UM_FLY_PAINT is used to arrange for this handler to get
        // attached to all IWindow children of a window with the
        // handler attached.  This is needed so that we can route
        // mouse events to the tooltips control.   There are 2 flavors:
        // Query and Set
        //  Query:   lParam == 0.  True is returned by this handler.
        //           in the message result.  This is used to detect the
        //           existance of this handler (or a derived one) on a
        //           control.  Performance optimization to avoid >1 handler
        //           per control doing the same thing.
        //  Set:     lParam == HWND of a control to start handling
        //           events for.   If the control is an IWindow,
        //           install this handler for the control.
        //
        // We use the user message IC_UM_FLY_PAINT to notify ourselves
        // of the creation of a child window.  We need to post it
        // to ourselves so that IWindow will have done its creation
        // logic when we process it.  WM_PARENTNOTIFY is sent
        // by CreateWindowEx, and therefore windowWithHandle will
        // always return 0.
        {
        IWindowHandle childHwnd = (HWND)event.parameter2().asUnsignedLong();
        if (childHwnd != 0)
           {
           // Set.  Add the handler.
           IWindow*      child = IWindow::windowWithHandle(childHwnd);
           if (child)
              {
              // Send the IC_UM_FLY_PAINT message with lParam == 0 to check
              // if this handler is already installed on the child.  If so,
              // don't install it again.  We need this check here since
              // WM_PARENTNOTIFY goes to all ancestor windows.
              if (!child->sendEvent( IC_UM_FLY_PAINT,
                                     0,
                                     0 ).asUnsignedLong() )
                 {
                 ITRACE_DEVELOP( IString("FlyHelpHandler -  new child hwnd=") +
                                 IString(childHwnd.asUnsigned()).d2x() );
                 this->handleEventsFor( child );
                 event.setResult(1);
                 }      // handler not installed
              }         // is IWindow
           }    // is set
        else
           {
           // Query.  Just set result to indicate we are here.
           event.setResult(1);
           }    // is query
        }  // case IC_UM_FLY_PAINT

        return true;       // done ... stop handlers

     default:
        break;
     }
#else
  if (event.eventId() == WM_CONTROLPOINTER || event.eventId() == WM_MOUSEMOVE)
  {
    if (!fFlyOverData->fTimer.isStarted() &&
        IISCHILD( event.handle(),
                  IQUERYACTIVEWINDOW( IWindow::desktopWindow()->handle() )))
    {
      // Kick the timer off.
      // Note: the interval is divided by two becuase it takes to timer pops
      //       in order to show fly over help for a window.
      fFlyOverData->fTimer.setInterval( fFlyOverData->fInitialTimerDelay/2 );
      fFlyOverData->fTimer.start(
        new ITimerMemberFn<IFlyOverData>(*fFlyOverData, IFlyOverData::timerpop));
      fFlyOverData->fPrevPos = IWindow::pointerPosition();
    }
    if (event.eventId() == WM_CONTROLPOINTER)
    {
      IControlEvent controlEvent(event);
      fFlyOverData->fHandle = controlEvent.controlHandle();
      if (!fFlyOverData->fHandle)
        fFlyOverData->fHandle = event.handle();
    }
    else
      fFlyOverData->fHandle = event.handle();
  }
  // If the window the handler is attached to is moving,
  // stop the timer and hide the flytext control.
  else if (event.eventId() == WM_ADJUSTWINDOWPOS)
  {
    PSWP pswp((PSWP)event.parameter1().asUnsignedLong());
    if (pswp->fl & SWP_MOVE)
    {
      if ( fFlyOverData->fFlytxt )
      {
         fFlyOverData->fFlytxt->hide();
      }
      fFlyOverData->fPrevHandle = 0;
    }
  }
#endif
  return false;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IFlyOverData::notifyEvent                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IFlyOverData::notifyEvent( tagNMHDR*            notifyHeader )
{
   IMODTRACE_ALL("IFlyOverData::notifyEvent");
   switch (notifyHeader->code)
      {
      case TTN_NEEDTEXT :
         {
         // The control is asking for text
         TOOLTIPTEXT* textHdr = (TOOLTIPTEXT*)notifyHeader;

         // insure fields are initialized.
         textHdr->hinst = 0;
         textHdr->lpszText = textHdr->szText;

         if (textHdr->uFlags & TTF_IDISHWND )
            fControlHandle = (HWND)textHdr->hdr.idFrom;
         else
            fControlHandle = 0;   // should not happen

         // Code below is very similar to that in timerpop.

         // Look through the handle string set to see if we have
         // a hit based on the window handle.

         if ( fgHandleStringSet.containsElementWithKey(
                                   fControlHandle.asUnsigned() ))
           {
           // We go a hit in the handle set.
           IHandleStringElement*
                    mouseOverElement(fgHandleStringSet.elementWithKey(
                                               fControlHandle.asUnsigned()));

           if ( mouseOverElement->fStrings )
             {
             if (fFlytxt)
               setFlyText(mouseOverElement->flyStr, textHdr->szText);
             if (fLongtxt)
               setLongText(mouseOverElement->longStr);
             }
           else
             {
             unsigned long id(mouseOverElement->flyResId);
             if (id && fFlytxt)
               setFlyText(fFly->resourceLibrary().tryToLoadString(id),
                          textHdr->szText);
             id = mouseOverElement->longResId;
             if (id && fLongtxt)
               setLongText(fFly->resourceLibrary().tryToLoadString(id));
             }
           }
         else
           {
           unsigned long controlId = IIDOF(fControlHandle) ;

           if ( fFlytxt )
             {
             IString flyInfo = IString();
             flyInfo = fFly->resourceLibrary().tryToLoadString(controlId + fFlyOffset);
             if (flyInfo.length() == 0)
               setFlyText(fFly->defaultText(), textHdr->szText );
             else
               setFlyText(flyInfo, textHdr->szText );
             }

           if ( fLongtxt )
             {
             IString longInfo = IString();
             longInfo = fFly->resourceLibrary().tryToLoadString(controlId + fLongOffset);
             if (longInfo.length() == 0)
               setLongText(fFly->defaultText());
             else
               setLongText(longInfo);
             }
           }
         break;
         } //case TTN_NEEDTEXT
      case TTN_POP :
         {
         // Fly help is being hidden
         if (fLongtxt)
           setLongText( fDefault );
         }
         break;

      default:
         break;
      }  // switch
   return;
}
#endif

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IFlyOverData::timerpop                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IFlyOverData::timerpop( unsigned long timerId )
{
  // Get the mouse location.

  IPoint mousePoint(IWindow::pointerPosition());
  POINTL pt(mousePoint.asPOINTL());

  // If either of the controls is no longer valid, return.

  if (fFlytxt && !fFlytxt->isValid())
    return;
  if (fLongtxt && !fLongtxt->isValid())
    return;

  // Get the handle of the control that is under the mouse.
  // Note:  controls that are disabled will not be returned since
  //        they do not respond to hit-test.
  fControlHandle = IWINDOWFROMPOINT( IWindow::desktopWindow()->handle(), &pt, true );

  // If the control that is under the mouse is a descendant of the
  // handle that was in the control pointer or mousemove event or
  // the mouse is over the fly text control (this prevents the fly text
  // control from being removed once displayed for a window) and
  // the control is a child of the active window then proceed to show
  // fly over help.

  Boolean overFlyTextWindow(fFlytxt && (fFlytxt->handle() == fControlHandle));

  if (( IISCHILD( fControlHandle, fHandle ) &&
        IISCHILD( fControlHandle, IQUERYACTIVEWINDOW( IWindow::desktopWindow()->handle() ))) ||
        overFlyTextWindow )
  {
    if (overFlyTextWindow)
      return;
    if (pt.x == fPrevPos.x() && pt.y == fPrevPos.y())  // If the mouse has been
                                                       // at the same location.
    {
      fControlWindow = IWindow::windowWithHandle( fControlHandle );

      // If the controlWindow is null, then we are dealing with
      // an aggregate control that is not wrappered.

      if (!fControlWindow)
      {
        // Query the handle's parent and then try to get the
        // IWindow pointer that corresponds to the parent handle.

        IWindowHandle parentHandle(IPARENTOF(fControlHandle));
        fControlWindow = IWindow::windowWithHandle( parentHandle );
      }

      if (fControlWindow && (fControlWindow != fFlytxt))
      {
        // Look through the handle string set to see if we have
        // a hit based on the window handle.

        if ( fgHandleStringSet.containsElementWithKey(
                                  fControlHandle.asUnsigned() ))
        {
          // We go a hit in the handle set.
          IHandleStringElement*
                   mouseOverElement(fgHandleStringSet.elementWithKey(
                                              fControlHandle.asUnsigned()));

          if ( mouseOverElement->fStrings )
          {
            if (fFlytxt)
              setFlyText(mouseOverElement->flyStr);
            if (fLongtxt)
              setLongText(mouseOverElement->longStr);
          }
          else
          {
            unsigned long id(mouseOverElement->flyResId);
            if (id && fFlytxt)
              setFlyText(fFly->resourceLibrary().tryToLoadString(id));
            id = mouseOverElement->longResId.id();
            if (id && fLongtxt)
              setLongText(fFly->resourceLibrary().tryToLoadString(id));
          }
        }
        else
        {
          unsigned long controlId(fControlWindow->id());

          if ( fFlytxt )
          {
            IString flyInfo = IString();
            flyInfo = fFly->resourceLibrary().tryToLoadString(controlId + fFlyOffset);
            if (flyInfo.length() == 0)
              setFlyText(fFly->defaultText());
            else
              setFlyText(flyInfo);
          }

          if ( fLongtxt )
          {
            IString longInfo = IString();
            longInfo = fFly->resourceLibrary().tryToLoadString(controlId + fLongOffset);
            if (longInfo.length() == 0)
              setLongText(fFly->defaultText());
            else
              setLongText(longInfo);
          }
        }
      }

      // Set the previous handle to the current handle for the next timerpop.
      fPrevHandle = fControlHandle;

      // set the interval to 1/3 second so we can tell "quickly" if we've
      // moved off of the window.
      fTimer.setInterval(333);
    }
    else // The mouse has moved.
    {
      // If we are no longer over the same control.
      if (fPrevHandle != fControlHandle)
      {
        // Reset previous handle to null and reset the interval.
        fPrevHandle = 0;
        fTimer.setInterval(fTimerDelay/2);
        if (fFlytxt)
          fFlytxt->hide();
        if (fLongtxt)
          setLongText( fDefault );
      }
    }
    // Update the previous point to the current point.
    fPrevPos.setX( pt.x ).setY( pt.y );
  }
  else
  {
    // Mouse is no longer over a window that is a child of the
    // window the flyOverHelpHandler what attached to.

    if (fFlytxt)
    {
      fFlytxt->hide();
    }
    if (fLongtxt)
    {
      setLongText( fDefault );
    }
    fPrevHandle = 0;
    fTimer.stop();
  }
}
#endif

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setFlyText                                              |
|                                                                              |
| Copys the specified helpText into the szText buffer.  Max length: 79 bytes   |
------------------------------------------------------------------------------*/
IFlyOverData& IFlyOverData::setFlyText( const IString& helpText,
                                        char*          szText  )
{
  unsigned lengthToCopy = helpText.length();
  if(lengthToCopy != 0)
    {
    if (lengthToCopy > 79)
       {
       // Adjust length back so it will fit, insuring that MBCS is preserved
       lengthToCopy = 80;
       while ( lengthToCopy > 1 &&
               helpText.charType( lengthToCopy ) > IStringEnum::mbcs1 )
          lengthToCopy--;
       lengthToCopy--;
       }
    if (lengthToCopy)
      memcpy( szText, (char*)helpText, lengthToCopy);
    }
  szText[lengthToCopy] = '\0';
  return *this;
}

#else   // !IC_WIN

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setFlyText                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverData& IFlyOverData::setFlyText( const IString& helpText )
{
  if(helpText.size()!=0)
  {
    if(fFlytxt->text()!=helpText)
    {
      fFlytxt->hide();
      fFlytxt->setText(IString::change(helpText, (char*)"\xA", (char*)""));
    }
    // If the previous handle is null, this means the mouse has moved
    // onto a new window so show the fly text control.
    if ( fPrevHandle == 0 )
    {
      positionTextWindow();
    }
  }
  else
  {
    if(fFlytxt->isVisible())
      fFlytxt->hide();
  }
  return *this;
}
#endif

/*------------------------------------------------------------------------------
| IFlyOverHelpHandler::setLongText                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFlyOverData& IFlyOverData::setLongText( const IString& helpText )
{
  Boolean preventLongTextUpdate(fLongtxt->sendEvent(IC_UM_QRY_PREVENTUPDATE));
  if (!preventLongTextUpdate)
  {
    if (helpText.size()!=0)
      fLongtxt->setText( helpText );
    else
      if ( fDefault.length() != 0 )
        fLongtxt->setText( fDefault );
      else
        // At a minimum, set the long text to a single blank to avoid
        // the frame handler from hiding the long text control if the
        // long text control happens to be a frame extension.
        fLongtxt->setText( " " );
  }
  return *this;
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IFlyOverData::positionTextWindow                                             |
|                                                                              |
| Reposition and show the flyText window if the mouse has moved over a         |
| new window since the last timer pop.                                         |
|                                                                              |
| Note:  The window handle of the window the mouse is over and the IWindow*    |
|        is passed into this function.  The window handle will not correspond  |
|        to the IWindow object in the case of aggregate controls (spin button, |
|        combo box, mle, etc).  The IWindow pointer is needed to query the     |
|        layoutAdjustment for such controls as combo box.                      |
------------------------------------------------------------------------------*/
IFlyOverData& IFlyOverData::positionTextWindow( )
{
  if (fPrevHandle != fControlHandle)
  {
    fFlytxt->hide();

    // Get the location and size of the window.

#ifdef IC_PM
    SWP swp;
    IQUERYWINDOWPOS(fControlHandle, &swp);
    ISize winsize(swp.cx, swp.cy);
#endif
#ifdef IC_WIN
    RECT rect;
    IQUERYWINDOWPOS(fControlHandle, &rect);
    ISize winsize(rect.left - rect.right, rect.bottom - rect.top);
#endif

    // Map the location of the window the mouse is over to desktop
    // coordinates.
    IPoint pt(0,0);
    if ( ICoordinateSystem::applicationOrientation() ==
         ICoordinateSystem::originUpperLeft )
        {
        pt.setY( fControlWindow->parentSize().height() - winsize.height() );
        }

    pt = IWindow::mapPoint(pt, fControlHandle,
                           IWindow::desktopWindow()->handle() );

    IRectangle controlRect( pt, winsize );

    // Get the control's visible rectangle.
    IRectangle visibleRect(fControlWindow->visibleRectangle());

    // if the visible rectangle does not equal the rectangle
    // from query the control (as in the case of a drop down
    // combo box, then convert the origin point of the
    // visible rectangle to desktop coordinates and position
    // the fly text relative to it.
    if (controlRect.size() == visibleRect.size())
      fFlytxt->setRelativeWindowRect( controlRect );
    else
    {
      pt = IWindow::mapPoint( visibleRect.minXMinY(),
                              fControlWindow->parent()->handle(),
                              IWindow::desktopWindow()->handle());
      fFlytxt->setRelativeWindowRect( IRectangle( pt, visibleRect.size() ));
    }
    fFlytxt->postEvent(IC_UM_FLY_PAINT);
  }
  return *this;
}
#endif

