/*******************************************************************************
* FILE NAME: ihandle2.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions for Bitmap      *
*   handles declared ihandle.hpp.                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINPOINTERS
  #define INCL_GPIBITMAPS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  }

#include <ihandle.hpp>

#include <icconst.h>
#include <iexcept.hpp>
#include <ireslib.hpp>
#include <istring.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHANDLE2_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
 Note that bitmap handles use references to reduce
  duplicate bitmaps for better storage utilization.
  See IReference, IRefCounted classes for more info.
------------------------------------------------------------------------------*/

ISystemBitmapHandle :: ISystemBitmapHandle(Identifier bitmapId)
         : IBitmapHandle(0)
{
  long lPMId(0);
  switch(bitmapId)
  {
    case  systemMenu:
          lPMId = SBMP_SYSMENU;
          break;
    case  systemMenuPressed:
          lPMId = SBMP_SYSMENUDEP;
          break;
    case  scrollBarUpArrow:
          lPMId = SBMP_SBUPARROW;
          break;
    case  scrollBarUpArrowPressed:
          lPMId = SBMP_SBUPARROWDEP;
          break;
    case  scrollBarUpArrowDisabled:
          lPMId = SBMP_SBUPARROWDIS;
          break;
    case  scrollBarDownArrow:
          lPMId = SBMP_SBDNARROW;
          break;
    case  scrollBarDownArrowPressed:
          lPMId = SBMP_SBDNARROWDEP;
          break;
    case  scrollBarDownArrowDisabled:
          lPMId = SBMP_SBDNARROWDIS;
          break;
    case  scrollBarRightArrow:
          lPMId = SBMP_SBRGARROW;
          break;
    case  scrollBarRightArrowPressed:
          lPMId = SBMP_SBRGARROWDEP;
          break;
    case  scrollBarRightArrowDisabled:
          lPMId = SBMP_SBRGARROWDIS;
          break;
    case  scrollBarLeftArrow:
          lPMId = SBMP_SBLFARROW;
          break;
    case  scrollBarLeftArrowPressed:
          lPMId = SBMP_SBLFARROWDEP;
          break;
    case  scrollBarLeftArrowDisabled:
          lPMId = SBMP_SBLFARROWDIS;
          break;
    case  menuCheckMark:
          lPMId = SBMP_MENUCHECK;
          break;
    case  menuAttached:               /* cascading menu mark */
          lPMId = SBMP_MENUATTACHED;
          break;
    case  checkBoxCheck:
          lPMId = SBMP_CHECKBOXES;
          break;
    case  comboBoxDownArrow:
          lPMId = SBMP_COMBODOWN;
          break;
    case  pushButtonCorners:
          lPMId = SBMP_BTNCORNERS;
          break;
    case  minimizeButton:
          lPMId = SBMP_MINBUTTON;
          break;
    case  minimizeButtonPressed:
          lPMId = SBMP_MINBUTTONDEP;
          break;
    case  maximizeButton:
          lPMId = SBMP_MAXBUTTON;
          break;
    case  maximizeButtonPressed:
          lPMId = SBMP_MAXBUTTONDEP;
          break;
    case  restoreButton:
          lPMId = SBMP_RESTOREBUTTON;
          break;
    case  restoreButtonPressed:
          lPMId = SBMP_RESTOREBUTTONDEP;
          break;
    case  childSystemMenu:
#ifdef IC_WIN
          lPMId = SBMP_SYSMENU;
#else
          lPMId = SBMP_CHILDSYSMENU;
#endif
          break;
    case  childSystemMenuPressed:
#ifdef IC_WIN
          lPMId = SBMP_SYSMENUDEP;
#else
          lPMId = SBMP_CHILDSYSMENUDEP;
#endif
          break;
#ifndef IC_WIN
    case  drive:
          lPMId = SBMP_DRIVE;
          break;
    case  file:
          lPMId = SBMP_FILE;
          break;
    case  folder:
          lPMId = SBMP_FOLDER;
          break;
    case  collapsedTree:
          lPMId = SBMP_TREEPLUS;
          break;
    case  expandedTree:
          lPMId = SBMP_TREEMINUS;
          break;
    case  program:
          lPMId = SBMP_PROGRAM;
          break;
#endif  // !IC_WIN
    case  sizeBox:
          lPMId = SBMP_SIZEBOX;
          break;
  } /* end switch */
  if (lPMId)
  {
    handle = (IHandle::Value)IResourceLibrary::systemBitmap(lPMId);
    IResourceLibrary::addRef(IBitmapHandle(*this));
  }
  else
  {
#ifdef IC_WIN
    Boolean       useIcon(false);
    Boolean       useShell32(false);
#endif
    switch(bitmapId)
      {
      case  smallSystemMenu:
        // Load from IOC resource lib
        lPMId = IC_SMALLSYSTEMMENU;
        break;
#ifdef IC_WIN
      case  drive:
        lPMId =  IC_DRIVE_SHELL32;
        useShell32 = true;
        useIcon = true;
        break;
      case  file:
        lPMId = IC_FILE_SHELL32;
        useShell32 = true;
        useIcon = true;
        break;
      case  folder:
        lPMId = IC_FOLDER_SHELL32;
        useShell32 = true;
        useIcon = true;
        break;
      case  collapsedTree:
        lPMId = IC_COLLAPSEICON;
        useIcon = true;
        break;
      case  expandedTree:
        lPMId = IC_EXPANDICON;
        useIcon = true;
        break;
      case  program:
        lPMId = IC_PROGRAM_SHELL32;
        useShell32 = true;
        useIcon = true;
        break;
#endif
      } // switch


#ifdef IC_WIN
    if (useIcon)
       {
       IPointerHandle hicon(0);
       try
          {
          if (useShell32)
             hicon =
                IDynamicLinkLibrary( IC_SHELL32_DLL ).tryToLoadIcon( lPMId );
          else
             hicon =
                IApplication::current().resourceLibrary().tryToLoadIcon( lPMId );
          }
       catch (...)
          {
          // this is to catch nonexistant DLL errors
          }
       if (!hicon)
          {
          // could not load ... use a default
          hicon =
            ISystemPointerHandle( ISystemPointerHandle::standardApplication) ;
          }
       // extract the bitmap portion of the icon
       ICONINFO    pointerInfo;
       IQUERYPOINTERINFO( hicon, &pointerInfo );
       handle = pointerInfo.hbmColor;
       DeleteObject( pointerInfo.hbmMask );     // throw away mask part
       // Add a reference since this one must be deleted.
       IResourceLibrary::addRef(IBitmapHandle(*this));
       }
    else
#endif
       {
       handle = IApplication::current().resourceLibrary().loadBitmap( lPMId );
       }
  } /* endif */
}

ISystemBitmapHandle :: ~ISystemBitmapHandle( ) {}


IBitmapHandle::IBitmapHandle( Value value)
  : IHandle( value )
{
   if(value!=0)
     IResourceLibrary::addRef(*this);
}

IBitmapHandle::IBitmapHandle(const IBitmapHandle& bitmapHandle)
  : IHandle( bitmapHandle )
{
   if(bitmapHandle!=0)
     IResourceLibrary::addRef(bitmapHandle);
}

IBitmapHandle::~IBitmapHandle()
{
   if(handle!=0)
     IResourceLibrary::removeRef(*this);
}

IBitmapHandle& IBitmapHandle:: operator =(const IBitmapHandle& bitmapHandle)
{
   /* Note: Ensure self assignment works */
     if(this != &bitmapHandle)
     {
        if(bitmapHandle!=0)
            IResourceLibrary::addRef(bitmapHandle);
        if(*this!=0)
          IResourceLibrary::removeRef(*this);     // drop ref to current value
        this->handle = bitmapHandle;             // set self to new value
     }
     return *this;
}
