/*******************************************************************************
* FILE NAME: imle.cpp                                                          *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imle.hpp.                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define IUSING_OLE
#ifdef WIN32_LEAN_AND_MEAN
  #undef WIN32_LEAN_AND_MEAN
#endif
#ifdef _OLE2_H_
  #undef _OLE2_H_
#endif

extern "C" {
  #define INCL_WINMLE
  #define INCL_WINCLIPBOARD
  #define INCL_WININPUT             // WM_CHAR, etc.
  #define INCL_WINMESSAGEMGR        // for WNDPARAM struct
  #define INCL_WINWINDOWMGR         // WinQueryWindowULong
  #define INCL_GPIPRIMITIVES
  #define INCL_WINSYS
  #include <iwindefs.h>

  #include <sys\stat.h>
  #include <stdio.h>              // include standard IO routines
#ifdef IC_WIN
  #include <io.h>                 // include windows IO routines
  #include <fcntl.h>              // include windows IO routines
#endif
  #include <string.h>             // include string handling routines
  #include <stdlib.h>             // include memory handling routines
}

#include <imle.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <icolor.hpp>
#include <ifont.hpp>
#include <itrace.hpp>
#include <imphdr.hpp>
#include <iplatfrm.hpp>

#ifdef IC_WIN
#ifndef _IMCEVT_
  #include <imcevt.hpp>
#endif

#ifndef _ICOORDSY_
  #include <icoordsy.hpp>
#endif
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMLE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public multiple-line entry field control styles                              |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IMultiLineEdit::Style
#ifndef IC_WIN
  IMultiLineEdit::wordWrap          = MLS_WORDWRAP,
#endif
#ifdef IC_WIN
  IMultiLineEdit::wordWrap          ( 0, IMLS_WORDWRAP ),
#endif
  IMultiLineEdit::border3D          ( 0, IWS_BORDER3D ),
  IMultiLineEdit::ignoreTab         = MLS_IGNORETAB,
  IMultiLineEdit::border            = MLS_BORDER,
  IMultiLineEdit::verticalScroll    = MLS_VSCROLL,
  IMultiLineEdit::horizontalScroll  = MLS_HSCROLL,
  IMultiLineEdit::readOnly          = MLS_READONLY,
  IMultiLineEdit::classDefaultStyle ( WS_VISIBLE  |
                                      MLS_BORDER  |
                                      MLS_VSCROLL
#ifndef IC_WIN
                                      | MLS_WORDWRAP );
#endif
#ifdef IC_WIN
                                      , IMLS_WORDWRAP | IWS_BORDER3D );
#endif

const long IMultiLineEdit::deselect = -1;
const long IMultiLineEdit::end = -2;
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IMultiLineEdit::Style
  IMultiLineEdit::currentDefaultStyle ( WS_VISIBLE  |
                                        MLS_BORDER  |
                                        MLS_VSCROLL
#ifndef IC_WIN
                                      | MLS_WORDWRAP );
#endif
#ifdef IC_WIN
                                      , IMLS_WORDWRAP | IWS_BORDER3D );
#endif

#pragma data_seg()

#ifdef IC_WIN
// Workaround for NT bug with not returning WS_BORDER style
#define IMLS_BORDERNTHACK           0x00000008ul
#endif

#ifdef IC_WIN
class IMultiLineEditData : public IBase {
/*------------------------------------------------------------------------------
| The IMultiLineEditData class encapsulates private data and functions         |
| used by the IMultiLineEdit class.  An object of this class is created in the |
| IMultiLineEdit constructors.  This is the os2/windows version.               |
------------------------------------------------------------------------------*/
public:
  IMultiLineEditData ( );

virtual
 ~IMultiLineEditData ( );

HBRUSH
  bkgrndbrush;

IBase::Boolean
  pointInSelectedText(const IEvent& evt);

IBase::Boolean
  bCheckingRightDragStart;
IPoint
  pointRButtonDown;
}; // IMultiLineEditData
/*------------------------------------------------------------------------------
| IMultiLineEditData::IMultiLineEditData                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IMultiLineEditData::IMultiLineEditData ( )
  : bkgrndbrush( 0 )
   , bCheckingRightDragStart(false)
   , pointRButtonDown()
{ }

/*------------------------------------------------------------------------------
| IMultiLineEditData::~IMultiLineEditData                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMultiLineEditData::~IMultiLineEditData ( )
{
  if ( bkgrndbrush != 0 )
    DeleteObject(bkgrndbrush);
}

class IMultiLineHandler : public IHandler {
typedef IHandler
  Inherited;
public:

/*------------------------------- Constructors -------------------------------*/
  IMultiLineHandler ( );

virtual
  ~IMultiLineHandler ( );

/*----------------------------- Event Processing -----------------------------*/
virtual Boolean
  dispatchHandlerEvent ( IEvent& event );

}; // class IMultiLineHandler

/*------------------------------------------------------------------------------
| IMultiLineHandler::IMultiLineHandler                                                 |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IMultiLineHandler::IMultiLineHandler()
{ }

/*------------------------------------------------------------------------------
| IMultiLineHandler::~IMultiLineHandler                                                |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IMultiLineHandler::~IMultiLineHandler()
{ }

/*------------------------------------------------------------------------------
| IMultiLineHandler::dispatchHandlerEvent                                          |
|                                                                              |
| Intercepts control color event                                               |
|    This handler returns false for all other events.                          |
------------------------------------------------------------------------------*/
IBase::Boolean IMultiLineHandler::dispatchHandlerEvent( IEvent& event )
{
  Boolean
    result = false;

  switch( event.eventId() )
  {

    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORSTATIC:
         {
         IWindow* ctlwin =
           IWindow::windowWithHandle(IWindowHandle(event.parameter2()) );
         IColor clrFgnd =
           ((IMultiLineEdit *) ctlwin)->foregroundColor();
         IColor clrBgnd =
           ((IMultiLineEdit *) ctlwin)->backgroundColor();

         // Set foreground and background text colors in device context
         SetBkColor( (HDC)event.parameter1(), clrBgnd.asRGBLong() );
         SetTextColor( (HDC)event.parameter1(), clrFgnd.asRGBLong() );

         // Return HBRUSH for painting controls background
         if ( ((IMultiLineEdit *) ctlwin)->fMultiLineEditData->bkgrndbrush != 0 )
           DeleteObject( ((IMultiLineEdit *) ctlwin)->fMultiLineEditData->bkgrndbrush);
         ((IMultiLineEdit *) ctlwin)->fMultiLineEditData->bkgrndbrush =
           CreateSolidBrush (NEARESTPALETTECOLOR(clrBgnd.asRGBLong()));
         event.setResult( ((IMultiLineEdit *) ctlwin)->fMultiLineEditData->bkgrndbrush );

         result = true;
         }
      break;

    default:
      break;
  }
  return result;
}

/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointer to the object |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator IMultiLineHandler* ();
IMultiLineHandler
 *operator -> ()
    {
    return *this;
    }
 ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
IMultiLineHandler
 *ptr;
} defaultHandler;

DefaultHandler::operator IMultiLineHandler* ( )
  {
  if ( !ptr )
    ptr = new IMultiLineHandler;
  return ptr;
  }
#endif


/*------------------------------------------------------------------------------
| IMultiLineEdit::IMultiLineEdit                                               |
|                                                                              |
| Constructor to create a MLE control on a standard window.                    |
------------------------------------------------------------------------------*/
IMultiLineEdit::IMultiLineEdit( unsigned long ulId,
                                IWindow* pwndParent,
                                IWindow* pwndOwner,
                                const IRectangle& rectInit,
                                const Style& style )
                                 : bRefresh ( true )
#ifdef IC_WIN
                                 , fMultiLineEditData( new IMultiLineEditData )
                                 , fcharLimit( 30000 )
                                 , flimitSet( false )
#else
                                 , fMultiLineEditData ( 0 )
#endif
{
  // assertions on input parms
  IASSERTPARM(pwndParent!=0);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  /****************************************************************/
  /* Borland has bug with embedded ? operator so below is needed. */
  /****************************************************************/
#ifdef __BORLANDC__
  IWindowHandle owner;
  if (pwndOwner == 0)
    owner = IWindowHandle(0);
  else
    owner = pwndOwner->handle();
#else
  IWindowHandle owner = (pwndOwner == 0) ? IWindowHandle(0) : pwndOwner->handle();
#endif

  IWindowHandle whMLE(
      this -> create( ulId,
                      0,
                      convertToGUIStyle( style ),
                      WC_MLE,
                      pwndParent->handle(),
                      owner,
                      rectInit,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( style, true ) ) );

  startHandlingEventsFor(whMLE);
#ifndef IC_WIN
  setEditRegion();
#endif
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::IMultiLineEdit                                               |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
IMultiLineEdit::IMultiLineEdit( unsigned long ulId,
                                IWindow* pwndParent )
                                 : bRefresh ( true )
#ifdef IC_WIN
                                 , fMultiLineEditData( new IMultiLineEditData )
                                 , fcharLimit( 30000 )
                                 , flimitSet( false )
#else
                                 , fMultiLineEditData ( 0 )
#endif
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pwndParent);
#ifndef IC_WIN
  setEditRegion();
#endif
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

  // Get the system style and look for extended styles.  Set if found.
  unsigned long sysStyle = style();

  // Extended style should always be zero
  unsigned long extStyle = extendedStyle();
#ifdef IC_WIN
  // Since we can't query the WS_BORDER style due to a NT bug, assume the
  // wrappered control always has a border (will be correct in most cases)
  extStyle |= IMLS_BORDERNTHACK;
#endif

  // Check for the 3D border style - will be WS_EX_CLIENTEDGE in NT and zero
  // otherwise
  if ((sysStyle & IWS_BORDER3D) || (sysStyle & MLS_BORDER))
     extStyle |= IWS_BORDER3D;

#ifdef IC_WIN
  // If auto horizontal scroll is not set, then we want wordwrap in NT
  if (!(sysStyle & ES_AUTOHSCROLL))
     extStyle |= IMLS_WORDWRAP;
#endif

  // Set the extended style
  setExtendedStyle (extStyle);
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::IMultiLineEdit                                               |
|                                                                              |
| Constructor to instantiate from an existing MLE control.                     |
------------------------------------------------------------------------------*/
IMultiLineEdit::IMultiLineEdit ( const IWindowHandle& wh )
                                 : bRefresh ( true )
#ifdef IC_WIN
                                 , fMultiLineEditData( new IMultiLineEditData )
                                 , fcharLimit( 30000 )
                                 , flimitSet( false )
#else
                                 , fMultiLineEditData ( 0 )
#endif
{
  setAutoDestroyWindow(false);
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  startHandlingEventsFor(wh);
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

  // Get the system style and look for extended styles.  Set if found.
  unsigned long sysStyle = style();

  // Extended style should always be zero
  unsigned long extStyle = extendedStyle();
#ifdef IC_WIN
  // Since we can't query the WS_BORDER style due to a NT bug, assume the
  // wrappered control always has a border (will be correct in most cases)
  extStyle |= IMLS_BORDERNTHACK;
#endif

  // Check for the 3D border style - will be WS_EX_CLIENTEDGE in NT and zero
  // otherwise
  if ((sysStyle & IWS_BORDER3D) || (sysStyle & MLS_BORDER))
     extStyle |= IWS_BORDER3D;

#ifdef IC_WIN
  // If auto horizontal scroll is not set, then we want wordwrap in NT
  if (!(sysStyle & ES_AUTOHSCROLL))
     extStyle |= IMLS_WORDWRAP;
#endif

  // Set the extended style
  setExtendedStyle (extStyle);
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::~IMultiLineEdit                                              |
|                                                                              |
| Remove the default mouse pointer handler                                     |
------------------------------------------------------------------------------*/
IMultiLineEdit::~IMultiLineEdit()
{
  IMousePointerHandler::defaultHandler()->stopHandlingEventsFor(this);

#ifdef IC_WIN
  removeHandler(defaultHandler);
  if ( fMultiLineEditData != 0 )
    delete fMultiLineEditData;
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::Style  IMultiLineEdit::defaultStyle                          |
|                                                                              |
| Return the default style for new entry field objects.                        |
------------------------------------------------------------------------------*/
IMultiLineEdit::Style  IMultiLineEdit::defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::setDefaultStyle                                              |
|                                                                              |
| Replace the default style for new entry field objects.                       |
------------------------------------------------------------------------------*/
void  IMultiLineEdit::setDefaultStyle(const Style& mlsStyle)
{
  currentDefaultStyle = mlsStyle;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::convertToGUIStyle                                            |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                  Boolean bExtOnly) const
{
  // Obtain the style from the class (ITextControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & ( IS_EXTMASK | ITC_EXTGUIMASK);
#ifdef IC_WIN
    if ( !(guiStyle.asUnsignedLong() & IMultiLineEdit::border.asUnsignedLong()))
    {
      // if border is not specified then remove border3D
      ulStyle &= ~IWS_BORDER3D;
    }
    else
    // Hack around bug in Windows NT 3.51 which does not allow the WS_BORDER
    // style to be queried back after use on create - Use our own style for now
      ((IMultiLineEdit*)this)->setExtendedStyle( extendedStyle() |
         IMLS_BORDERNTHACK );
#endif

  }
  else
  {
    // Let the clean MLS_ styles, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IMLS_MASK;

#ifdef IC_WIN
    // For Windows it is always a child and always has multiline style
    ulStyle |= WS_CHILD | ES_MULTILINE | ES_AUTOVSCROLL;

    // If extended style of IMLS_WORDWRAP is NOT set, then add auto hscroll
    if (!( guiStyle.asExtendedUnsignedLong() & IMLS_WORDWRAP))
      ulStyle |= ES_AUTOHSCROLL;
#endif
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setTextChangedFlag                                           |
|                                                                              |
| Set a flag to indicate that the MLE contents have changed.                   |
------------------------------------------------------------------------------*/
IMultiLineEdit&   IMultiLineEdit::setTextChangedFlag( Boolean changed )
{
   handle().sendEvent(MLM_SETCHANGED, IEventParameter1(changed),
                                      IEventParameter2(0));
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::resetTextChangedFlag                                         |
|                                                                              |
| Reset the flag that indicates that the MLE contents have changed.            |
------------------------------------------------------------------------------*/
IMultiLineEdit&   IMultiLineEdit::resetTextChangedFlag( )
{
   setTextChangedFlag( false );
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::hasTextChanged                                               |
|                                                                              |
| Queries whether the MLE text has changed since the last time the changed     |
| flag was cleared.                                                            |
------------------------------------------------------------------------------*/
Boolean  IMultiLineEdit::hasTextChanged( ) const
{
  IEventResult evt(handle().sendEvent(MLM_QUERYCHANGED,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));

  return ( evt.asUnsignedLong() != 0 );
}



/*------------------------------------------------------------------------------
| IMultiLineEdit::disableDataUpdate                                            |
|                                                                              |
| Set the MLE to read-only mode, preventing updates to the text.               |
------------------------------------------------------------------------------*/
IMultiLineEdit&   IMultiLineEdit::disableDataUpdate()
{
   enableDataUpdate(false);
   return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::enableDataUpdate                                             |
|                                                                              |
| Sets the MLE to read-write mode which allows the user to update the MLE text.|
------------------------------------------------------------------------------*/
IMultiLineEdit&  IMultiLineEdit::enableDataUpdate ( Boolean update )
{
  Boolean writeable = this->isWriteable();

  /***************************/
  /* Only set if we need to. */
  /***************************/
  if ((update && !writeable) || (!update && writeable))
  {
    handle().sendEvent(MLM_SETREADONLY, IEventParameter1(!update),
                                        IEventParameter2(0));
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::isWriteable                                                  |
------------------------------------------------------------------------------*/
Boolean IMultiLineEdit::isWriteable() const
{
#ifdef IC_WIN
  if ( style() & readOnly.asUnsignedLong() )
#else
  IEventResult evt(handle().sendEvent(MLM_QUERYREADONLY,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
  if (evt.asUnsignedLong())
#endif
     return false;
  else
     return true;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::disableUpdate                                                |
|                                                                              |
| Disable screen updates.                                                      |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::disableUpdate()
{
  enableUpdate( false );
  return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::enableUpdate                                                 |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::enableUpdate( Boolean update )
{
#ifdef IC_WIN
  // No special code needed on Windows, just use the IWindow version
  Inherited::enableUpdate( update );
#endif
#ifdef IC_PM
  if ( update )
  {
     IEventResult evt(handle().sendEvent(MLM_ENABLEREFRESH,
                                         IEventParameter1(0),
                                         IEventParameter2(0)));
     if (evt.asUnsignedLong())
        bRefresh=true;
  } else {
     IEventResult evt(handle().sendEvent(MLM_DISABLEREFRESH,
                                         IEventParameter1(0),
                                         IEventParameter2(0)));
     if (evt.asUnsignedLong())
        bRefresh=false;
  } /* endif */
#endif
  return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::setText                                                      |
|                                                                              |
| Changes the control text given a string.                                     |
------------------------------------------------------------------------------*/
IMultiLineEdit&  IMultiLineEdit::setText(const char* pszText)
{
  setText((char*)pszText,0);
  return *this;

}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setText                                                      |
|                                                                              |
| Changes the control text using the string table.                             |
------------------------------------------------------------------------------*/
IMultiLineEdit&  IMultiLineEdit::setText(const IResourceId& residText)
{
  IString str(residText.resourceLibrary().loadString(residText));
  setText((char*)str,0);
  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setText                                                      |
|                                                                              |
| Changes the control text to the text passed in the buffer.                   |
------------------------------------------------------------------------------*/
IMultiLineEdit&  IMultiLineEdit::setText( const char*   pszBuffer,
                                          unsigned long ulBufferSize )
{
   /*******************************************************************/
   /* If no buffer size passed, then use the entire character buffer  */
   /*******************************************************************/
   if (pszBuffer && !ulBufferSize)
     ulBufferSize = strlen(pszBuffer);

#ifdef IC_WIN
   if (ulBufferSize <= limit())
   {
      /*******************************************************************/
      /* Create a temporary buffer, so that we can take into account the */
      /*  buffer size (in case a truncation is necessary)                */
      /*******************************************************************/
      IString textString( pszBuffer, ulBufferSize );

      /**************************************************************/
      /* No import support in Windows so just set the text          */
      /**************************************************************/
      Inherited::setText( textString );
   }
   else
   {
      ITHROWLIBRARYERROR( IC_INVALIDSIZE_LIMIT,
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable );
   } /* endif */
#endif
#ifdef IC_PM
   removeAll();                   // First clear out the MLE text

   /**************************************************************/
   /* Handle setting the MLE text to NULL                        */
   /**************************************************************/
   if ( !pszBuffer )
     return *this;

   /**************************************************************/
   /* Specify the origin of the text to be imported from the     */
   /* buffer as the current cursor position.                     */
   /**************************************************************/
   long ipt = -1;

   /**************************************************************/
   /* Specify the buffer for MLE object import operation.        */
   /**************************************************************/
   import(&ipt,ulBufferSize,pszBuffer);   // Import text from buffer
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::text                                                         |
|                                                                              |
| Returns the current contents of the MLE                                      |
------------------------------------------------------------------------------*/
IString IMultiLineEdit::text() const
{
#ifdef IC_WIN
   /**************************************************************/
   /* No export support in Windows so just get the text string   */
   /**************************************************************/
   return Inherited::text();
#else
   unsigned long numChars;             // Number of characters exported
   long ipt(0);

   unsigned long ulMyBufSize(textLengthAfterFormat() + 1);
                                  // Set buffer length

   char* pszNewBuf = new char[ulMyBufSize];
                                  // Create buffer

   numChars = ((IMultiLineEdit *)this)->export(&ipt,ulMyBufSize-1,pszNewBuf);
                                  // Export text to buffer

   *(pszNewBuf + numChars) = '\0';// Add a null terminator
   IString str( pszNewBuf );
   delete pszNewBuf;
   return str;
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::text                                                         |
|                                                                              |
| Returns the current contents of the MLE at the specified line number.        |
------------------------------------------------------------------------------*/
IString IMultiLineEdit::text(unsigned long lLineNumber) const
{
   LONG  numChars;                // Number of chars exported
   long  ipt(0);
   char* pszNewBuf;
   unsigned long ulMyBufSize;

   if (lLineNumber < numberOfLines())
   {
      ipt = (long)((unsigned long) handle().sendEvent(MLM_CHARFROMLINE,
                                 IEventParameter1(lLineNumber),
                                 IEventParameter2(0)));
   } else {
      return IString();
   }
   ulMyBufSize = (unsigned long) handle().sendEvent(MLM_QUERYLINELENGTH,
                              IEventParameter1((unsigned long)ipt),
                              IEventParameter2(0));

   pszNewBuf = new char[ulMyBufSize + 1]; // Create buffer
#ifdef IC_WIN
   // Windows GetLine function requires buffer size to be in first word
   // so set it there and pass line number and buffer for query
   pszNewBuf[0] = (char)ulMyBufSize;
   numChars = handle().sendEvent( EM_GETLINE,
                                  IEventParameter1(lLineNumber),
                                  IEventParameter2(pszNewBuf) );
#else
   numChars = ((IMultiLineEdit *)this)->export(&ipt,ulMyBufSize,pszNewBuf);
                                      // Export text to buffer
#endif
   *(pszNewBuf + numChars) = '\0';    // Add a null terminator
   IString str( pszNewBuf );
   delete pszNewBuf;
   return str;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::addAtOffset                                                  |
|                                                                              |
| Insert the specified text into the MLE at the specified insertion point.     |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::addAtOffset( const char*   pszBuffer,
                                             unsigned long lInsertionPoint,
                                             unsigned long ulBufferSize,
                                             EOLFormat     type )
{
   long ipt;
   if (ulBufferSize == 0)                 // If no size specified,
      ulBufferSize = strlen(pszBuffer);   // Assume null-terminated string
#ifdef IC_PM
   long lNumInserted;
   long iptCursor;

   ipt = lInsertionPoint;                 // Input to zero-based
   switch (type)
   {
     case cfText:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_CFTEXT),
                                         IEventParameter2(0));
          break;
     case noTran:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_NOTRANS),
                                         IEventParameter2(0));
          break;
     case MLEFormat:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_WINFMT),
                                         IEventParameter2(0));
          break;
   } /* endswitch */

   /**************************************************************/
   /* Import text from the buffer into the MLE                   */
   /**************************************************************/
   lNumInserted = (long)import(&ipt,ulBufferSize,pszBuffer);
   IEventResult evt(handle().sendEvent(MLM_QUERYSEL,
                                       IEventParameter1(MLFQS_CURSORSEL),
                                       IEventParameter2(0)));
   iptCursor = evt.asUnsignedLong();
   if (iptCursor > lInsertionPoint) {      // If text was inserted before
      iptCursor += lNumInserted;           // the cursor location, reset the
      setCursorPosition( iptCursor );      // cursor position
   } /* endif */
#endif
#ifdef IC_WIN
   IRange  initMarked, newMarked;
   Boolean bMustRestore = true;

   /*******************************************************************/
   /* If selected text exists, save range away to restore later       */
   /*******************************************************************/
   if (hasSelectedText())
   {
     bMustRestore = true;
     initMarked = selectedRange();
     newMarked = initMarked;
   }

   /*******************************************************************/
   /* Disable refresh while modifications are made to avoid flashing  */
   /*******************************************************************/
   Boolean bOldRefresh=bRefresh;
   if (bRefresh)
      disableUpdate();

   /*******************************************************************/
   /* Place cursor at indicated position and insert text              */
   /*******************************************************************/
   ipt = lInsertionPoint;
   setCursorPosition( ipt );
   handle().sendEvent( EM_REPLACESEL,
                       IEventParameter1(0),
                       IEventParameter2((char*)pszBuffer) );

   /*******************************************************************/
   /* If a range was selected before, then restore that range         */
   /*******************************************************************/
   if (bMustRestore)
   {
     /*****************************************************************/
     /* If text inserted is before selected range, move range up      */
     /*****************************************************************/
     if ( initMarked.lowerBound() > ipt )
       newMarked.setLowerBound( initMarked.lowerBound() + ulBufferSize);

     /*****************************************************************/
     /* If text is inserted into the selected area, then change upper */
     /*****************************************************************/
     else if ( (initMarked.lowerBound() <= ipt) &&
               (initMarked.upperBound() > ipt) )
     {
       newMarked.setUpperBound(initMarked.upperBound() + ulBufferSize);
     }

     /*****************************************************************/
     /* If inserted above, then original range is still valid so in   */
     /* any case, must restore selection.                             */
     /*****************************************************************/
     selectRange( newMarked );
   }
   else
     setCursorPosition(ipt+ulBufferSize);

   /*******************************************************************/
   /* Restore refresh (if required) after all modifications complete  */
   /*******************************************************************/
   if (bOldRefresh)
     enableUpdate();
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::addAsLast                                                    |
|                                                                              |
| Insert the specified text into the MLE at the end of the existing text.      |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::addAsLast( const char*   pszBuffer,
                                           unsigned long ulBufferSize,
                                           EOLFormat     type)
{
   if (ulBufferSize == 0)                 // If no size specified,
      ulBufferSize = strlen(pszBuffer);   // Assume null-terminated string
#ifdef IC_PM
   long ipt(textLengthAfterFormat());

   switch (type)
   {
     case cfText:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_CFTEXT),
                                         IEventParameter2(0));
          break;
     case noTran:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_NOTRANS),
                                         IEventParameter2(0));
          break;
     case MLEFormat:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_WINFMT),
                                         IEventParameter2(0));
          break;
   } /* endswitch */

   /**************************************************************/
   /* Import text from the buffer into the MLE                   */
   /**************************************************************/
   import(&ipt,ulBufferSize,pszBuffer);
#endif
#ifdef IC_WIN
   IRange  initMarked;
   Boolean bMustRestore = true;

   /*******************************************************************/
   /* If selected text exists, save range away to restore later       */
   /*******************************************************************/
   if (hasSelectedText())
   {
     bMustRestore = true;
     initMarked = selectedRange();
   }

   /*******************************************************************/
   /* Disable refresh while modifications are made to avoid flashing  */
   /*******************************************************************/
   Boolean bOldRefresh=bRefresh;
   if (bRefresh)
     disableUpdate();

   /*******************************************************************/
   /* Place cursor at end of text and insert                          */
   /*******************************************************************/
   long ipt = textLength();
   setCursorPosition( ipt + 1 );
   handle().sendEvent( EM_REPLACESEL,
                       IEventParameter1(0),
                       IEventParameter2((char*)pszBuffer) );

   /*******************************************************************/
   /* If original selection, restore, otherwise put cursor after text */
   /*******************************************************************/
   if (bMustRestore)
     selectRange( initMarked );
   else
     setCursorPosition(ipt+ulBufferSize);

   /*******************************************************************/
   /* Restore refresh (if required) after all modifications complete  */
   /*******************************************************************/
   if (bOldRefresh)
     enableUpdate();
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::add                                                          |
|                                                                              |
| Insert the specified text into the MLE at the current cursor position.       |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::add( const char*   pszBuffer,
                                     unsigned long ulBufferSize,
                                     EOLFormat     type )
{
   if (ulBufferSize == 0)                 // If no size specified,
      ulBufferSize = strlen(pszBuffer);   // Assume null-terminated string
   long ipt(-1);                          // Insert at current cursor position
#ifdef IC_WIN
   ipt = cursorPosition();
   setCursorPosition(ipt);                // Deselect any current selection
   handle().sendEvent( EM_REPLACESEL,
                       IEventParameter1(0),
                       IEventParameter2((char*)pszBuffer) );
   setCursorPosition(ipt+ulBufferSize);   // Position cursor after newly
                                          // inserted text
#else
   switch (type)
   {
     case cfText:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_CFTEXT),
                                         IEventParameter2(0));
          break;
     case noTran:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_NOTRANS),
                                         IEventParameter2(0));
          break;
     case MLEFormat:
          handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_WINFMT),
                                         IEventParameter2(0));
          break;
   } /* endswitch */

   /**************************************************************/
   /* Import text from the buffer into the MLE                   */
   /**************************************************************/
   import( &ipt, ulBufferSize, pszBuffer );

   setCursorPosition( ipt );     // Position cursor after newly inserted text
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::importFromFile                                               |
|                                                                              |
| Insert the contents of the specified file into the MLE at the current cursor |
| position.                                                                    |
| Notes: Prior to importing a file to the MLE, the file must exist (in binary  |
|          format).                                                            |
------------------------------------------------------------------------------*/
unsigned long  IMultiLineEdit::importFromFile( const char* pszFileName,
                                               EOLFormat   type )
{
   FILE*          filePtr;

   /***********************************************************************/
   /* Open the specified file.                                            */
   /***********************************************************************/
   struct stat buf;
   if (stat(pszFileName,&buf))
      buf.st_size=0;

#ifdef IC_WIN
   // If a limit has NOT been set specifically by application
   if (!flimitSet)
   {
     int fh;
     // Then get the overall size of the file
     if (( fh = open( pszFileName, O_RDONLY )) == -1 )
       ITHROWGUIERROR("_open");
     long length = filelength( fh );
     close( fh );

     // And if the size is greater than default, change the size
     if (length > fcharLimit)
     {
       IEventResult evt(handle().sendEvent( EM_LIMITTEXT,
                                            IEventParameter1( (int)length ),
                                            IEventParameter2( 0 ) ));
     }
   }
#endif
   if ((filePtr = fopen(pszFileName, "rb")) == (FILE *)0)
      ITHROWGUIERROR("fopen");
   IString data(0,(unsigned int)(buf.st_size+1));

   /***********************************************************************/
   /* Read a line from the file till the end of file                      */
   /***********************************************************************/
   Boolean bOldRefresh=bRefresh;
   if (bRefresh)
      disableUpdate();

   if (!(fread((char*)data,sizeof(char),buf.st_size-1,filePtr)))
      ITHROWGUIERROR("fread");
   add(data,buf.st_size-1,type);

// unsigned long  ulNumChars(0);
// char line[1025];
// int  i;
// int  j;
// for (j=0;j<buf.st_size ; )
// {
//    for (i=0;(i<=1023 && j+i<buf.st_size);i++)
//       if ((INT)(line[i]=fgetc(filePtr)) == 26)
//       {
//          i--;
//          j++;
//       }
//    if (i)
//    {
//       line[i] = 0;      // Null terminate (for Windows)
//       add(&line[0],i,type);
//    }
//    ulNumChars+=i;
//    j+=i;
// } /* endfor */

   if (bOldRefresh)
     enableUpdate();

   fclose(filePtr);
   return buf.st_size-1;
// return ulNumChars;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::exportToFile                                                 |
|                                                                              |
| Save the contents of the MLE to the specified file.                          |
------------------------------------------------------------------------------*/
unsigned long  IMultiLineEdit::exportToFile( const char* pszFileName,
                                             EOLFormat   type )
{
   if (textLengthAfterFormat())
   {
#ifdef IC_WIN
      handle().sendEvent( EM_FMTLINES,
                          IEventParameter1(type == MLEFormat),
                          IEventParameter2(0));
#else //IC_PM
      switch (type)
      {
        case cfText:
             handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_CFTEXT),
                                            IEventParameter2(0));
             break;
        case noTran:
             handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_NOTRANS),
                                            IEventParameter2(0));
             break;
        case MLEFormat:
             handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_WINFMT),
                                            IEventParameter2(0));
             break;
      } /* endswitch */
#endif //IC_PM

      return exportFile(pszFileName, FALSE);

   } else {

      return 0;

   } /* endif */
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::exportSelectedTextToFile                                     |
|                                                                              |
| Export the marked area of MLE to a file.                                     |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::exportSelectedTextToFile ( const char* fileName,
                                                         EOLFormat   type)
{
  IASSERTSTATE(hasSelectedText());
#ifdef IC_PM
  switch (type)
  {
    case cfText:
         handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_CFTEXT),
                                        IEventParameter2(0));
         break;
    case noTran:
         handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_NOTRANS),
                                        IEventParameter2(0));
         break;
    case MLEFormat:
         handle().sendEvent(MLM_FORMAT, IEventParameter1(MLFIE_WINFMT),
                                        IEventParameter2(0));
         break;
  } /* endswitch */
#endif
  return exportFile(fileName, TRUE);
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::setLimit                                                     |
|                                                                              |
| Set the limit on the number of bytes that can be input into the MLE.         |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setLimit( unsigned long lNewLimit )
{
  IASSERTSTATE(text().length() <= lNewLimit);

  /********************************************************************/
  /* If the limit is already the value passed, don't do anything      */
  /********************************************************************/
  if (lNewLimit != this->limit())
  {
    /******************************************************************/
    /* Now set the new edit limit for the mle                         */
    /******************************************************************/
    IEventResult evt(handle().sendEvent(MLM_SETTEXTLIMIT,
                                        IEventParameter1(lNewLimit),
                                        IEventParameter2(0)));
#ifdef IC_WIN
    fcharLimit = lNewLimit;
    flimitSet = true;
#endif
#ifdef IC_PM
    // Check error code returned only on PM
    if (evt.asUnsignedLong())
    {
      ITHROWGUIERROR("MLM_SETTEXTLIMIT");
    }
#endif
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::limit                                                        |
|                                                                              |
| Returns the maximum number of bytes allowed that the MLE can hold.           |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::limit() const
{
#ifdef IC_PM
  IEventResult evt(handle().sendEvent(MLM_QUERYTEXTLIMIT,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));

  if (evt.asLong() < 0)
     return ULONG_MAX;
  else
     return evt.asUnsignedLong();
#endif
#ifdef IC_WIN
  return fcharLimit;    // Must keep locally since Windows does not
                        // support querying of current limit set
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::selectRange                                                  |
|                                                                              |
| Specifies a range of characters.                                             |
| Notes: If upper range is not specified, default is last character in the MLE.|
------------------------------------------------------------------------------*/
IMultiLineEdit&  IMultiLineEdit::selectRange( const IRange& range,
                                              unsigned long timestamp )
{
   long start(range.lowerBound());
   long stop(range.upperBound());

   /*******************************************************************/
   /* If first parameter is deselect, then set up parameters to do so */
   /*******************************************************************/
   if (start == deselect)
   {
#ifdef IC_WIN
     stop = 0;
#endif
#ifdef IC_PM
     start = cursorPosition();
     stop = start;
#endif
   }
   /*******************************************************************/
   /* otherwise we are selecting, so validate parameters passed       */
   /*******************************************************************/
   else
   {
     if (start == end)
        start = textLength()+1;

     if (stop == end)
        stop = textLength()+1;

     if ( stop  < 0)
        stop  = 0;

     if ( start < 0)
        start = 0;

     if (start > stop )
        start++;   //Need to move cursor past character so we can select it
     else
        stop++;    //Need to move cursor past character so we can select it
   }

   handle().sendEvent(MLM_SETSEL, IEventParameter1((unsigned long)start),
                                  IEventParameter2((unsigned long)stop));

   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::selectedRange                                                |
|                                                                              |
| Gets the bound of the marked or selected text.  Index of marked characters   |
| is returned as 0-based.                                                      |
------------------------------------------------------------------------------*/
IRange  IMultiLineEdit::selectedRange() const
{
  IASSERTSTATE(hasSelectedText());
#ifdef IC_WIN
  IEventResult minSel, maxSel;
  handle().sendEvent(EM_GETSEL,
                     IEventParameter1(&minSel),
                     IEventParameter2(&maxSel));
#else
  IEventResult minSel( handle().sendEvent(MLM_QUERYSEL,
                                          IEventParameter1(MLFQS_MINSEL),
                                          IEventParameter2(0)));
  IEventResult maxSel( handle().sendEvent(MLM_QUERYSEL,
                                          IEventParameter1(MLFQS_MAXSEL),
                                          IEventParameter2(0)));
#endif
  IRange rangMarked(minSel.asUnsignedLong(), maxSel.asUnsignedLong());
  if (rangMarked.lowerBound() != rangMarked.upperBound())
     rangMarked.setUpperBound(rangMarked.upperBound() - 1);
                                  // Make upper the last selected character
  return rangMarked;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::selectedText                                                 |
|                                                                              |
| Gets the marked (selected) text.                                             |
------------------------------------------------------------------------------*/
IString IMultiLineEdit::selectedText() const
{
  unsigned long length = selectedTextLength();
  if (!length)
  {                               // No selected text
     IString str("\0");
     return str;
  }
  else
  {
#ifndef IC_WIN
    char  *pchMarkedText;           // Source text
    IString str(IString(0,length+1));
    pchMarkedText = (char *)str;
    handle().sendEvent( MLM_QUERYSELTEXT,
                        IEventParameter1((unsigned long) pchMarkedText),
                        IEventParameter2(0) );
#endif
#ifdef IC_WIN
    /******************************************************************/
    /* No easy way to only get the selected text.                     */
    /* Must first get all text and build substring of it              */
    /******************************************************************/
    IRange rangeMarked = selectedRange();
    IString str( (char*)text() + rangeMarked.lowerBound(),
                 (unsigned int)length, '\0' );
#endif
    return str;
  } /*endif */
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::hasSelectedText                                              |
|                                                                              |
| Query whether any text is currently marked.                                  |
------------------------------------------------------------------------------*/
Boolean IMultiLineEdit::hasSelectedText() const
{
  if (selectedTextLength() == 0)
    return( false );
  else
    return( true );
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::isUndoable                                                   |
|                                                                              |
| Queries whether any undoable actions have been performed on the contents of  |
| the MLE.                                                                     |
------------------------------------------------------------------------------*/
Boolean IMultiLineEdit::isUndoable() const
{
  IEventResult evt(handle().sendEvent(MLM_QUERYUNDO,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
  if (evt.asUnsignedLong())
     return true;
  else
     return false;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::cut                                                          |
|                                                                              |
| Copy any marked text to the clipboard, and then delete the marked text from  |
| the MLE.                                                                     |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::cut(unsigned long timestamp)
{
  IASSERTSTATE(hasSelectedText());

  IEventResult evt(handle().sendEvent(MLM_CUT,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
#ifdef IC_PM
  if (!(evt.asUnsignedLong()))
     ITHROWGUIERROR("MLM_CUT");
#endif

  return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::copy                                                         |
|                                                                              |
| Copy any marked text to the clipboard.                                       |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::copy(unsigned long timestamp)
{
  IASSERTSTATE(hasSelectedText());

  IEventResult evt(handle().sendEvent(MLM_COPY,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
#ifdef IC_PM
  if (!(evt.asUnsignedLong()))
     ITHROWGUIERROR("MLM_COPY");
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::paste                                                        |
|                                                                              |
| Paste text from the clipboard into the MLE at the current cursor position    |
| in the text.                                                                 |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::paste()
{
  IASSERTSTATE(clipboardHasTextFormat());

  IEventResult evt(handle().sendEvent(MLM_PASTE,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
#ifdef IC_PM
  if (!(evt.asUnsignedLong()))
     ITHROWGUIERROR("MLM_PASTE");
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::clear                                                        |
|                                                                              |
| Replace the marked area with blanks.                                         |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::clear(unsigned long timestamp)
{
  char    *pchMarkedText,              // Source text
          *pchText;
  ULONG    num_char, i;                // char and loop count

  IASSERTSTATE(hasSelectedText());
  IString str(selectedText());
  pchMarkedText = (char *)str;
  pchText = (char *) text();

  /***********************************************************************/
  /* Get the size of user marked area.                                   */
  /***********************************************************************/
  num_char = selectedTextLength();

  /******************************************************************/
  /* Replace the text with blanks, leaving CR and LF chars intact.  */
  /******************************************************************/
  for (i = 0; i < num_char; i++)
      if (*(pchMarkedText + i) != 0x0D && *(pchMarkedText +i) != 0x0A)
         *(pchMarkedText + i) = ' ';

  /******************************************************************/
  /* Replace the marked area with the blanked out text.             */
  /******************************************************************/
#ifdef IC_WIN
  handle().sendEvent( EM_REPLACESEL,
                      IEventParameter1(1),  // set undo flag to true
                      IEventParameter2((unsigned long)pchMarkedText));
  pchText = (char *) text();
#else
  Boolean bOldRefresh=bRefresh;
  if (bRefresh)
     disableUpdate();

  handle().sendEvent( MLM_INSERT,
                      IEventParameter1((unsigned long)pchMarkedText),
                      IEventParameter2(0));

  if (bOldRefresh)
     enableUpdate();
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::discard                                                      |
|                                                                              |
| Erase the marked text.                                                       |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::discard()
{
  IASSERTSTATE(hasSelectedText());

  IEventResult evt(handle().sendEvent(MLM_CLEAR,
                                      IEventParameter1(0),
                                      IEventParameter2(0)));
#ifdef IC_PM
  if (!(evt.asUnsignedLong()))
     ITHROWGUIERROR("MLM_CLEAR");
#endif

  return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::undo                                                         |
|                                                                              |
| Restores the MLE contents to the state they were in before the last change.  |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::undo()
{
   /********************************************************************/
   /* Request system to undo the last operation, if there was one.     */
   /********************************************************************/
   IEventResult evt(handle().sendEvent(MLM_UNDO,
                                       IEventParameter1(0),
                                       IEventParameter2(0)));

#ifndef IC_WIN
   if (!(evt.asUnsignedLong()))
      /*****************************************************************/
      /* Undo failed, so lets reset the MLE undo flag.                 */
      /*****************************************************************/
      handle().sendEvent(MLM_RESETUNDO, IEventParameter1(0),
                                        IEventParameter2(0));
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::removeAll                                                    |
|                                                                              |
| Deletes the entire contents of the MLE.                                      |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::removeAll()
{
#ifdef IC_WIN
   // No delete all function in Windows so just set contents to NULL string
   setText((char*)IString(),0);
#else
   Boolean bOldRefresh(bRefresh);

   if (bRefresh)
      disableUpdate();

   handle().sendEvent(MLM_DELETE, IEventParameter1(0),
                                  IEventParameter2(textLengthAfterFormat()));

   if (bOldRefresh)
      enableUpdate(bOldRefresh);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setEditRegionWidth                                           |
|                                                                              |
| Set the width (in pixels) of the MLE text region.  The edit region defines   |
| the area in which text will be visible.  It can be less than the actual      |
| MLE window size.                                                             |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setEditRegionWidth(long sWidth)
{
#ifdef IC_WIN
   RECT mleRect;

   handle().sendEvent( EM_GETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect) );
   // Based on ref text, adjust returned rect by border size for set to work
   // correctly.
   if ((style() & border.asUnsignedLong()) ||
       (extendedStyle() & IMLS_BORDERNTHACK))
   {
      mleRect.left -= IQUERYSYSVALUE(SV_CXBORDER);
      mleRect.top -= IQUERYSYSVALUE(SV_CYBORDER);
      mleRect.right = mleRect.left + sWidth;
      mleRect.bottom += IQUERYSYSVALUE(SV_CYBORDER);
   }
   else
      mleRect.right = mleRect.left + sWidth;

   // Note no exception thrown since message does not return value
   IEventResult evt( handle().sendEvent(EM_SETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect)));
#else
   MLEFORMATRECT mleformrc;
   mleformrc.cxFormat = sWidth;
   Boolean bOldRefresh=bRefresh;

   if (bRefresh)
      disableUpdate();

   IEventResult evt(
           handle().sendEvent(MLM_SETFORMATRECT,
                              IEventParameter1((unsigned long)(&mleformrc)),
                              IEventParameter2((unsigned long)
                                                        MLFFMTRECT_LIMITHORZ)));
   if (bOldRefresh)
      enableUpdate();
   if (!(evt.asUnsignedLong()))
      ITHROWLIBRARYERROR(IC_INVALIDSIZE,IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setEditRegionHeight                                          |
|                                                                              |
| Set the height of the text region inside of the MLE control.                 |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setEditRegionHeight(long sHeight)
{
#ifdef IC_WIN
   RECT mleRect;

   handle().sendEvent( EM_GETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect) );
   // Based on ref text, adjust returned rect by border size for set to work
   // correctly.
   if ((style() & border.asUnsignedLong()) ||
       (extendedStyle() & IMLS_BORDERNTHACK))
   {
      mleRect.left -= IQUERYSYSVALUE(SV_CXBORDER);
      mleRect.top -= IQUERYSYSVALUE(SV_CYBORDER);
      mleRect.right += IQUERYSYSVALUE(SV_CXBORDER);
      mleRect.bottom = mleRect.top + sHeight;
   }
   else
      mleRect.bottom = mleRect.top + sHeight;

   // Note no exception thrown since message does not return value
   IEventResult evt( handle().sendEvent(EM_SETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect)));
#else
   MLEFORMATRECT mleformrc;
   mleformrc.cyFormat = sHeight;
   Boolean bOldRefresh=bRefresh;

   if (bRefresh)
      disableUpdate();

   IEventResult evt(
           handle().sendEvent(MLM_SETFORMATRECT,
                              IEventParameter1((unsigned long)(&mleformrc)),
                              IEventParameter2((unsigned long)
                                                     MLFFMTRECT_LIMITVERT)));
   if (bOldRefresh)
      enableUpdate();
   if (!(evt.asUnsignedLong()))
      ITHROWLIBRARYERROR(IC_INVALIDSIZE,IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setEditRegion                                                |
|                                                                              |
| Set the width and the height of the text region inside of the MLE control.   |
| Method returns TRUE if the text fits within new format-rectangle dimension.  |
| Otherwise, it returns FALSE, indicating that the format rectangle was not    |
| set.                                                                         |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setEditRegion( const ISize& sizeNew )
{
#ifdef IC_WIN
   RECT mleRect;
   GetClientRect ( handle(), &mleRect );

   mleRect.right = mleRect.left + (long)sizeNew.width();
   mleRect.bottom = mleRect.top + (long)sizeNew.height();

   // Note no exception thrown since message does not return value
   IEventResult evt( handle().sendEvent(EM_SETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect)));
#else
   MLEFORMATRECT mleformrc;
   mleformrc.cxFormat = (long)sizeNew.width();
   mleformrc.cyFormat = (long)sizeNew.height();
   Boolean bOldRefresh=bRefresh;

   if (bRefresh)
      disableUpdate();

   IEventResult evt(
           handle().sendEvent(MLM_SETFORMATRECT,
                              IEventParameter1((unsigned long)(&mleformrc)),
                              IEventParameter2((unsigned long)
                                                  (MLFFMTRECT_LIMITVERT |
                                                   MLFFMTRECT_LIMITHORZ))));
   if (bOldRefresh)
      enableUpdate();

   if (!(evt.asUnsignedLong()))
      ITHROWLIBRARYERROR(IC_INVALIDSIZE,IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setEditRegion                                                |
|                                                                              |
| Set the edit region to be the size of the entire MLE window (minus margins). |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setEditRegion()
{
#ifdef IC_WIN
   RECT mleRect;
   GetClientRect ( handle(), &mleRect );

   // Note no exception thrown since message does not return value
   IEventResult evt( handle().sendEvent(EM_SETRECT,
                     IEventParameter1(0),
                     IEventParameter2((unsigned long)&mleRect)));

#else
   MLEFORMATRECT mleformrcNew;
   Boolean bOldRefresh=bRefresh;
   if (bRefresh)
      disableUpdate();
   IEventResult evt(
           handle().sendEvent(MLM_SETFORMATRECT,
                              IEventParameter1((unsigned long)(&mleformrcNew)),
                              IEventParameter2((unsigned long)
                                                      MLFFMTRECT_MATCHWINDOW)));
   if (bOldRefresh)
      enableUpdate();

   if (!(evt.asUnsignedLong()))
      ITHROWLIBRARYERROR(IC_INVALIDSIZE,IBaseErrorInfo::accessError,
                         IException::recoverable);
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::editRegionWidth                                              |
|                                                                              |
| Returns the width of the MLE editing area.                                   |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::editRegionWidth() const
{
#ifdef IC_WIN
   RECT mleRect;

   handle().sendEvent( EM_GETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect) );

   // Based on ref text, adjust returned rect by border size for set to work
   // correctly.
   if ((style() & border.asUnsignedLong()) ||
       (extendedStyle() & IMLS_BORDERNTHACK))
   {
      mleRect.left -= IQUERYSYSVALUE(SV_CXBORDER);
      mleRect.right += IQUERYSYSVALUE(SV_CXBORDER);
   }
   return (mleRect.right - mleRect.left);
#else
   MLEFORMATRECT mleformrc;
   unsigned long         ulFormatFlag =  MLFFMTRECT_LIMITHORZ;

   /***********************************************************************/
   /* Get the extent of the edit area.                                    */
   /***********************************************************************/
   handle().sendEvent(MLM_QUERYFORMATRECT,
                      IEventParameter1((unsigned long)(&mleformrc)),
                      IEventParameter2((unsigned long)(&ulFormatFlag)));

   return (mleformrc.cxFormat);
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::editRegionHeight                                             |
|                                                                              |
| Returns the height of the MLE editing area.                                  |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::editRegionHeight() const
{
#ifdef IC_WIN
   RECT mleRect;

   handle().sendEvent( EM_GETRECT,
                       IEventParameter1(0),
                       IEventParameter2((unsigned long)&mleRect) );

   // Based on ref text, adjust returned rect by border size for set to work
   // correctly.
   if ((style() & border.asUnsignedLong()) ||
       (extendedStyle() & IMLS_BORDERNTHACK))
   {
      mleRect.top -= IQUERYSYSVALUE(SV_CYBORDER);
      mleRect.bottom += IQUERYSYSVALUE(SV_CYBORDER);
   }
   return (mleRect.bottom - mleRect.top);
#else
   MLEFORMATRECT mleformrc;
   unsigned long         ulFormatFlag =  MLFFMTRECT_LIMITVERT;

   /***********************************************************************/
   /* Get the extent of the edit area.                                    */
   /***********************************************************************/
   handle().sendEvent(MLM_QUERYFORMATRECT,
                      IEventParameter1((unsigned long)(&mleformrc)),
                      IEventParameter2((unsigned long)(&ulFormatFlag)));
   return (mleformrc.cyFormat);
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setTab                                                       |
|                                                                              |
| Sets the interval at which tab stops are placed (expressed in pixels).       |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setTab(unsigned long lTabPixelInterval)
{
#ifdef IC_WIN
   // Get Dialog base units and mask in lower value (horizontal base units)
   LONG gdbu = GetDialogBaseUnits() & 0x0000FFFF;

   // Calculate dialog units based on pixel interval passed in
   unsigned long lTabDistance = (lTabPixelInterval * 4) / gdbu;
   IEventResult evt(
           handle().sendEvent(EM_SETTABSTOPS,
                              IEventParameter1(1),
                              IEventParameter2(&lTabDistance)));

   if (evt.asLong() == 0)
      ITHROWGUIERROR("MLM_SETTABSTOP");
#else
   IEventResult evt(
           handle().sendEvent(MLM_SETTABSTOP,
                              IEventParameter1((unsigned long)
                                                       lTabPixelInterval),
                              IEventParameter2(0)));

   if (evt.asLong() < 0)
      ITHROWGUIERROR("MLM_SETTABSTOP");
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::isWordWrap                                                   |
------------------------------------------------------------------------------*/
Boolean IMultiLineEdit::isWordWrap() const
{
#ifdef IC_PM
   IEventResult evt(handle().sendEvent(MLM_QUERYWRAP,
                                       IEventParameter1(0),
                                       IEventParameter2(0)));
   if (evt.asUnsignedLong())
      return true;
   else
      return false;
#endif
#ifdef IC_WIN
  // Check word wrap from style bits (extended) since only set on create
  return ( extendedStyle() & wordWrap.asExtendedUnsignedLong() );
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::enableWordWrap                                               |
|                                                                              |
| Puts the MLE in word-wrap mode, which means lines will not be scrolled       |
| right or left.                                                               |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::enableWordWrap( Boolean enable )
{
#ifdef IC_PM
  Boolean wordWrap(this->isWordWrap());

  /***************************/
  /* Only set if we need to. */
  /***************************/
  if ((enable && !wordWrap) || (!enable && wordWrap))
  {
    IEventResult evt(handle().sendEvent(MLM_SETWRAP,
                                        IEventParameter1(enable),
                                        IEventParameter2(0)));

    if (!(evt.asUnsignedLong()))
    {
      ITHROWGUIERROR("MLM_SETWRAP");
    }
  }
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::disableWordWrap                                              |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::disableWordWrap()
{
   enableWordWrap(false);
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::foregroundColor                                              |
|                                                                              |
| Returns the foreground color of the IMultiLineEdit control.                  |
------------------------------------------------------------------------------*/
IColor IMultiLineEdit::foregroundColor() const
{
  IGUIColor guiColor(IGUIColor::windowText);

  if (!isWriteable())
  {
    guiColor = IGUIColor(IGUIColor::outputText);
  }
  return (IWindow::color(PP_FOREGROUNDCOLOR, guiColor));
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::backgroundColor                                              |
|                                                                              |
| Returns the background color of the IMultiLineEdit control.                  |
------------------------------------------------------------------------------*/
IColor IMultiLineEdit::backgroundColor() const
{
  return (IWindow::color(PP_BACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::entryFieldBgnd)));
}

#ifndef IC_WIN
/*------------------------------------------------------------------------------
| IMultiLineEdit::setFont                                                      |
|                                                                              |
| Change the font used for this MLE object.                                    |
| Notes: The font point size set by this routine is the best fit size          |
|          determined by the system instead of the size specified by the user. |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setFont(const IFont& fntm)
{
   _FATTRS* pfat = (_FATTRS*)fntm.fattrs();

   if (pfat->fsSelection & FATTR_SEL_OUTLINE)
   {
      pfat->fsFontUse |=FATTR_FONTUSE_OUTLINE;
   } /* endif */

   if (fntm.isVectorOnly()) {
      pfat->fsFontUse |=FATTR_FONTUSE_TRANSFORMABLE;
      long  lxFontResolution, lyFontResolution;
      HDC   hDC;
      HPS hps = WinGetPS(handle());
      hDC = GpiQueryDevice(hps);
      DevQueryCaps( hDC, CAPS_HORIZONTAL_FONT_RES,
                    (long)1, &lxFontResolution);
      DevQueryCaps( hDC, CAPS_VERTICAL_FONT_RES,
                    (long)1, &lyFontResolution);
      pfat->lMaxBaselineExt = fntm.pointSize() * lyFontResolution / 72;
      pfat->lAveCharWidth   = fntm.pointSize() * lxFontResolution / 72;
      WinReleasePS(hps);
      }

   IEventResult evt(handle().sendEvent(MLM_SETFONT,
                                       IEventParameter1((unsigned long)pfat),
                                       IEventParameter2(0)));

   if (!(evt.asUnsignedLong()))
      ITHROWGUIERROR("MLM_SETFONT");
   return *this;
}
#endif

/*------------------------------------------------------------------------------
| IMultiLineEdit::numberOfLines                                                |
|                                                                              |
| Query the number of lines of text in the MLE.                                |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::numberOfLines() const
{
   unsigned long ulRetVal,ipt;

   /******************************************************************/
   /* Query number of lines in the MLE                               */
   /******************************************************************/
   ulRetVal = (unsigned long) handle().sendEvent( MLM_QUERYLINECOUNT,
                                                  IEventParameter1(0),
                                                  IEventParameter2(0));

   /******************************************************************/
   /* Get the position of the first character on the last line       */
   /******************************************************************/
   ipt = (unsigned long) handle().sendEvent( MLM_CHARFROMLINE,
                                             IEventParameter1(ulRetVal-1),
                                             IEventParameter2(0) );

   /******************************************************************/
   /* Check last line, if line length is 0, subtract one from count  */
   /******************************************************************/
   if ((unsigned long) handle().sendEvent(MLM_QUERYLINELENGTH,
                  IEventParameter1(ipt), IEventParameter2(0)) != 0)
     return ulRetVal;
   else
     return ulRetVal-1;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::addLineAsLast                                                |
|                                                                              |
| Insert the specified NULL-terminated text at the last line.                  |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::addLineAsLast( const char* pszLineText,
                                               EOLFormat   type )
{
   addLine(pszLineText,numberOfLines()+1,type);
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::addLine                                                      |
|                                                                              |
| Insert the specified line of text at the line location specified.            |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::addLine( const char*   pszLineText,
                                         unsigned long lLineNumber,
                                         EOLFormat     type )
{
   long ipt;
   unsigned long  ulBuffSize;
   char * pszTempBuff;
   char pszEolChar[3];
   unsigned long ulFormat;
   Boolean bTextSel;
   IRange  selRange;

   strcpy( pszEolChar, "\r\n");
   switch (type)
   {
     case cfText:
        ulFormat = MLFIE_CFTEXT;
        break;
     case noTran:
        ulFormat = MLFIE_NOTRANS;
        strcpy( pszEolChar, "\n");
        break;
     case MLEFormat:
        ulFormat = MLFIE_WINFMT;
        break;
   } /* endswitch */

#ifdef IC_WIN
   handle().sendEvent( EM_FMTLINES,
                       IEventParameter1(ulFormat == MLFIE_WINFMT),
                       IEventParameter2(0));
#else
   handle().sendEvent( MLM_FORMAT,
                       IEventParameter1(ulFormat),
                       IEventParameter2(0));
#endif

   if (bTextSel = hasSelectedText())
     selRange = selectedRange();

   // If line number to add is before end of file, calculate insertion point
   if (lLineNumber < numberOfLines())
   {
      ipt = (long)((unsigned long) handle().sendEvent(MLM_CHARFROMLINE,
                              IEventParameter1(lLineNumber),
                              IEventParameter2(0)));
      ulBuffSize = strlen(pszLineText) + strlen(pszEolChar) + 1;
      pszTempBuff = new char[ulBuffSize];        // Allocate storage

      strcpy( pszTempBuff, pszLineText);
      strcat( pszTempBuff, pszEolChar);          // Add line terminator

#ifdef IC_WIN
      handle().sendEvent( EM_SETSEL,
                          IEventParameter1((unsigned long)ipt),
                          IEventParameter2((unsigned long)ipt) );
      handle().sendEvent( EM_REPLACESEL,
                          IEventParameter1(0),
                          IEventParameter2(pszTempBuff) );
#else
      import(&ipt,ulBuffSize,pszTempBuff);                   // Import text,

      handle().sendEvent( MLM_DELETE,
                          IEventParameter1((unsigned long)ipt-1),
                          IEventParameter2(1));
#endif
      if (bTextSel)
        selectRange( selRange );
      delete pszTempBuff;

   }
   // otherwise line is to be added at the end of the file
   else
   {
      ipt = textLength();
      if (ipt!=0)
      {
#ifdef IC_WIN
         handle().sendEvent( EM_SETSEL,
                             IEventParameter1((unsigned long)ipt),
                             IEventParameter2((unsigned long)ipt) );
         handle().sendEvent( EM_REPLACESEL,
                             IEventParameter1(0),
                             IEventParameter2(pszEolChar) );
         ipt += 2;
#else
         import(&ipt, strlen(pszEolChar)+1, pszEolChar );    // Import an eol char
         handle().sendEvent(MLM_DELETE,
                                 IEventParameter1((unsigned long)ipt-1),
                                 IEventParameter2(1));
         ipt--;
#endif
      } /* endif */

#ifdef IC_WIN
      handle().sendEvent( EM_SETSEL,
                          IEventParameter1((unsigned long)ipt),
                          IEventParameter2((unsigned long)ipt) );
      handle().sendEvent( EM_REPLACESEL,
                          IEventParameter1(0),
                          IEventParameter2((char*)pszLineText) );
#else
      import(&ipt,strlen(pszLineText),pszLineText);  // import text,
#endif
      if (bTextSel)
        selectRange( selRange );
   } /* endif */

   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::removeLine                                                   |
|                                                                              |
| Deletes the text on the specified line of the MLE.                           |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::removeLine( unsigned long lLineNumber )
{
   unsigned long ipt;

   if (lLineNumber <= numberOfLines())
      ipt = (unsigned long) handle().sendEvent(MLM_CHARFROMLINE,
                                 IEventParameter1(lLineNumber),
                                 IEventParameter2(0));
   else
      return *this;
#ifdef IC_WIN
   IRange  initMarked, newMarked;
   Boolean bMustRestore = true;

   /*******************************************************************/
   /* If selected text exists, save range away to restore later       */
   /*******************************************************************/
   if (hasSelectedText())
   {
     bMustRestore = true;
     initMarked = selectedRange();
     newMarked = initMarked;
   }

   /*******************************************************************/
   /* Get ending point of area to delete                              */
   /*******************************************************************/
   unsigned long ept = (unsigned long)handle().sendEvent(
                                 MLM_CHARFROMLINE,
                                 IEventParameter1(lLineNumber+1),
                                 IEventParameter2(0));

   /*******************************************************************/
   /* Disable refresh while modifications are made to avoid flashing  */
   /*******************************************************************/
   Boolean bOldRefresh=bRefresh;
   if (bRefresh)
      disableUpdate();

   /*******************************************************************/
   /* Modify selection to select the line indicated and delete it     */
   /*******************************************************************/
   handle().sendEvent( EM_SETSEL,
                       IEventParameter1(ipt),
                       IEventParameter2(ept) );
   handle().sendEvent( EM_REPLACESEL,
                       IEventParameter1(0),
                       IEventParameter2((char*)""));

   /*******************************************************************/
   /* If a range was selected before, then restore that range         */
   /*******************************************************************/
   if (bMustRestore)
   {
     /*****************************************************************/
     /* If deleted line is totally inside selected area, shrink range */
     /*****************************************************************/
     if ((initMarked.lowerBound() <= ipt) && (initMarked.upperBound() >= ept))
       newMarked.setUpperBound(initMarked.upperBound() - (ept-ipt));

     /*****************************************************************/
     /* If deleted line is all below selected area, move range down   */
     /*****************************************************************/
     else if (initMarked.lowerBound() >= ept)
       newMarked -= (ept-ipt);

     /*****************************************************************/
     /* If deleted line contains selection start, but not end, change */
     /*  start point to be just after line deleted                    */
     /*****************************************************************/
     else if ( (initMarked.lowerBound() > ipt) &&
               (initMarked.lowerBound() < ept) &&
               (initMarked.upperBound() >= ept) )
     {
       newMarked.setLowerBound(ipt);
       newMarked.setUpperBound(initMarked.upperBound() - (ept-ipt));
     }

     /*****************************************************************/
     /* If deleted line contains selection end, but not start, change */
     /*  end point to be just before line deleted                     */
     /*****************************************************************/
     else if ( (initMarked.lowerBound() < ipt) &&
               (initMarked.upperBound() >= ipt) &&
               (initMarked.upperBound() < ept) )
       newMarked.setUpperBound( ipt-1 );

     /*****************************************************************/
     /* Now, reselect the range, for any of the above conditions or if*/
     /*  range was totally below deletion.  Only if the range wasn't  */
     /*  totally contained in the deleted area should we not reselect */
     /*****************************************************************/
     if (!((initMarked.lowerBound() >= ipt) &&
           (initMarked.upperBound() <= ept)))
       selectRange( newMarked );
   }

   /*******************************************************************/
   /* Restore refresh (if required) after all modifications complete  */
   /*******************************************************************/
   if (bOldRefresh)
     enableUpdate();
#endif
#ifdef IC_PM
   unsigned long ulNumChars;
   ulNumChars = (unsigned long) handle().sendEvent(MLM_QUERYLINELENGTH,
                                  IEventParameter1(ipt),
                                  IEventParameter2(0));

   handle().sendEvent(MLM_DELETE, IEventParameter1(ipt),
                                  IEventParameter2(ulNumChars));
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::setTop                                                       |
|                                                                              |
| Make the specified line be the first visible line of the MLE on the screen.  |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setTop( unsigned long lLineNumber )
{

#ifdef IC_WIN
   unsigned long scrollAmount = lLineNumber - top();

   handle().sendEvent( EM_LINESCROLL,
                       IEventParameter1(0),
                       IEventParameter2(scrollAmount) );
#else
   IEventResult evt(handle().sendEvent(MLM_CHARFROMLINE,
                                       IEventParameter1(lLineNumber),
                                       IEventParameter2(0)));

   IEventResult evt2(handle().sendEvent(MLM_SETFIRSTCHAR,
                                        IEventParameter1(evt.asUnsignedLong()),
                                        IEventParameter2(0)));

   if (!(evt2.asUnsignedLong()))
      ITHROWGUIERROR("MLM_SETFIRSTCHAR");
#endif

   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::top                                                          |
|                                                                              |
| Returns the line number of the line currently displayed as the first visible |
| line on the MLE.                                                             |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::top() const
{
#ifdef IC_WIN
   IEventResult evt2(handle().sendEvent(EM_GETFIRSTVISIBLELINE,
                                        IEventParameter1(0),
                                        IEventParameter2(0)));
#else
   IEventResult evt(handle().sendEvent(MLM_QUERYFIRSTCHAR,
                                       IEventParameter1(0),
                                       IEventParameter2(0)));

   IEventResult evt2(handle().sendEvent(MLM_LINEFROMCHAR,
                                        IEventParameter1(evt.asUnsignedLong()),
                                        IEventParameter2(0)));
#endif
   return evt2.asUnsignedLong();
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setCursorLinePosition                                        |
|                                                                              |
| Position the cursor on the line specified.                                   |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setCursorLinePosition(unsigned long lLineNumber)
{
   IEventResult evt(handle().sendEvent(MLM_CHARFROMLINE,
                                       IEventParameter1(lLineNumber),
                                       IEventParameter2(0)));

   handle().sendEvent(MLM_SETSEL, IEventParameter1(evt.asUnsignedLong()),
                                  IEventParameter2(evt.asUnsignedLong()));
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::cursorLinePosition                                           |
|                                                                              |
| Returns the line number where the cursor is currently positioned.            |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::cursorLinePosition() const
{
   IEventResult evt(handle().sendEvent( MLM_LINEFROMCHAR,
                                        IEventParameter1(cursorPosition()),
                                        IEventParameter2(0)));

   return evt.asUnsignedLong();
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::cursorPosition                                               |
|                                                                              |
| Returns the character position of where the cursor currently is.             |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::cursorPosition() const
{
#ifdef IC_WIN
   unsigned long endPos;
   IEventResult evt(handle().sendEvent(MLM_QUERYSEL,
                                       IEventParameter1(0),
                                       IEventParameter2(&endPos)));
   return endPos;
#else
   IEventResult evt(handle().sendEvent(MLM_QUERYSEL,
                                       IEventParameter1(MLFQS_CURSORSEL),
                                       IEventParameter2(0)));
   return evt.asUnsignedLong();
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::setCursorPosition                                            |
|                                                                              |
| Position the cursor in front of the character specified.                     |
------------------------------------------------------------------------------*/
IMultiLineEdit& IMultiLineEdit::setCursorPosition( unsigned long lCursorChar )
{
   handle().sendEvent(MLM_SETSEL, IEventParameter1(lCursorChar),
                                  IEventParameter2(lCursorChar));
   return *this;
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::visibleLines                                                 |
|                                                                              |
| Returns the number of visible lines that will fit in the MLE using the       |
| current font.                                                                |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::visibleLines() const
{
   long lRetVal=0;

   /*******************************************************************/
   /* Get the font and then the maximum height of one character.      */
   /*******************************************************************/
#ifdef IC_PM
   FATTRS fattrs;

   /*******************************************************************/
   /* In OS/2, must use API to get font since MLE is different        */
   /*******************************************************************/
   handle().sendEvent(MLM_QUERYFONT,
                      IEventParameter1((unsigned long)(PFATTRS(&fattrs))),
                      IEventParameter2(0));

   long lCharHeight = fattrs.lMaxBaselineExt;
#endif
#ifdef IC_WIN
   long lCharHeight = font().maxCharHeight();
#endif
   /*******************************************************************/
   /* Get the MLEs visible edit region                                */
   /*******************************************************************/
   long lRegionHeight = editRegionHeight();

   /*******************************************************************/
   /* If either character height or region is 0, then no visible lines*/
   /*******************************************************************/
   if ((lRegionHeight==0) || (lCharHeight==0))
      return 0;

   /*******************************************************************/
   /* Otherwise number of lines is overall region divided by character*/
   /* height.  Calculate it and then return that value                */
   /*******************************************************************/
   lRetVal = ( lRegionHeight / lCharHeight );
   return lRetVal;
}


/*------------------------------------------------------------------------------
| Private methods                                                              |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| IMultiLineEdit::exportFile                                                   |
|                                                                              |
| Exports either the entire file, or only the marked area, to a file.          |
------------------------------------------------------------------------------*/
long IMultiLineEdit::exportFile( const char* pszFileName,
                                 Boolean     flMarkedAreaOnly )
{
   unsigned long      ulBufSize;        // Buffer size
   long        iptOrigin;               // Origin of the MLE object
   FILE        *filePtr;                // File pointer of specified file
   unsigned long      ulNumChars;       // Number of chars exported by MLE
   int         numItems=1;

   /***********************************************************************/
   /* Open file in binary mode.  If file exists, erase its contents.  If  */
   /* file does not exist, create it for writing or reading.              */
   /* If file cannot be opened for writing (i.e. readonly) return 0 as    */
   /* number of bytes written                                             */
   /***********************************************************************/
   if ((filePtr = fopen(pszFileName, "wb")) == (FILE *)0)
     return 0;

   /***********************************************************************/
   /* Check if we need to export all text or only selected text           */
   /***********************************************************************/
   if (flMarkedAreaOnly == FALSE)
   {
     ulNumChars = ulBufSize  = textLengthAfterFormat();
     iptOrigin = 0;
   }
   /***********************************************************************/
   /* If selected, determine area the user has marked area for exporting. */
   /***********************************************************************/
   else
   {
     IRange rang = selectedRange();
     iptOrigin = rang.lowerBound();
     ulNumChars = ulBufSize = selectedTextLength();
   }

   /***********************************************************************/
   /* Allocate Maximum buffer for the import/export operation.            */
   /***********************************************************************/
   IString pchBuf(0,(unsigned int)ulBufSize+1);

#ifdef IC_PM
   /***********************************************************************/
   /* Export MLE to the specified buffer - it returns number of characters*/
   /* actually copied to the buffer including CR and LF.                  */
   /***********************************************************************/
   ulNumChars = export( &iptOrigin, ulBufSize, (char*)pchBuf );
#else  //IC_WIN
   if (flMarkedAreaOnly == FALSE)
     pchBuf = text();
   else
     pchBuf = selectedText();
   ulNumChars = pchBuf.length();
#endif  //IC_WIN

   /***********************************************************************/
   /* Write number of chars from the buffer to the specified file.        */
   /***********************************************************************/
   if (ulNumChars)
     numItems = fwrite((char*)pchBuf, ulNumChars, 1, filePtr);

   /***********************************************************************/
   /* Write out an end of file character and close the file               */
   /***********************************************************************/
   long endfile=26;
   fwrite((char*) &endfile,1,1,filePtr);
   fclose(filePtr);

   /***********************************************************************/
   /* If the original write failed, then return 0 as count written        */
   /***********************************************************************/
   if (numItems != 1)
     return 0;

   return ulNumChars;                  // return number of chars exported
}

#ifdef IC_PM
/*------------------------------------------------------------------------------
| IMultiLineEdit::export                                                       |
|                                                                              |
| Export the block of text from the insertion point passed into the specified  |
| buffer                                                                       |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::export( long*          ipt,
                                      unsigned long  length,
                                      const char*    buffer )
{
  char* exportBuffer;
  char* tempBuffer = (char*)buffer;
  unsigned long numberToExport, origExport, totalExported=0;

  /********************************************************************/
  /* Allocate the memory to be used to transfer from MLE to buffer    */
  /* MLE can only export in chunks of less than 64K                   */
  /********************************************************************/
  if ( DosAllocMem( (PPVOID)&exportBuffer, 63005,
                     PAG_WRITE | PAG_COMMIT | OBJ_TILE ))
    ITHROWGUIERROR("DosAllocMem");

  /********************************************************************/
  /* Point the export to our recently allocated block of temp storage */
  /********************************************************************/
  handle().sendEvent( MLM_SETIMPORTEXPORT,
                      IEventParameter1(exportBuffer),
                      IEventParameter2(63000));

  /********************************************************************/
  /* As long as there is text to export, do the following             */
  /********************************************************************/
  while (length != 0)
  {
    /******************************************************************/
    /* If remainder is less than largest buffer, use it as export size*/
    /******************************************************************/
    numberToExport = 63000;
    if ( length < 63000 )
      numberToExport = length;

    /******************************************************************/
    /* Save original export size since MLM_EXPORT modifies it and then*/
    /* export the text into the buffer                                */
    /******************************************************************/
    origExport = numberToExport;
    IEventResult evt = handle().sendEvent( MLM_EXPORT,
                                   IEventParameter1(ipt),
                                   IEventParameter2(&numberToExport));

    /******************************************************************/
    /* Now copy the data out of the temp buffer to the final buffer   */
    /******************************************************************/
    memcpy( tempBuffer, exportBuffer, origExport );

    /******************************************************************/
    /* Update external pointer to where to write next block of text   */
    /******************************************************************/
    tempBuffer += origExport;

    /******************************************************************/
    /* Modify counts of text left to export and total exported        */
    /******************************************************************/
    length -= origExport;
//  totalExported += evt.asUnsignedLong();  // Bug in PM causes the return
                                            // value to be one less than
                                            // actual number exported
    totalExported += origExport;            // so use original requested #
  } /* endwhile */
  /********************************************************************/
  /* Free the memory allocated above                                  */
  /********************************************************************/
  DosFreeMem( exportBuffer );

  /********************************************************************/
  /* Return count of text exported                                    */
  /********************************************************************/
  return totalExported;
}


/*------------------------------------------------------------------------------
| IMultiLineEdit::import                                                       |
|                                                                              |
| Import the specified block of text from a buffer into an MLE at the          |
| desired insertion point.                                                     |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::import( long*         ipt,
                                      unsigned long length,
                                      const char*   buffer )
{
  char* importBuffer;
  char* tempBuffer = (char*)buffer;
  unsigned long numberToImport=63000, totalImported=0;

  /********************************************************************/
  /* Verify that we will not exceed the currently set limit on the    */
  /* MLE before trying to import text                                 */
  /********************************************************************/
  if (length+textLength() <= limit())
  {
     /*****************************************************************/
     /* Allocate the memory to be used to transfer from buffer to MLE */
     /* MLE can only import in chunks of less than 64K that are       */
     /* segment aligned                                               */
     /*****************************************************************/
     if ( DosAllocMem( (PPVOID)&importBuffer, 63005,
                       PAG_WRITE | PAG_COMMIT | OBJ_TILE ))
       ITHROWGUIERROR("DosAllocMem");

     /*****************************************************************/
     /* Set up the buffer to import from to the newly allocated one   */
     /*****************************************************************/
     handle().sendEvent( MLM_SETIMPORTEXPORT,
                         IEventParameter1(importBuffer),
                         IEventParameter2(63005));

     /*****************************************************************/
     /* As long as there is data to import, bring it in               */
     /*****************************************************************/
     while (length != 0)
     {
        /**************************************************************/
        /* If smaller than the buffer, change from max to smaller val */
        /**************************************************************/
        if ( length < 63000 )
          numberToImport = length;

        /**************************************************************/
        /* Copy from the external buffer to the segment aligned buffer*/
        /**************************************************************/
        memcpy(importBuffer,tempBuffer,numberToImport);

        /**************************************************************/
        /* Update the external pointer to the next block of text      */
        /**************************************************************/
        tempBuffer += numberToImport;

        /**************************************************************/
        /* Import the text from the segment aligned block to the MLE  */
        /**************************************************************/
        IEventResult evt = handle().sendEvent( MLM_IMPORT,
                                      IEventParameter1(ipt),
                                      IEventParameter2(numberToImport));

        /**************************************************************/
        /* Modify counts of text left to import and total imported    */
        /**************************************************************/
        length -= numberToImport;
        totalImported += evt.asUnsignedLong();
     }

     /*****************************************************************/
     /* Free the memory allocated above                               */
     /*****************************************************************/
     DosFreeMem( importBuffer );
  }
  /********************************************************************/
  /* If limit exceeded, than throw exception stating this             */
  /********************************************************************/
  else
  {
     ITHROWLIBRARYERROR( IC_INVALIDSIZE_LIMIT,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable );
  } /* endif */
  /********************************************************************/
  /* Return count of text imported                                    */
  /********************************************************************/
  return totalImported;
}
#endif

/*------------------------------------------------------------------------------
| IMultiLineEdit::textLength                                                   |
|                                                                              |
| Returns the current length in bytes (containing CR/LF) of the MLE text.      |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::textLength() const
{
#ifdef IC_WIN
   return Inherited::textLength();
#else
   IEventResult evt(handle().sendEvent(MLM_QUERYTEXTLENGTH,
                                       IEventParameter1(0),
                                       IEventParameter2(0)));
   return evt.asUnsignedLong();
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::textLengthAfterFormat                                        |
|                                                                              |
| Returns current length in bytes (containing CR/LF) of the MLE text.          |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::textLengthAfterFormat() const
{
#ifdef IC_WIN
   // Approximate format count by adding an additional CR per line
   //  since we cannot query current format
   return (textLength() + numberOfLines());
#else
   IEventResult evt(handle().sendEvent(MLM_QUERYFORMATTEXTLENGTH,
                                 IEventParameter1(0),
                                 IEventParameter2((unsigned long) 0xFFFFFFFF)));
   return evt.asUnsignedLong();
#endif
}

/*------------------------------------------------------------------------------
| IMultiLineEdit::selectedTextLength                                           |
|                                                                              |
| Returns the size of the user marked area in bytes (containing CR/LF). Does   |
| not count '\0'.                                                              |
------------------------------------------------------------------------------*/
unsigned long IMultiLineEdit::selectedTextLength() const
{
#ifdef IC_WIN
   unsigned long startPt;                          // Start point of marked area
   unsigned long endPt;                            // End point of marked area

   handle().sendEvent( EM_GETSEL,
                       IEventParameter1( &startPt ),
                       IEventParameter2( &endPt ) );
#else
   IEventResult minSel( handle().sendEvent(MLM_QUERYSEL,
                                           IEventParameter1(MLFQS_MINSEL),
                                           IEventParameter2(0)));
   IEventResult maxSel( handle().sendEvent(MLM_QUERYSEL,
                                           IEventParameter1(MLFQS_MAXSEL),
                                           IEventParameter2(0)));
   unsigned long startPt(minSel.asUnsignedLong());   // Start point of marked area
   unsigned long endPt(maxSel.asUnsignedLong());     // End point of marked area
#endif

   if (startPt == endPt)
     return 0;
   else
   {
#ifdef IC_WIN
   /*******************************************************************/
   /* Get character index for start point and end point and subtract  */
   /* them.  Assume byte count for now by using 1 for each character  */
   /*******************************************************************/
   return (endPt - startPt);
#else
      IEventResult evt(handle().sendEvent(MLM_QUERYFORMATTEXTLENGTH,
                                          IEventParameter1(startPt),
                                          IEventParameter2(endPt - startPt)));
      return evt.asUnsignedLong();
#endif
   }
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IMultiLineEdit::isDragStarting                                               |
|                                                                              |
| Windows does not generate drag messages, so therefore it is up to a          |
| a control to decide if the conditions exist to start a drag. A drag          |
| occurs if text is selected in the edit control and a button down occurs      |
| whose position is within the selected text rectangle.                        |
|                                                                              |
| Note: if you don't want the event to continue up the owner chain,            |
|       then set the event result to true.                                     |
------------------------------------------------------------------------------*/
Boolean IMultiLineEdit::isDragStarting( IEvent &event )
{
  Boolean brc = false;
  switch ( event.eventId() )
    {
    case WM_LBUTTONDOWN:
      {
      ITRACE_DEVELOP("Got a button down, start to look for drag");
      // only allow drag operation if text selected
      if (hasSelectedText())
        {
        if (fMultiLineEditData->pointInSelectedText(event))
          {
          IMouseClickEvent mevt( event );
          ITRACE_DEVELOP("Mouse is in selected range");
          ITRACE_DEVELOP("We have a drag" );
          // Instead of setting the return code to true, we will start the
          // drag and drop from here. This is so the text is selected when
          // the drag is occurring.
          event.window()->sendEvent( WM_BEGINDRAG,
               IEventParameter1((unsigned short)mevt.mousePosition().x(),
                                (unsigned short)mevt.mousePosition().y()),
               IEventParameter2( 0 ));
          selectRange(IRange(-1,0));
          }
        }
      break;
      }
    case WM_RBUTTONDOWN:
      {
      ITRACE_DEVELOP("Got a right button down, start to look for drag");
      // only allow drag operation if text selected
      if (hasSelectedText())
        {
        if (fMultiLineEditData->pointInSelectedText(event))
          {
          ITRACE_DEVELOP("Mouse is in selected range");
          if ( ::GetCapture() == NULL )
            {
            IMouseClickEvent mevt( event );
            ::SetCapture( event.handle());
            fMultiLineEditData->bCheckingRightDragStart = True;
            fMultiLineEditData->pointRButtonDown = mevt.mousePosition();
            }
          }
        }
      break;
      }
    case WM_RBUTTONUP:
      {
      if ( fMultiLineEditData->bCheckingRightDragStart )
        {
        ::ReleaseCapture();
        SetCapture( NULL );
        fMultiLineEditData->bCheckingRightDragStart = False;
        fMultiLineEditData->pointRButtonDown = IPoint(0,0);
        }
      break;
      }
    case WM_MOUSEMOVE:
      {
      if ( fMultiLineEditData->bCheckingRightDragStart )
        {
        ITRACE_DEVELOP("Mouse move while checking drag start");
        IMouseClickEvent mevt( event );
        IPoint mPoint( fMultiLineEditData->pointRButtonDown );

        if ( abs(mevt.mousePosition().x() - mPoint.x()) >
                 DD_DEFDRAGMINDIST ||
             abs(mevt.mousePosition().y() - mPoint.y()) >
                 DD_DEFDRAGMINDIST )
          {
          /*-----------------------------------------
          | We are going to start the drag from here instead
          | of returning true, because we need to control
          | the lifetime of the text being selected.
          -----------------------------------------*/
          ITRACE_DEVELOP("Starting a drag" );
          ::ReleaseCapture();
          fMultiLineEditData->bCheckingRightDragStart = FALSE;
          event.window()->sendEvent( WM_BEGINDRAG,
             IEventParameter1((unsigned short)mevt.mousePosition().x(),
                              (unsigned short)mevt.mousePosition().y()),
             IEventParameter2( 0 ));
          selectRange(IRange(-1,0));
          }
        }
      break;
      }
    }
  return brc;
}
#endif

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IMultiLineEditData::pointInSelectedText                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IMultiLineEditData::pointInSelectedText(const IEvent& evt)
{
  IMODTRACE_DEVELOP("IDMSourceHandlerData::pointInSelectedText");
  // Get the handle to the edit control
  IMultiLineEdit *pMLECntrl = (IMultiLineEdit *)evt.window();

  RECT selRect;
  POINT pt;
  SIZE sizeSel;
  sizeSel.cx = sizeSel.cy = 0;
  IBase::Boolean bPtinSel = false;

  IMouseClickEvent mevt( evt );

  // Get mouse position when button down
  IPoint pmPt = mevt.mousePosition();

  // convert if necessary.
  if ( ICoordinateSystem::applicationOrientation() ==
       ICoordinateSystem::originLowerLeft )
    {
    pmPt.setY(evt.window()->size().height() - mevt.mousePosition().y());
    }

  pt.x = pmPt.x();
  pt.y = pmPt.y();

  ITRACE_DEVELOP("Mouse position x: " + IString(pt.x));
  ITRACE_DEVELOP("Mouse position y: " + IString(pt.y));

  // get the range of the selected text
  IRange range = pMLECntrl->selectedRange();

  if ((IPlatform::isWin95()) || (IPlatform::isNTNewShell()))
    {
    // Retrieve the zero-based character index of the character nearest the
    // mouse point on the button down.
    // Note:  This message is currently supported only in Windows 95 and NT Newshell
    IEventResult charEvt = pMLECntrl->handle().sendEvent(EM_CHARFROMPOS, 0,
                                                IEventData(pmPt.x(), pmPt.y()));

    long position = (long)charEvt.number1();

    ITRACE_DEVELOP("Character position of mouse in MLE: " + IString(position));

    // Return true if the mouse position is within the selected range
    return (range.includes(position));
    }

  HDC hDC = GetDC(pMLECntrl->handle());
  TEXTMETRIC tm;
  GetTextMetrics(hDC, &tm);

  // initialize rectangles
  selRect.left = selRect.right = selRect.top = 0;
  // bottom of rectangle is height of first line;
  long lineHeight = tm.tmHeight + tm.tmExternalLeading;
  selRect.bottom = lineHeight;

  // get the line number of the first visible line
  unsigned long top = pMLECntrl->top();

  unsigned long lineNum = 0, accumLength = 0;
  IString text;

  // The accumulated length has to account for not only the visible portion of
  // text in the MLE, but all text in the mle leading up the to the selected
  // text.  This is necessary to accurately compute the position of the
  // selected text based on the offsets from the range.
  while (lineNum < top)
    {
    text = pMLECntrl->text(lineNum);
    accumLength += text.length();

    // Only add additional length if not the first line in the MLE.
    // selectedRange() appears to add two additional characters for CRLF
    // per line, so we have to adjust the accumulated length accordingly.
    if (lineNum)
      accumLength += 2;

    ITRACE_DEVELOP("accumLength: " + IString(accumLength));
        lineNum++;
        }

  // get the number of visible lines.  Subtract for zero based index
  unsigned long numVisible = pMLECntrl->visibleLines()-1;
  ITRACE_DEVELOP("Number of visible lines (visible lines -1) " + IString(numVisible));

  // get the text for the first visible line
  text = pMLECntrl->text(top);
  ITRACE_DEVELOP("1st visible line of text: " + text);
  IString selText = pMLECntrl->selectedText();

  // update accumulated text length to account for current line
  accumLength += text.length();

  // Only add additional length if not the first line in the MLE.
  // selectedRange() appears to add two additional characters for CRLF
  // per line, so we have to adjust the accumulated length accordingly.
  if (lineNum)
    accumLength += 2;

  ITRACE_DEVELOP("accumLength: " + IString(accumLength));

  ITRACE_DEVELOP("Line num: " + IString(lineNum));
  // Find the first line with selected text
  while ( (accumLength < range.lowerBound()) && (lineNum < (top + numVisible)))
    {
    lineNum++;
    ITRACE_DEVELOP("Line num: " + IString(lineNum));
    text = pMLECntrl->text(lineNum);
    accumLength += text.length()+2;
    ITRACE_DEVELOP("accumLength " + IString(accumLength));

    // Update top and bottom of rectangle for each new line
    selRect.top = selRect.bottom+1;
    selRect.bottom+=lineHeight;
    }

  // 1st visible line may be partially selected
  // handle as a special case

  // Determine the starting point of the selected text, which is either the
  // lower bound of the range or the beginning of the line (zero).  The only
  // case where it is zero is if there is text selected above the topmost
  // visible line
  ITRACE_DEVELOP("1st line with selected text: " + text);
  long lineStartPos = accumLength - text.length();
  ITRACE_DEVELOP("line start pos: " + IString(lineStartPos));

  // Calculate the selected text start position within the initial line
  // NOTE:  There is a need to determine the position relative to a line
  // and relative to the entire selected text range (which may be prior to
  // the current line). selTextStart is relative to the given line.
  long selTextStart = (range.lowerBound() < lineStartPos) ? 0 :
                      range.lowerBound() - lineStartPos;
  ITRACE_DEVELOP("selTextStart: " + IString(selTextStart));
  // get length of selected text
  long lenSelText = (long)( (accumLength >= range.upperBound()) ?
                            ( (range.upperBound()-(accumLength-text.length())) - selTextStart + 1 ) :
                            ( text.length() - selTextStart + 1 ) );

  ITRACE_DEVELOP("lenSelText value " + IString(lenSelText));

  ITRACE_DEVELOP("Range lower bound: " + IString(range.lowerBound()));
  ITRACE_DEVELOP("Range upper bound: " + IString(range.upperBound()));

  // calcute the relative position of the selected text starting point
  // for the 1st visible line with selected text.
  long selRangeStartPos = (range.lowerBound() < lineStartPos) ?
                      lineStartPos - range.lowerBound() : 0;
  IString topSelText = selText.subString((unsigned)selRangeStartPos,
                                         (unsigned)max(0,lenSelText));
  ITRACE_DEVELOP("Selected text on first line: " + topSelText);

  SIZE sizeLead;
  sizeLead.cx = sizeLead.cy = 0;

  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s: Call GetTextExtentPoint not GetTextExtentPoint32");
    GetTextExtentPoint(hDC, topSelText, topSelText.length(), &sizeSel);
  }
  else
  {
    GetTextExtentPoint32(hDC, topSelText, topSelText.length(), &sizeSel);
  }

  IString leadingText("");
  if (range.lowerBound() > 0)
    {
    // Calculate the text on the line up to where selection starts, which
    // is the accumlength of previous lines to the beginning of selected range
    leadingText = text.subString(0,
                      (unsigned)max(0,(long)(range.lowerBound()-(accumLength-text.length()))));
    ITRACE_DEVELOP("Leading text: " + leadingText);
    }

  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s: Call GetTextExtentPoint not GetTextExtentPoint32");
    GetTextExtentPoint(hDC, leadingText, leadingText.length(), &sizeLead);
  }
  else
  {
    GetTextExtentPoint32(hDC, leadingText, leadingText.length(), &sizeLead);
  }

  // update the left and right positions to equate to the selection rectangle
  selRect.left = sizeLead.cx;
  selRect.right = sizeLead.cx + sizeSel.cx;

  ITRACE_DEVELOP("Selected rectangle (left,right,top,bottom)");
  ITRACE_DEVELOP("\tLeft: " + IString(selRect.left));
  ITRACE_DEVELOP("\tRight: " + IString(selRect.right));
  ITRACE_DEVELOP("\tTop: " + IString(selRect.top));
  ITRACE_DEVELOP("\tBottom: " + IString(selRect.bottom));

  // If the mouse is in the selected rectangle
  if (PtInRect(&selRect, pt))
    bPtinSel = true;

  IBase::Boolean bSelTextProcessed = false;

  // if point not in current rectangle and end of selected text not reached
  if (!bPtinSel && accumLength < range.upperBound())
    {
    lineNum++;
    // Get the next line of text
    text = pMLECntrl->text(lineNum);
    accumLength += text.length()+2;
    ITRACE_DEVELOP("accumLength: " + IString(accumLength));

    // Update top and bottom of rectangle for each new line
    selRect.top = selRect.bottom+1;
    selRect.bottom+=lineHeight;

    // Now handle case of middle lines where all the text is selected
    while (!bPtinSel && (lineNum < (top + numVisible)) && (accumLength < range.upperBound()))
      {
      // Left always 0 since all of the line is selected
      selRect.left = 0;

      ITRACE_DEVELOP("next line of text: " + text);
      if (IPlatform::isWin32S())
      {
        ITRACE_DEVELOP("Win32s: Call GetTextExtentPoint not GetTextExtentPoint32");
        GetTextExtentPoint(hDC, text, text.length(), &sizeSel);
      }
      else
      {
        GetTextExtentPoint32(hDC, text, text.length(), &sizeSel);
      }
      selRect.right = sizeSel.cx;

      ITRACE_DEVELOP("Selected rectangle (left,right,top,bottom)");
      ITRACE_DEVELOP("\tLeft: " + IString(selRect.left));
      ITRACE_DEVELOP("\tRight: " + IString(selRect.right));
      ITRACE_DEVELOP("\tTop: " + IString(selRect.top));
      ITRACE_DEVELOP("\tBottom: " + IString(selRect.bottom));

      if (PtInRect(&selRect, pt))
        bPtinSel = true;

      lineNum++;
      // Get the next line of text
      text = pMLECntrl->text(lineNum);
      accumLength += text.length()+2;
      ITRACE_DEVELOP("accumLength: " + IString(accumLength));

      // Update top and bottom of rectangle for each new line
      selRect.top = selRect.bottom+1;
      selRect.bottom+=lineHeight;
      }
    }
  else
    bSelTextProcessed = true;

  if (!bPtinSel && !bSelTextProcessed)
    {
    // Last line is all that remains.  Special case because the last line
    // could be partially selected.
    selRect.left = 0;

    SIZE sizeTrail;
    sizeTrail.cx = sizeTrail.cy = 0;

    ITRACE_DEVELOP("Last line of text: " + text);
    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: Call GetTextExtentPoint not GetTextExtentPoint32");
      GetTextExtentPoint(hDC, text, text.length(), &sizeSel);
    }
    else
    {
      GetTextExtentPoint32(hDC, text, text.length(), &sizeSel);
    }

    IString trailingText("");
    // Get the trailing text - which is end of selected text to the end
    // of the last line
    trailingText = text.subString((unsigned)max(0,(long)(range.upperBound()-(accumLength-text.length()))));
    ITRACE_DEVELOP("Trailing text: " + trailingText);

    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: Call GetTextExtentPoint not GetTextExtentPoint32");
      GetTextExtentPoint(hDC, trailingText, trailingText.length(), &sizeTrail);
    }
    else
    {
      GetTextExtentPoint32(hDC, trailingText, trailingText.length(), &sizeTrail);
    }

    // set right side of rectangle
    selRect.right = max(0,sizeSel.cx-sizeTrail.cx);

    ITRACE_DEVELOP("Selected rectangle (left,right,top,bottom)");
    ITRACE_DEVELOP("\tLeft: " + IString(selRect.left));
    ITRACE_DEVELOP("\tRight: " + IString(selRect.right));
    ITRACE_DEVELOP("\tTop: " + IString(selRect.top));
    ITRACE_DEVELOP("\tBottom: " + IString(selRect.bottom));

    PtInRect(&selRect, pt) ? bPtinSel = true : bPtinSel = false;
    }

  // release device context
  ReleaseDC(pMLECntrl->handle(), hDC);
  return bPtinSel;
}
#endif

