/* NOSHIP */
#ifndef _ISLIDERW_
   #define _ISLIDERW_
/*******************************************************************************
* FILE NAME: isliderw.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*                                                                              *
* Implementation Notes:                                                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <icolor.hpp>
#include <icontrol.hpp>
#include <igbitmap.hpp>
#include <iglist.hpp>
#include <igstring.hpp>
#include <ihandler.hpp>
#include <islider.hpp>
#include <itimer.hpp>
extern "C"
{
  #include <iwindefs.h>
  #include <commctrl.h>
}
#pragma info(none)
  #include <iset.h>
#pragma info(restore)

class INativeSliderHandler;


/*----------------------------------------------------------------------------*/
/* Align classes on four byte boundary.                                       */
/*----------------------------------------------------------------------------*/
#pragma pack(4)
#pragma info(none)

typedef ISet<unsigned long> TickSet;

/*------------------------------------------------------------------------------
| Private data class for IProgressIndicator                                    |
|                                                                              |
| This class contains functions and data necessary for the use of the Windows  |
| trackbar in the IProgressIndicator class.                                    |
------------------------------------------------------------------------------*/
class IProgressIndicatorData : public IBase {
public:
  IProgressIndicatorData();
  ~IProgressIndicatorData();

IWindowHandle
  trackbar;
unsigned long
  scaleFactor,
  tickPos0,
  numberTicks[2],
  tickSpacing[2],
  pixelRange;
TickSet
  visibleTicks[2];
IPoint
  shaftPosition;
IGList
  tickText[2],
  buttons;
Boolean
  textVisible;
static Boolean
  fRegistered;

IGBitmap
  *leftDown,
  *leftDownPressed,
  *leftDownDisabled,
  *rightUp,
  *rightUpPressed,
  *rightUpDisabled;

ITimer
  *timer;

WNDPROC
  fdefaultTrackbarProc;

static void
  registerSlider();

void
  createNativeSlider( const IWindowHandle& parent,
                      const IWindowHandle& owner,
                      unsigned long        ulStyle,
                      unsigned long        extStyle ),
  destroyNativeSlider( ),
  setupNativeScale( IProgressIndicator::Scale scale ),
  layoutNativeSlider( ),
  drawRibbon( unsigned long newPosition ),
  setButtonState( ),
  buttonPressed( IGBitmap *button ),
  buttonReleased( ),
  startTimer( ),
  cancelTimer( ),
  moveArm( );

unsigned long
  armRangeOffset( ),
  trackbarRange( );

IProgressIndicatorData
  &setArmRangeOffset( unsigned long newPosition );

private:
IProgressIndicatorData
   ( const IProgressIndicatorData& );
IProgressIndicatorData
 &operator=     ( const IProgressIndicatorData& );
}; // IProgressIndicatorData


class INativeSliderHandler : public IHandler
{
typedef IHandler
  Inherited;
/*******************************************************************************
* The INativeSliderHandler class implements a default handler for native       *
* Windows sliders.                                                             *
*******************************************************************************/
public:

/*------------------------ Constructors ----------------------------------------
| The only way to construct an instance of this class is to use the default.   |
------------------------------------------------------------------------------*/
  INativeSliderHandler( );
 ~INativeSliderHandler( );

protected:
/*----------------------------- Implementation ---------------------------------
| This function is used to implement the class:                                |
|   dispatchHandlerEvent - Routes the size/trackbar events to the trackbar.    |
------------------------------------------------------------------------------*/
virtual Boolean
  dispatchHandlerEvent  ( IEvent& event       );
};


/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointed               |
| to object when the wrapper is destructed.                                    |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator INativeSliderHandler* ();
INativeSliderHandler
 *operator->()
    {
    return *this;
    }
  ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
INativeSliderHandler
 *ptr;
} defaultHandler;

#pragma pack()
#endif  /* _ISLIDERW_ */
