/*******************************************************************************
* FILE NAME: itemid.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     ItemID                                                                   *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define UNICODE
extern "C"
{
  #define IUSING_OLE
  #ifdef WIN32_LEAN_AND_MEAN
    #undef WIN32_LEAN_AND_MEAN
  #endif
  #ifdef _OLE2_H_
    #undef _OLE2_H_
  #endif
  #include <iwindefs.h>
  #include <shlobj.h>
}

#include "itemid.hpp"

#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#pragma data_seg(ICLNonConst)
static HIMAGELIST sysImageList = 0;
#pragma data_seg()


/*------------------------------------------------------------------------------
| ItemID::ItemID                                                               |
|                                                                              |
| Constructor.                                                                 |
|   Creates an ItemID of the size cb, it automatically adds the null           |
|   terminator at the end of the ItemID.                                       |
------------------------------------------------------------------------------*/
ItemID::ItemID(unsigned int cbSize)
  : thePidl( 0 )
  , lpMalloc(0)
{
  int cbTotal = cbSize + sizeof(thePidl->mkid.cb ); // allow for null terminator
  if (SUCCEEDED(SHGetMalloc(&lpMalloc) ))
  {
    thePidl=(LPITEMIDLIST)lpMalloc->Alloc( cbTotal);
    if (thePidl)
    {
      _fmemset(thePidl, 0, cbTotal);
      thePidl->mkid.cb = cbSize;
    }
  }
  else
    lpMalloc = 0;
}

/*------------------------------------------------------------------------------
| ItemID::ItemID                                                               |
|                                                                              |
| Constructor.                                                                 |
|   Makes a copy of the pidl.                                                  |
------------------------------------------------------------------------------*/
ItemID::ItemID( LPITEMIDLIST pidl)
  : thePidl( 0 )
  , lpMalloc(0)

{
  if ( SUCCEEDED(SHGetMalloc(&lpMalloc) ))
  {
    thePidl = pidl;
    if ( thePidl)
    {
      int cbTotal = sizeof(thePidl->mkid.cb);
      LPSTR lpMem;
      while ( thePidl->mkid.cb )
      {
        cbTotal += thePidl->mkid.cb;
        LPSTR lpMem = (LPSTR)thePidl;

        lpMem += thePidl->mkid.cb;

        thePidl = (LPITEMIDLIST) lpMem;
      }
      thePidl=(LPITEMIDLIST)lpMalloc->Alloc( cbTotal );
      CopyMemory( thePidl, pidl, cbTotal );
    }
  }
}

/*------------------------------------------------------------------------------
| ItemID::ItemID                                                               |
|                                                                              |
| Constructor.                                                                 |
|   Copy Constructor.                                                          |
------------------------------------------------------------------------------*/
ItemID::ItemID( const ItemID &other )
{
  if (SUCCEEDED(SHGetMalloc(&lpMalloc)))
  {
    if ( other.thePidl )
    {
      unsigned int cbSize = other.size();
      thePidl=(LPITEMIDLIST)lpMalloc->Alloc( cbSize);

      if (thePidl)
      {
        CopyMemory( thePidl, other.thePidl, cbSize );
      }
    }
    else
      thePidl = NULL;
  }
  else
    lpMalloc = 0;
}

/*------------------------------------------------------------------------------
| ItemID::ItemID                                                               |
|                                                                              |
| Constructor.                                                                 |
|   Default Constructor                                                        |
------------------------------------------------------------------------------*/
ItemID::ItemID( )
  : thePidl( 0 )
  , lpMalloc(0)
{
  if (!SUCCEEDED(SHGetMalloc(&lpMalloc) ))
    lpMalloc = 0;
}

/*------------------------------------------------------------------------------
| ItemID::ItemID                                                               |
|                                                                              |
| Destructor.                                                                 |
------------------------------------------------------------------------------*/
ItemID::~ItemID( )
{
  if (lpMalloc)
  {
    if ( thePidl)
      lpMalloc->Free( thePidl );
    lpMalloc->Release();
  }
}

/*------------------------------------------------------------------------------
| ItemID::operator=                                                            |
|                                                                              |
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
ItemID& ItemID::operator= ( const ItemID& other )
{
  if (&other != this)
  {
    if (thePidl && lpMalloc )
      lpMalloc->Free( thePidl );
    unsigned int cbSize = other.size();

    thePidl=(LPITEMIDLIST)lpMalloc->Alloc( cbSize);

    if (thePidl)
      CopyMemory( thePidl, other.thePidl, cbSize );

  }
  return *this;
}

/*------------------------------------------------------------------------------
| ItemID::size                                                              |
|                                                                              |
| Returns the size of the ItemID. The size includes the null terminator.       |
------------------------------------------------------------------------------*/
unsigned int ItemID::size( ) const

{
  LPITEMIDLIST pidl;
  unsigned int cbTotal = sizeof(thePidl->mkid.cb);

  pidl = thePidl;

  if ( pidl)
  {
    LPSTR lpMem;
    while (pidl->mkid.cb)
    {
      cbTotal += pidl->mkid.cb;
      LPSTR lpMem = (LPSTR)pidl;

      lpMem += pidl->mkid.cb;

      pidl = (LPITEMIDLIST) lpMem;
      }
    }
  return cbTotal;
}


/*------------------------------------------------------------------------------
| ItemID::concatItemIDs                                                        |
|                                                                              |
| Concatenates two ItemIDs returning a new Item.                               |
------------------------------------------------------------------------------*/
ItemID ItemID::concatItemIDs( ItemID& itemID )
{
  unsigned int cb1, cb2;

  if ( itemID.thePidl)  //May be NULL
    cb1 = itemID.size( ) - sizeof( itemID.thePidl->mkid.cb);
  else
    cb1 = 0;

  cb2 = size( );

  ItemID newItemId(cb1 + cb2);
  if ( itemID.thePidl )
    memcpy( newItemId.asLPITEMIDLIST(), itemID.thePidl, cb1);
  memcpy(( (LPSTR) newItemId.asLPITEMIDLIST() ) + cb1, thePidl, cb2);

  return newItemId;
}


/*------------------------------------------------------------------------------
| ItemID::operator+=                                                            |
|                                                                              |
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
ItemID& ItemID::operator+= ( const ItemID& nextItem )
{
  unsigned int cb1, cb2;

  cb1 = size( ) - sizeof( thePidl->mkid.cb);

  if ( nextItem.thePidl)  //May be NULL
    cb2 = nextItem.size( );
  else
    cb2 = sizeof( thePidl->mkid.cb);


  LPITEMIDLIST tempPidl= (LPITEMIDLIST)lpMalloc->Alloc( cb1 + cb2 );


  memcpy( (LPSTR) tempPidl      , thePidl, cb1);
  memcpy( (LPSTR) tempPidl + cb1, nextItem.thePidl, cb2 );

  lpMalloc->Free( thePidl );
  thePidl = tempPidl;

  return *this;
}


/*------------------------------------------------------------------------------
| ItemID::asLPITEMIDLIST                                                       |
|                                                                              |
| Returns pidl                                                                 |
------------------------------------------------------------------------------*/
LPITEMIDLIST ItemID::asLPITEMIDLIST( )
{
  return thePidl;
}

/*------------------------------------------------------------------------------
| ItemID::name                                                              |
|                                                                              |
| Returns the display name of the item id.                                     |
------------------------------------------------------------------------------*/
IString ItemID::name(IShellFolder* lpsf,
                                      DWORD         dwFlags )
{
  STRRET Str;

  IString displayName;
  if (SUCCEEDED( lpsf->GetDisplayNameOf( thePidl,
                                         dwFlags,
                                         &Str)) )
  {
    switch (Str.uType)
    {
      case STRRET_WSTR:
      {
        int cch = WideCharToMultiByte(CP_ACP, 0,
                    Str.pOleStr, -1, NULL, 0, NULL, NULL);
        LPSTR lpsz = (LPSTR) lpMalloc->Alloc( cch );

        if (lpsz != NULL )
        {
          int cch2 = WideCharToMultiByte( CP_ACP, 0,
                    Str.pOleStr, -1, lpsz, cch, NULL, NULL );
          if (cch2 == 0)
          {
            DWORD lastError = GetLastError();
          }
          displayName = IString( (char *) lpsz);

          lpMalloc->Free( lpsz );
        }
        break;
      }

      case STRRET_OFFSET:
      {
        displayName = IString( (char *)((char *) thePidl) + Str.uOffset);
        break;
      }

      case STRRET_CSTR:
      {
        displayName = IString( (char *)Str.cStr );
        break;
      }
    }
  }
  return displayName;
}

/*------------------------------------------------------------------------------
| ItemID::fullyQualifiedPidl                                                     |
|                                                                              |
| Returns an fully qualified pidl for the item id.                             |
------------------------------------------------------------------------------*/
ItemID ItemID::fullyQualifiedPidl(LPSHELLFOLDER lpsf )
{
  OLECHAR szOleChar[MAX_PATH];
  LPSHELLFOLDER lpsfDeskTop;
  LPITEMIDLIST  lpifq;
  ULONG ulEaten, ulAttribs;
  HRESULT hr;

  IString lname(name(lpsf, SHGDN_FORPARSING));

  if ( SUCCEEDED(SHGetDesktopFolder(&lpsfDeskTop) ))
  {
    MultiByteToWideChar(CP_ACP,
                        MB_PRECOMPOSED,
                        (char *) lname,
                                                      -1,
                        szOleChar,
                        sizeof(szOleChar));

    if ( FAILED(lpsfDeskTop->ParseDisplayName( NULL, NULL,
                                                  szOleChar,
                                                  &ulEaten, &lpifq,
                                                  &ulAttribs) ))
      lpifq = NULL;
    lpsfDeskTop->Release( );
  }
  return ItemID( lpifq );
}

/*------------------------------------------------------------------------------
| ItemID::attributes                                                           |
|                                                                              |
| Gets the attributes of the itemId.                                           |
------------------------------------------------------------------------------*/
HRESULT ItemID::attributes( LPSHELLFOLDER lpsf,
                         int itemNumber,
                         unsigned long &ulAttrs)
{

  unsigned long temp = ulAttrs;
  HRESULT hr = lpsf->GetAttributesOf( itemNumber,
                                      (LPCITEMIDLIST *)&thePidl,
                                      &temp);
  ulAttrs = temp;
  return hr;
}


/*------------------------------------------------------------------------------
| ItemID::nameFromDesktop                                                      |
|                                                                              |
| Returns full name relative to the desktop.                                   |
------------------------------------------------------------------------------*/
IString ItemID::nameFromDesktop( )
{
  IString fullName;
  LPSHELLFOLDER pFolder;
  if (SUCCEEDED(SHGetDesktopFolder( &pFolder)))
  {
    fullName = name( pFolder, SHGDN_FORPARSING );
  }
  return fullName;
}

/*------------------------------------------------------------------------------
| ItemID::itemIDFromDisplayName                                                |
|                                                                              |
| Returns an ItemID for a fully qualified path name.                           |
------------------------------------------------------------------------------*/
ItemID ItemID::itemIDFromDisplayName( const IString &displayName )
{
  ITRACE_DEVELOP("ItemID::itemIDFromDisplayName");
  LPSHELLFOLDER pFolder;
  ItemID temp;
  if (SUCCEEDED(SHGetDesktopFolder( &pFolder)))
  {
    TCHAR dNameW[MAX_PATH];
    int cch = MultiByteToWideChar( CP_ACP,
                         0,
                         (char *) displayName,
                         displayName.length() + 1,  // include null
                         dNameW,
                         MAX_PATH );
    if (cch)
    {
      unsigned long chEaten;
      LPITEMIDLIST pidl;
      unsigned long attributes;
      if (SUCCEEDED(pFolder->ParseDisplayName( NULL,
                                               NULL,
                                               dNameW,
                                               &chEaten,
                                               &pidl,
                                               &attributes )))
      {
        ITRACE_DEVELOP("Successfull");
        temp = ItemID( pidl );
      }
    }
  }
  return temp;
}

/*------------------------------------------------------------------------------
| ItemID::itemIcon                                                             |
|                                                                              |
| Returns returns an icon from the system image list for the itemid. Pass in   |
| true to get the open icon.
------------------------------------------------------------------------------*/
IPointerHandle ItemID::itemIcon( Boolean openIcon )
{
  SHFILEINFOA    sfi;
  if (!sysImageList)
  {
    sysImageList = (HIMAGELIST) SHGetFileInfoA( "C:\\",
                   0,
                   &sfi,
                   sizeof(SHFILEINFOA),
                   SHGFI_SYSICONINDEX
//                   | SHGFI_SMALLICON
                   | SHGFI_LARGEICON
                   );

  }

  if ( thePidl )
  {
    unsigned long flags = SHGFI_PIDL | SHGFI_SYSICONINDEX |
//                          SHGFI_SMALLICON;
                          SHGFI_LARGEICON;
    if ( openIcon )
      flags |= SHGFI_OPENICON;

    SHGetFileInfoA( (LPCSTR) thePidl,
                    0,
                    &sfi,
                    sizeof(SHFILEINFOA),
                    flags);

  }
  else
    // If the the pidl is null assume that the desktop icon should be returned.
    // Really need a better way to get the desktop icon!
    sfi.iIcon = 34;


  return ImageList_GetIcon( sysImageList, sfi.iIcon, ILD_NORMAL );
}

/*------------------------------------------------------------------------------
| ItemID::fullyQualifiedName                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString ItemID::fullyQualifiedName( )
{
  IString fullName;
  LPSHELLFOLDER pFolder;
  if (SUCCEEDED(SHGetDesktopFolder( &pFolder)))
  {
    fullName = name( pFolder, SHGDN_FORPARSING );
    pFolder->Release();
  }
  return fullName;
}

/*------------------------------------------------------------------------------
| ItemID::itemType                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString ItemID::itemType( )
{
  SHFILEINFOA sfi;

  IString typeName;

  if ( thePidl )
  {
    setToLast();
    if ( currentIsValid() )
    {
      SHGetFileInfoA( (LPCSTR) currentPidl,
                      0,
                      &sfi,
                      sizeof(SHFILEINFOA),
                      SHGFI_PIDL |
                      SHGFI_TYPENAME );

      typeName = IString( sfi.szTypeName );
    }
  }
  return typeName;
}


/*------------------------------------------------------------------------------
| ItemID::setToFirst
|
| Sets current pidl pointer to first pidl
------------------------------------------------------------------------------*/
ItemID& ItemID::setToFirst( )
{
  currentPidl = thePidl;
  return *this;
}

/*------------------------------------------------------------------------------
| ItemID::setToLast
|
| Sets current pidl pointer to last pidl
------------------------------------------------------------------------------*/
ItemID& ItemID::setToLast( )
{
  LPITEMIDLIST lastGood;
  currentPidl = thePidl;
  while (currentPidl->mkid.cb)
  {
    lastGood = currentPidl;
    LPSTR lpMem=(LPSTR)currentPidl;

    lpMem+=currentPidl->mkid.cb;

    currentPidl = (LPITEMIDLIST) lpMem;

  }
  currentPidl = lastGood;
  return *this;
}

/*------------------------------------------------------------------------------
| ItemID::setToNext
|
| Sets current pidl pointer to next pidl
------------------------------------------------------------------------------*/
ItemID& ItemID::setToNext( )
{
  LPSTR lpMem=(LPSTR)currentPidl;

  lpMem+=currentPidl->mkid.cb;

  currentPidl = (LPITEMIDLIST) lpMem;
  return *this;
}

/*------------------------------------------------------------------------------
| ItemID::current
|
| returns a copy of the current pidl;
------------------------------------------------------------------------------*/
ItemID ItemID::current()
{
  return ItemID( currentPidl );
}

/*------------------------------------------------------------------------------
| ItemID::currentIsValid
|
| returns true if current pidl is valid
------------------------------------------------------------------------------*/
Boolean ItemID::currentIsValid()
{
  return (currentPidl->mkid.cb) ? true : false;
}

